package dds.control;

import java.applet.Applet;
import java.awt.Image;
import java.awt.Toolkit;
import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.imageio.ImageIO;

import processing.core.PApplet;
import processing.core.PFont;
import processing.core.PImage;
import dds.control.simulator.SimController;
import dds.control.user.AbstractModule;

/**
 * The DDSApplet class is the main class for the Defensive Driving Simulation.
 * It initializes and then directs the execution of all the subcomponents which
 * form this application. An instance of this class is launched by the Java
 * OpenGL Applet Laucher, which allows the PApplet to use a PGraphicsOpenGL
 * object to render frames. At present this class mainly functions as a
 * Processing sketch would by defining the main program flow in the setup() and
 * draw() methods. The setup() method is called first, and it's function within
 * a DDSApplet is to prepare the initial RenderableContext and InputHandler. The
 * DDSApplet's current RenderableContext is responsible for drawing each frame,
 * and the current InputHandler is responsible for interpreting keyboard and
 * mouse actions. These actions are mapped to the current InputHandler through
 * the various key and mouse methods. The setup() method prepares a
 * RenderingContext and InputHandler for the main menu. After the DDSApplet has
 * been setup, it will begin its rendering loop, which consists of calling the
 * draw() to render a single frame. The draw() method is a mapping to the
 * current RenderingContext's drawOn(DDSApplet) method, and that is what does
 * the drawing onto the screen. The DDSApplet is also responsible for overriding
 * certain aspects of PApplet functionality. Currently this is limited to the
 * loading and management of images. Our enhancements allow the
 * loadImage(String) method to utilize files directly from our ResourceLoader.
 * This gives us access to the files in the local, jar-level, and remote data
 * repositories.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public class DDSApplet extends PApplet {
    /**
     * The applet is a static reference to the first instance of DDSApplet that
     * has had its setup run.
     */
    private static DDSApplet applet;

    /**
     * The DEFAULT_HEIGHT for the application, this is the value used if it
     * cannot be loaded from an applet context.
     */
    public static final int DEFAULT_HEIGHT = 480;

    /**
     * The DEFAULT_WIDTH for the application, this is the value used if it
     * cannot be loaded from an applet context.
     */
    public static final int DEFAULT_WIDTH = 640;

    /**
     * The RENDER_MODE. We are now using our own extension of OPENGL because of
     * a bug that existed in the old P3D renderer (Processing versions Beta
     * 0115-0118) that affected z-ordering in perspective mode. We additionally
     * extended Processing's OPENGL implementation because it did not support
     * repeating textures.
     */
    public static final String RENDER_MODE = "dds.control.DDSGraphicsGLDirect";

    /**
     * The serialVersionUID id the unique ID to be used for Java serialization.
     */
    private static final long serialVersionUID = 1L;

    // /**
    // * The NULL_APPLET is used mainly for initialization and checking to see
    // if
    // * a DDSApplet has been created or not.
    // */
    // private static final DDSApplet NO_APPLET = null;

    /**
     * The NO_MODULE constant is used for setting AbstractModule variables to
     * null so that it is easy to read.
     */
    public static final AbstractModule NO_MODULE = null;

    /**
     * The DISPOSE_INDEX is the index for the dispose RegisteredMethods.
     */
    private static final int DISPOSE_INDEX = 0;

    /**
     * The DRAW_INDEX is the index for the draw RegisteredMethods.
     */
    private static final int DRAW_INDEX = 1;

    /**
     * The KEY_INDEX is the index for the key RegisteredMethods.
     */
    private static final int KEY_INDEX = 2;

    /**
     * The MOUSE_INDEX is the index for the mouse RegisteredMethods.
     */
    private static final int MOUSE_INDEX = 3;

    /**
     * The POST_INDEX is the index for the post RegisteredMethods.
     */
    private static final int POST_INDEX = 4;

    /**
     * The PRE_INDEX is the index for the pre RegisteredMethods.
     */
    private static final int PRE_INDEX = 5;

    /**
     * The SIZE_INDEX is the index for the size RegisteredMethods.
     */
    private static final int SIZE_INDEX = 6;

    // /**
    // * The camPosition contains the xyz coordinates for the location of the
    // * camera's eye.
    // */
    // private final Vector3D camPosition;
    //
    // /**
    // * The camAimPosition contains the xyz coordinates for the location that
    // the
    // * camera should look at.
    // */
    // private final Vector3D camAimPosition;
    //
    // /**
    // * The upVector contains the xyz coordinates that describe the unit vector
    // * that the camera should use for it's upward direction. This vector
    // should
    // * always be a unit vector.
    // */
    // private final Vector3D upVector;
    //
    // /**
    // * The clippingNear describes the distance closest to the camera that
    // causes
    // * clipping.
    // */
    // private float clippingNear;
    //
    // /**
    // * The clippingFar describes the distance farthest to the camera that
    // causes
    // * clipping.
    // */
    // private float clippingFar;
    //
    // /**
    // * The fieldOfView describes the viewing angle to be used for perspective
    // * projection in radians.
    // */
    // private float fieldOfView;
    //
    // /**
    // * The aspectRatio is the ratio of width to height to be used for
    // * perspective projection.
    // */
    // private float aspectRatio;

    /**
     * The loadedImages field is a cache of PImages that get loaded so that the
     * DDSApplet doesn't have to repeatedly establish IO links for already
     * accessed images.
     */
    private final transient Map<String, PImage> loadedImages;

    /**
     * Getter for static applet. This applet is the first one to have its
     * setup() run within this JVM.
     * 
     * @return the applet
     */
    // TODO Try to do away with this functionality by assigning the appropriate
    // applet as a parent to the dependent components.
    public static final DDSApplet getApplet() {
        return applet;
    }

    /**
     * Setter for static reference to the first DDSApplet created within this
     * JVM. This will only set the DDSAppet.applet field the first time it is
     * run, provided the first call does not set it to null, which should never
     * happen, in fact, don't let that happen because it gets confusing.
     * 
     * @param theApplet the first DDSApplet instance.
     */
    public static final void setApplet(final DDSApplet theApplet) {
        // if (DDSApplet.applet == NO_APPLET) {
        if (DDSApplet.applet != null) {
            DDSApplet.applet.destruct();
        }
        DDSApplet.applet = theApplet;
        // }
    }

    /**
     * The destruct method performs cleanup to prevent running out of memory.
     */
    private void destruct() {
        // Clear any previously loaded OBJModels so they don't glitch out OpenGL
        // if the browser is reloaded.
        DDSOBJModel.clearCache();

        SimController.getSimControl().destruct();
        System.gc();
    }

    /**
     * The moduleLoader field contains this DDSApplet's module loader, which is
     * responsible for managing the module to use for the application.
     */
    private final transient DDSModuleLoader moduleLoader;

    /**
     * Constructor for DDSApplet. This should be where the start-up handlers and
     * initial application instantiation should occur. The setup method should
     * be used to connect the new instances and begin Object interaction.
     */
    public DDSApplet() {
        super();

        // Initialize component size to prevent an annoying resizing glitch on
        // startup.
        this.setSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);

        // // Initialize Vector3D objects to be used for storing camera
        // positioning
        // // information.
        // this.camPosition = new Vector3D();
        // this.camAimPosition = new Vector3D();
        // this.upVector = new Vector3D();
        //
        // // Initialize variables that store information about the perspective
        // // projection being used.
        // this.clippingFar = 0;
        // this.clippingNear = 0;
        // this.fieldOfView = 0;
        // this.aspectRatio = 1;

        // Set the loadImageFormats variable with a String[] that contains the
        // file extensions for image formats that Java is able to load.
        // TODO Investigate removing this by setting it to null. We may have
        // done away with the need to use this through our creation and use of
        // the DDSImageFileType enumeration class
        this.loadImageFormats = ImageIO.getReaderFormatNames();

        // Set up a Map that allows us to cache PImages when they are loaded to
        // prevent extraneous file io.
        this.loadedImages = new HashMap<String, PImage>();

        // Instantiate a ModuleLoader for this DDSApplet.
        this.moduleLoader = new DDSModuleLoader(this);
    }

    /**
     * The resetRegisteredMethods method removes all methods registered with
     * this extension of PApplet. This method needs to be removed when contexts
     * using Interfacia components are swapped out/destroyed. This is because
     * Interfacia uses registered methods to hook its component managers into
     * the PApplet function calls.
     */
    public final void resetRegisteredMethods() {
        // By setting all of the registered method fields to new, blank
        // instances all previously registered methods are cleaned out and no
        // longer used. Do this for all registered method fields.
        this.disposeMethods = new RegisteredMethods();
        this.drawMethods = new RegisteredMethods();
        this.keyEventMethods = new RegisteredMethods();
        this.mouseEventMethods = new RegisteredMethods();
        this.postMethods = new RegisteredMethods();
        this.preMethods = new RegisteredMethods();
        this.sizeMethods = new RegisteredMethods();
        // TODO Look into using these RegisteredMethod fields as a way to do
        // InputHandlers and rendering contexts.
    }

    /**
     * The METHODS_COUNT constant is the number of RegisteredMethods objects
     * that need to be tracked during module switches.
     */
    public static final int METHODS_COUNT = 7;

    /**
     * The getRegisteredMethods returns all methods registered with this
     * DDSApplet. This is used during module changes so that modules which use
     * RegisteredMethods hooks can have their hook reestablished when returned
     * to.
     * 
     * @return An array containing the current RegisteredMethods.
     */
    public final RegisteredMethods[] getRegisteredMethods() {
        // Create storage for the result.
        final RegisteredMethods[] result = new RegisteredMethods[METHODS_COUNT];

        // Copy the RegisteredMethods into their appropriate places in the
        // result array.
        result[DISPOSE_INDEX] = this.disposeMethods;
        result[DRAW_INDEX] = this.drawMethods;
        result[KEY_INDEX] = this.keyEventMethods;
        result[MOUSE_INDEX] = this.mouseEventMethods;
        result[POST_INDEX] = this.postMethods;
        result[PRE_INDEX] = this.preMethods;
        result[SIZE_INDEX] = this.sizeMethods;

        // Return the result array.
        return result;
    }

    /**
     * The setRegisteredMethods sets all the methods registered with this
     * DDSApplet. This is used during module changes so that modules which use
     * RegisteredMethods hooks can have their hook reestablished when returned
     * to.
     * 
     * @param methods The methods to set n the DDSApplet.
     */
    public final void setRegisteredMethods(final RegisteredMethods[] methods) {
        // This prevents a checkstyle error about storing a
        // user-supplied array directly. Luckily this shouldn't happen too often
        // so it shouldn't become a memory optimization issue.
        final RegisteredMethods[] methodsCopy = methods.clone();

        // Set all of the RegisteredMethods fields.
        this.disposeMethods = methodsCopy[DISPOSE_INDEX];
        this.drawMethods = methodsCopy[DRAW_INDEX];
        this.keyEventMethods = methodsCopy[KEY_INDEX];
        this.mouseEventMethods = methodsCopy[MOUSE_INDEX];
        this.postMethods = methodsCopy[POST_INDEX];
        this.preMethods = methodsCopy[PRE_INDEX];
        this.sizeMethods = methodsCopy[SIZE_INDEX];
    }

    /**
     * The draw method is called to draw each frame of the display per
     * Processing specifications. This should pass the drawing responsibility
     * off to the current RenderableContext.
     * 
     * @see processing.core.PApplet#draw()
     */
    public final void draw() {
        // Check to see if the ModuleLoader is busy.
        if (this.moduleLoader.isBusy()) {
            // It is so use the loading screen.
            this.moduleLoader.drawLoadingScreen();
        } else {
            // It's not so use the current module to draw the frame.
            this.moduleLoader.getCurrentModule().drawOn(this);
        }
    }

    /**
     * The keyPressed. This Processing input-handling method should be mapped to
     * the current InputHandler.
     * 
     * @see processing.core.PApplet#keyPressed()
     */
    public final void keyPressed() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass key info to the current InputHandler.
            this.moduleLoader.getCurrentModule().keyPressed(this.key,
                    this.keyCode);
        }
    }

    /**
     * The keyReleased. This Processing input-handling method should be mapped
     * to the current InputHandler.
     * 
     * @see processing.core.PApplet#keyReleased()
     */
    public final void keyReleased() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass key info to the current InputHandler.
            this.moduleLoader.getCurrentModule().keyReleased(this.key,
                    this.keyCode);
        }
    }

    /**
     * The mouseDragged. This Processing input-handling method should be mapped
     * to the current InputHandler.
     * 
     * @see processing.core.PApplet#mouseDragged()
     */
    public final void mouseDragged() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass mouse info to the current InputHandler.
            this.moduleLoader.getCurrentModule().mouseDragged(this.mouseX,
                    this.mouseY, this.mouseButton);
        }
    }

    /**
     * The mouseMoved. This Processing input-handling method should be mapped to
     * the current InputHandler.
     * 
     * @see processing.core.PApplet#mouseMoved()
     */
    public final void mouseMoved() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass mouse info to the current InputHandler.
            this.moduleLoader.getCurrentModule().mouseMoved(this.mouseX,
                    this.mouseY);
        }
    }

    /**
     * The mousePressed. This Processing input-handling method should be mapped
     * to the current InputHandler.
     * 
     * @see processing.core.PApplet#mousePressed()
     */
    public final void mousePressed() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass mouse info to the current InputHandler.
            this.moduleLoader.getCurrentModule().mousePressed(this.mouseX,
                    this.mouseY, this.mouseButton);
        }
    }

    /**
     * The mouseReleased. This Processing input-handling method should be mapped
     * to the current InputHandler.
     * 
     * @see processing.core.PApplet#mouseReleased()
     */
    public final void mouseReleased() {
        // Check to make sure there's an InputHandler set before we try to pass
        // information to it.
        if (!this.moduleLoader.isBusy()) {
            // Pass mouse info to the current InputHandler.
            this.moduleLoader.getCurrentModule().mouseReleased(this.mouseX,
                    this.mouseY, this.mouseButton);
        }
    }

    /**
     * The setup method is the first method called per Processing
     * specifications. Its purpose it to set up the Definsive Driving Simulation
     * up for use by a user. It's primary actions are to assert and set the size
     * of the application screen size and then initialize the main menu context
     * and handlers.
     * 
     * @see processing.core.PApplet#setup()
     */
    @Override
    public final void setup() {

        // By disabling the depth test we save a few cycles, especially since
        // OpenGL will handle the depth ordering on its own.
        this.hint(DISABLE_DEPTH_TEST);

        // Grab the Java OpenGL Applet container for this DDSApplet.
        final Applet jogla = (Applet) this.getParent();

        // Set width and height to the defaults.
        int wide = DEFAULT_WIDTH;
        int high = DEFAULT_HEIGHT;
        if (this.online) {
            try {
                // Try to get new width and height from the webpage.
                wide = Integer.parseInt(jogla.getParameter("width"));
                high = Integer.parseInt(jogla.getParameter("height"));
            } catch (NumberFormatException e) {
                // Integers were not available as parameters.
                // TODO Add logger here.
                System.err.println(e);
            }

            // Set OpenGL container's size to prevent a lockup that can happen
            // in
            // the Eclipse applet run configuration.
            jogla.setSize(wide, high);
            // jogla.setPreferredSize(new Dimension(wide, high));
        }

        // Set a static reference to this DDSApplet.
        DDSApplet.setApplet(this);

        // Set the size and mode for rendering.
        this.size(wide, high, DDSApplet.RENDER_MODE);

        // Dan: I hooked the Administration option on the main menu to start the
        // trap editor.
        this.moduleLoader.setup();
    }

    /**
     * The loadModule method provides access to loading modules via this
     * DDSApplet's ModuleLoader.
     * 
     * @param moduleArg The argument for loading the module.
     */
    public final void loadModule(final Object moduleArg) {
        this.moduleLoader.loadModule(moduleArg);
    }

    /**
     * The loadImage method is reponsible for producing a PImage from a
     * specified file. The original method was copied directly from Processing
     * source code on 10/20/2006 and then streamlined for our own use.
     * 
     * @param fileName The file to load into a PImage
     * @return A new PImage or null if it does not load.
     * @see processing.core.PApplet#loadImage(java.lang.String)
     */
    public final PImage loadImage(final String fileName) {
        // Use this as storage for the resulting PImage, and set it to null for
        // now to indicate that it has not been successfully created yet.
        PImage result = null;

        // Check to see if this file has already been loaded and therefor placed
        // in our loaded images cache.
        if (this.loadedImages.containsKey(fileName)) {
            // If it was then just grab it from the cache.
            result = this.loadedImages.get(fileName);
        } else {
            // Otherwise parse the filename for the type of image file that
            // we'll be trying to load.
            final DDSImageFileType imageFileType = DDSImageFileType
                    .parseImageFileType(fileName);

            // Branch on type of image file since there are different ways to
            // handle different image formats.
            switch (imageFileType) {
            case TGA:
                // Load TIFF file using a Processing-based method.
                result = this.safeLoadImageTGA(fileName);
                break;
            case TIFF:
                // Get image file data from repository
                final byte[] bytes = loadImageBytes(fileName);
                // Load TIFF using wrapper to Processing's TIFF protected image
                // parser.
                // TODO Consider using reflection here so that we can remove the
                // DDSImage wrapper class since that would alleviate the
                // responsibility of keeping up with another Processing
                // implementation.
                result = DDSImage.publicLoadTIFF(bytes);
                break;
            case GIF:
            case JPEG:
            case PNG:
                // Load file using Processing's approach to certain standard
                // file types, which is supposedly faster than using Java's
                // ImageIO classes.
                result = this.loadImageProcessingStyle(fileName, imageFileType);
                break;
            case JAVA_IMAGEIO:
                // Load file using Processing's wrapper to Java ImageIO classes.
                // TODO Make sure this works with the repository.
                result = this.loadImageIO(fileName);
                break;
            case UNSUPPORTED:
            default:
                // Unsupported image file type detected.
                // TODO Consider logger and/or throwing an exception.
            }

            // Check to see if a PImage was succesfully created.
            if (result != null) {
                // If so, it must be added to the cache.
                this.loadedImages.put(fileName, result);
            }
        }

        // Return the result after all that
        // NOTE: The result may be null, so calling methods will need to check
        // for this.
        return result;
    }

    /**
     * The loadImageProcessingStyle method is an abstraction of the way
     * Processing tries to load JPEG, GIF and PNG images. The method relies on
     * Java's AWT-based createImage method, which is supposed to be faster than
     * going through Java's ImageIO classes.
     * 
     * @param fileName The name of the image file to load.
     * @param imageFileType The image file format type.
     * @return A loaded PImage or null.
     */
    private PImage loadImageProcessingStyle(final String fileName,
            final DDSImageFileType imageFileType) {
        // Use a repository-based image-loading method to create a
        // byte[] with the correct file's data.
        final byte[] bytes = loadImageBytes(fileName);

        // Storage for the resulting PImage.
        PImage result = null;

        // For jpeg, gif, and png, load them using createImage(),
        // because the javax.imageio code was found to be much slower,
        // see
        // <A
        // HREF="http://dev.processing.org/bugs/show_bug.cgi?id=392">Bug
        // 392</A>.
        try {
            // Use Java's AWT-based createImage method to load in image data.
            final Image awtImage = Toolkit.getDefaultToolkit().createImage(
                    bytes);
            // Convert Java's Image data into a PImage.
            final PImage pimage = this.loadImageSync(awtImage);
            // Since there was no way to access the PImage.checkAlpha
            // method, we had to wrap it in an public method to use.
            // TODO Consider using reflection here so that we can remove the
            // DDSImage wrapper class since that would alleviate the
            // responsibility of keeping up with another Processing
            // implementation.
            final DDSImage image = new DDSImage(pimage);
            // Check to see if the file's type supports transparency.
            if (imageFileType.isTransparent()) {
                // If so, check the data to see if it's used and enable it if it
                // is.
                image.publicCheckAlpha();
            }
            // Set the result to the DDSImage we loaded.
            result = image;
        } catch (Exception e) {
            // TODO Add logger.
            System.err.println(e);
        }
        return result;
    }

    /**
     * The safeLoadImageTGA method simplifies calls to the standard PApplet
     * loadImageTGA(String) method by wrapping that in an appropriate try-catch
     * block.
     * 
     * @param fileName The name of the TGA-formatted image file.
     * @return A PImage for the specified file or null if the operation fails or
     *         throws an exception.
     */
    private PImage safeLoadImageTGA(final String fileName) {
        PImage result = null;
        try {
            // Try to use Processing's TGA image loader.
            // TODO Determine if TGA loading works from the data
            // repository
            result = loadImageTGA(fileName);
        } catch (IOException e) {
            // TODO Add logger.
            System.err.println(e);
        }
        return result;
    }

    /**
     * The loadImageBytes method is used to load up a byte[] for an image file
     * from the data repository. The basis for this has been borrowed from
     * Processing source code from 10/20/2006.
     * 
     * @param filename The name of the image file to try to load.
     * @return A byte[] with the image data or null if an image could not be
     *         loaded.
     */
    private byte[] loadImageBytes(final String filename) {
        byte[] result = null;
        try {
            // Get resource URL from the ResourceLoader.
            final URL fileURL = ResourceManager.getLOADER().getFile(filename,
                    ResourceType.IMAGE);
            // Create streams to grab and save data.
            final ByteArrayOutputStream out = new ByteArrayOutputStream();
            final BufferedInputStream bis = new BufferedInputStream(fileURL
                    .openStream());
            // Read data from input stream
            int aChar = bis.read();
            // Continue reading until there is no more data.
            while (aChar != -1) {
                out.write(aChar);
                aChar = bis.read();
            }
            // Get an array of the collected data stream.
            result = out.toByteArray();
        } catch (FileNotFoundException e) {
            // TODO Add logger
            System.err.println(e);
        } catch (IOException e) {
            // TODO Add logger
            System.err.println(e);
        }
        return result;
    }

    // TODO Setup other processing IO functions to utilize the data repository
    // through the Resource loader

    /**
     * The transformToScreen method sets up a transformation that places
     * coordinates on the z=0 plane directly on the screen plane. This method
     * needs to be followed by a call to transformFromScreen() because this
     * invokes pushMatrix() and the matrix stack needs to be managed properly.
     */
    public final void tranformToScreen() {
        ((DDSGraphicsGLDirect) g).transformToScreen();
    }

    /**
     * The transformFromScreen method reverts the transformation caused by a
     * previous call to transformFromScreen(), which must be called before this
     * is because this invokes popMatrix() and the matrix stack needs to be
     * managed properly.
     */
    public final void tranformFromScreen() {
        ((DDSGraphicsGLDirect) g).transformFromScreen();
    }

    /**
     * The beginList method begins the definition of an OpenGL display list.
     * 
     * @param list The list ID.
     */
    public final void beginList(final int list) {
        ((DDSGraphicsGLDirect) g).beginList(list);
    }

    /**
     * The endList method completes the definition of an OpenGL display list.
     */
    public final void endList() {
        ((DDSGraphicsGLDirect) g).endList();
    }

    /**
     * The drawList method invokes a previously defined OpenGL display list.
     * 
     * @param list The list ID.
     */
    public final void drawList(final int list) {
        ((DDSGraphicsGLDirect) g).drawList(list);
    }

    /**
     * The openStream method tries to open a stream on a file located somewhere
     * in the repository. Since this method cannot know what type of file is
     * expected to be accessed this method finds files in the repository root
     * folder. Therefore and resoucre accesses should have the appropriate
     * folder prepended to the filename.
     * 
     * @param arg0 Filename of the resource fom the base folder of the
     *            repository.
     * @return An InputStream if this works, null if it doesn't.
     * @see processing.core.PApplet#openStream(java.lang.String)
     */
    @Override
    public final InputStream openStream(final String arg0) {
        InputStream result = null;
        try {
            final URL resource = ResourceManager.getLOADER().getFile(arg0,
                    ResourceType.DIRECT);
            result = resource.openStream();
        } catch (FileNotFoundException e) {
            // TODO Add logger.
            System.err.println(e);
        } catch (IOException e) {
            // TODO Add logger.
            System.err.println(e);
        }
        return result;
    }

    /**
     * The loadFont method loads a font as a PFont so it can be used for
     * rendering text using Processing methods. This method has been overridden
     * to make fonts load from storage in the repository. Just use the name of
     * the font. Fonts must be in the Processing format (.vlw).
     * 
     * @param arg0 The font .vlw file name.
     * @return The appropriate PFont.
     * @see processing.core.PApplet#loadFont(java.lang.String)
     */
    @Override
    public final PFont loadFont(final String arg0) {
        return super.loadFont(ResourceType.FONT.getDirectory() + arg0);
    }

    /**
     * The isStandAlone method returns if we're running as an applet or as a
     * stand-alone application.
     * 
     * @return True if being run as a stand-alone app.
     */
    public final boolean isStandAlone() {
        return this.getAppletContext() instanceof DDSLauncher;
    }

    /**
     * Getter for moduleLoader.
     * 
     * @return the moduleLoader
     */
    public final DDSModuleLoader getModuleLoader() {
        return moduleLoader;
    }

    /**
     * Save the current frame as a .tif or .tga image.
     * <P>
     * The String passed in can contain a series of # signs that will be
     * replaced with the screengrab number.
     * 
     * <PRE>
     * 
     * i.e. saveFrame("blah-####.tif"); // saves a numbered tiff image,
     * replacing the // #### signs with zeros and the frame number
     * 
     * </PRE>
     * 
     * @param what The nae of the image to save to.
     * @see processing.core.PApplet#saveFrame(java.lang.String)
     */
    public final void saveFrame(final String what) {
        if (!this.isStandAlone()) {
            // TODO Add logger.
            System.err.println("Can't use saveFrame()"
                    + " when running in a browser.");
            return;
        }

        g.save(savePath(insertFrame(what)));
    }
}
