package dds.control;

import javax.media.opengl.GL;
import javax.media.opengl.GLContext;

import processing.core.PApplet;
import processing.opengl.PGraphicsOpenGL;

import com.sun.opengl.util.BufferUtil;

/**
 * The DDSGraphicsOpenGL class is an extension of the standard Processing OpenGL
 * implementation that allows access to OpenGL accelerations. Since the main
 * goal of this class is to introduce increases in render rates, much of
 * Processing's original code will be replaced since Processing tends to do
 * things that OpenGL will do on it's own. It's hoped that by removing the extra
 * work done by Processing and allowing access to more advanced OpenGL
 * accelerations we will be able to better guarantee decent framerates on most
 * platforms. Due to limitations of OpenGL and the way in which Processing
 * interacts with OpenGL, repeating texture images MUST have both a width and
 * height that is a power of two. If a texture image does not have dimensions
 * that are powers of two and is given uv coordinates that require repetition,
 * the result will be undesireable. It will probably result in gaps in the
 * texturing.
 * 
 * @author Jacob Bolton
 */
public class DDSGraphicsGLDirect extends PGraphicsOpenGL {

    /**
     * The EIGHT constant has a value of 8.
     */
    private static final int EIGHT = 8;

    /**
     * The ELEVEN constant has a value of 11.
     */
    private static final int ELEVEN = 11;

    /**
     * The FIFTEEN constant has a value of 15.
     */
    private static final int FIFTEEN = 15;

    /**
     * The FIVE constant has a value of 5.
     */
    private static final int FIVE = 5;

    /**
     * The FOUR constant has a value of 4.
     */
    private static final int FOUR = 4;

    /**
     * The FOURTEEN constant has a value of 14.
     */
    private static final int FOURTEEN = 14;

    /**
     * The HALF constant has a value of 0.5.
     */
    private static final float HALF = 0.5f;

    /**
     * The NINE constant has a value of 9.
     */
    private static final int NINE = 9;

    /**
     * The ONE_EIGHTY constant has a value of 180.
     */
    private static final float ONE_EIGHTY = 180;

    /**
     * The SEVEN constant has a value of 7.
     */
    private static final int SEVEN = 7;

    /**
     * The SIX constant has a value of 6.
     */
    private static final int SIX = 6;

    /**
     * The SIXTEEN constant has a value of 16.
     */
    private static final int SIXTEEN = 16;

    /**
     * The TEN constant has a value of 10.
     */
    private static final int TEN = 10;

    /**
     * The THIRTEEN constant has a value of 13.
     */
    private static final int THIRTEEN = 13;

    /**
     * The THREE constant has a value of 3.
     */
    private static final int THREE = 3;

    /**
     * The TWELVE constant has a value of 12.
     */
    private static final int TWELVE = 12;

    /**
     * The TWO constant has a value of 2.
     */
    private static final int TWO = 2;

    /**
     * The ZERO constant has a value of 0.
     */
    private static final int ZERO = 0;

    /**
     * The dim1 field stores the current first axis being used for the current
     * polygon's calculations.
     */
    private transient int dim1;

    /**
     * The dim2field stores the current second axis being used for the current
     * polygon's calculations.
     */
    private transient int dim2;

    /**
     * Constructor for DDSGraphicsGLDirect.
     * 
     * @param width Width of this graphics.
     * @param height Height of this graphics.
     * @param parent The parent PApplet.
     */
    public DDSGraphicsGLDirect(final int width, final int height,
            final PApplet parent) {
        // Just call the superclass since there's nothing else to initialize at
        // this point, and it's necessary to provide a constructor since the
        // default is inaccessible.
        super(width, height, parent);

        // Create these now so they don't have to be created during the render
        // loop.
        projectionFloats = new float[SIXTEEN];
        projectionFloatBuffer = BufferUtil.newFloatBuffer(SIXTEEN);

        // this.defaults() needs to be run on some platforms, but not on Macs as
        // far as I can tell.
        if (!NativeLibInfo.detectPlatform().isMacOS()) {
            this.defaults();
        }
    }

    /**
     * The beginDraw.
     * 
     * @see processing.opengl.PGraphicsOpenGL#beginDraw()
     */
    @Override
    public final void beginDraw() {
        super.beginDraw();
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glEnable(GL.GL_DEPTH_TEST);
    }

    /**
     * The beginList method begins the definition of an OpenGL display list.
     * 
     * @param list The list ID.
     */
    public final void beginList(final int list) {
        gl.glNewList(list, GL.GL_COMPILE_AND_EXECUTE);
    }

    /**
     * The calcPolyArea method calculates the area of a polygon based on the
     * current dimensions.
     * 
     * @return The calculated area.
     */
    private float calcPolyArea() {
        float area = 0;
        int pVert = vertex_end - 1;
        int qVert = vertex_start;
        for (; qVert < vertex_end; pVert = qVert++) {
            area += vertices[qVert][dim1] * vertices[pVert][dim2]
                    - vertices[pVert][dim1] * vertices[qVert][dim2];
        }
        return area;
    }

    /**
     * The camera method sets up the camera used to view the scene.
     * 
     * @param eyeX x coordinate for the eye.
     * @param eyeY y coordinate for the eye.
     * @param eyeZ z coordinate for the eye.
     * @param centerX x coordinate for the center of the scene.
     * @param centerY y coordinate for the center of the scene.
     * @param centerZ z coordinate for the center of the scene.
     * @param upX x component of the upwards directional vector.
     * @param upY y component of the upwards directional vector.
     * @param upZ z component of the upwards directional vector.
     * @see processing.core.PGraphics3D#camera(float, float, float, float,
     *      float, float, float, float, float)
     */
    @Override
    public final void camera(final float eyeX, final float eyeY,
            final float eyeZ, final float centerX, final float centerY,
            final float centerZ, final float upX, final float upY,
            final float upZ) {
        super
                .camera(eyeX, eyeY, eyeZ, centerX, centerY, centerZ, upX, upY,
                        upZ);
        if (GLContext.getCurrent() != null) {
            glu.gluLookAt(eyeX, eyeY, eyeZ, centerX, centerY, centerZ, upX,
                    upY, upZ);
        }
    }

    /**
     * The createLines method takes care of generating stroked lines for a
     * shape.
     * 
     * @param mode The shape closing mode.
     */
    private void createLines(final int mode) {
        switch (shape) {

        case POINTS:
            this.strokePoints();
            break;

        case LINES:
            // case LINE_STRIP:
            // case LINE_LOOP:
            this.strokeLines(mode);
            break;

        case TRIANGLES:
            this.strokeTriangles();
            break;

        case TRIANGLE_STRIP:
            this.strokeTriangleStrip();
            break;

        case TRIANGLE_FAN:
            this.strokeTriangleFan();
            break;

        case QUADS:
            this.strokeQuads();
            break;

        case QUAD_STRIP:
            this.strokeQuadStrip();
            break;

        case POLYGON:
            // case CONCAVE_POLYGON:
            // case CONVEX_POLYGON:
            this.strokePolygon(mode);
            break;
        default:
        }
    }

    /**
     * The createTriangles method takes care of generating filled triangles for
     * a shape.
     */
    private void createTriangles() {
        switch (shape) {
        case TRIANGLE_FAN:
            this.triangulateTriangleFan();
            break;

        case QUADS:
            this.triangulateQuads();
            break;

        case QUAD_STRIP:
            this.triangulateQuadStrip();
            break;

        case POLYGON:
            // case CONCAVE_POLYGON:
            // case CONVEX_POLYGON:
            triangulatePolygon();
            break;

        case TRIANGLES:
        case TRIANGLE_STRIP:
        default:
            this.triangulateTriangles();
            break;
        }
    }

    /**
     * The drawList method invokes a previously defined OpenGL display list.
     * 
     * @param list The list ID.
     */
    public final void drawList(final int list) {
        gl.glCallList(list);
    }

    /**
     * The endList method completes the definition of an OpenGL display list.
     */
    public final void endList() {
        gl.glEndList();
    }

    /**
     * The endShape method takes care of finishing up a shape's verts and then
     * renders it.
     * 
     * @param mode The shape closing mode.
     * @see processing.core.PGraphics3D#endShape(int)
     */
    public final void endShape(final int mode) {
        vertex_end = vertexCount;
        vertex_end_including_clip_verts = vertex_end;

        // don't try to draw if there are no vertices
        // (fixes a bug in LINE_LOOP that re-adds a nonexistent vertex)
        if (vertexCount == 0) {
            shape = 0;
            return;
        }

        // ------------------------------------------------------------------
        // CREATE LINES
        if (stroke) {
            this.createLines(mode);
        }

        // ------------------------------------------------------------------
        // CREATE TRIANGLES
        if (fill) {
            this.createTriangles();
        }

        // ------------------------------------------------------------------
        // TRANSFORM / LIGHT / CLIP
        if (lightCount > 0 && fill) {
            handle_lighting();
        } else {
            handle_no_lighting();
        }

        // ------------------------------------------------------------------
        // RENDER SHAPES FILLS HERE
        if (fill) {
            render_triangles();
        }
        if (stroke) {
            render_lines();
        }

        shape = 0;
    }

    /**
     * The findPolyAxis method attempts to determine the orthoganal orientation
     * that will allow for optimal analysis of the polygon's verts.
     * 
     * @return True if an orientation could be determined.
     */
    private boolean findPolyAxis() {
        boolean result = true;
        boolean foundValidX = false;
        boolean foundValidY = false;
        for (int i = vertex_start; i < vertex_end; i++) {
            if (vertices[i][MX] != 0) {
                foundValidX = true;
            }
            if (vertices[i][MY] != 0) {
                foundValidY = true;
            }
            // TODO Determine if this is a useful optimization.
            if (foundValidX && foundValidY) {
                // System.out.println("optimized");
                break;
            }
        }
        if (foundValidX && !foundValidY) {
            // d1 = MX; // already the case
            dim2 = MZ;
        } else if (!foundValidX && foundValidY) {
            // ermm.. which is the proper order for cw/ccw here?
            dim1 = MY;
            dim2 = MZ;
        } else {
            // screw it, this polygon is just f-ed up
            // return null;
            result = false;
        }
        return result;
    }

    /**
     * The flush.
     * 
     * @see processing.core.PGraphics3D#flush()
     */
    @Override
    protected void flush() {
        // TODO Auto-generated method stub
        // super.flush();
    }

    /**
     * The frustum method sets a perspective matrix defined through the
     * parameters.
     * 
     * @param left Left component of the clipping plane.
     * @param right Rigth component of the clipping plane.
     * @param bottom Bottom component of the clipping plane.
     * @param top Top component of the clipping plane.
     * @param near Near component of the clipping plane.
     * @param far Far component of the clipping plane.
     * @see processing.core.PGraphics3D#frustum(float, float, float, float,
     *      float, float)
     */
    @Override
    public final void frustum(final float left, final float right,
            final float bottom, final float top, final float near,
            final float far) {
        super.frustum(left, right, bottom, top, near, far);
        if (GLContext.getCurrent() != null) {
            gl.glMatrixMode(GL.GL_PROJECTION);
            gl.glLoadIdentity();
            gl.glFrustum(left, right, bottom, top, near, far);
            gl.glMatrixMode(GL.GL_MODELVIEW);
        }
    }

    /**
     * The loadProjection method loads the current Processing projection matrix
     * into OpenGL.
     */
    private void loadProjection() {
        gl.glMatrixMode(GL.GL_PROJECTION);
        projectionFloats[ZERO] = projection.m00;
        projectionFloats[(int) ONE] = projection.m10;
        projectionFloats[TWO] = projection.m20;
        projectionFloats[THREE] = projection.m30;

        projectionFloats[FOUR] = projection.m01;
        projectionFloats[FIVE] = projection.m11;
        projectionFloats[SIX] = projection.m21;
        projectionFloats[SEVEN] = projection.m31;

        projectionFloats[EIGHT] = projection.m02;
        projectionFloats[NINE] = projection.m12;
        projectionFloats[TEN] = projection.m22;
        projectionFloats[ELEVEN] = projection.m32;

        projectionFloats[TWELVE] = projection.m03;
        projectionFloats[THIRTEEN] = projection.m13;
        projectionFloats[FOURTEEN] = projection.m23;
        projectionFloats[FIFTEEN] = projection.m33;
        projectionFloatBuffer.put(projectionFloats);
        projectionFloatBuffer.rewind();
        gl.glLoadMatrixf(projectionFloatBuffer);

        gl.glMatrixMode(GL.GL_MODELVIEW);
    }

    /**
     * The makePolyTriangles method performs the actual triangluation of the
     * polygon.
     */
    private void makePolyTriangles() {
        // remove vc-2 Vertices, creating 1 triangle every time
        int vertexCount = vertex_end - vertex_start;
        int count = 2 * vertexCount; // complex polygon detection

        // for (int m = 0, v = vertexCount - 1; vertexCount > 2;) {
        for (int vVert = vertexCount - 1; vertexCount > 2;) {
            boolean snip = true;

            // if we start over again, is a complex polygon
            if (0 >= (count--)) {
                break; // triangulation failed
            }

            // get 3 consecutive vertices <u,v,w>
            int uVert = vVert;
            if (vertexCount <= uVert) {
                uVert = 0; // previous
            }
            vVert = uVert + 1;
            if (vertexCount <= vVert) {
                vVert = 0; // current
            }
            int wVert = vVert + 1;
            if (vertexCount <= wVert) {
                wVert = 0; // next
            }

            // triangle A B C
            final float pointAx = -vertices[vertex_order[uVert]][dim1];
            final float pointAy = vertices[vertex_order[uVert]][dim2];
            final float pointBx = -vertices[vertex_order[vVert]][dim1];
            final float pointBy = vertices[vertex_order[vVert]][dim2];
            final float pointCx = -vertices[vertex_order[wVert]][dim1];
            final float pointCy = vertices[vertex_order[wVert]][dim2];

            // first we check if <u,v,w> continues going ccw
            if (EPSILON > ((pointBx - pointAx) * (pointCy - pointAy))
                    - ((pointBy - pointAy) * (pointCx - pointAx))) {
                continue;
            }

            for (int p = 0; p < vertexCount; p++) {
                // float ax, ay, bx, by, cx, cy, apx, apy, bpx, bpy, cpx, cpy;
                // float cCROSSap, bCROSScp, aCROSSbp;

                if ((p == uVert) || (p == vVert) || (p == wVert)) {
                    continue;
                }

                final float pointPx = -vertices[vertex_order[p]][dim1];
                final float pointPy = vertices[vertex_order[p]][dim2];

                final float avx = pointCx - pointBx;
                final float avy = pointCy - pointBy;
                final float bvx = pointAx - pointCx;
                final float bvy = pointAy - pointCy;
                final float cvx = pointBx - pointAx;
                final float cvy = pointBy - pointAy;
                final float apx = pointPx - pointAx;
                final float apy = pointPy - pointAy;
                final float bpx = pointPx - pointBx;
                final float bpy = pointPy - pointBy;
                final float cpx = pointPx - pointCx;
                final float cpy = pointPy - pointCy;

                final float aCROSSbp = avx * bpy - avy * bpx;
                final float cCROSSap = cvx * apy - cvy * apx;
                final float bCROSScp = bvx * cpy - bvy * cpx;

                if ((aCROSSbp >= 0.0f) && (bCROSScp >= 0.0f)
                        && (cCROSSap >= 0.0f)) {
                    snip = false;
                }
            }

            if (snip) {
                add_triangle(vertex_order[uVert], vertex_order[vVert],
                        vertex_order[wVert]);

                // remove v from remaining polygon
                // TODO Make sure that this code is properly replaced by the
                // arraycopy below.
                // for (int s = v, t = v + 1; t < vc; s++, t++) {
                // vertex_order[s] = vertex_order[t];
                // }
                System.arraycopy(vertex_order, vVert + 1, vertex_order, vVert,
                        vertexCount - vVert - 1);
                vertexCount--;

                // reset error detection counter
                count = 2 * vertexCount;
            }
        }
    }

    /**
     * The ortho method sets up an orthographic projection matrix.
     * 
     * @param left Left plane of the clipping volume.
     * @param right Right plane of the clipping volume.
     * @param bottom Bottom plane of the clipping volume.
     * @param top Top plane of the clipping volume.
     * @param near Maximum distance from the origin to the viewer.
     * @param far Maximum distance from the origin away from the viewer.
     * @see processing.core.PGraphics3D#ortho(float, float, float, float, float,
     *      float)
     */
    @Override
    public final void ortho(final float left, final float right,
            final float bottom, final float top, final float near,
            final float far) {
        super.ortho(left, right, bottom, top, near, far);
        if (GLContext.getCurrent() != null) {
            this.loadProjection();
        }
    }

    /**
     * The perspective method sets up a perspective projection matrix.
     * 
     * @param fov Field-of-view angle (in radians) for vertical direction.
     * @param aspect Ratio of width to height.
     * @param zNear z-position of nearest clipping plane.
     * @param zFar z-position of nearest farthest plane.
     * @see processing.core.PGraphics3D#perspective(float, float, float, float)
     */
    @Override
    public final void perspective(final float fov, final float aspect,
            final float zNear, final float zFar) {
        super.perspective(fov, aspect, zNear, zFar);
    }

    /**
     * The popMatrix method pops off the current transformation matrix.
     * 
     * @see processing.core.PGraphics3D#popMatrix()
     */
    @Override
    public final void popMatrix() {
        super.popMatrix();
        gl.glPopMatrix();
    }

    /**
     * The pushMatrix method pushes the current tranformation matrix onto the
     * stack.
     * 
     * @see processing.core.PGraphics3D#pushMatrix()
     */
    @Override
    public final void pushMatrix() {
        super.pushMatrix();
        gl.glPushMatrix();
    }

    /**
     * The rotate method performs a rotation transformation around the z-axis.
     * 
     * @param angle The angle of rotation in radians.
     * @see processing.core.PGraphics3D#rotate(float)
     */
    @Override
    public final void rotate(final float angle) {
        this.rotate(angle, 0, 0, 1);
    }

    /**
     * The rotate method performs a rotation transformation around an arbitrary
     * vector.
     * 
     * @param angle The angle of rotation in radians.
     * @param xVec The x component of the rotation vector.
     * @param yVec The y component of the rotation vector.
     * @param zVec The z component of the rotation vector.
     * @see processing.core.PGraphics3D#rotate(float, float, float, float)
     */
    @Override
    public final void rotate(final float angle, final float xVec,
            final float yVec, final float zVec) {
        super.rotate(angle, xVec, yVec, zVec);
        gl.glRotatef(angle / PI * ONE_EIGHTY, xVec, yVec, zVec);

    }

    /**
     * The rotateX method performs a rotation transformation around the x-axis.
     * 
     * @param angle The angle of rotation in radians.
     * @see processing.core.PGraphics3D#rotateX(float)
     */
    @Override
    public final void rotateX(final float angle) {
        this.rotate(angle, 1, 0, 0);
    }

    /**
     * The rotateY method performs a rotation transformation around the y-axis.
     * 
     * @param angle The angle of rotation in radians.
     * @see processing.core.PGraphics3D#rotateY(float)
     */
    @Override
    public final void rotateY(final float angle) {
        this.rotate(angle, 0, 0, 1);
    }

    /**
     * The rotateZ method performs a rotation transformation around the z-axis.
     * 
     * @param angle The angle of rotation in radians.
     * @see processing.core.PGraphics3D#rotateZ(float)
     */
    @Override
    public final void rotateZ(final float angle) {
        this.rotate(angle, 0, 0, 1);
    }

    /**
     * The scale method performs a scaling transformation.
     * 
     * @param scaleFactor The scaling factor.
     * @see processing.core.PGraphics3D#scale(float)
     */
    @Override
    public final void scale(final float scaleFactor) {
        super.scale(scaleFactor);
        gl.glScalef(scaleFactor, scaleFactor, scaleFactor);

    }

    /**
     * The scale method performs a scaling transformation.
     * 
     * @param xScale The x-axis scaling factor.
     * @param yScale The y-axis scaling factor.
     * @see processing.core.PGraphics3D#scale(float, float)
     */
    @Override
    public final void scale(final float xScale, final float yScale) {
        super.scale(xScale, yScale);
        gl.glScalef(xScale, yScale, 1);

    }

    /**
     * The scale method performs a scaling transformation.
     * 
     * @param xScale The x-axis scaling factor.
     * @param yScale The y-axis scaling factor.
     * @param zScale The z-axis scaling factor.
     * @see processing.core.PGraphics3D#scale(float, float, float)
     */
    @Override
    public final void scale(final float xScale, final float yScale,
            final float zScale) {
        super.scale(xScale, yScale, zScale);
        gl.glScalef(xScale, yScale, zScale);

    }

    /**
     * The setup_vertex methods prepares a vertex for later rendering.
     * 
     * @param xPos The x coordinate.
     * @param yPos The y coordinate.
     * @param zPos The z coordinate.
     * @see processing.core.PGraphics3D#setup_vertex(float, float, float)
     */
    // TODO Move away from using this and to a pure GL implementation for shapes
    // and verts.
    @Override
    protected final void setup_vertex(final float xPos, final float yPos,
            final float zPos) {
        if (vertexCount == vertices.length) {
            final float[][] temp = new float[vertexCount << 1][VERTEX_FIELD_COUNT];
            System.arraycopy(vertices, 0, temp, 0, vertexCount);
            vertices = temp;
            final int[] temp2 = new int[vertexCount << 1];
            System.arraycopy(vertex_order, 0, temp2, 0, vertexCount);
            vertex_order = temp2;
        }

        float[] vertex = vertices[vertexCount];

        // if (polygon.redundantVertex(x, y, z)) return;
        if (vertexCount > 0) {
            final float[] pvertex = vertices[vertexCount - 1];
            if ((Math.abs(pvertex[MX] - xPos) < EPSILON)
                    && (Math.abs(pvertex[MY] - yPos) < EPSILON)
                    && (Math.abs(pvertex[MZ] - zPos) < EPSILON)) {
                // this vertex is identical, don't add it,
                // because it will anger the triangulator
                return;
            }
        }

        // user called vertex(), so that invalidates anything queued
        // up for curve vertices. if this is internally called by
        // spline_segment, then splineVertexCount will be saved and restored.
        splineVertexCount = 0;

        vertex[MX] = xPos;
        vertex[MY] = yPos;
        vertex[MZ] = zPos;

        vertex[VX] = xPos;
        vertex[VY] = yPos;
        vertex[VZ] = zPos;
        vertex[VW] = 1;

        if (fill) {
            vertex[R] = fillR;
            vertex[G] = fillG;
            vertex[B] = fillB;
            vertex[A] = fillA;

            vertex[AR] = ambientR;
            vertex[AG] = ambientG;
            vertex[AB] = ambientB;

            vertex[SPR] = specularR;
            vertex[SPG] = specularG;
            vertex[SPB] = specularB;
            vertex[SPA] = specularA;

            vertex[SHINE] = shininess;

            vertex[ER] = emissiveR;
            vertex[EG] = emissiveG;
            vertex[EB] = emissiveB;
        }

        if (stroke) {
            vertex[SR] = strokeR;
            vertex[SG] = strokeG;
            vertex[SB] = strokeB;
            vertex[SA] = strokeA;
            vertex[SW] = strokeWeight;
        }

        if (textureImage != null) {
            vertex[U] = textureU;
            vertex[V] = textureV;
        }

        vertex[NX] = normalX;
        vertex[NY] = normalY;
        vertex[NZ] = normalZ;

        vertex[BEEN_LIT] = 0;

        vertexCount++;
    }

    /**
     * The sortPolyVerts method sorts the current polygon's verts according to
     * whether they've been added CW or CCW accourding to the sign of the area.
     * 
     * @param area The area of the polygon, used to determine vert order.
     */
    private void sortPolyVerts(final float area) {
        // int j = 0;
        if (area > 0) {
            for (int i = vertex_start, j = 0; i < vertex_end; i++) {
                j = i - vertex_start;
                vertex_order[j] = i;
            }
        } else {
            for (int i = vertex_start, j = 0; i < vertex_end; i++) {
                j = i - vertex_start;
                vertex_order[j] = (vertex_end - 1) - j;
            }
        }
    }

    /**
     * The strokeLines method creates lines for a LINES shape.
     * 
     * @param mode The shape closing mode.
     */
    private void strokeLines(final int mode) {
        // store index of first vertex
        final int first = lineCount;
        final int stop = vertex_end - 1;

        final int increment;
        if (shape == LINES) {
            increment = 2;
        } else {
            increment = 1;
        }

        // for LINE_STRIP and LINE_LOOP, make this all one path
        if (shape != LINES) {
            add_path();
        }

        for (int i = vertex_start; i < stop; i += increment) {
            // for LINES, make a new path for each segment
            if (shape == LINES) {
                add_path();
            }
            add_line(i, i + 1);
        }

        // for LINE_LOOP, close the loop with a final segment
        // if (shape == LINE_LOOP) {
        if (mode == CLOSE) {
            add_line(stop, lines[first][VERTEX1]);
        }
    }

    /**
     * The strokePoints method creates dots for a POINTS shape.
     */
    private void strokePoints() {
        for (int i = vertex_start; i < vertex_end; i++) {
            add_path(); // total overkill for points
            add_line(i, i);
        }
    }

    /**
     * The strokePolygon method creates lines for a POLYGON shape.
     * 
     * @param mode Indicates if the polygon stroke is to be closed.
     */
    private void strokePolygon(final int mode) {
        // store index of first vertex
        // int first = lineCount;
        final int stop = vertex_end - 1;

        add_path();
        for (int i = vertex_start; i < stop; i++) {
            add_line(i, i + 1);
            // System.out.println("adding line " + i);
        }
        if (mode == CLOSE) {
            // draw the last line connecting back to the first point in
            // poly
            add_line(stop, vertex_start); // lines[first][VERTEX1]);
        }
    }

    /**
     * The strokeQuads method creates lines for a QUADS shape.
     */
    private void strokeQuads() {
        for (int i = vertex_start; i < vertex_end; i += FOUR) {
            add_path();
            // counter = i - vertex_start;
            add_line(i + 0, i + 1);
            add_line(i + 1, i + 2);
            add_line(i + 2, i + THREE);
            add_line(i + THREE, i + 0);
        }
    }

    /**
     * The strokeQuadStrip method creates lines for a QUAD_STRIP shape.
     */
    private void strokeQuadStrip() {
        for (int i = vertex_start; i < vertex_end - THREE; i += 2) {
            add_path();
            add_line(i + 0, i + 2);
            add_line(i + 2, i + THREE);
            add_line(i + THREE, i + 1);
            add_line(i + 1, i + 0);
        }
    }

    /**
     * The strokeTriangleFan method creates lines for a TRIANGLE_FAN shape.
     */
    private void strokeTriangleFan() {
        // this just draws a series of line segments
        // from the center to each exterior point
        for (int i = vertex_start + 1; i < vertex_end; i++) {
            add_path();
            add_line(vertex_start, i);
        }

        // then a single line loop around the outside.
        add_path();
        for (int i = vertex_start + 1; i < vertex_end - 1; i++) {
            add_line(i, i + 1);
        }
        // closing the loop
        add_line(vertex_end - 1, vertex_start + 1);
    }

    /**
     * The strokeTriangles method creates lines for a TRIANGLES shape.
     */
    private void strokeTriangles() {
        for (int i = vertex_start; i < vertex_end - 2; i += THREE) {
            add_path();
            // counter = i - vertex_start;
            add_line(i + 0, i + 1);
            add_line(i + 1, i + 2);
            add_line(i + 2, i + 0);
        }
    }

    /**
     * The strokeTriangleStrip method creates lines for a TRIANGLE_STRIP shape.
     */
    private void strokeTriangleStrip() {
        // first draw all vertices as a line strip
        int stop = vertex_end - 1;

        add_path();
        for (int i = vertex_start; i < stop; i++) {
            // counter = i - vertex_start;
            add_line(i, i + 1);
        }

        // then draw from vertex (n) to (n+2)
        stop = vertex_end - 2;
        for (int i = vertex_start; i < stop; i++) {
            add_path();
            add_line(i, i + 2);
        }
    }

    /**
     * The textureVertex method sets (U, V) coords for the next vertex in the
     * current shape. We've overridden this so that it no longer clamps uv
     * coordinates between 0 and 1. This allows for repeating textures. However,
     * this functionality comes at the price of requiring all texture images
     * have their height and width be in powers of 2. Any other size should
     * appear normally provided the uv coordinates do not exceed the boundaries
     * of 0 and 1, but if they do the results will probably be undesired.
     * 
     * @param uCoord The U coordinate for texturing.
     * @param vCoord The V coordinate for texturing.
     * @see processing.core.PGraphics#textureVertex(float, float)
     */
    @Override
    protected final void textureVertex(final float uCoord, final float vCoord) {
        // Start with the provided uv coordinates.
        float uResult = uCoord;
        float vResult = vCoord;

        // If there's no texture set then complain about it.
        if (textureImage == null) {
            // TODO Consider logging this instead since it may not be a mission
            // critical failure.
            // throw new RuntimeException("need to set an image with texture() "
            // + "before using u and v coordinates");
            throw new UnsupportedOperationException(
                    "need to set an image with texture() "
                            + "before using u and v coordinates");
        }

        // Check to see if Processing is expecting uv coordinates in pixels.
        if (textureMode == IMAGE) {
            // Scale pixel coordinates into normal coordinates.
            uResult /= (float) textureImage.width;
            vResult /= (float) textureImage.height;
        }

        // Store the coordinates for storage with a vertex. The original
        // implementation of this clamped the normal coordinates between 0 and
        // 1.
        textureU = uResult;
        textureV = vResult;
    }

    /**
     * The transformFromScreen method reverts the transformation caused by a
     * previous call to transformFromScreen(), which must be called before this
     * is because this invokes popMatrix() and the matrix stack needs to be
     * managed properly.
     */
    public final void transformFromScreen() {
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPopMatrix();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPopMatrix();
    }

    /**
     * The transformToScreen method sets up a transformation that places
     * coordinates on the z=0 plane directly on the screen plane. This method
     * needs to be followed by a call to transformFromScreen() because this
     * invokes pushMatrix() and the matrix stack needs to be managed properly.
     */
    public final void transformToScreen() {
        gl.glPushMatrix();
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glScalef(2f, 2f, 1f);
        gl.glTranslatef(-HALF, HALF, 0f);
        gl.glScalef(1f / (float) this.width, -1f / (float) this.height, 1f);
    }

    /**
     * The translate performs a translation transformation.
     * 
     * @param deltaX The x-axis translation.
     * @param deltaY The y-axis translation.
     * @see processing.core.PGraphics3D#translate(float, float)
     */
    @Override
    public final void translate(final float deltaX, final float deltaY) {
        super.translate(deltaX, deltaY);
        gl.glTranslatef(deltaX, deltaY, 0);
    }

    /**
     * The translate performs a translation transformation.
     * 
     * @param deltaX The x-axis translation.
     * @param deltaY The y-axis translation.
     * @param deltaZ The z-axis translation.
     * @see processing.core.PGraphics3D#translate(float, float, float)
     */
    @Override
    public final void translate(final float deltaX, final float deltaY,
            final float deltaZ) {
        super.translate(deltaX, deltaY, deltaZ);
        gl.glTranslatef(deltaX, deltaY, deltaZ);

    }

    /**
     * Triangulate the current polygon. <BR>
     * <BR>
     * Simple ear clipping polygon triangulation adapted from code by John W.
     * Ratcliff (jratcliff at verant.com). Presumably <A
     * HREF="http://www.flipcode.org/cgi-bin/fcarticles.cgi?show=63943">this</A>
     * bit of code from the web.
     */
    private void triangulatePolygon() {
        // this clipping algorithm only works in 2D, so in cases where a
        // polygon is drawn perpendicular to the z-axis, the area will be zero,
        // and triangulation will fail. as such, when the area calculates to
        // zero, figure out whether x or y is empty, and calculate based on the
        // two dimensions that actually contain information.
        // http://dev.processing.org/bugs/show_bug.cgi?id=111
        dim1 = MX;
        dim2 = MY;
        // this brings up the nastier point that there may be cases where
        // a polygon is irregular in space and will throw off the
        // clockwise/counterclockwise calculation. for instance, if clockwise
        // relative to x and z, but counter relative to y and z or something
        // like that.. will wait to see if this is in fact a problem before
        // hurting my head on the math.

        // first we check if the polygon goes clockwise or counterclockwise
        float area = calcPolyArea();

        // rather than checking for the perpendicular case first, only do it
        // when the area calculates to zero. checking for perpendicular would be
        // a needless waste of time for the 99% case.
        if (area == 0) {
            // figure out which dimension is the perpendicular axis
            final boolean axisFound = this.findPolyAxis();

            if (!axisFound) {
                return;
            }

            // re-calculate the area, with what should be good values
            area = this.calcPolyArea();
        }

        // don't allow polygons to come back and meet themselves,
        // otherwise it will anger the triangulator
        // http://dev.processing.org/bugs/show_bug.cgi?id=97
        final float[] vfirst = vertices[vertex_start];
        final float[] vlast = vertices[vertex_end - 1];
        if ((Math.abs(vfirst[MX] - vlast[MX]) < EPSILON)
                && (Math.abs(vfirst[MY] - vlast[MY]) < EPSILON)
                && (Math.abs(vfirst[MZ] - vlast[MZ]) < EPSILON)) {
            vertex_end--;
        }

        // then sort the vertices so they are always in a counterclockwise order
        this.sortPolyVerts(area);

        this.makePolyTriangles();

    }

    /**
     * The triangulateQuads method creates triangles for a QUADS shape.
     */
    private void triangulateQuads() {
        final int stop = vertexCount - THREE;
        for (int i = vertex_start; i < stop; i += FOUR) {
            // first triangle
            add_triangle(i, i + 1, i + 2);
            // second triangle
            add_triangle(i, i + 2, i + THREE);
        }
    }

    /**
     * The triangulateQuadStrip method creates triangles for a QUAD_STRIP shape.
     */
    private void triangulateQuadStrip() {
        final int stop = vertexCount - THREE;
        for (int i = vertex_start; i < stop; i += 2) {
            // first triangle
            add_triangle(i + 0, i + 2, i + 1);
            // second triangle
            add_triangle(i + 2, i + THREE, i + 1);
        }
    }

    /**
     * The triangulateTriangleFan method creates triangles for a TRIANGLE_FAN
     * shape.
     */
    private void triangulateTriangleFan() {
        final int stop = vertex_end - 1;
        for (int i = vertex_start + 1; i < stop; i++) {
            add_triangle(vertex_start, i, i + 1);
        }
    }

    /**
     * The triangulateTriangles method creates triangles for a TRIANGLES shape.
     */
    private void triangulateTriangles() {
        final int stop = vertex_end - 2;

        final int increment;
        if (shape == TRIANGLES) {
            increment = THREE;
        } else {
            increment = 1;
        }

        for (int i = vertex_start; i < stop; i += increment) {
            // have to switch between clockwise/counter-clockwise
            // otherwise the feller is backwards and renderer won't draw
            if ((i % 2) == 0) {
                add_triangle(i, i + 2, i + 1);
            } else {
                add_triangle(i, i + 1, i + 2);
            }
        }
    }
}
