/**
 * DDSImageFileType.java.
 */
package dds.control;

import java.util.Locale;

import javax.imageio.ImageIO;

/**
 * The DDSImageFileType enum class is responsible for providing some very handy
 * functionality pertaining to different types of image formats. This class is
 * mainly used by DDSApplet during the execution of it's loadImage(String)
 * method. The enumerations of this class contain mappings of filename
 * extensions to image file types as well as indicating if this format supports
 * transparency. The file extensions defined in the String[] should all be in
 * lowercase so that we can avoid an extra case conversion during the extension
 * check.
 * 
 * @author Jacob Bolton
 */
public enum DDSImageFileType {
    /**
     * The TGA enumeration represents files that are in the TARGA format.
     */
    TGA(false, new String[] { "tga", "tpic" }),

    /**
     * The TIFF enumeration represents files that are in the Tagged Image File
     * Format.
     */
    TIFF(false, new String[] { "tif", "tiff" }),

    /**
     * The JPEG enumeration represents files that are in the Joint Photographic
     * Experts Group format.
     */
    JPEG(false, new String[] { "jpg", "jpeg", "jpe", "jfif", "jfi" }),

    /**
     * The GIF enumeration represents files that are in the Graphics Interchange
     * Format.
     */
    GIF(true, new String[] { "gif" }),

    /**
     * The PNG enumeration represents files that are in the Portable Network
     * Graphics format.
     */
    PNG(true, new String[] { "png" }),

    /**
     * The JAVA_IMAGEIO enumeration represents files that are in various formats
     * that can be read via Java's ImageIO classes.
     */
    JAVA_IMAGEIO(false, ImageIO.getReaderFormatNames()),

    /**
     * The UNSUPPORTED enumeration represents files that were unable to be
     * recognized.
     */
    UNSUPPORTED(false, null);

    /**
     * The fileExtensions.
     */
    private String[] fileExtensions;

    /**
     * The transparency field describes if this ImageFileType supports
     * transparency or not.
     */
    private boolean transparent;

    /**
     * Constructor for DDSImageFileType.
     * 
     * @param hasTransparency True if the format supports transparency.
     * @param fileNameEndings A String[] of filename extensions that are common
     *            to this format. The extensions should NOT include the . that
     *            usually precedes the extension.
     */
    private DDSImageFileType(final boolean hasTransparency,
            final String[] fileNameEndings) {
        // Set the transparency indicator and list of known extensions.
        this.transparent = hasTransparency;
        this.fileExtensions = fileNameEndings;
    }

    /**
     * The parseImageFileType takes a filename in as a String and attempts to
     * match it to one of the defined image format types defined in this
     * enumeration class.
     * 
     * @param fileName The name of the file to attempt format matching on.
     * @return The DDSImageFileType corresponding to the filename provided, even
     *         if the resulting type if UNDEFINED.
     */
    public static DDSImageFileType parseImageFileType(final String fileName) {
        // Set the result to UNDEFINED in case we can't determine the type.
        DDSImageFileType result = UNSUPPORTED;

        // Loop through all of the defined types.
        for (DDSImageFileType type : DDSImageFileType.values()) {
            // Skip over JAVA_IMAGEIO since that's a last resort. Then check to
            // see if the filename is one recognized as one of the current type.
            if ((!type.equals(JAVA_IMAGEIO)) && (type.isThisType(fileName))) {
                // Set the result to this type and break out of the loop.
                result = type;
                break;
            }
        }

        // Check to see if any type was detected, and if not then try to match
        // with JAVA_IMAGEIO.
        if ((result.equals(UNSUPPORTED)) 
                && (JAVA_IMAGEIO.isThisType(fileName))) {
            // If it IS JAVA_IMAGEIO then set the return type appropriately.
            result = JAVA_IMAGEIO;
        }

        // TODO Consider logging the result being UNDEFINED at this point.

        // Return the resulting type even if it hasn't changed from UNDEFINED.
        return result;
    }

    /**
     * The isThisType method checks a given filename String to see if it has an
     * extension matching one defined in this enumeration's array of recognized
     * file types.
     * 
     * @param fileName The filename to check.
     * @return True if the filename's extension is in the list.
     */
    public boolean isThisType(final String fileName) {
        // Get the lowercase version of the filename in case the extension was
        // given in UPPERCASE.
        final String lower = fileName.toLowerCase(Locale.getDefault());

        // Set the result to false in case we are unable to detect a match.
        boolean result = false;

        // Check to see if we even have a list to chack against. The UNSUPPORTED
        // type does not and this alleviates an extra check during the iteration
        // of defined enumerations in the parseImageFileType method.
        if (this.fileExtensions != null) {
            // Iterate through the defined file extensions for this type.
            for (int i = 0; i < this.fileExtensions.length; i++) {
                // Check to see if the filename ends with the current extension.
                if (lower.endsWith("." + this.fileExtensions[i])) {
                    // If it does, set our result to true and break from this
                    // loop.
                    result = true;
                    break;
                }
            }
        }

        // Return the result.
        return result;
    }

    /**
     * Getter for transparent.
     * 
     * @return the transparent
     */
    public final boolean isTransparent() {
        return transparent;
    }
}
