/**
 * DDSModuleLoader.java.
 */
package dds.control;

import processing.core.PImage;
import dds.control.user.AbstractModule;
import dds.control.user.ModuleType;

/**
 * The DDSModuleLoader class is a helper child to the DDSApplet that is
 * primarily responsible for managing the loading and switching of DDS modules.
 * It will also provide means for making inter-module communication easier,
 * especially for menus that are used. This will be accomplished by keeping a
 * reference to the previously used module. This class will also be responsible
 * for providing a visually interesting and smooth transition between diffrent
 * modules when they are being changed.
 * 
 * @author Jacob Bolton
 */
public class DDSModuleLoader extends AbstractDDSAppletChild 
        implements Runnable {

    /**
     * The NO_MODULE constant is an easily readable way to set the cureent and
     * previous modules to null.
     */
    private static final AbstractModule NO_MODULE = null;

    /**
     * The currentModule field stores the currently active module for the
     * application.
     */
    private transient AbstractModule currentModule;

    // /**
    // * The lastModule field stores the previously used module for the
    // * application.
    // */
    // private transient AbstractModule lastModule;

    /**
     * The newModuleType field is used to store the ModuleType for the next
     * module that is to be instantiated.
     */
    private transient ModuleType newModuleType;

    /**
     * The loadingImage field is used to store a loaded PImage of the graphic
     * displayed during module transitions.
     */
    private transient PImage loadingImage;

    // /**
    // * The MAX_FADEOUT constant defines the maximum alpha value for fading.
    // */
    // private static final int MAX_FADEOUT = 255;
    //
    // /**
    // * The MIN_FADEOUT constant defines the minimum alpha value for fading.
    // */
    // private static final int MIN_FADEOUT = 16;

    /**
     * The FILL_COLOR is the fill color to use for drawing the background of the
     * loading screen.
     */
    private static final int FILL_COLOR = 0x10000000;

    // /**
    // * The fadeout field is used to keep track of how much alpha to use for
    // the
    // * renderng of a frame of the loading screen bwteen transitions.
    // */
    // private transient int fadeout;

    /**
     * The busy boolean field is used to help determine is this ModuleLoader is
     * busy working on something. It is used by the DDSApplet to determine if it
     * needs to render the loading screen instead of the current module, which
     * may be null at this time since the busy flag should only be set when this
     * ModuleLoader is in the process of switching modules.
     */
    private transient boolean busy;

    /**
     * Constructor for DDSModuleLoader.
     * 
     * @param applet The parent DDSApplet.
     */
    public DDSModuleLoader(final DDSApplet applet) {
        super(applet);
        // Set the busy flag since we stil don't have a module to be active.
        this.busy = true;

        // Set the current fadeout for rendering the loading screen.
        // this.fadeout = MIN_FADEOUT;
    }

    /**
     * The destruct method is responsible for breaking this object apart when
     * it's done being used.
     * 
     * @see dds.control.AbstractDDSAppletChild#destruct()
     */
    @Override
    public final void destruct() {
        // TODO Auto-generated method stub

    }

    /**
     * The drawLoadingScreen method takes care of drawing the loading screen
     * when this ModuleLoader is busy switching or loading modules.
     */
    public final void drawLoadingScreen() {
        this.getParent().perspective();
        this.getParent().camera();
        // Set the fill color for the background of the loading screen.
//        this.getParent().background(FILL_COLOR);
        this.getParent().fill(FILL_COLOR);

        // Draw the background of the loading screen.
        this.getParent().rect(0, 0, this.getParent().width,
                this.getParent().height);

        // Draw the loading graphic in the middle of the DDSApplet screen.
        this.getParent().image(this.loadingImage,
                (this.getParent().width - this.loadingImage.width) / 2,
                (this.getParent().height - this.loadingImage.height) / 2);
    }

    /**
     * The isBusy method returns if this is busy doing something or is unable to
     * provide a valid module for rendering a frame.
     * 
     * @return True if there is no useable module for rendering.
     */
    public final boolean isBusy() {
        // Return true if the busy flag is set OR the current module is set as
        // NO_MODULE.
        return this.busy || (this.currentModule == null);
    }

    /**
     * Getter for currentModule.
     * 
     * @return the currentModule
     */
    public final AbstractModule getCurrentModule() {
        return currentModule;
    }

    /**
     * The setup method is used to perform any necessary operations that this
     * objects needs to execute in order to ready itself for proper use by the
     * DDSApplet parent.
     * 
     * @see dds.control.AbstractDDSAppletChild#setup()
     */
    @Override
    public final void setup() {
        this.loadingImage = this.getParent().loadImage("loading.png");
        this.drawLoadingScreen();
        // Dan: I hooked the Administration option on the main menu to start the
        // trap editor.
        this.loadModule(ModuleType.MAIN_MENU);
    }

    /**
     * The loadModule method is what takes care of trying to load a module. It
     * is capable of taking in a variety of arguments to determine which module
     * to load. It is able to use a String that is the full name of the module
     * class to instantiate, e.g. dds.control.user.modules.SimulationModule.
     * This is accomplished by resolving the String to a ModuleType, which is
     * another type of argument that can be used to instantiate a new module.
     * The method can also take in an already instantiated module. The busy flag
     * is set to true during the operation of this method, which means the
     * loading screen is displayed during the transition. If the loading
     * operation needs to instantiate a new module, the instantiation proces is
     * handled in a separate thread, which frees up the display to produce nicer
     * transitional effects.
     * 
     * @param moduleArg The argument used to load a module. Currently it can be
     *            either a String naming the class to use, the ModuleType
     *            representing the module, or an instantiated module.
     */
    public final void loadModule(final Object moduleArg) {
        // Check for null arguments to prevent unnecessary work.
        if (moduleArg == null) {
            // TODO Consider about or complaining/logging this case since it
            // shouldn't occur.
            return;
        }
        // Handle the RegisteredMethods fields of DDSApplet.
        // TODO Have this actually retain these methods for later restoration if
        // necesary.
        this.getParent().resetRegisteredMethods();

        // Set the busy flag to true so that the work done here doesn't cause
        // the DDSApplet to try to draw null or invalid modules.
        this.busy = true;

        // Preserve the current module.
        final AbstractModule lastModule = this.currentModule;

        // Set the current module to null.
        this.currentModule = NO_MODULE;

        // Check the type of argument supplie to this method to determine what
        // action to take.
        if (moduleArg instanceof AbstractModule) {
            // The argument is an AbstractModule instance, so just switch this
            // into place.
            this.currentModule = (AbstractModule) moduleArg;
            // Reset the registered methods.
            // TODO get this to save/restore registered methods so that modules
            // making use of these can be switched back in without having
            // problems.
            if (this.currentModule == null) {
                this.currentModule = lastModule;
            } else {
                this.getParent().resetRegisteredMethods();
            }
        } else {
            // Try resolving the argument to a ModuleType
            final ModuleType type = ModuleType.parseType(moduleArg);

            // Check to see if the type resolved is suppoerted or not.
            if (type == ModuleType.UNSUPPORTED) {
                // Complain about an unsuppoerted module.
                System.err.println("Unsupported module: " + moduleArg);
            } else {
                // Set the instantiation type for the next module to use.
                this.newModuleType = type;

                // Start the loading thread for instatiation.
                new Thread(this).start();
                // this.run();
            }
        }

        // We've finished up the work here so unset he busy flag.
        this.busy = false;
    }

    /**
     * The run method is the threaded way in which newly instantiated objects
     * are instantiated and then put into place.
     * 
     * @see java.lang.Runnable#run()
     */
    public final void run() {
        // TODO Consider getting this to also work for loading already
        // instantiated modules.
        System.gc();

        // Use the specified ModuleType to create a new module instance.
        final AbstractModule module = this.newModuleType.constructNew(this
                .getParent());

        // Set the module up for use.
        module.setup();

        // Set the current module to the one just created.
        this.currentModule = module;
        System.gc();
    }
}
