/**
 * DDSOBJModel.java.
 */
package dds.control;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import saito.objloader.Debug;

/**
 * The DDSOBJModel class is a wrapper for the saito.objloader.OBJModel class,
 * which is what we use to load and display 3D models created from and external
 * graphics editor (such as Blender) and saved in the Alias Wavefront (.obj)
 * format. It has been noticed that this class performs poorly when the model
 * file relies on polygon faces; it tends not to draw certain faces. Therefore,
 * it is crucial to have the model files use triangle faces. Our extension
 * allows for proper access to files stored in the data repository via the
 * dds.control.ResourceLoader. This is accomplished by overriding the
 * getBufferedReader(String) method to pull from the models stored in the
 * repository rather than from the Processing data folder. We've also privatized
 * the constructor so that instances of the same model can use the same instance
 * from a filename-based cache. This cache helps to improve on our memory usage.
 * 
 * @author Jacob Bolton
 */
public final class DDSOBJModel extends saito.objloader.OBJModel {

    /**
     * The OBJ_CACHE is used to store a single instance of an DDSOBJModel for
     * each file.
     */
    private static final Map<String, DDSOBJModel> OBJ_CACHE;

    static {
        // Create the cache for DDSOBJModel instances.
        OBJ_CACHE = new HashMap<String, DDSOBJModel>();
    }

    /**
     * The getOBJModel method is used to get instances of DDSOBJModel. By using
     * this method instead of a public constructor, we prevent the creation of
     * redundant instances.
     * 
     * @param filename The filename from which the .OBJ data is to be loaded
     *            from.
     * @return A reference to an DDSOBJModel.
     */
    public static DDSOBJModel getOBJModel(final String filename) {
        // Try to get an already loaded instance from the cache.
        DDSOBJModel result = OBJ_CACHE.get(filename);

        // Check to see if we got anything.
        if (result == null) {
            // Since we didn't, create a new instance and load the data.
            result = new DDSOBJModel();
            result.load(filename);

            // Place new instance into the cache for reuse.
            OBJ_CACHE.put(filename, result);
        }

        // Return the result.
        return result;
    }

    /**
     * The loaded field is used to determine if an instance has already loaded
     * data from a file or not.
     */
    private transient boolean loaded;

    /**
     * The listCompiled field indicates if this OBJModel has had its GL commands
     * compiled into a list.
     */
    private transient boolean listCompiled;

    /**
     * Constructor for DDSOBJModel. Sets the default rendermode to TRIANGLES so
     * all models must be fully triangulated. There were problems with missing
     * faces when redering in POLYGONS mode so TRIANGLES are a must.
     */
    private DDSOBJModel() {
        // Use the superclass constructor, which attachs this DDSOBJModel to a
        // DDSApplet. However, this behavior forces this object to be attached
        // to the view AND model causing poor abstraction.
        // TODO Perhaps invent a way to have this class be more independent from
        // the view. Unfortunately this may be difficult since it both models an
        // object and also renders it.
        super(DDSApplet.getApplet());

        // Set the rendering mode to use triangle faces because polygons have
        // problems rendering certain faces. Make sure ALL model files use ONLY
        // triagle faces to ensure prope rendering.
        this.drawMode(DDSApplet.TRIANGLES);

        // Try to disable the DDSOBJModel print outs.
        try {
            Field debugField = saito.objloader.OBJModel.class
                    .getDeclaredField("debug");
            debugField.setAccessible(true);
            Debug debugger = (Debug) debugField.get(this);
            debugger.enabled = false;
        } catch (SecurityException e) {
            // TODO Add logger
            System.out.println(e);
        } catch (NoSuchFieldException e) {
            // TODO Add logger
            System.out.println(e);
        } catch (IllegalArgumentException e) {
            // TODO Add logger
            System.out.println(e);
        } catch (IllegalAccessException e) {
            // TODO Add logger
            System.out.println(e);
        }

        // Set the loaded field to false since we haven't loaded any data yet.
        this.loaded = false;

        // Set the listCompiled to false since the GL commands haven't been
        // compiled yet.
        this.listCompiled = false;
    }

    /**
     * The getBufferedReader method returns a BufferedReader for the specified
     * file. This was overridden because we needed to have all .obj models come
     * from the repository. This implementation takes in a file name and then
     * goes to the repository to get the appropriate data from the models
     * directory.
     * 
     * @param fileName The String name of the file to be loaded (must exist in
     *            /data/models)
     * @return A BufferedReader for that resource or null if resource access
     *         failure.
     * @see saito.objloader.OBJModel#getBufferedReader(java.lang.String)
     */
    @Override
    public BufferedReader getBufferedReader(final String fileName) {
        // Create a URL for receiving the resource locatio from the repository.
        // Initialize it to null so that it has been initialized in case the
        // repository returns no results.
        URL file = null;

        // Try to get a URL to the requested resource from the Resource
        // loader.
        file = ResourceManager.getLOADER()
                .getFile(fileName, ResourceType.MODEL);

        // Create a BufferedReader for receiving the result of opening the .obj
        // file.
        BufferedReader fileReader = null;
        try {
            // Try to get an InputStream from the URL and wrap it in a
            // BufferedReader
            fileReader = new BufferedReader(new InputStreamReader(file
                    .openStream()));
        } catch (IOException e) {
            // TODO Add logger
            System.err.println(e);
        }

        // Return the resulting StreamReader.
        return fileReader;
    }

    /**
     * The load method is responsible for loading the data in from the .obj
     * file. This extension of it prevents this instance from attempting to load
     * data from more than one file, i.e. more than once.
     * 
     * @param filename The filename from which the .OBJ data is to be loaded
     *            from.
     * @see saito.objloader.OBJModel#load(java.lang.String)
     */
    @Override
    public void load(final String filename) {
        // Check to see if we've loaded data already.
        if (this.loaded) {
            // We have, and that makes this call illegal.
            new UnsupportedOperationException(
                    "Only one load operation is permitted per instance.");
        } else {
            // We haven't, so set the loaded flag to true and load the specified
            // file.
            super.load(filename);
            this.loaded = true;
        }
    }

    /**
     * The draw method takes care of the drawing of this DDSOBJModel. It is now
     * using a GL display list to try to accelerate the drawing process.
     * 
     * @see saito.objloader.OBJModel#draw()
     */
    @Override
    public void draw() {
        // Check to see if a display list has already been compiled.
        if (this.listCompiled) {
            // It has so just use it to draw this object.
            DDSApplet.getApplet().drawList(this.hashCode());
        } else {
            // It hasn't so begin a new one.
            DDSApplet.getApplet().beginList(this.hashCode());

            // Draw the model.
            super.drawModel();

            // Finish the list definition.
            DDSApplet.getApplet().endList();

            // Indicate that this object's list has been compiled.
            this.listCompiled = true;
        }
    }

    /**
     * The clearCache method clears the cache of OBJ models.
     */
    public static void clearCache() {
        OBJ_CACHE.clear();
    }
}
