/**
 * This java class is distributed under the BSD license. Copyright 2005 Lilian
 * Chamontin. contact lilian.chamontin at f r e e . f r
 */

/**
 * Portions Copyright (c) 2003 Sun Microsystems, Inc. All Rights Reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: -
 * Redistribution of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer. - Redistribution in binary
 * form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided
 * with the distribution. Neither the name of Sun Microsystems, Inc. or the
 * names of contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission. This software is
 * provided "AS IS," without a warranty of any kind. ALL EXPRESS OR IMPLIED
 * CONDITIONS, REPRESENTATIONS AND WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE
 * HEREBY EXCLUDED. SUN MICROSYSTEMS, INC. ("SUN") AND ITS LICENSORS SHALL NOT
 * BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING,
 * MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL
 * SUN OR ITS LICENSORS BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR
 * DIRECT, INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES,
 * HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING OUT OF THE
 * USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF SUN HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGES. You acknowledge that this software is not
 * designed or intended for use in the design, construction, operation or
 * maintenance of any nuclear facility.
 */
package dds.control;

import java.util.Locale;

/**
 * The NativeLibInfo enumeration class is responsible for providing metadata for
 * the native libraries that need to be loaded in order to use OpenGL on a given
 * platform.
 * 
 * @author Jacob Bolton
 */
public enum NativeLibInfo {
    /**
     * The WIN_X86 defines the properties used to recognize Windows OS running
     * on an x86 platform.
     */
    WIN_X86("win", "x86", "jogl-natives-windows-i586.jar", "", ".dll"),

    /**
     * The MAC_PPC defines the properties used to recognize MacOS running on a
     * PPC platform.
     */
    MAC_PPC("mac", "ppc", "jogl-natives-macosx-ppc.jar", "lib", ".jnilib"),

    /**
     * The MAC_I386 defines the properties used to recognize MacOS running on an
     * i386 platform.
     */
    MAC_I386("mac", "i386", "jogl-natives-macosx-universal.jar", "lib",
            ".jnilib"),

    /**
     * The LINUX_I386 defines the properties used to recognize Linux OS running
     * on an i386 platform.
     */
    LINUX_I386("linux", "i386", "jogl-natives-linux-i586.jar", "lib", ".so"),

    /**
     * The LINUX_X86 defines the properties used to recognize Linux OS running
     * on an x86 platform.
     */
    LINUX_X86("linux", "x86", "jogl-natives-linux-i586.jar", "lib", ".so"),

    /**
     * The SUN_SPARC defines the properties used to recognize Sun/Solaris OS
     * running on a SPARC platform.
     */
    SUN_SPARC("sunos", "sparc", "jogl-natives-solaris-sparc.jar", "lib", ".so"),

    /**
     * The SUN_X86 defines the properties used to recognize Sun/Solaris OS
     * running on an x86 platform.
     */
    SUN_X86("sunos", "x86", "jogl-natives-solaris-i586.jar", "lib", ".so"),

    /**
     * The UNSUPPORTED.
     */
    UNSUPPORTED(null, null, null, null, null);

    /**
     * The osName field stores this enumeration's expected operating system name
     * String.
     */
    private String osName;

    /**
     * The osArch field stores this enumeration's expected system architecture
     * name String.
     */
    private String osArch;

    /**
     * The nativeJar field stores this enumeration's native library Jar file
     * name.
     */
    private String nativeJar;

    /**
     * The nativePrefix field stores this enumeration's native library file
     * prefix.
     */
    private String nativePrefix;

    /**
     * The nativeSuffix field stores this enumeration's native library file
     * suffix.
     */
    private String nativeSuffix;

    /**
     * Constructor for NativeLibInfo.
     * 
     * @param osNameStr The expected operating system name String.
     * @param osArchStr The expected system architecture name String.
     * @param nativeJarFile The native library Jar file name.
     * @param prefixStr The native library file prefix.
     * @param suffixStr The native library file suffix.
     */
    private NativeLibInfo(final String osNameStr, final String osArchStr,
            final String nativeJarFile, final String prefixStr,
            final String suffixStr) {
        this.osName = osNameStr;
        this.osArch = osArchStr;
        this.nativeJar = nativeJarFile;
        this.nativePrefix = prefixStr;
        this.nativeSuffix = suffixStr;
    }

    /**
     * The matchesOSAndArch detects if this enumeration's information maches the
     * current system's configuration.
     * 
     * @return True if the OS name and architecture match.
     */
    public boolean matchesOSAndArch() {
        // Get the system's OS and architecture.
        final String name = System.getProperty("os.name").toLowerCase();
        final String arch = System.getProperty("os.arch").toLowerCase();
        
        return (name.startsWith(this.osName))
                && ((this.osArch == null) || (arch.startsWith(this.osArch)));
    }

    /**
     * The matchesNativeLib checks a library file name to make sure it has the
     * expected ending.
     * 
     * @param nativeLibraryName The library filename to check.
     * @return True if the suffix matches.
     */
    public boolean matchesNativeLib(final String nativeLibraryName) {
        return nativeLibraryName.toLowerCase(Locale.getDefault()).endsWith(
                nativeSuffix);
    }

    /**
     * The getNativeJarName is an accessor for th nativeJar field.
     * 
     * @return The nativeJar field.
     */
    public String getNativeJarName() {
        return nativeJar;
    }

    /**
     * The getNativeLibName method returns the provided name pre and postpended
     * with the prefix and suffix.
     * 
     * @param baseName The base name to modify.
     * @return The modified base name.
     */
    public String getNativeLibName(final String baseName) {
        return nativePrefix + baseName + nativeSuffix;
    }

    /**
     * The isMacOS method indicates if this enumertion is for a Mac platform.
     * 
     * @return True if the OS name is mac.
     */
    public boolean isMacOS() {
        return "mac".equals(osName);
    }

    /**
     * The detectPlatform method tries to resolve the current system's OS and
     * architecture againt on of the enumerations in this class. It will return
     * UNSUPPORTED if there's nothing else to attach to.
     * 
     * @return The enumeration for this system configuration.
     */
    public static NativeLibInfo detectPlatform() {
        NativeLibInfo result = UNSUPPORTED;

        for (NativeLibInfo platform : NativeLibInfo.values()) {
            if ((platform != UNSUPPORTED) && platform.matchesOSAndArch()) {
                result = platform;
                break;
            }
        }
        
        return result;
    }

    /**
     * The isSupported method returnes true when this enumeration is NOT
     * UNSUPPORTED.
     * 
     * @return True when this enumeration is NOT UNSUPPORTED.
     */
    public boolean isSupported() {
        return !UNSUPPORTED.equals(this);
    }
}
