package dds.control;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;

/**
 * The ResourceLoader class is responsible for locating and providing means of
 * access to any and all file resources that might exist either within the
 * application bundle or in the data repository. The data repository consists of
 * three distinct components: local, jar-level, and remote. The local repository
 * should be located on the machine running the the application in the user's
 * home directory as specified by the HOME system environment variable. This is
 * where user-generated files and local file caching can occur because it should
 * be the first place to be searched for resources. The jar-level repository is
 * really just a collection of files that come bundled in the application's jar
 * file. These are all static files containing all of the basics needed to run a
 * vanilla distribution. The remote repository currently consists of reading
 * files from the website the serves the applet page. Eventually it should
 * support the ability to upload files either through webdav, ftp, or some other
 * file transfer protocol. This functionality would be for administrators who
 * want to customize the delivery of the applet. When being used to find files
 * the levels of the repository are searched in the following order: local,
 * jar-level, then remote. This order is intended to reduce latency from
 * accessing distant resources. Once a resouce is located it is passed along as
 * a URL since they seem to be the most convenient way to pass resource data
 * without causing problems related to the Java sandbox. These problems are
 * related to file access permissions. We tried to address these problems by
 * signing our applet's jar, but this may not be good enough since our
 * certificate is not recognized as trusted by default. This class is
 * implemented as a singleon because it needs to be easy to access from many
 * aspects of the application, and there is no purpose in having multiple
 * instances since they would all be the same with respect to internal data and
 * functionality.
 * 
 * @author Jacob Bolton
 * @author krystian
 * @author Mike
 * @author Dan
 */
public final class ResourceManager {

    /**
     * The NULL constant is being used to get rid of checkstyle errors.
     */
    private static final URL NULL = null;

    /**
     * The loader is storage field for the singleton instance of this class. It
     * is loaded as part of the static intialization of this class and is final
     * to prevent problems with unsynchronized access when tryng to first create
     * and access the singleton instance.
     */
    private static final ResourceManager LOADER = new ResourceManager();

    /**
     * The getLoader method provides the singleton instance of this class.
     * 
     * @return The singleton loader.
     */
    public static ResourceManager getLOADER() {
        // Check to see if the remote repository URL has been set. It should be
        // noted that because this class is a singleton the remote URL should be
        // stored in the instance rather than the class.
        if (LOADER.getRepository() == null) {
            // The remote repository has yet to be detected and set so detect
            // and set it.
            LOADER.inferRepository();
        }

        // Return the singleton instance of this class.
        return LOADER;
    }

    /**
     * The remote repository location as a URL.
     */
    private URL repository;

    /**
     * Constructor for ResourceLoader.
     */
    private ResourceManager() {
        super();
    }

    /**
     * The getAvailableResources method searches both the entire repository for
     * all files of a specific type and then relays this information. It should
     * search the local repository, then the jar and finally the remote
     * repository for the resources to be listed. If there are duplicate
     * resources in different repositories then the first one detected should be
     * the result. This means the the closest version of a file will be the one
     * used. This design is to ensure as little latency as possible when
     * retrieving data.
     * 
     * @param type is a ResourceType representing the directory to search for
     *            resources.
     * @return resources is a COllection of available file names.
     */
    public Collection<String> getAvailableResources(final ResourceType type) {

        // the availableResources in the remote directory
        final Collection<String> resources = new ArrayList<String>();

        // the path to the directory that needs to be searched
        final String resourcePath = this.repository.toString()
                + type.getDirectory();

        try {

            // the URL of the resources...
            final URL resourceURL = new URL(resourcePath);

            // a File representing the directory of the resources...
            final File file = new File(resourceURL.getFile());

            // if the file is a directory...
            if (file.isDirectory()) {

                // get the files in the directory...
                final File[] files = file.listFiles();

                // and add their file names to the Collection.
                for (int i = 0; i < files.length; ++i) {
                    final String fileName = files[i].getName();
                    resources.add(fileName);
                } // end for

            } // end if

        } catch (MalformedURLException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } // end try-catch

        // return the Collection of availableResources
        return resources;

    } // end method getAvailableResources()

    /**
     * The getFile method searches both the entire repository for the specified
     * file of the specific type and then relays this information. It should
     * search the local repository, then the jar and finally the remote
     * repository for the resources to be listed. If there are duplicate
     * resources in different repositories then the first one detected should be
     * the result. This means the the closest version of a file will be the one
     * used. This design is to ensure as little latency as possible when
     * retrieving data.
     * 
     * @param fileName The location of the file to retrieve.
     * @param resourceType The type of resource.
     * @return file is the File in the remote directory.
     */
    public URL getFile(final String fileName, final ResourceType resourceType) {

        // the path to the file
        final String filePath = resourceType.getDirectory() + fileName;

        // create a URL and try to get it from a local directory...
        URL file = null;
        if (DDSApplet.getApplet().isStandAlone()) {
            file = this.getLocalFile(filePath);
        }

        // if the file is still null...
        if (file == null) {
            // Try to get the resource from the jar-level.
            file = this.getJarredFile(filePath);
        } // end if

        // if the file is still null...
        if (file == null) {
            // Try to resolve it from the remote repository.
            file = this.getRemoteFile(filePath);
        } // end if

        // if the file is still null...
        if (file == null) {
            // inform the user that the file was not found and shut down.
            System.err.println("Could not find: " + filePath
                    + "\nShutting down...\n");
            System.exit(1);
        } // end if

        // Return the resulting resource URL.
        return file;

    } // end method getFile()

    /**
     * The getLocalFile method attempts to find the needed file in a local
     * directory in the user's file system.
     * 
     * @param fileName is a String representing the needed file.
     * @return toReturn is a URL to the file on the local system.
     */
    private URL getLocalFile(final String fileName) {

        // the URL to be returned
        URL toReturn = null;

        // the path of the file
        // TODO Refactor this into a constant.
        final String filePath = System.getProperty("user.home") + "/.dds/data/"
                + fileName;

        // first, create a File from the filePath...
        final File localFile = new File(filePath);
        // if that file exists...
        if (localFile.exists()) {
            try {
                // create a URL from the file.
                toReturn = new URL("file:" + filePath);
            } catch (MalformedURLException e) {
                System.err.println(e.getMessage());
                System.err.println(e.getStackTrace());
            } // end try-catch
        } // end if

        // if the file doesn't exist, the URL returned will be null.
        return toReturn;

    } // end method getLocalFile()

    /**
     * The getJarredFile method tries to resolve the given file against the
     * contents of the jar file. This is done by creating a simple path for the
     * resource and using Java's getResource method to resolve it on the
     * classpath.
     * 
     * @param fileName The name of the file to resolve.
     * @return The corresponding File object.
     */
    private URL getJarredFile(final String fileName) {

        // Create a path that will be able to check within the jar for the
        // resource.
        final String filePath = "/data/" + fileName;

        // Try to resolve the resource fom the classpath and return the result.
        return this.getClass().getResource(filePath);

    } // end method getJarredFile()

    /**
     * The getRemoteFile method tries to resolve the given file against the
     * contents of the remote repository. At the moment this merely creates a
     * URL to the resource but does NOT validate it.
     * 
     * @param fileName The name of the file to resolve.
     * @return The corresponding File object.
     */
    private URL getRemoteFile(final String fileName) {

        // Create a path that will be able to check the remote repository for
        // the resource.
        final String filePath = this.repository.toString() + fileName;

        // Storage for the reulting URL for the resource.
        URL resultURL = null;

        try {
            // create a URL from the filePath...
            resultURL = new URL(filePath);
            // open a connection and try to read from the URL. This is done to
            // see if the new URL is valid. If there is a problem with the URL
            // a FileNotFoundException will be thrown by this method call. If
            // that happens, then resultURL should be set to null before it is
            // returned to the caller.
            resultURL.openStream().read();
        } catch (MalformedURLException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } catch (IOException e) {
            // There was a problem with the newly created URL, so the
            // resultURL should be null.
            resultURL = NULL;
        }

        // Return the resulting URL even if it's null.
        return resultURL;

    } // end method getRemoteFile()

    /**
     * The writeFile method is used to write a given file to the user's home
     * directory.
     * 
     * @param outputFile is the File object to be saved.
     * @return writeSuccess is a boolean which indicates if the File object was
     *         written successfully or not.
     */
    public boolean writeFile(final File outputFile) {

        // writeSuccess should be false initially.
        boolean writeSuccess = false;

        // the filePath includes the path to the user's home directory
        // and the name of the file.
        final String filePath = System.getProperty("user.home") + "/"
                + outputFile.getName();

        try {

            // readers and writers for the File object and the file being
            // written to the local directory
            final BufferedWriter out = new BufferedWriter(new FileWriter(
                    filePath));
            final BufferedReader fromFile = new BufferedReader(new FileReader(
                    outputFile));

            // write the output file one character at a time.
            for (int i = 0; i < outputFile.length(); ++i) {
                out.write(fromFile.read());
            } // end for

            // close the reader and writer
            out.close();
            fromFile.close();
            // and finally, change writeSuccess to true.
            writeSuccess = true;

        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } // end try-catch

        return writeSuccess;

    } // end method writeFile()

    /**
     * Getter for repository.
     * 
     * @return the repository
     */
    public URL getRepository() {
        return repository;
    } // end method getRepository()

    /**
     * The inferRepository method tries to resolve the location of the data
     * repository (a.k.a. the data folder). The current funtionality tries to
     * get the parameter from a repository parameter set on the webpage of this
     * applet. If that fails it tries to get the path of the current working
     * directory/classpath?, and if that all fails it reverts to a data folder
     * off of the webpage's document base? However, the way this method really
     * should work is to scan for the local, jar-level and remote repositories
     * and set URL fieldsfor them all. This way we will always have references
     * to the different locations where files should be stored.
     */
    private void inferRepository() {
        // Start with a null location for the repository.
        String repoStr = null;

        // Try to assert the repository location from parameters passed from the
        // applet's website. This is useful for users who wish to have their
        // repository exist on a separate server of site address from the one
        // that serves the applet.
        repoStr = DDSApplet.getApplet().getParameter("repository");

        // Check to see if the parameters were there.
        if (repoStr == null) {
            // Since there were no parameters set the choose the applet's base
            // location as assumed from the page it is being loaded from?
            repoStr = DDSApplet.getApplet().getCodeBase().toString() + "data/";
        }

        // Set the repository location field.
        LOADER.setRepository(repoStr);

    }

    /**
     * Setter for repository.
     * 
     * @param repo the repository to set
     */
    public void setRepository(final String repo) {
        // Storage for the resulting URL
        URL repoURL = null;

        try {
            // Try to create a URL from the String provided.
            repoURL = new URL(repo);
        } catch (MalformedURLException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } // end try-catch

        // Check if a valid URL was able to be created from the input.
        if (repoURL != null) {
            // If so then set the repository.
            this.repository = repoURL;
        } // end if

    } // end method setRepository()

    /**
     * Setter for repository.
     * 
     * @param repo the repository to set
     */
    public void setRepository(final URL repo) {
        this.repository = repo;
    } // end method setRepository()

    /**
     * The writeFile method attempts to write a file to the local repository.
     * 
     * @param fileName the name of the file to write.
     * @param resourceType The type of resource to write.
     * @param data the data that gets written.
     */
    public void writeFile(final String fileName,
            final ResourceType resourceType, final Object data) {
        if (!DDSApplet.getApplet().isStandAlone()) {
            throw new UnsupportedOperationException();
        }
        // the path of the file
        // TODO Refactor this into a constant.
        final String filePath = System.getProperty("user.home") + "/.dds/data/"
                + resourceType.getDirectory();

        final File outPath = new File(filePath);
        if (!outPath.exists() && !outPath.mkdirs()) {
            throw new UnsupportedOperationException(filePath);
        }
        final File outFile = new File(filePath + fileName);

        try {
            final FileWriter fileWriter = new FileWriter(outFile);
            fileWriter.write(data.toString());
            fileWriter.close();
        } catch (IOException e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }

        // System.out.println("--------BEGIN FILE: " + fileName + " ---------");
        // System.out.println(data.toString());
        // System.out.println("--------END FILE: " + fileName + " ---------");
    }

} // end class definition ResourceLoader
