package dds.control;

import java.io.IOException;
import java.net.URL;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.w3c.dom.Document;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * The XMLLoader class.
 * 
 * @author gte189u
 */
public final class XMLLoader implements ErrorHandler {

    /**
     * The NULL_DOC.
     */
    private static final Document NULL_DOC = null;

    /**
     * The mapGen.
     */
    private static XMLLoader loader = new XMLLoader();

    /**
     * The factory.
     */
    private final transient DocumentBuilderFactory factory;

    /**
     * The mapBuilder.
     */
    private final transient DocumentBuilder mapBuilder;

    /**
     * The scenarioBuilder.
     */
    private final transient DocumentBuilder scenarioBuilder;

    /**
     * The errors.
     */
    private transient int errors;

    /**
     * Constructor for MapGenerator.
     */
    private XMLLoader() {
        // TODO: Even though this will now allow the appropriate schema to be
        // loaded independently of the schema definition in the xml (maybe?) it
        // does NOT seem to be recognizing errors in the xml.
        super();
        // set up the DocumentBuilderFactory to read the XML
        // file into a Document Object Model.
        factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(false);
        factory.setNamespaceAware(true);
        this.mapBuilder = this.getBuilderFrom("dds.xsd", ResourceType.MAP);
        this.scenarioBuilder = this.getBuilderFrom("scenario.xsd",
                ResourceType.SCENARIO);
    } // end constructor MapGenerator()

    /**
     * The getBuilderFrom method creates a DocumentBuilder that uses the
     * specified XSD schema file.
     * 
     * @param file The XSD schema file to use.
     * @param type The resource type of the specified schema file.
     * @return A new DocumentBuilder.
     */
    private DocumentBuilder getBuilderFrom(final String file,
            final ResourceType type) {
        DocumentBuilder builder = null;
        final Schema schema = this.getSchemaFrom(file, type);
        factory.setSchema(schema);
        // create a builder to create the DOM
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
        builder.setErrorHandler(this);
        return builder;
    }

    /**
     * The getSchemaFrom.
     * 
     * @param file The name of the XSD file to load into a Schema.
     * @param type The resource type, which corresponds to the type of
     *            resource this Schema will be used to validate.
     * @return A Schema object from the specified XSD file.
     */
    private Schema getSchemaFrom(final String file, final ResourceType type) {
        URL xsd = null;
        Schema result = null;
        xsd = ResourceManager.getLOADER().getFile(file, type);
        // System.out.println(xsd + " exists: " + xsd.exists());
        Source schemaFile = null;
        try {
            schemaFile = new StreamSource(xsd.openStream());
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
        final SchemaFactory sfact = SchemaFactory
                .newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
        try {
            result = sfact.newSchema(schemaFile);
        } catch (SAXException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
        return result;
    }

    /**
     * The loadXML method creates a new Document from the specified XML file and
     * resource type.
     * 
     * @param fileName The name of the XML file.
     * @param resourceType The type of resource (e.g. MAP, SCENARIO, etc.)
     * @return A new Document object.
     */
    public Document loadXML(final String fileName,
                            final ResourceType resourceType) {
        final DocumentBuilder builder;
        if (resourceType == ResourceType.SCENARIO) {
            builder = this.scenarioBuilder;
        } else {
            builder = this.mapBuilder;
        }
        Document doc = NULL_DOC;

        // TODO implement changes to the map based on the difficulty setting.
        this.errors = 0;
        URL file = null;
        
        file = ResourceManager.getLOADER().getFile(fileName, resourceType);
        
        try {
            doc = builder.parse(file.openStream());
        } catch (SAXException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }

        if (this.errors > 0) {
           System.err.println("Errors occured loading " + fileName + ". "
                    + this.errors);
            doc = NULL_DOC;
        }

        return doc;

    }

    /**
     * The getLoader the singleton XMLLoader.
     * 
     * @return The singleton XMLLoader.
     */
    public static XMLLoader getLoader() {
        return loader;
    } // end method getMapGen()

    /**
     * The error method catches errors during the parse process and handles them
     * here.
     * 
     * @param exception The SAXParseException that is thrown.
     * @throws SAXException A SAXException.
     * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
     */
    public void error(final SAXParseException exception) throws SAXException {
        // Apparently most parse errors will come through here. So do our error
        // handling/counting here.
        this.errors++;
    }

    /**
     * The fatalError method catches errors during the parse process and handles
     * them here.
     * 
     * @param exception The SAXParseException that is thrown.
     * @throws SAXException A SAXException.
     * @see org.xml.sax.ErrorHandler#fatalError(org.xml.sax.SAXParseException)
     */
    public void fatalError(final SAXParseException exception)
            throws SAXException {
        // Haven't seen any errors come through here yet, but make error
        // counting work here just in case.
        this.errors++;
    }

    /**
     * The warning method catches errors during the parse process and handles
     * them here.
     * 
     * @param exception The SAXParseException that is thrown.
     * @throws SAXException A SAXException.
     * @see org.xml.sax.ErrorHandler#warning(org.xml.sax.SAXParseException)
     */
    public void warning(final SAXParseException exception) throws SAXException {
        // Haven't seen any errors come through here yet, but make error
        // counting work here just in case.
        this.errors++;
    }

}
