package dds.control.simulator;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.concurrent.ConcurrentLinkedQueue;

import dds.model.simulation.AbstractSimObject;

/**
 * The SimCommand class is responsible for representing an action that has been
 * requested to take place during the execution of the simulation. A command is
 * comprised of a reference to the target object in the for of the target
 * object's ID String. It also has a String of the name of the method to execute
 * and the parameters to be passsed to that method. The last part is an optional
 * specification of a time when this command should be executed. This part is
 * optional because at the moment the only SimCommands that utilize this are the
 * ones that get cued within AIScripts. The SimCommand class is now also used to
 * avoid excessive Object creation/destruction in the form of SimCommands. The
 * "pile" is a ConcurrentLinkedQueue of SimCommands, chosen since this
 * Collection is Thread safe. The purpose of this class is to only create new
 * SimCommand Objects when needed. A class needing a SimCommand will call the
 * getCommand() method with the desired parameters for the SimCommand passed in
 * as arguments. If a SimCommand is available on the pile, it will have its
 * attributes set to the new arguments and be returned. If the pile is empty, a
 * new SimCommand will be instantiated with the given arguments and returned.
 * Once a SimCommand has been executed in the Simulation, it is returned to the
 * pile via the returnCommand() method.
 * 
 * @author Mike Sanders
 * @author Jacob Bolton
 * @author krystian
 */
public final class SimCommand implements Comparable {

    /**
     * The NO_ARGS constant is the easy way to pass a set of no arguments to a
     * SimCommand.
     */
    public static final Object[] NO_ARGS = new Object[0];

    /**
     * The NO_COMMAND is an easy-to-read way to set something to null.
     */
    private static final String NO_COMMAND = null;

    /**
     * The NO_OBJECT is an easy-to-read way to set something to null.
     */
    private static final String NO_OBJECT = null;

    /**
     * The NO_VALUES is an easy-to-read way to set something to null.
     */
    private static final Object[] NO_VALUES = null;

    /**
     * The TIME_RESOLUTION is the resolution to use for comparing different
     * SimCommand times.
     */
    public static final float TIME_RESOLUTION = 100.0f;

    /**
     * The pile of SimCommand objects available for use.
     */
    private static final ConcurrentLinkedQueue<SimCommand> PILE 
            = new ConcurrentLinkedQueue<SimCommand>();

    /**
     * The getCommand method returns a SimCommand from the pile if one is
     * available or creates a new SimCommand if the pile is empty. The command
     * that is returned will have its attributes set to those that were passed
     * in to the method.
     * 
     * @param idStr the ID of the command's target object
     * @param cmd the command descriptor
     * @param args the arguments of the command
     * @return the command that was returned or created
     */
    public static SimCommand getCommand(final String idStr, final String cmd,
            final Object[] args) {
        return getCommand(idStr, cmd, args, 0);
    } // end method getCommand()

    /**
     * The getCommand method returns a SimCommand from the pile if one is
     * available or creates a new SimCommand if the pile is empty. The command
     * that is returned will have its attributes set to those that were passed
     * in to the method.
     * 
     * @param idStr the ID of the command's target object.
     * @param cmd the command descriptor.
     * @param args the arguments of the command.
     * @param time The execution time for the command.
     * @return the command that was returned or created
     */
    public static SimCommand getCommand(final String idStr, final String cmd,
            final Object[] args, final float time) {
        // see if a SimCommand is available...
        SimCommand command = PILE.poll();

        // if not...
        if (command == null) {
            // create a new SimCommand with the given attributes...
            command = new SimCommand(idStr, cmd, args, time);
        } else {
            // otherwise, set the attributes of the SimCommand that was
            // retrieved from the pile to those that were passed in...
            command.setTargetObject(idStr);
            command.setCommand(cmd);
            command.setParamValues(args);
            command.setTime(time);
        } // end if-else

        // return the newly configured SimCommand
        return command;

    } // end method getCommand()

    /**
     * The returnCommand method returns the given SimCommand to the pile once it
     * has been executed.
     * 
     * @param command is the SimCommand to be reused
     */
    public static void returnCommand(final SimCommand command) {
        PILE.add(command);
    } // end method returnCommand()

    /**
     * The parseType method is used to extract the appropriate Class to use for
     * the array of parameter types that assist in reflection for grabbing the
     * desired method for execution.
     * 
     * @param object The object to determine the appropriate type.
     * @return The appropriate type for parameter type data.
     */
    private static Class parseType(final Object object) {
        Class result = null;
        if (object instanceof Boolean) {
            result = Boolean.TYPE;
        } else if (object instanceof Character) {
            result = Character.TYPE;
        } else if (object instanceof Byte) {
            result = Byte.TYPE;
        } else if (object instanceof Short) {
            result = Short.TYPE;
        } else if (object instanceof Integer) {
            result = Integer.TYPE;
        } else if (object instanceof Long) {
            result = Long.TYPE;
        } else if (object instanceof Float) {
            result = Float.TYPE;
        } else if (object instanceof Double) {
            result = Double.TYPE;
        } else {
            result = object.getClass();
        }
        return result;
    }

    /**
     * The command field stores a copy of the method name that is to be
     * executed.
     */
    private String command;

    /**
     * The paramValues field stores the set of parameters that are to be set for
     * this trap.
     */
    private Object[] paramValues;

    /**
     * The targetObject field stores a copy of the ID String for the target
     * object.
     */
    private String targetObject;

    /**
     * The time field stores the optional time for execution of this SimCommand.
     */
    private float time;

    /**
     * Constructor for SimCommand.
     */
    private SimCommand() {
        super();
        this.targetObject = NO_OBJECT;
        this.command = NO_COMMAND;
        this.paramValues = NO_VALUES;
        this.time = 0;
    }

    /**
     * Constructor for SimCommand.
     * 
     * @param idStr Name/ID of the target object.
     * @param cmd Name of method to execute on target object.
     * @param args Params to pass to the method.
     * @param execTime The time for this SimCommand to be executed.
     */
    private SimCommand(final String idStr, final String cmd,
            final Object[] args, final float execTime) {
        this.targetObject = idStr;
        this.command = cmd;
        this.paramValues = args;
        this.time = execTime;
    }

    /**
     * The compareTo method is used to compare two SimCommands.
     * 
     * @param obj The object to compare with. It must be a SimCommand.
     * @return The result of the comparison.
     */
    public int compareTo(final Object obj) {
        // Cast the Object passed in as a SimCommand.
        final SimCommand cmd = (SimCommand) obj;

        // Try calculating a result from the difference between execution times.
        int result = (int) ((this.time - cmd.time) * TIME_RESOLUTION);

        // If that result is 0, we should check against hash codes. This will
        // allow the SortedSet to see two diferent SimCommands with the same
        // execution time as different objects.
        if (result == 0) {
            result = this.hashCode() - cmd.hashCode();
        }

        // Return the result.
        return result;
    }

    /**
     * The executeOn method attempts to execute this command on a given
     * SimObject. This is accomplished by using reflection on the provided
     * SimObject to find and execute the method specified by the command.
     * 
     * @param simObj SimObject to execute command on.
     */
    public void executeOn(final AbstractSimObject simObj) {
        
        if (simObj == null) {
            return;
        }
        
        // Get the Class of the SimObject that is targeted for execution.
        final Class classToActivate = simObj.getClass();

        // Create a Class[] for storing the types of the parameters.
        final Class[] paramTypes = new Class[this.paramValues.length];

        // Propagate the Class[] for the parameter types.
        for (int i = 0; i < paramValues.length; i++) {
            // Process a parameter's type.
            paramTypes[i] = parseType(this.paramValues[i]);
        }

        try {
            // Try to find the method to be executed.
            final Method methodMan = classToActivate.getMethod(getCommand(),
                    paramTypes);

            // Set the method's accessibility so that the SimCommand can execute
            // it.
            // TODO Determine if we really need this because the getMethod()
            // method used to find the Method to execute only returns publicly
            // accessible methods.
            methodMan.setAccessible(true);

            // Try to invoke the method on the target object using the given
            // parameters.
            methodMan.invoke(simObj, getParamValues());
        } catch (IllegalArgumentException e) {
            // TODO Add logger.
            System.err.println(e);
        } catch (IllegalAccessException e) {
            // TODO Add logger.
            System.err.println(e);
        } catch (InvocationTargetException e) {
            // TODO Add logger.
            System.err.println(e);
        } catch (SecurityException e) {
            // TODO Add logger.
            System.err.println(e);
        } catch (NoSuchMethodException e) {
            // TODO Add logger.
            System.err.println(e);
        }

    }

    /**
     * Getter for command.
     * 
     * @return the command
     */
    public String getCommand() {
        return command;
    }

    /**
     * Getter for paramValues.
     * 
     * @return the paramValues
     */
    public Object[] getParamValues() {
        return paramValues.clone();
    }

    /**
     * Getter for targetObject.
     * 
     * @return the targetObject
     */
    public String getTargetObject() {

        return targetObject;
    }

    /**
     * Getter for time.
     * 
     * @return the time
     */
    public float getTime() {
        return time;
    }

    // /**
    // * The hashCode method returns a hash for this object.
    // *
    // * @return The hash code.
    // * @see java.lang.Object#hashCode()
    // */
    // @Override
    // public int hashCode() {
    // return super.hashCode();
    // }

    /**
     * Setter for command.
     * 
     * @param cmdStr the command to set
     */
    public void setCommand(final String cmdStr) {
        this.command = cmdStr;
    }

    /**
     * Setter for paramValues.
     * 
     * @param args the paramValues to set
     */
    public void setParamValues(final Object[] args) {
        // TODO Figure out a way to not store this directly to prevent a
        // checkstyle issue.
        // TODO Consider removing this set by just parsing in the values and
        // then forgetting about them.
        this.paramValues = args;
    }

    /**
     * Setter for targetObject.
     * 
     * @param obj the targetObject to set
     */
    public void setTargetObject(final String obj) {
        this.targetObject = obj;
    }

    /**
     * Setter for time.
     * 
     * @param execTime the time to set
     */
    public void setTime(final float execTime) {
        this.time = execTime;
    }

    /**
     * The toString method creates a String representation of this object.
     * 
     * @return The String representation.
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return "SimCommand on " + this.targetObject + ": " + this.command;
    }
}
