/**
 * SimCommandQueue.java.
 */
package dds.control.simulator;

import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * The SimCommandQueue class is responsible for collecting and returning
 * SimCommands generated through the course of the simulation in a thread-safe
 * manner. SimCommands that need to be executed are added to the back of the
 * SimCommandQueue. The SimCommandQueue is emptied by the SimulationController
 * during each time cycle of the Simulation, and each of the SimCommands are
 * executed on their targetObjects and returned to the SimCommandPile. (Please
 * see the documentation for SimCommand and SimCommandPile for further
 * clarification)
 * 
 * @author Jacob Bolton
 * @author Mike Sanders
 */
public final class SimCommandQueue extends ConcurrentLinkedQueue<SimCommand> {

    /**
     * The serialVersionUID is an ID used for serialization.
     */
    private static final long serialVersionUID = 1L;

    /**
     * The queue of SimCommands.
     */
    private static SimCommandQueue queue = new SimCommandQueue();

    /**
     * Constructor for SimCommandQueue.
     */
    private SimCommandQueue() {
        super();
    }

    /**
     * The getInstance.
     * 
     * @return queue the instance of the SimCommandQueue
     */
    public static SimCommandQueue getInstance() {
        return queue;
    }

    /**
     * The enQueue method is used to add a SimCommand to the back of the
     * SimCommandQueue to be placed in line for execution.
     * 
     * @param command is the SimCommand to be added to the queue
     */
    public static void enQueue(final SimCommand command) {
        // Add the command to the queue.
        queue.add(command);
    } // end method enQueue()

    /**
     * The reset method clears the queue by returning all remaing SimCommands
     * and then clearing the queue's old references.
     */
    public static void reset() {
        // Loop through all commands left in the queue.
        for (SimCommand command : queue) {
            // Return the command to the SimCommand class.
            SimCommand.returnCommand(command);
        }

        // Clear out the queue's references.
        queue.clear();
    }

} // end class definition SimCommandQueue()
