package dds.control.simulator;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import javax.swing.Timer;

import org.w3c.dom.Document;

import dds.control.DDSApplet;
import dds.control.ResourceType;
import dds.control.XMLLoader;
import dds.control.simulator.ai.AIManager;
import dds.control.simulator.ai.AIScript;
import dds.control.simulator.ai.ConcreteAIFactory;
import dds.control.user.ModuleType;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.ConcreteSimulationFactory;
import dds.model.simulation.Intersection;
import dds.model.simulation.Road;
import dds.model.simulation.Simulation;

/**
 * The SimulationController class controls the behavior of the simulation. The
 * Simulation is created through the calls to the generateScripts() method,
 * which reads in the information stored in the appropriate scenario.xml file,
 * and the setupSimulation() method, which creates a Simulation object through a
 * call to the ConcreteSimulationFactory. These methods must be called in the
 * order given here due to the fact that the name of the mapFile used to create
 * the Simulation object is defined in the scenario.xml file and stored in the
 * AIManager. The AIManager also stores other objects and variables that are
 * needed by the Simulation to complete the setup process. Calling
 * setupSimulation() before making a call to generateScripts() will result in a
 * NullPointerException. This class controls gameplay while the simulation is
 * running through the use of a Timer and several boolean flags. (See the
 * Javadoc for the class variables for a more detailed description) The
 * SimulationController is also responsible for randomly generating the driver's
 * path while the Simulation is running through calls to the generatePathPoint()
 * method. Calls to this method are made in the actionPerformed() method, which
 * receives actionEvents from the timer variable. There is also an int variable
 * named pathType, which is used to let the SimulationController know if the
 * path is predefined in the scenario.xml file or if it should be randomly
 * generated by the SimulationController. By default, the path is being randomly
 * generated; however, it can be defined by changing the value of
 * "intersections" attribute in the driverpath tag of the scenario.xml file from
 * '="random"' to a list of intersection id's. (IE: '="i10 i5 i20 etc..." NOTE:
 * the intersection id's must be sequential for this implementation to work!)
 * 
 * @author Mike Sanders
 */
public final class SimController implements ActionListener {

    /**
     * The simControl is the instance of the SimulationController singleton.
     */
    private static SimController simControl = new SimController();

    /**
     * The TIME_DELAY is a constant which represents the delay between timer
     * ticks.
     */
    private static final float TIME_DELAY = 20.0f;

    /**
     * The RANDOM constant is used to define the pathType variable.
     */
    private static final int RANDOM = 1;

    /**
     * The DEFINED constant is used to define the pathType variable.
     */
    private static final int DEFINED = 2;

    /**
     * The NO_MANAGER constant is an easy-to-read way to set the current AI
     * manager to null.
     */
    private static final AIManager NO_MANAGER = null;

    /**
     * The NO_SIMULATION constant is an easy-to-read way to set the current
     * simulation to null.
     */
    private static final Simulation NO_SIMULATION = null;

    /**
     * The AFTER_POINTS constant defines the number of path points to generate
     * after the path is completed?
     */
    // TODO Verify comment.
    private static final int AFTER_POINTS = 3;

    /**
     * The pathType indicates whether the driver's path is defined or needs to
     * be randomly generated on the fly.
     */
    private transient int pathType;

    /**
     * The currentSim variable is the instance of the currently configured
     * Simulation object.
     */
    private Simulation currentSim;

    /**
     * The currentAIManager variable is the instance of the AIManager that is
     * currently managing the Simulation's AI objects.
     */
    private AIManager currentAIManager;

    /**
     * The nextPathPoint is the next Intersection that the driver must cross to
     * continue the path.
     */
    private transient AbstractSimObject nextPathPoint;

    /**
     * The pathComplete flag indicates when the driver completes the assigned
     * path.
     */
    private transient boolean pathNotComplete;

    /**
     * The driverCrashed flag indicate if the driver has crashed into another
     * car or a building.
     */
    private transient boolean driverNotCrashed;

    /**
     * The hazardsComplete flag indicates when the driver has completed all of
     * the road hazards in the AIManager.
     */
    private transient boolean hazardsComplete;

    /**
     * The detectCollisions flag indicates if the driver should be penalized for
     * colliding with buildings and cars.
     */
    private transient boolean detectCollisions;

    /**
     * The timer is the heartbeat timer for the Simulation.
     */
    private final transient Timer timer;

    /**
     * The lastLocation variable stores the AbstractSimObject that the driver
     * drove over last. This should only be an Intersection or a Road object
     * since other objects are currently ignored.
     */
    private transient AbstractSimObject lastLocation;

    /**
     * The lastTime in the last time in milliseconds that the last simulation
     * update occured. This is used to determine the amount of real time passed
     * so that the simulation can emulate real time.
     */
    private transient long lastTime;

    /**
     * The pointsRemaining field declares how many path points remain?
     */
    // TODO Verify comment.
    private transient int pointsRemaining;

    /**
     * Constructor for SimulationController.
     */
    private SimController() {
        super();

        // A timer for this controller.
        this.timer = new Timer((int) TIME_DELAY, this);
    } // end construction SimulationController()

    /**
     * The setupSimulation method gets a new Simulation object from the
     * ConcreteSimulationFactory based on the mapFile name that was passed in.
     * This method also sets the initial values of the necessary attributes that
     * are used to run the currentSimulation. NOTE: the generateScripts() method
     * MUST be called before setupSimulation() in order for this method to
     * execute properly!
     * 
     * @param mapFile Location of the map file to load.
     */
    public void setupSimulation(final String mapFile) {

        // then create a Document Object Model from the mapFile that was
        // passed in...
        final Document simModel = MapGenerator.getMapGen().loadMap(mapFile);

        // set up a new simulation from the DOM that was just created...
        currentSim = ConcreteSimulationFactory.getFactory().createSimulation(
                simModel);
        // set the driver of the new simulation...
        currentSim.setDriver(currentAIManager.getDriver());
        // set the time limit for the driver to complete the path...
        currentSim.setTimeLimit(currentAIManager.getTimeLimit());
        // Run setup for the AIScripts
        this.currentAIManager.setup();

        if (currentAIManager.getNextPathPoint().equals("random")) {
            pathType = RANDOM;
            // we need to call detectCollision before
            // the first call to simulate()
            for (AbstractSimObject nextObj : currentSim.getStaticObjects()) {
                currentSim.getDriver().detectCollision(nextObj);
                if (!(nextObj instanceof Road)) {
                    currentSim.getDriver().handleCollision(nextObj, null);
                } // end if
            } // end for
            // set the lastLocation to the location
            // the driver is starting on...
            lastLocation = currentSim.getDriver().getObjectUnderMe();

        } else {
            pathType = DEFINED;
            // set the next intersection on the path the driver must follow...
            nextPathPoint = (Intersection) currentSim
                    .getObject(currentAIManager.getNextPathPoint());
        } // end if-else

        // indicate that the driver hasn't completed the path yet,
        pathNotComplete = true;
        // hasn't crashed into another car or building,
        driverNotCrashed = true;
        // and hasn't completed all of the road hazards.
        hazardsComplete = false;
        // enable collision detection by default.
        detectCollisions = true;

        pointsRemaining = AFTER_POINTS;

        // Have the Simulation do all of its pre-simulation calculations.
        this.currentSim.calculateAll();

        // Set the last time of simulation to now.
        this.lastTime = System.currentTimeMillis();

    } // end method setupSimulation()

    /**
     * The generateScripts method passes the script file to the
     * ConcreteAIFactory to create the AIManager. This method must be called
     * before the setupSimulation() method because the driver object and map
     * file name will be stored in the AIManager object that this method
     * creates.
     * 
     * @param scriptFile is the file that lists the traps
     */
    public void generateScripts(final String scriptFile) {

        // create the Document Object Model from the scenario file...
        final Document aiModel = XMLLoader.getLoader().loadXML(scriptFile,
                ResourceType.SCENARIO);
        // and create the new AIManager from the DOM that was just loaded.
        currentAIManager = ConcreteAIFactory.getFactory().generateScripts(
                aiModel);
    } // end method generateScripts()

    /**
     * The runSimulation method starts the simulation by starting the
     * SimulationController and Simulation timers.
     */
    public void runSimulation() {
        this.lastTime = System.currentTimeMillis();
        timer.start();
        // currentSim.startTimer();
    } // end method runSimulation()

    /**
     * The pauseSimulation method pauses the simulation by stopping the
     * simulation timers if the simulation is running and starting the
     * simulation timers if the simulation is paused.
     */
    public void pauseSimulation() {

        if (timer.isRunning()) {
            timer.stop();
            // currentSim.stopTimer();
        } else {
            runSimulation();
        } // end if-else

    } // end method pauseSimulation()

    /**
     * The endSimulation method ends the current simulation by stopping the
     * simulation timers.
     */
    public void endSimulation() {
        this.destruct();
    } // end method endSimulation()

    /**
     * The actionPerformed method receives timer events from the
     * SimulationController's timer, and updates each cycle of the simulation
     * accordingly.
     * 
     * @param evt is the event received
     * @see java.awt.event.ActionListener
     *      #actionPerformed(java.awt.event.ActionEvent)
     */
    // TODO Consider turning this into our own thread so we don't crowd the
    // Timer thread pool.
    public void actionPerformed(final ActionEvent evt) {
        // System.out.println("Timer ticked at " + System.currentTimeMillis());
        if (this.currentSim == null || this.currentAIManager == null) {
            return;
        }
        // if the driver hasn't completed the simulation path...
        if (pathNotComplete && driverNotCrashed) {
            // Simulate the actions of the AIScripts.
            this.currentAIManager.simulate(System.currentTimeMillis()
                    - this.lastTime);

            // Execute the awaiting SimCommands...
            executeCommandsInQueue();

            // update the driver's path...
            if (pathType == DEFINED && !currentSim.isTimesUp()) {
                // Use a predefined path.
                updateDriverPath();
            } else {
                // Use a random path.
                generateNextPathPoint();
            } // end if-else

            // Tell the simulation to go simulate itself for the amount of time
            // that has passed since the last simulation iteration.
            currentSim.simulate(System.currentTimeMillis() - this.lastTime);

            // Check to see if we're detecting collisions to end the simulation.
            if (detectCollisions) {
                // Since we are, make sure we're not coliding with anything that
                // shouldn't kill us?
                // NOTE: PMD says that bitwise inversion is faster than boolean
                // inversion, hence the ^ true.
                driverNotCrashed = (currentSim.getDriver()
                        .isDriverHitBuilding() || currentSim.getDriver()
                        .isDriverHitCar()) ^ true;
            } // end if
        } else {
            // Stop and clean up after the simulation.
            endSimulation();

            // Return to the main menu.
            DDSApplet.getApplet().loadModule(ModuleType.MAIN_MENU);
        } // end if-else

        // Save the current system time to be used for detemining how much time
        // has elapsed since the last simulation iteration completed.
        this.lastTime = System.currentTimeMillis();

    } // end method actionPerformed()

    /**
     * The updateDriverPath method removes the next intersection from the path
     * variable in AIManager as the player crosses it.
     */
    private void updateDriverPath() {

        // find out if the driver has entered the next intersection
        // on the path
        if (currentSim.getDriver().detectCollision(
                (Intersection) this.nextPathPoint)) {
            // see if the driver is at the end of the path
            if (currentAIManager.getNextPathPoint().equals("")) {
                // if so, set the pathComplete flag to true. This
                // will end the simulation.
                pathNotComplete = false;
            } else { // if not...
                // update the nextPathPoint to be the next intersection
                // in the list...
                nextPathPoint = (Intersection) currentSim
                        .getObject(currentAIManager.getNextPathPoint());
                // and remove the last intersection idString from the
                // path in AIManager
                final LinkedList<String> updatedPath = currentAIManager
                        .getPath();
                updatedPath.removeFirst();
                currentAIManager.setPath(updatedPath);
            } // end if-else
        } // end if

    } // end method updateDriverPath()

    /**
     * The generateNextPathPoint method will scan the roads leaving the current
     * Intersection for hazards. If no hazards are found on the exiting roads
     * then the Intersections on the other ends of those roads are scanned for
     * hazards. If any of these objects contains a road hazard the driver will
     * be directed towards it. Otherwise, one of the Intersections will be
     * selected randomly, and the driver will be directed down the road that
     * leads to it.
     */
    private void generateNextPathPoint() {
        // get the Object under the car to identify its location...
        final AbstractSimObject currentLocation = currentSim.getDriver()
                .getObjectUnderMe();

        if (hazardsComplete && pointsRemaining == 0) {
            if ((currentLocation.getIdStr()).equals(nextPathPoint.getIdStr())) {
                pathNotComplete = false;
            } // end if
        } else if (!((currentLocation.getIdStr()).equals(lastLocation
                .getIdStr()))) {

            // if the driver has completed the hazards...
            if (hazardsComplete) {
                // reduce the number of path points the driver has to
                // go to finish the simulation...
                --pointsRemaining;
            } // end if

            // if the currentLocation is a road...
            if (currentLocation instanceof Road) {
                // scan the Road and set the nextPathPoint to the Intersection
                // opposite the one the driver entered this Road from...
                scanRoad((Road) currentLocation);
            } else if (currentLocation instanceof Intersection) {
                // scan the Intersection for road hazards on the
                // exiting roads and the Intersections on the other
                // side of them...
                final boolean stillNotSet = this
                        .scanIntersection((Intersection) currentLocation);

                // if none of the roads lead to a hazard
                // then pick one randomly.
                if (stillNotSet) {
                    pickRandomExit((Intersection) currentLocation);
                } // end if
            } else {
                System.err.println("Not sure what's under me!");
            } // end if-else
            // Finally, set the last location to the Object we're on
            // for the following generateNextPathPoint() call.
            lastLocation = currentLocation;

        } // end if

    } // end method generateNextPathPoint()

    /**
     * The setPathPoint method sets the nextPathPoint variable to the
     * AbstractSimObject that is passed in. This method also scans the AIScripts
     * in the AIManager to see if the AbstractSimObject is the location of one
     * of the road hazards and remove it if it is.
     * 
     * @param obj is an AbstractSimObject that the nextPathPoint should be set
     *            too.
     */
    private void setPathPoint(final AbstractSimObject obj) {
        // set the nextPathPoint ot the given Object...
        nextPathPoint = obj;

        // if all of the road hazards have been traversed...
        if (currentAIManager.getScripts().isEmpty()) {
            // let the SimulationController know.
            hazardsComplete = true;
        }

    } // end method setPathPoint()

    /**
     * The isLocationAScript method checks to see if the object identified by
     * objIdString is a location of one of the AIScript objects.
     * 
     * @param objIdString is a String that represents the idString of the object
     *            we need to check.
     * @return contains is a boolean that indicates whether the locations is the
     *         location of an AIScript or not.
     */
    private boolean isLocationAScript(final String objIdString) {
        // used to indicate if the location contains a road hazard...
        boolean contains = false;

        // iterate through the list of AIScripts...
        final Iterator iterate = currentAIManager.getScripts().iterator();
        while (iterate.hasNext()) {
            // if the objIdString matches the locationId of the AIScript...
            if (objIdString.equals(((AIScript) iterate.next())
                    .getLocationId())) {
                // indicate the location contains a road hazard...
                contains = true;
                // and exit the loop.
                break;
            } else {
                // Otherwise, the objIdString is not the location
                // of a road hazard...
                contains = false;
            } // end if-else
        } // end while

        return contains;

    } // end method isLocationAScript()

    /**
     * The scanRoad method scans a Road the driver is currently on and sets the
     * nextPathPoint to the intersection opposite the one we entered the road
     * on.
     * 
     * @param currentLocation is the Road object the driver is on.
     */
    private void scanRoad(final Road currentLocation) {

        // if we entered the road from the "start" intersection...
        if (((Road) currentLocation).getStart().equals(lastLocation)) {
            // the nextPathPoint should be the "end" intersection...
            setPathPoint(((Road) currentLocation).getEnd());
        } else {
            // otherwise, the nextPathPoint should be the "start"
            // intersection...
            setPathPoint(((Road) currentLocation).getStart());
        } // end if-else

    } // end method scanRoad()

    /**
     * The scanIntersection method scans the Intersection the driver is on and
     * looks for roads leaving the Intersection that contain road hazards. If
     * none are found, the method then scans the Intersections on the other end
     * of the exiting Roads to see if they contain road hazards. If none of the
     * Roads leaving the Intersection and none of the Intersections they lead to
     * contain road hazards, then one of the exiting Roads is picked at random
     * and set as the nextPathPoint.
     * 
     * @param currentLocation is an Intersection that the driver is currently
     *            on.
     * @return stillNotSet is a boolean which indicates if the nextPathPoint
     *         variable has been set or not.
     */
    private boolean scanIntersection(final Intersection currentLocation) {
        boolean stillNotSet = true;
        // iterate through each exiting road and look for hazards...
        for (Road exitRoad : currentLocation.getRoads()) {
            // if the road I'm looking at isn't the one I came into
            // the intersection on...
            if (!((exitRoad.getIdStr()).equals(lastLocation.getIdStr()))) {
                // set the nextPathPoint to that road if it contains
                // a road hazard...
                if (isLocationAScript(exitRoad.getIdStr())
                        && directionValid(exitRoad, currentLocation)) {
                    setPathPoint(exitRoad);
                    stillNotSet = false;
                    break;
                } else {
                    // otherwise, look at the intersections on the
                    // other side of the exiting road for hazards...

                    // if the intersection I'm on is the "start"
                    // intersection of the exiting road...
                    if ((currentLocation.getIdStr()).equals(exitRoad.getStart()
                            .getIdStr())) {
                        // see if the "end" intersection contains a
                        // road hazard, and set the nextPathPoint to
                        // the exiting road if it does...
                        if (isLocationAScript(exitRoad.getEnd().getIdStr())
                                && directionValid(exitRoad, currentLocation)) {
                            setPathPoint(exitRoad);
                            stillNotSet = false;
                            break;
                        }
                    } else {
                        // otherwise, see if the "start"
                        // intersection contains a road hazard,
                        // and set the nextPathPoint to the exiting
                        // road if it does...
                        if (isLocationAScript(exitRoad.getStart().getIdStr())
                                && directionValid(exitRoad, currentLocation)) {
                            setPathPoint(exitRoad);
                            stillNotSet = false;
                            break;
                        } // end if
                    } // end if-else
                } // end if-else
            } // end if
        } // end for-each
        return stillNotSet;
    } // end method scanIntersection()

    /**
     * The directionValid method is used to determine if the driver should be
     * allowed to drive down the given Road. This is mainly to ensure that the
     * nextPathPoint will never be set to lead the driver down a one-way Road in
     * the wrong direction.
     * 
     * @param exitRoad is the Road which is being chedcked for direction.
     * @param currentLocation is the Intersection the driver is in.
     * @return valid is a boolean indicating if the direction of the given Road
     *         is valid.
     */
    private boolean directionValid(final Road exitRoad,
            final Intersection currentLocation) {

        // used to indicate if the direction is valid
        boolean valid;

        // if the Road is bi-directional, the direction is valid...
        if (exitRoad.getDirection().equals(Road.TRAFFIC_BI)) {
            valid = true;
        } else if (exitRoad.getDirection().equals(Road.TRAFFIC_A_TO_B)) {
            // if the traffic should go from "start" to "end"...
            // if the driver is at the "start" side, the direction is valid...
            if ((currentLocation.getIdStr()).equals(exitRoad.getStart()
                    .getIdStr())) {
                valid = true;
            } else {
                // otherwise, the direction is invalid...
                valid = false;
            } // end if-else
        } else if (exitRoad.getDirection().equals(Road.TRAFFIC_B_TO_A)) {
            // if the traffic should go from "end" to "start"...
            // if the driver is at the "end" side, the direction is valid...
            if ((currentLocation.getIdStr()).equals(exitRoad.getEnd()
                    .getIdStr())) {
                valid = true;
            } else {
                // otherwise, the direction is invalid...
                valid = false;
            } // end if-else
        } else {
            // if the direction is undeterminable, then it is invalid.
            valid = false;
        } // end if-else
        return valid;

    } // end method directionValid()

    /**
     * The pickRandomExit method picks a random Road to exit the Intersection on
     * in the event that none of the exit roads contain a road hazard and none
     * of the intersections that those roads lead to contain a road hazard.
     * 
     * @param currentLocation is the Intersection that the driver is currently
     *            driving on.
     */
    private void pickRandomExit(final Intersection currentLocation) {
        // indicates that an exit still has not been picked...
        boolean stillNotSet = true;

        final Random index = new Random();
        do {
            // randomly pick a road that exits the Intersection...
            final int roadIndex = index
                    .nextInt(((Intersection) currentLocation)
                            .getRoads().size());
            final Road exit = ((Intersection) currentLocation)
                    .getRoad(roadIndex);
            // if the Road that was picked is not the Road we entered on...
            if (!(lastLocation.equals(exit))
                    && directionValid(exit, currentLocation)) {
                // set that Road to be the nextPathPoint...
                setPathPoint(exit);
                // and exit the loop...
                stillNotSet = false;
            } // end if
        } while (stillNotSet);

    } // end method pickRandomExit()

    /**
     * The executeCommandsInQueue method retrieves the new commands from the
     * SimCommandQueue and executes them on their respective target objects.
     */
    private void executeCommandsInQueue() {

        // create an iterator so we can cycle through the commands...
        final Iterator iterate = SimCommandQueue.getInstance().iterator();
        // then iterate the list and execute each SimCommand on its
        // targetObject
        while (iterate.hasNext()) {
            final SimCommand cmd = (SimCommand) iterate.next();
            // System.out.println("Executing " + cmd);
            cmd.executeOn(currentSim.getObject(cmd.getTargetObject()));
            SimCommand.returnCommand(cmd);
            iterate.remove();
        } // end while

    } // end method executeCommandsInQueue()

    /**
     * Getter for simControl.
     * 
     * @return the simControl
     */
    public static SimController getSimControl() {
        return simControl;
    } // end method getSimControl()

    /**
     * The getSimulation returns the current Simulation.
     * 
     * @return A Simulation.
     */
    public Simulation getSimulation() {
        return this.currentSim;
    }

    /**
     * Getter for currentAIManager.
     * 
     * @return the currentAIManager
     */
    public AIManager getCurrentAIManager() {
        return currentAIManager;
    }

    /**
     * Setter for currentAIManager.
     * 
     * @param manager the currentAIManager to set
     */
    public void setCurrentAIManager(final AIManager manager) {
        this.currentAIManager = manager;
    }

    /**
     * Getter for currentSim.
     * 
     * @return the currentSim
     */
    public Simulation getCurrentSim() {
        return currentSim;
    }

    /**
     * Setter for currentSim.
     * 
     * @param sim the currentSim to set
     */
    public void setCurrentSim(final Simulation sim) {
        this.currentSim = sim;
    }

    /**
     * Getter for nextPathPoint.
     * 
     * @return the nextPathPoint
     */
    public AbstractSimObject getNextPathPoint() {
        return nextPathPoint;
    }

    /**
     * Setter for nextPathPoint.
     * 
     * @param next the nextPathPoint to set
     */
    public void setNextPathPoint(final AbstractSimObject next) {
        this.nextPathPoint = next;
    }

    /**
     * Getter for detectCollisions.
     * 
     * @return detectCollisions.
     */
    public boolean getDetectCollisions() {
        return detectCollisions;
    }

    /**
     * Setter for detectCollisions.
     * 
     * @param toSet is a boolean to set detectCollisions too
     */
    public void setDetectCollisions(final boolean toSet) {
        currentSim.getDriver().setDriverHitBuilding(false);
        currentSim.getDriver().setDriverHitCar(false);
        detectCollisions = toSet;
    } // end method setDetectCollisions

    /**
     * The destruct.
     */
    public void destruct() {
        timer.stop();
        // for now we destroy the old sim, but this will change when we start to
        // implement replay and such
        if (this.currentAIManager != null) {
            this.currentAIManager.destruct();
            this.currentAIManager = NO_MANAGER;
        }
        if (this.currentSim != null) {
            this.currentSim.destruct();
            this.currentSim = NO_SIMULATION;
        }
        SimCommandQueue.reset();
    }

} // end class definition getSimControl()
