package dds.control.simulator.ai;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.concurrent.ConcurrentLinkedQueue;

import dds.model.simulation.Car;

/**
 * The AIManager class manages the execution of the AIScripts for the Simulation
 * and also holds other key variables that are used to create the Simulation.
 * The AIManager holds the "path" variable, which is a LinkedList of String
 * objects representing intersection idString's. This is used in the case that
 * the driver's path is defined in the scenario.xml file. If the "intersections"
 * attribute of the "driverpath" file is '="random"', then the path list holds
 * the String "random". This lets the SimulationController know that it needs to
 * randomly generate the driver's path as the Simulation is running. Other key
 * variables this class holds are the mapFile, which is a String representing
 * the file name of the map file used to create the Simulation map, and the
 * driver, which is a Car object that represents the player of the Simulation.
 * These variables MUST be instantiated before the setupSimulation method is
 * called in the SimulationController to avoid NullPointerExceptions.
 * 
 * @author Jacob Bolton
 * @author Mike Sanders
 */
public class AIManager {

    /**
     * The NO_DRIVER constant is an easy-to-read way to set the driver reference
     * to null.
     */
    private static final Car NO_DRIVER = null;

    /**
     * The mapFile String represents the name of the mapFile.
     */
    private String mapFile;

    /**
     * The scripts variable is a collection of AIScript objects.
     */
    private Collection<AIScript> scripts;

    /**
     * The scripts variable is a collection of AIScript objects.
     */
    private final transient Collection<AIScript> completedScripts;

    /**
     * The driver is the player represented as an object.
     */
    private Car driver;

    /**
     * The path described by a list of intersection id strings.
     */
    private LinkedList<String> path;

    /**
     * The timeLimit specifies how much time the player has to complete the
     * path.
     */
    private Float timeLimit;

    /**
     * Constructor for AIManager .
     */
    public AIManager() {
        super();
        this.scripts = new ConcurrentLinkedQueue<AIScript>();
        this.completedScripts = new ConcurrentLinkedQueue<AIScript>();
        this.path = new LinkedList<String>();
    }

    /**
     * The addScript method ads a new script to the collection.
     * 
     * @param aiScript AIScript to add.
     */
    public final void addScript(final AIScript aiScript) {
        this.scripts.add(aiScript);
    }

    /**
     * The addPathPoint method adds intersection id's to the path.
     * 
     * @param pathPoint is the idString of an intersection to add.
     */
    public final void addPathPoint(final String pathPoint) {
        path.add(pathPoint);
    }

    /**
     * Getter for scripts.
     * 
     * @return the scripts
     */
    public final Collection<AIScript> getScripts() {
        return scripts;
    }

    /**
     * Setter for scripts.
     * 
     * @param aiScripts the scripts to set
     */
    public final void setScripts(final Collection<AIScript> aiScripts) {
        this.scripts = aiScripts;
    }

    /**
     * Getter for mapFile.
     * 
     * @return the mapFile
     */
    public final String getMapFile() {
        return mapFile;
    }

    /**
     * Setter for mapFile.
     * 
     * @param file the mapFile to set
     */
    public final void setMapFile(final String file) {
        this.mapFile = file;
    }

    /**
     * Getter for path.
     * 
     * @return the path
     */
    public final LinkedList<String> getPath() {
        return path;
    }

    /**
     * Getter for timeLimit.
     * 
     * @return the timeLimit
     */
    public final Float getTimeLimit() {
        return timeLimit;
    }

    /**
     * Setter for timeLimit.
     * 
     * @param time the timeLimit to set
     */
    public final void setTimeLimit(final Float time) {
        this.timeLimit = time;
    }

    /**
     * Getter for driver.
     * 
     * @return the driver
     */
    public final Car getDriver() {
        return driver;
    }

    /**
     * Setter for driver.
     * 
     * @param drvr the driver to set
     */
    public final void setDriver(final Car drvr) {
        this.driver = drvr;
    }

    /**
     * The getNextPathPoint method gets the next point on the driver path that
     * the driver must move towards and returns it as a String.
     * 
     * @return next is the idString of the next intersection on the driver path.
     */
    public final String getNextPathPoint() {

        // the idString of the next intersection on
        // the path
        String next;

        // get the next idString if there is one...
        try {
            next = path.getFirst();
        } catch (NoSuchElementException e) {
            // or get an empty string if the
            // list is empty
            next = "";
        } // end try-catch

        // return the idString of the next intersection
        // on the path
        return next;

    } // end method getNextPathPoint()

    /**
     * The toString method returns a string that contains the mapFile name, the
     * driver information, the timeLimit, and information about all of the
     * AIScript objects that this AIManager manages.
     * 
     * @return info is a String containing the information about this AIManager.
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        // TODO Decide if this is too elaborate cuz it looks pretty involved.
        String info = "AIManager mapFile: " + mapFile + " driver: "
                + driver.toString() + " timeLimit: " + timeLimit + "\n";

        info = info + "path: ";
        final Iterator iterator = path.iterator();
        while (iterator.hasNext()) {
            info = info + iterator.next() + " ";
        }
        info = info + "\nscripts\n-------\n\n";

        final Iterator scriptIterator = scripts.iterator();
        while (scriptIterator.hasNext()) {
            info = info + scriptIterator.next().toString() + "\n";
        }
        info = info + "-------\n";
        return info;
    } // end method toString()

    /**
     * Setter for path.
     * 
     * @param driverPath the path to set
     */
    public final void setPath(final LinkedList<String> driverPath) {
        this.path = driverPath;
    }

    /**
     * The setup method takes care of setting up all of the AIScripts so that
     * they are ready to be simulated.
     */
    public final void setup() {
        // Loop through all of the AIScripts in the scripts list.
        for (AIScript script : this.scripts) {
            // Call setup on the AIScript.
            script.setup();
        }
    }

    /**
     * The simulate method takes care of simulating all of the AIScripts managed
     * by this manager for the specified amount of time.
     * 
     * @param time The amount of time to simulate.
     */
    public final void simulate(final float time) {
        // Loop through all of the AIScripts in the scripts list.
        for (AIScript script : this.scripts) {
            // Call simulate on the AIScript.
            script.simulate(time);

            // Check to see if the script is completed.
            if (script.isCompleted()) {
                // Put the completed script in the list of completed scripts so
                // that it can be cleaned up later.
                this.completedScripts.add(script);
            }
        }

        // Iterate through all of the completed scripts.
        for (AIScript script : this.completedScripts) {
            // Remove the completed scripts from the list of scripts that get
            // simulated.
            this.scripts.remove(script);

            // Call on the scripts destruct method to do the script's
            // cleanup.
            script.tearDown();
        }

        // Reset the list of completed scripts since we don't need to keep track
        // of then after they've finished.
        this.completedScripts.clear();
    }

    /**
     * The destruct method performs cleanup on the AIManager when it's done
     * being used. This should help with garbage collection.
     */
    public final void destruct() {
        // Call destruct on all AIScripts.
        for (AIScript script : this.scripts) {
            script.destruct();
        }

        // Clear the list of AIScripts.
        this.scripts.clear();
        // this.path = null;

        // Clear the driver path.
        this.path.clear();

        // Detach the driver.
        this.driver = NO_DRIVER;
    }

} // end class definition AIManager
