package dds.control.simulator.ai;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import bsh.EvalError;
import bsh.Interpreter;
import dds.control.ResourceManager;
import dds.control.ResourceType;
import dds.control.simulator.SimCommand;
import dds.control.simulator.SimCommandQueue;
import dds.control.simulator.SimController;
import dds.model.simulation.Vector3D;

/**
 * The AIScript class is responsible for modelling and executing script files
 * for the simulation. Script files are a scripted Java language that is
 * compatible with the Beanshell interpreter. This allows us to extend the
 * Beanshell Interpreter class to get the basic functionality for AIScripts.
 * 
 * @author Jacob Bolton
 * @author Mike Sanders
 */
public class AIScript extends Interpreter implements Comparator<SimCommand> {

    /**
     * The NO_COMMAND is an easy-to-read way to set SimCommand variables to
     * null.
     */
    private static final SimCommand NO_COMMAND = null;

    /**
     * The serialVersionUID is used for object serialization.
     */
    private static final long serialVersionUID = 1L;

    /**
     * The behaviorFile is the name of the script file that this AIScript loads
     * and executes.
     */
    private String behaviorFile;

    /**
     * The commandCues field stores a list of SimCommands that are cued to
     * execute at specific times. This list is sorted by the execution time
     * field of the SimCommand.
     */
    private final transient SortedSet<SimCommand> commandCues;

    // /**
    // * The location field stores the specified location of the trap.
    // */
    // private Vector3D location;

    /**
     * The locationId field stores the name of the SimObject that this trap in
     * located on/near.
     */
    private String locationId;

    /**
     * The parameters fields stores a set of parameters for the script.
     */
    private final transient Collection<AIParameter> parameters;

    /**
     * The ID is the identifier for the script object.
     */
    private final transient String scriptId;

    /**
     * The targetObject is the ID of the default object this script acts on.
     */
    private String targetObject;

    /**
     * The timeElapsed field stores the amount of time elapsed while the
     * AIScript is triggered.
     */
    private float timeElapsed;

    /**
     * Constructor for AIScript.
     * 
     * @param loc Location provided to script.
     * @param sId is the identifier of the scriptl
     * @param file is the file name of the behavior file for this script.
     */
    public AIScript(final String sId, final Vector3D loc, final String file) {
        super();
        // Assign fields.
        this.scriptId = sId;
        // this.location = loc;
        // // Initialize the trap's location
        this.setLocation(loc);
        this.behaviorFile = file;

        // Load parameters.
        // TODO Move away from using the AIParameter class, and instead just use
        // direct access to the BeanShell environment to store parameters.
        parameters = new ArrayList<AIParameter>();

        // Create storage for the cued SimCommands.
        this.commandCues = new TreeSet<SimCommand>();

        // Load and parse the AI script file.
        this.loadScript(file);
    } // end constructor AIScript()

    /**
     * The addParameter method adds a parameter to the set of parameters.
     * 
     * @param param is the parameter to be added.
     */
    // TODO Refactor this to no longer depend on the AIParameter class and
    // instead store the parameter directly to the BeanShell environment.
    public final void addParameter(final AIParameter param) {
        this.parameters.add(param);
    }

    /**
     * The compare method is used to do the sorting of SimCommands.
     * 
     * @param cmd1 The first SimCommand to compare.
     * @param cmd2 The first SimCommand to compare.
     * @return The result of the comparison.
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public final int compare(final SimCommand cmd1, final SimCommand cmd2) {
        return (int) ((cmd1.getTime() - cmd2.getTime()) * SimCommand.TIME_RESOLUTION);
    }

    /**
     * The cueCommand method adds a SimCommand o the list of cued SimCommands.
     * 
     * @param command The SimCommand to add.
     */
    public final void cueCommand(final SimCommand command) {
        this.commandCues.add(command);
    }

    /**
     * The destruct method prepares this object for garbage collection by
     * removing all references to child objects so that they can be garbage
     * collected easier.
     */
    public final void destruct() {
        // Loop through the awaiting SimCommands.
        for (SimCommand command : this.commandCues) {
            // Return the SimCommand for reuse.
            SimCommand.returnCommand(command);
        }

        // Clear the old references from the list.
        this.commandCues.clear();
    }

    /**
     * Getter for behaviorFile.
     * 
     * @return the behaviorFile
     */
    public final String getBehaviorFile() {
        return behaviorFile;
    }

    /**
     * Getter for id.
     * 
     * @return the id
     */
    public final String getId() {
        return scriptId;
    }

    /**
     * Getter for location.
     * 
     * @return the location
     */
    public final Vector3D getLocation() {
        Vector3D location = null;
        try {
            location = (Vector3D) this.get("trapLocation");
        } catch (EvalError e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return location;
    }

    /**
     * Getter for locationId.
     * 
     * @return the locationId
     */
    public final String getLocationId() {
        return locationId;
    }

    /**
     * Getter for targetObject.
     * 
     * @return the targetObject
     */
    public final String getTargetObject() {
        return targetObject;
    }

    /**
     * Getter for timeElapsed.
     * 
     * @return the timeElapsed
     */
    public final float getTimeElapsed() {
        return timeElapsed;
    }

    /**
     * The hasCuedCommands method returns if this AIScript has commands left in
     * the list of time-cued SimCommands.
     * 
     * @return True if there are commands left in commandCues.
     */
    public final boolean hasCuedCommands() {
        return !this.commandCues.isEmpty();
    }

    /**
     * The isCompleted method is supposed to return true when the script is
     * completed. It should map directly to the scripted method isCompleted() in
     * the trap script file.
     * 
     * @return True if the script is completed.
     */
    public final boolean isCompleted() {
        // Assume that we're not complete.
        boolean result = false;

        // Evaluate the isCompleted() script method.
        try {
            this.eval("$_ = isCompleted();");
            final Object completed = this.get("$_");
            if (completed != null) {
                result = (Boolean) completed;
            }
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e);
        }

        // Return the result.
        return result;
    }

    /**
     * The loadScript method loads a script for the simulation.
     * 
     * @param file The script file to load.
     */
    private void loadScript(final String file) {
        try {
            // Import the commands we've defined.
            this.eval("importCommands(\"dds.control.simulator.ai.commands\");");

            // Set up a reference back to this script.
            this.set("script", this);

            // // Initialize the trap's location
            // this.set("trapLocation", this.location);

            // Load up the basescript template that ensures all function calls
            // are executed properly.
            URL fileURL = ResourceManager.getLOADER().getFile("basescript.bsh",
                    ResourceType.SCRIPT);
            this.eval(new InputStreamReader(fileURL.openStream()));

            // Load up the user's script file.
            fileURL = ResourceManager.getLOADER().getFile(file,
                    ResourceType.SCRIPT);
            this.eval(new InputStreamReader(fileURL.openStream()));

        } catch (FileNotFoundException e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        } catch (IOException e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
    }

    /**
     * Setter for behaviorFile.
     * 
     * @param bFile the behaviorFile to set
     */
    public final void setBehaviorFile(final String bFile) {
        this.behaviorFile = bFile;
    }

    /**
     * Setter for location.
     * 
     * @param loc the location to set
     */
    public final void setLocation(final Vector3D loc) {
        // this.location = loc;

        // Try to set the trap's location in the script environment.
        try {
            this.set("trapLocation", loc);
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
    }

    /**
     * Setter for locationId.
     * 
     * @param locId the locationId to set
     */
    public final void setLocationId(final String locId) {
        this.locationId = locId;
    }

    /**
     * Setter for targetObject.
     * 
     * @param tObject the targetObject to set
     */
    public final void setTargetObject(final String tObject) {
        this.targetObject = tObject;
        // Set the targetObject for the script
        try {
            this.set("targetObject", this.targetObject);
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
    }

    /**
     * Setter for timeElapsed.
     * 
     * @param time the timeElapsed to set
     */
    public final void setTimeElapsed(final float time) {
        this.timeElapsed = time;
    }

    /**
     * The setup method is used to perform any setup that a script may need to
     * do for initialization.
     */
    public final void setup() {
        // Run the setup method of the script
        try {
            this.eval("setup();");
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }
    }

    /**
     * The simulate method is used to execute the script for a specified amount
     * of time.
     * 
     * @param time The amount of time to simulate by executing the script.
     */
    public final void simulate(final float time) {
        if ((SimController.getSimControl() != null)
                || (SimController.getSimControl().getCurrentSim() == null)) {
            try {
                // Set the driver's location variable.
                final Vector3D driverLocation = SimController.getSimControl()
                        .getCurrentSim().getDriver().getPosition();
                this.set("driverLocation", driverLocation);

                // Evaluate the isTriggered() script method.
                this.eval("$_ = isTriggered();");
                final Object result = this.get("$_");

                // Check to see if the script is saying it's triggered.
                if (!(Boolean) result) {
                    // It isn't so stop here and do nothing else;
                    return;
                }
            } catch (EvalError e) {
                // TODO Add logger.
                System.err.println(e.getMessage());
                System.err.println(e.getStackTrace());
            }

            // System.out.println(this.getId() + " is triggered.");

            // Incerment the time elapsed.
            this.timeElapsed += time;

            // Check to see if we have commands cued to execute.
            if (this.commandCues.size() > 0) {
                // Since we do, get the first one of the sorted list.
                SimCommand command = this.commandCues.first();

                // While the current command has an execution time indicationg
                // it should be or have been exeuted.
                while ((command != null)
                        && (command.getTime() < this.timeElapsed)) {
                    // Enqueue the command for simulation.
                    SimCommandQueue.enQueue(command);

                    // System.out.println(this.getId() + " sent command: " +
                    // command);

                    // Remove the command from the list because it's been sent
                    // for execution.
                    this.commandCues.remove(command);

                    // Check to see if we have more commands cued to execute.
                    if (this.commandCues.size() > 0) {
                        // We do, so get the next one.
                        command = this.commandCues.first();
                    } else {
                        // We don't so set to null to break the loop.
                        command = NO_COMMAND;
                    }
                }
            }

            try {
                // Execute the script-defined simulate method.
                this.eval("simulate(" + time + "f);");
            } catch (EvalError e) {
                // TODO Add logger.
                System.err.println(e.getMessage());
                System.err.println(e.getStackTrace());
            }
        }

    }

    /**
     * The tearDown method does cleanup for a completed script. Currently, it
     * just maps to the scripted tearDown() method.
     */
    public final void tearDown() {
        // Evaluate the tearDown() script method.
        try {
            this.eval("tearDown();");
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e);
        }
    }

    /**
     * The toString method returns a string representation of the object.
     * 
     * @return info is a string containing information about this AIScript.
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        super.toString();
        String info = "AIScript scriptid: " + scriptId + " targetObject: "
                + targetObject + " behaviorFile: " + behaviorFile + "\n";

        final Iterator iterator = parameters.iterator();
        while (iterator.hasNext()) {
            info = info + iterator.next().toString() + "\n";
        }

        return info;

    }
}
