package dds.control.simulator.ai;

import java.util.StringTokenizer;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import dds.control.InvalidTypeException;
import dds.model.simulation.Car;
import dds.model.simulation.Vector3D;

/**
 * The ConcreteAIFactory class is the concrete implementation of the AIFactory
 * interface. (For further clarification, please see the factory design 
 * pattern) This class parses a DocumentObjectModel (DOM) that was created by 
 * the ResourceLoader, instantiates the appropriate Objects from the Nodes of 
 * the DOM, instantiates the AIManager used by the Simulation, and loads the 
 * appropriate variables into the AIManager for use in running the simulation. 
 * 
 * @author jbolton
 * @author Mike Sanders
 */
public final class ConcreteAIFactory {

    /**
     * The TEXT_NODE constant is used for detecting empty
     * nodes that are labeled "#text" in the XML file.
     */
    private static final String TEXT_NODE = "#text";

    /**
     * The TRAP constant is used to identify the trap XML tag.
     */
    private static final String TRAP = "trap";

    /**
     * The PARAM constant is used to identify the param XML tag.
     */
    private static final String PARAM = "param";

    /**
     * The instance variable contains the only reference to the
     * ConcreteAIFactory.
     */
    private static ConcreteAIFactory instance = new ConcreteAIFactory();

    /**
     * The manager variable contains the AIManager that is being
     * created and setup by the ConcreteAIFactory.
     */
    private final transient AIManager manager;

    /**
     * Constructor for ConcreteAIFactory.
     */
    private ConcreteAIFactory() {
        super();
        manager = new AIManager();
    }

    /**
     * The getFactory method returns the instance of ConcreteAIFactory.
     * 
     * @return The singleton factory.
     */
    public static ConcreteAIFactory getFactory() {
        return instance;
    }

    /**
     * The generateScripts method starts the process of creating the
     * AIScript objects and adding them to the AIManager that is being
     * configured.
     * @param aiModel is the DOM containing the hazard scripts.
     * @return manager is the newly created AIManager with the AIScripts
     * loaded.
     */
    public AIManager generateScripts(final Document aiModel) {

        parseTree(aiModel);
        return manager;

    } // end method generateScripts()

    /**
     * The parseTree method is the second step in the process
     * of creating the AIManager. This method parses the Document
     * Object Model for nodes that contain information relevant to
     * the simulation.
     * @param aiModel is the DOM to be parsed
     */
    private void parseTree(final Document aiModel) {

        // the list of the root node's children
        final NodeList children = aiModel.getChildNodes().item(0)
                .getChildNodes();

        // the current node being examined
        Node currentChild;

        // index for iterating the node list
        int index = 0;

        // cycle through each node and parse them if necessary
        for (; index < children.getLength(); ++index) {
            currentChild = children.item(index);

            // parse the node if it is named
            if (!currentChild.getNodeName().equals(TEXT_NODE)) {

                try {
                    parseNode(currentChild);
                } catch (InvalidTypeException e) {
                    System.err.println(e.getMessage());
                } // end try-catch

            } // end if

        } // end for

    } // end method generateScripts()

    /**
     * The parseNode method is the third step in the process of configuring
     * the AIManager. This method checks the type of the node that is
     * currently being parsed and calls the appropriate setup method on the
     * node.
     * @param rootNode is the root of the tree to be parsed
     * @throws InvalidTypeException if the node type can't be determined.
     */
    private void parseNode(final Node rootNode) throws InvalidTypeException {

        if (rootNode.getNodeName().equals("driverpath")) {
            addDriverPath(rootNode);
        } else if (rootNode.getNodeName().equals("driver")) {
            addDriverObject(rootNode);
        } else if (rootNode.getNodeName().equals("map")) {
            setMap(rootNode);
        } else if (rootNode.getNodeName().equals("traps")) {
            addTraps(rootNode);
        } else {
            throw new InvalidTypeException("Invalid type for XML node!");
        } // end if-else

    } // end method parseNode()

    /**
     * The addDriverPath sets the driver path and time limit in the
     * AIManager object.
     * @param rootNode is the root of the tree to be parsed.
     */
    private void addDriverPath(final Node rootNode) {

        // get the attributes
        final NamedNodeMap attributes = rootNode.getAttributes();

        try {
            // get the time limit for the path and set it
            final float timeLimit = Float.parseFloat(attributes.getNamedItem(
                    "timelimit").getTextContent());
            manager.setTimeLimit(timeLimit);

            if (attributes.getNamedItem("intersections").getTextContent()
                    .equals("random")) {
                manager.addPathPoint("random");
            } else {
                // get the idStrings of the intersections that describe
                // the driver's path and add them to the list
                final StringTokenizer tokenizer = new StringTokenizer(
                        attributes.getNamedItem("intersections")
                                .getTextContent());

                while (tokenizer.hasMoreTokens()) {
                    manager.addPathPoint(tokenizer.nextToken(" "));
                } // end while                
            } // end if-else

        } catch (NumberFormatException e) {
            System.err.println(e.getMessage());
        } // end try-catch

    } // end method addDriverPath()

    /**
     * The addDriverObject method parses the driver information from
     * the rootNode that is passed in, creates the driver Object, and
     * adds it to the AIManager being configured. 
     * @param rootNode is the root node of the tree to be parsed
     */
    private void addDriverObject(final Node rootNode) {

        final Car driver = new Car("driver");

        // get the attributes
        final NamedNodeMap attributes = rootNode.getAttributes();

        try {

            // get the start location of the driver
            final float xPos = Float.parseFloat(attributes.getNamedItem("x")
                    .getTextContent());
            final float yPos = Float.parseFloat(attributes.getNamedItem("y")
                    .getTextContent());
            final float zPos = Float.parseFloat(attributes.getNamedItem("z")
                    .getTextContent());

            // get the starting orientation of the driver
            final float xOrient = Float.parseFloat(attributes
                    .getNamedItem("ox").getTextContent());
            final float yOrient = Float.parseFloat(attributes
                    .getNamedItem("oy").getTextContent());
            final float zOrient = Float.parseFloat(attributes
                    .getNamedItem("oz").getTextContent());

            // set the position and orientation
            driver.setPosition(new Vector3D(xPos, yPos, zPos));
            driver.setOrientation(new Vector3D(xOrient, yOrient, zOrient));

            manager.setDriver(driver);

        } catch (NumberFormatException e) {
            System.err.println(e.getMessage());
            System.err.println("Driver position or location not set!\n");
        } // end try-catch

    } // end method addDriverObject()

    /**
     * The setMap method sets the map that this scenario files
     * applies too.
     * @param rootNode the root node of the tree
     */
    private void setMap(final Node rootNode) {
        manager.setMapFile(rootNode.getAttributes().getNamedItem("file")
                .getTextContent());
    } // end method setMap()

    /**
     * The addScripts method parses the rootNode passed in for "trap" tags,
     * creates an AIScript that represents each tag, creates AIParameter
     * objects for the AIScript Objects when appropriate, and adds the new
     * Objects to the AIManager being configured.
     * @param rootNode is the root node for the tree of hazard files
     */
    private void addTraps(final Node rootNode) {

        // the list of the root node's children
        final NodeList children = rootNode.getChildNodes();

        // the current node being examined
        Node currentChild;

        // index for iterating the node list
        int index = 0;

        // cycle through each node and parse them if necessary
        for (; index < children.getLength(); ++index) {
            currentChild = children.item(index);

            // parse the node if it is named
            if (!currentChild.getNodeName().equals(TEXT_NODE)
                    && currentChild.getNodeName().equals(TRAP)) {
                // get the attributes
                final NamedNodeMap attributes = currentChild.getAttributes();

                // get the id string
                final String idStr = attributes.getNamedItem("id")
                        .getTextContent();

                // create the Vector3D object for the location
                final float xValue = Float.parseFloat(attributes.getNamedItem(
                        "x").getTextContent());
                final float yValue = Float.parseFloat(attributes.getNamedItem(
                        "y").getTextContent());
                final float zValue = Float.parseFloat(attributes.getNamedItem(
                        "z").getTextContent());

                final Vector3D location = getVector3D(xValue, yValue, zValue);

                // get the behavior file for the script
                final String behaviorFile = attributes.getNamedItem("file")
                        .getTextContent();

                // create the new script
                final AIScript script = getAIScript(idStr, location,
                        behaviorFile);

                // set the locationId String of the hazard location in the map
                script.setLocationId(attributes.getNamedItem("locationid")
                        .getTextContent());

                // set the target object of the new script
                script.setTargetObject(attributes.getNamedItem("object")
                        .getTextContent());
                
                // if the trap has any parameters, handle them
                if (currentChild.hasChildNodes()) {
                    final NodeList params = currentChild.getChildNodes();
                    Node currParam;
                    int innerIndex = 0;

                    // iterate through the parameters, create them,
                    // and add them to the new script 
                    for (; innerIndex < params.getLength(); ++innerIndex) {

                        currParam = params.item(innerIndex);

                        if (!currParam.getNodeName().equals(TEXT_NODE)
                                && currParam.getNodeName().equals(PARAM)) {

                            final String name = currParam.getAttributes()
                                    .getNamedItem("name").getTextContent();

                            final String value = currParam.getAttributes()
                                    .getNamedItem("value").getTextContent();

                            // add the parameter to the AIScript
                            script.addParameter(getAIParameter(name, value));

                        } // end if

                    } // end for

                } // end if

                manager.addScript(script);

            } // end if

        } // end for

    } // end method addTraps()
    
    /**
     * The getVector3D method returns a new Vector3D object.
     * 
     * @param xValue is the x location of the new Vector3D.
     * @param yValue is the y location of the new Vector3D.
     * @param zValue is the z location of the new Vector3D.
     * @return the new Vector3D that was created with the given arguments.
     */
    private Vector3D getVector3D(final float xValue, 
                                 final float yValue,
                                 final float zValue) {
        return new Vector3D(xValue, yValue, zValue);
        
    } // end method getVector3D()
    
    /**
     * The getAIScript method returns a new AIScript object.
     * @param idStr is the id String of the AIScript.
     * @param location is the Vector3D location of the AIScript.
     * @param behaviorFile is the file name String of the behavior file
     * for this AIScript.
     * @return the new AIScript that was created with the given arguments.
     */
    private AIScript getAIScript(final String idStr,
                                 final Vector3D location,
                                 final String behaviorFile) {
        
        return new AIScript(idStr, location, behaviorFile);
        
    } // end method getAIScript()
    
    /**
     * The getAIParameter method returns a new AIParameter object.
     * @param name is a String representing the AIParameter name.
     * @param value is a String representing the AIParameter value.
     * @return the new AIParameter that was created with the given arguments.
     */
    private AIParameter getAIParameter(final String name, final String value) {
        
        return new AIParameter(name, value);
        
    } // end method getAIParameter()

} // end class definition ConcreteAIFactory
