/**
 * bezierPosition.java.
 */
package dds.control.simulator.ai.commands;

import dds.model.simulation.Vector3D;
import bsh.CallStack;
import bsh.EvalError;
import bsh.Interpreter;

/**
 * The BezierPosition class defines the BezierPosition script command for
 * AIScript files.
 * 
 * @author Jacob Bolton
 */
public final class BezierPosition {

    /**
     * The NINE is a constant for 9.
     */
    private static final float NINE = 9.0f;

    /**
     * The SIX is a constant for 6.
     */
    private static final float SIX = 6.0f;

    /**
     * The THREE is a constant for 3.
     */
    private static final float THREE = 3.0f;

    /**
     * The TWELVE is a constant for 12.
     */
    private static final float TWELVE = 12.0f;

    /**
     * The cubicBezier method performs a cubic bezier interpolation in a single
     * dimension.
     * 
     * @param start The starting point of the bezier curve.
     * @param control1 The first control point.
     * @param control2 The second control point.
     * @param end The ending point of the bezier curve.
     * @param time The duration of the curve.
     * @return The result of the interpolation.
     */
    private static float cubicBezier(final float start, final float control1,
            final float control2, final float end, final float time) {
        // Find the point p12 on line from p1 to p2
        // at time t.
        final float p12 = (control1 - start) * time + start;

        // Find the point p23 on line from p2 to p3
        // at time t.
        final float p23 = (control2 - control1) * time + control1;

        // Find the point p24 on line from p3 to p4
        // at time t.
        final float p34 = (end - control2) * time + control2;

        // Find the point p123 on line from p12 to p23
        // at time t.
        final float p123 = (p23 - p12) * time + p12;

        // Find the point p234 on line from p23 to p34
        // at time t.
        final float p234 = (p34 - p23) * time + p23;

        // Find the point on the Bezier curve from the
        // line between p123 and p234 at time t.
        final float p1234 = (p234 - p123) * time + p123;

        // Return the position.
        return p1234;
    }

    /**
     * The cubicBezierDirection method performs the differential of a cubic
     * bezier interpolation in a single dimension.
     * 
     * @param start The starting point of the bezier curve.
     * @param control1 The first control point.
     * @param control2 The second control point.
     * @param end The ending point of the bezier curve.
     * @param time The duration of the curve.
     * @return The result of the interpolation.
     */
    private static float cubicBezierDirection(final float start,
            final float control1, final float control2, final float end,
            final float time) {
        // Calculate the derivative of the first point's contribution.
        final float point1 = (-THREE + SIX * time - THREE * time * time)
                * start;

        // Calculate the derivative of the second point's contribution.
        final float point2 = (THREE - TWELVE * time + NINE * time * time)
                * control1;

        // Calculate the derivative of the third point's contribution.
        final float point3 = (SIX * time - NINE * time * time) * control2;

        // Calculate the derivative of the fourth point's contribution.
        final float point4 = (THREE * time * time) * end;

        // Return the sum of all point's contributions.
        return point1 + point2 + point3 + point4;
    }

    /**
     * The cueCommands.
     * 
     * @param env The execution environment. - Not needed for script call.
     * @param callstack The call stack. - Not needed for script call.
     * @param target The ID of the target object, i.e. the one o be positioned.
     * @param xPos The x-axis position.
     * @param yPos The y-axis position.
     * @param zPos The z-axis position.
     * @param xOrient The x-axis orientation.
     * @param yOrient The y-axis orientation.
     * @param zOrient The z-axis orientation.
     * @param time The time to cue the commands.
     */
    private static void cueCommands(final Interpreter env,
            final CallStack callstack, final String target, final float xPos,
            final float yPos, final float zPos, final float xOrient,
            final float yOrient, final float zOrient, final float time) {
        // Cue commands to set the new position and orientation at the
        // appropriate time. This is done by calling on the CueSimCommand
        // script command.
        // TODO Consider refactoring CueSimCommand so that we can call on it
        // without having to use invoke and pass the BeasnShell callstack and
        // such.
        CueSimCommand.invoke(env, callstack, target, "setPosition",
                new Object[] { xPos, yPos, zPos }, time);
        CueSimCommand.invoke(env, callstack, target, "setOrientation",
                new Object[] { xOrient, yOrient, zOrient }, time);
    }

    /**
     * The invoke method defines a syntax for a BezierPosition command.
     * 
     * @param env The execution environment. - Not needed for script call.
     * @param callstack The call stack. - Not needed for script call.
     * @param target The ID of the target object, i.e. the one o be positioned.
     * @param startPoint The starting point of the bezier curve.
     * @param controlPoint1 The first control point.
     * @param controlPoint2 The second control point.
     * @param endPoint The ending point of the bezier curve.
     * @param startTime The time elapsed after the trap is triggered until the
     *            curve should start to execute.
     * @param timeElapsed The duration of the curve.
     * @param resolution The amount of time between interpolation samples.
     */
    public static void invoke(final Interpreter env, final CallStack callstack,
            final String target, final Vector3D startPoint,
            final Vector3D controlPoint1, final Vector3D controlPoint2,
            final Vector3D endPoint, final float startTime,
            final float timeElapsed, final float resolution) {

        // Counter to keep track of interpolated execution time.
        float execTime = 0;

        // Loop to interpolate over the entire curve.
        do {
            // Calculate a new position.
            final float xPos = cubicBezier(startPoint.getXpos(), controlPoint1
                    .getXpos(), controlPoint2.getXpos(), endPoint.getXpos(),
                    execTime / timeElapsed);
            final float yPos = cubicBezier(startPoint.getYpos(), controlPoint1
                    .getYpos(), controlPoint2.getYpos(), endPoint.getYpos(),
                    execTime / timeElapsed);
            final float zPos = cubicBezier(startPoint.getZpos(), controlPoint1
                    .getZpos(), controlPoint2.getZpos(), endPoint.getZpos(),
                    execTime / timeElapsed);

            // Calculate a new orientation.
            final float xOrient = cubicBezierDirection(startPoint.getXpos(),
                    controlPoint1.getXpos(), controlPoint2.getXpos(), endPoint
                            .getXpos(), execTime / timeElapsed);
            final float yOrient = cubicBezierDirection(startPoint.getYpos(),
                    controlPoint1.getYpos(), controlPoint2.getYpos(), endPoint
                            .getYpos(), execTime / timeElapsed);
            final float zOrient = cubicBezierDirection(startPoint.getZpos(),
                    controlPoint1.getZpos(), controlPoint2.getZpos(), endPoint
                            .getZpos(), execTime / timeElapsed);

            // Cue commands to set the new position and orientation at the
            // appropriate time.
            cueCommands(env, callstack, target, xPos, yPos, zPos, xOrient,
                    yOrient, zOrient, startTime + execTime);

            // // Cue commands to set the new position and orientation at the
            // // appropriate time. This is done by calling on the CueSimCommand
            // // script command.
            // CueSimCommand.invoke(env, callstack, target, "setPosition",
            // new Object[] { xPos, yPos, zPos }, startTime + execTime);
            // CueSimCommand.invoke(env, callstack, target, "setOrientation",
            // new Object[] { xOrient, yOrient, zOrient }, startTime
            // + execTime);

            // Increment time counter.
            execTime += resolution;
        } while (execTime <= timeElapsed);
        // Loop ends when the time counter has reached the end of the curves
        // specified duration.

    }

    /**
     * The invoke method defines a syntax for a BezierPosition command.
     * 
     * @param env The execution environment. - Not needed for script call.
     * @param callstack The call stack. - Not needed for script call.
     * @param startPoint The starting point of the bezier curve.
     * @param controlPoint1 The first control point.
     * @param controlPoint2 The second control point.
     * @param endPoint The ending point of the bezier curve.
     * @param startTime The time elapsed after the trap is triggered until the
     *            curve should start to execute.
     * @param timeElapsed The duration of the curve.
     * @param resolution The amount of time between interpolation samples.
     */
    public static void invoke(final Interpreter env, final CallStack callstack,
            final Vector3D startPoint, final Vector3D controlPoint1,
            final Vector3D controlPoint2, final Vector3D endPoint,
            final float startTime, final float timeElapsed,
            final float resolution) {

        String targetName;

        try {
            targetName = (String) env.get("targetObject");
            invoke(env, callstack, targetName, startPoint, controlPoint1,
                    controlPoint2, endPoint, startTime, timeElapsed, 
                    resolution);
        } catch (EvalError e) {
            System.err.println(e.getMessage());
            System.err.println(e.getStackTrace());
        }

    }

    /**
     * Constructor for BezierPosition to prevent instantiation.
     */
    private BezierPosition() {
        super();
    }

}
