package dds.control.simulator.ai.commands;

import bsh.CallStack;
import bsh.EvalError;
import bsh.Interpreter;
import dds.control.simulator.SimController;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Vector3D;

/**
 * The WithinDistance class is used as a shell command to detect when the driver
 * is within range of a road hazard script location.
 * 
 * @author Mike Sanders
 */
public final class WithinDistance {

    /**
     * Private constructor for WithinDistance to prevent instantiation.
     */
    private WithinDistance() {
        super();
    }

    /**
     * The invoke method is called to execute this command.
     * 
     * @param env is the Interpreter environment that the command is executed
     *            in.
     * @param callstack is the CallStack of the environment.
     * @param distance is a float representing the distance to check against.
     * @return True if the driver is within the specified distance from the
     *         trap's location.
     */
    public static boolean invoke(final Interpreter env,
            final CallStack callstack, final float distance) {

        boolean result = false;
        try {
            result = distanceReached(env, distance, (Vector3D) env
                    .get("trapLocation"));
        } catch (EvalError e) {
            // TODO Add logger.
            System.err.println(e);
        }
        return result;
    } // end method invoke()

    /**
     * The invoke method is called to execute this command.
     * 
     * @param env is the Interpreter environment that the command is executed
     *            in.
     * @param callstack is the CallStack of the environment.
     * @param targetID Is the SimObject from which the distanc to the driver
     *            should be calculated.
     * @param distance is a float representing the distance to check against.
     * @return True if the driver is within the specified distance from the
     *         object's location.
     */
    public static boolean invoke(final Interpreter env,
            final CallStack callstack, final String targetID,
            final float distance) {

        boolean result = false;
        final AbstractSimObject targetObj = SimController.getSimControl()
                .getCurrentSim().getObject(targetID);
        if (targetObj != null) {
            result = distanceReached(env, distance, targetObj.getPosition());
        }
        return result;

    } // end method invoke()

    /**
     * The distanceReached method computes the distance from the trap location
     * to the driver. If that distance is less than the distance passed in, the
     * method returns true; otherwise, the method returns false.
     * 
     * @param env is the Interpreter environment this command is executed in.
     * @param distance is a float representing the distance to see if we're
     *            closer than.
     * @param position The position away from the driver from which the distance
     *            should be calculated.
     * @return withinRange is a boolean indicating if the driver is within range
     *         of the trap or not.
     */
    private static boolean distanceReached(final Interpreter env,
            final float distance, final Vector3D position) {

        // the driver is not close enough by default.
        boolean withinRange = false;

        try {
            // get a vector representing the distance from the driver
            // to the trap location...
            final Vector3D driverToMe = Vector3D.subtract(position,
                    (Vector3D) env.get("driverLocation"));

            // if the driver is within range...
            if (driverToMe.magnitude() <= distance) {
                // set withinRange to true...
                withinRange = true;
            }
        } catch (EvalError e) {
            System.err.println(e.getStackTrace());
        } // end try-catch

        return withinRange;

    } // end method distanceReached()

} // end class definition WithinDistance
