/**
 * AbstractModule.java.
 */
package dds.control.user;

import processing.core.PConstants;
import processing.core.PApplet.RegisteredMethods;
import dds.control.AbstractDDSAppletChild;
import dds.control.DDSApplet;
import dds.view.RenderableContext;

/**
 * The AbstractModule class defines the standard way in which DDSApplet modules
 * operate. The module is basically a single mode of operation for application,
 * e.g. the simulation, the trap editor, etc.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractModule extends AbstractDDSAppletChild implements
        InputHandler, RenderableContext, PConstants {

    /**
     * The registeredMethods field stores the last used set of RegisteredMethod
     * objects that were attached to the DDSApplet by this module. They are kept
     * stored for future reuse.
     */
    private final RegisteredMethods[] registeredMethods;

    /**
     * Constructor for AbstractModule.
     * 
     * @param applet The parent DDSApplet.
     */
    public AbstractModule(final DDSApplet applet) {
        super(applet);
        // Create an array to store the registered methods upon detachment.
        this.registeredMethods = new RegisteredMethods[DDSApplet.METHODS_COUNT];
    }

    /**
     * Getter for registeredMethods.
     * 
     * @return the registeredMethods
     */
    public final RegisteredMethods[] getRegisteredMethods() {
        // Clone the set of RegisteredMethods so it doesn't expose the array.
        // TODO Look into a way to do this without relying on object creation.
        return registeredMethods.clone();
    }

    /**
     * Setter for registeredMethods.
     * 
     * @param methods the registeredMethods to set
     */
    public final void setRegisteredMethods(final RegisteredMethods[] methods) {
        System.arraycopy(methods, 0, this.registeredMethods, 0,
                DDSApplet.METHODS_COUNT);
    }

    /**
     * The drawOn method defines the appropriate drawing functions that are
     * executed to render each frame of the application.
     * 
     * @param applet The applet on which to draw the context.
     * @see dds.view.RenderableContext#drawOn(dds.control.DDSApplet)
     */
    public void drawOn(final DDSApplet applet) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The keyPressed method defines what actions take place whe keys are
     * pressed.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     * @see dds.control.user.InputHandler#keyPressed(char, int)
     */
    public void keyPressed(final char key, final int keyCode) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The keyReleased method defines what actions take place whe keys are
     * released.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     * @see dds.control.user.InputHandler#keyReleased(char, int)
     */
    public void keyReleased(final char key, final int keyCode) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The mouseDragged method defines what actions take place when mouse motion
     * is detected while buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.InputHandler#mouseDragged(int, int, int)
     */
    public void mouseDragged(final int mouseX, final int mouseY,
            final int mouseButton) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The mouseMoved method defines what actions take place when mouse motion
     * is detected when no buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @see dds.control.user.InputHandler#mouseMoved(int, int)
     */
    public void mouseMoved(final int mouseX, final int mouseY) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The mousePressed defines what actions take place when mouse buttons are
     * pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.InputHandler#mousePressed(int, int, int)
     */
    public void mousePressed(final int mouseX, final int mouseY,
            final int mouseButton) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

    /**
     * The mouseReleased method defines what actions take place when mouse
     * buttons are released.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.InputHandler#mouseReleased(int, int, int)
     */
    public void mouseReleased(final int mouseX, final int mouseY,
            final int mouseButton) {
        // This is left empty so it can be extended when needed, but not written
        // just for the heck of it.
    }

}
