package dds.control.user;

/**
 * The InputHandler interface defines the methods by which the DDSApplet is able
 * to pass the handling of user input off to different modules within the
 * program. When an InputHandler is set in the DDSApplet the key and mouse
 * functions (as defined in Processing) get passed directly to the InputHandler.
 * That input handler is responsible for performing the appropriate actions
 * depending on the context in which they are received. Currently the DDSApplet
 * has it's displaying and input handling functionality in separate classes,
 * which can make the processing of input that affect the display somewhat
 * tedious. Since BOTH InputHandler and RenderableContext are an interfaces,
 * there's no reason why a singular class can't implement both and be singular
 * modules for DDSApplet operation.
 * 
 * @author Jacob Bolton
 */
public interface InputHandler {

    // TODO Work on designing a singular abstract class, which implements both
    // InputHandler AND RenderableContext so that it is easier to design
    // operational modules for this application.

    /**
     * The mouseMoved method defines what actions take place when mouse motion
     * is detected when no buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     */
    void mouseMoved(int mouseX, int mouseY);

    /**
     * The mousePressed defines what actions take place when mouse buttons are
     * pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    void mousePressed(int mouseX, int mouseY, int mouseButton);

    /**
     * The mouseReleased method defines what actions take place when mouse
     * buttons are released.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    void mouseReleased(int mouseX, int mouseY, int mouseButton);

    /**
     * The mouseDragged method defines what actions take place when mouse motion
     * is detected while buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    void mouseDragged(int mouseX, int mouseY, int mouseButton);

    /**
     * The keyPressed method defines what actions take place whe keys are
     * pressed.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     */
    void keyPressed(char key, int keyCode);

    /**
     * The keyReleased method defines what actions take place whe keys are
     * released.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     */
    void keyReleased(char key, int keyCode);
}
