package dds.control.user;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import dds.control.DDSApplet;
import dds.control.user.modules.AdminMenuModule;
import dds.control.user.modules.MainMenuModule;
import dds.control.user.modules.PauseMenuModule;
import dds.control.user.modules.SimulationModule;
import dds.control.user.modules.TrapEditorModule;
import dds.control.user.modules.UploadMenuModule;

/**
 * The ModuleTypes enum class defines a nice set of enumerations to help with
 * the managing of DDSModules.
 * 
 * @author Jacob Bolton
 */
public enum ModuleType {
    /**
     * The SIMULATION enumeration defines information for a simulation module.
     */
    SIMULATION(true, SimulationModule.class),

    /**
     * The TRAP_EDITOR enumeration defines information for the trap editor
     * module.
     */
    TRAP_EDITOR(true, TrapEditorModule.class),

    /**
     * The MAIN_MENU enumeration defines information for a main menu module.
     */
    MAIN_MENU(true, MainMenuModule.class),

    /**
     * The PAUSE_MENU enumeration defines information for a pause menu module.
     */
    PAUSE_MENU(false, PauseMenuModule.class),

    /**
     * The ADMIN_MENU enumeration defines information for a admin menu module.
     */
    ADMIN_MENU(true, AdminMenuModule.class),

    /**
     * The UPLOAD_MENU enumeration defines information for a upload menu module.
     */
    UPLOAD_MENU(true, UploadMenuModule.class),

    /**
     * The UNSUPPORTED enumeration defines information for an unsupported module
     * type.
     */
    UNSUPPORTED(false, null);

    /**
     * The moduleClassString field stores the String name for the class that
     * this enumeration represents.
     */
    private String moduleClassString;

    /**
     * The moduleClass field stores the Class representation of the class that
     * this enumeration represents.
     */
    private Class moduleClass;

    /**
     * The moduleConstructor field stores the Constructor for the class that
     * this enumeration represents.
     */
    private Constructor moduleConstructor;

    /**
     * The moduleConstructorArgs field stores an Object array with one element
     * which will be reused to construct new instances of this module type.
     */
    private Object[] constructorArgs;

    /**
     * The useLoadingDisplay flag describes whether or not this module type
     * should use the loading display when switching to an already instantiated
     * module.
     */
    private boolean useLoadingDisplay;

    /**
     * Constructor for ModuleTypes.
     * 
     * @param loadDisplay Boolean flag that describes whether or not the loading
     *            display should be active during the loading of this module.
     * @param classObj Is the Class representation of the class to be used for
     *            this type of module.
     */
    private ModuleType(final boolean loadDisplay, final Class classObj) {
        // Initialize fields.
        this.moduleClass = classObj;
        this.useLoadingDisplay = loadDisplay;
        this.constructorArgs = new Object[1];
        this.moduleClassString = "";
        // this.moduleConstructor = null;

        // Make sure class is not null before proceding because this code would
        // cause null pointer exceptions.
        if (classObj != null) {
            this.moduleClassString = classObj.getName();
            try {
                Class[] classArray = new Class[] { DDSApplet.class };
                this.moduleConstructor = this.moduleClass
                        .getDeclaredConstructor(classArray);
            } catch (SecurityException e) {
                // Constructor is not accessible.
                // TODO Add logger
                System.out.println(e);
            } catch (NoSuchMethodException e) {
                // Constructor does not exist.
                // TODO Add logger
                System.out.println(e);
            }
        }
    }

    /**
     * Getter for moduleClassString.
     * 
     * @return the moduleClassString
     */
    public final String getModuleClassString() {
        return moduleClassString;
    }

    /**
     * Getter for useLoadingDisplay.
     * 
     * @return the useLoadingDisplay
     */
    public final boolean isUseLoadingDisplay() {
        return useLoadingDisplay;
    }

    /**
     * The parseString method tries to parse out the appropriate ModuleType
     * object from a given String.
     * 
     * @param classString The class string to return.
     * @return The related ModuleType, UNSUPPORTED if it is not recognized.
     */
    public static ModuleType parseString(final String classString) {
        // Set the result to UNSUPPORTED in case the following processing
        // doesn't change anything.
        ModuleType result = UNSUPPORTED;

        // Check for a null string provided.
        if (classString != null) {
            // Loop through the provided ModuleTypes.
            for (ModuleType type : ModuleType.values()) {
                // Check to see if the provided string is equal to the current
                // ModuleType's
                if (classString.equals(type.getModuleClassString())) {
                    // Set the result to this type and break the loop.
                    result = type;
                    break;
                }
            }
        }
        return result;
    }

    /**
     * The parseType method takes an Object and attempts to reconcile it against
     * one of the defined ModuleTypes.
     * 
     * @param moduleArg The Object to try to resolve against a ModuleType.
     * @return The appropriate ModuleType.
     */
    public static ModuleType parseType(final Object moduleArg) {
        ModuleType result = UNSUPPORTED;
        if (moduleArg instanceof AbstractModule) {
            result = ModuleType.parseString(moduleArg.getClass().toString());
        } else if (moduleArg instanceof ModuleType) {
            result = (ModuleType) moduleArg;
        } else if (moduleArg instanceof String) {
            result = ModuleType.parseString((String) moduleArg);
        }
        return result;
    }

    /**
     * Getter for moduleClass.
     * 
     * @return the moduleClass
     */
    public final Class getModuleClass() {
        return moduleClass;
    }

    /**
     * Getter for moduleConstructor.
     * 
     * @return the moduleConstructor
     */
    public final Constructor getModuleConstructor() {
        return moduleConstructor;
    }

    /**
     * The constructNew method initializes a new module of this type.
     * 
     * @param applet The parent DDSApplet for the new instance
     * @return A new module.
     */
    public AbstractModule constructNew(final DDSApplet applet) {
        AbstractModule module = null;
        this.constructorArgs[0] = applet;
        try {
            module = (AbstractModule) moduleConstructor
                    .newInstance(this.constructorArgs);
        } catch (SecurityException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (IllegalArgumentException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (InstantiationException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (IllegalAccessException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (InvocationTargetException e) {
            // TODO Add logger.
            System.out.println(e);
        }
        return module;
    }
}
