/**
 * AbstractListMenuModule.java.
 */
package dds.control.user.modules;

import java.util.ArrayList;
import java.util.List;

import processing.core.PImage;

import dds.control.DDSApplet;
import dds.control.user.AbstractModule;

/**
 * The AbstractListMenuModule class is responsible for defining a standard
 * behavior for menus that only present a list of options to choose from.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractListMenuModule extends AbstractModule {

    // TODO Figure out how to add hooks to alow for proper centering of the
    // menus so that the resizing of the window/applet will recenter the menu.

    // TODO Determine what abstract methods should be added to this class to
    // make it more useful.

    /**
     * The NO_IMAGE constant is easy-to-read way to set the background image to
     * null.
     */
    public static final PImage NO_IMAGE = null;

    /**
     * The background field contains a reference to the immage to use when
     * drawing the background.
     */
    private PImage background;

    /**
     * The backgroundColor field stores the value of the color that will be used
     * to wipe the background when no image is specified.
     */
    private int backgroundColor;

    /**
     * The DEFAULT_COLOR is the default background color to use.
     */
    private static final int DEFAULT_COLOR = 128;

    /**
     * The options fields stores the list of options that are prsented by this
     * menu.
     */
    private List<ListMenuOption> options;

    /**
     * Constructor for AbstractListMenuModule.
     * 
     * @param applet The parent DDSApplet.
     */
    public AbstractListMenuModule(final DDSApplet applet) {
        super(applet);

        // Create a list to store the options.
        this.options = new ArrayList<ListMenuOption>();

        this.background = NO_IMAGE;
        this.backgroundColor = DEFAULT_COLOR;
    }

    /**
     * The destruct method is used to destroy this object when the DDSApplet
     * parent is done with it to allow for easier garbage collection. It does
     * this by unsetting the references this object has to/with other objects.
     * 
     * @see dds.control.AbstractDDSAppletChild#destruct()
     */
    @Override
    public final void destruct() {
        // TODO Consider calling a destruct on all of the menu options too.
        this.options.clear();
    }

    /**
     * The addOption method adds a new option to the list for this menu.
     * 
     * @param option The new option to add.
     */
    public final void addOption(final ListMenuOption option) {
        if (option != null) {
            this.options.add(option);
        }
    }

    /**
     * The setup method is used to perform any necessary operations that this
     * objects needs to execute in order to ready itself for proper use by the
     * DDSApplet parent.
     * 
     * @see dds.control.AbstractDDSAppletChild#setup()
     */
    @Override
    public final void setup() {
        // Recalculate height
        // TODO include a width calculation as well?
        int height = 0;
        for (ListMenuOption option : this.options) {
            height += option.getHeight();
        }

        // Start vertical positioning so that all elements will be vertically
        // centered.
        int yPosition = (this.getParent().height / 2) - (height / 2);

        // Position the elements.
        for (ListMenuOption option : this.options) {
            // Position their width so that they're horizontally centered.
            option.setXPosition((this.getParent().width / 2)
                    - (option.getWidth() / 2));
            option.setYPosition(yPosition);
            yPosition += option.getHeight();
        }
    }

    /**
     * The drawOn method defines the appropriate drawing functions that are
     * executed to render each frame of the application.
     * 
     * @param applet The applet on which to draw the context.
     * @see dds.control.user.AbstractModule#drawOn(dds.control.DDSApplet)
     */
    @Override
    public final void drawOn(final DDSApplet applet) {
        // Set up the appropraite viewing coordinates.
        // TODO Consider putting this into a method in DDSApplet.
        applet.perspective();
        applet.camera();

        // Use this to recenter the menu options for now.
        // TODO Come up with a way to do this that doesn't use iteration through
        // the menu elements.
        this.setup();

        // Wipe the screen with the background color.
        applet.background(this.backgroundColor);
        
        // Check to see if there's an image to use for the background.
        if (this.background != null) {
            // There's an image set so stretch the image to use it as a
            // background.
            applet.image(this.background, 0, 0, applet.width, applet.height);
        }

        // TODO Add support for a title image.

        // Draw the options.
        for (ListMenuOption option : this.options) {
            option.drawOn(applet);
        }
    }

    /**
     * The mouseMoved method defines what actions take place when mouse motion
     * is detected when no buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @see dds.control.user.AbstractModule#mouseMoved(int, int)
     */
    @Override
    public final void mouseMoved(final int mouseX, final int mouseY) {
        for (ListMenuOption option : this.options) {
            option.isOver(mouseX, mouseY);
        }
    }

    /**
     * The mousePressed defines what actions take place when mouse buttons are
     * pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.AbstractModule#mousePressed(int, int, int)
     */
    @Override
    public final void mousePressed(final int mouseX, final int mouseY,
            final int mouseButton) {
        for (ListMenuOption option : this.options) {
            if (option.isOver(mouseX, mouseY)) {
                option.execute();
                break;
            }
        }
    }

    /**
     * Getter for options.
     * 
     * @return the options
     */
    public final List<ListMenuOption> getOptions() {
        return options;
    }

    /**
     * Setter for options.
     * 
     * @param list the options to set
     */
    public final void setOptions(final List<ListMenuOption> list) {
        this.options = list;
    }

    /**
     * Getter for background.
     * 
     * @return the background
     */
    public final PImage getBackground() {
        return background;
    }

    /**
     * Setter for background.
     * 
     * @param image the background to set
     */
    public final void setBackground(final PImage image) {
        this.background = image;
    }

    /**
     * Getter for backgroundColor.
     * 
     * @return the backgroundColor
     */
    public final int getBackgroundColor() {
        return backgroundColor;
    }

    /**
     * Setter for backgroundColor.
     * 
     * @param color the backgroundColor to set
     */
    public final void setBackgroundColor(final int color) {
        this.backgroundColor = color;
    }

}
