/**
 * ListMenuOption.java.
 */
package dds.control.user.modules;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import dds.control.DDSApplet;
import dds.view.RenderableContext;
import processing.core.PImage;

/**
 * The ListMenuOption class.
 * 
 * @author jbolton
 */
public class ListMenuOption implements RenderableContext {

    /**
     * The HIGHLIGHT_COLOR defines the color to use for highlighting an item.
     */
    private static final int HIGHLIGHT_COLOR = 0xff007777;

    /**
     * The parent field contains a reference to the parent ListMenuModule.
     */
    private AbstractListMenuModule parent;

    /**
     * The executionMethod field contains a reference to the method that should
     * be invoked when this item is executed.
     */
    private Method executionMethod;

    /**
     * The itemImage field contains a PImage to be used for drawing this item.
     */
    private PImage itemImage;

    /**
     * The xPosition field contains the x coordinate for positioning this item.
     */
    private int xPosition;

    /**
     * The yPosition field contains the y coordinate for positioning this item.
     */
    private int yPosition;

    /**
     * The highlighted field flags whether this item has been highlighted or
     * not.
     */
    private boolean highlighted;

    /**
     * Constructor for ListMenuOption.
     * 
     * @param listMenu The parent list menu which owns this item.
     * @param method The method to execute when this item is clicked on.
     * @param image The image to use to draw this item.
     */
    public ListMenuOption(final AbstractListMenuModule listMenu,
            final Method method, final PImage image) {
        super();
        this.parent = listMenu;
        this.executionMethod = method;
        this.itemImage = image;
        this.highlighted = false;
        this.xPosition = 0;
        this.yPosition = 0;
    }

    /**
     * The execute method attempts to execute the executionMethod on the parent
     * list menu of this item.
     */
    public final void execute() {
        try {
            this.executionMethod.invoke(this.parent, (Object[]) null);
        } catch (IllegalArgumentException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (IllegalAccessException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (InvocationTargetException e) {
            // TODO Add logger.
            System.out.println(e);
        }
    }

    /**
     * The isOver method determines if the current mouse coordinates are over
     * this list option or not. This is usefule to determine if we draw
     * highlighting or not.
     * 
     * @param xCoord The x mouse coordinate.
     * @param yCoord The x mouse coordinate.
     * @return True if the specified coordinates are within the bounds of this
     *         item. False otherwise.
     */
    public final boolean isOver(final int xCoord, final int yCoord) {
        this.highlighted = (xCoord >= this.xPosition)
                && (yCoord >= this.yPosition)
                && (xCoord <= (this.xPosition + this.itemImage.width))
                && (yCoord <= (this.yPosition + this.itemImage.height));
        return this.highlighted;
    }

    /**
     * The drawOn method defines the appropriate drawing functions that are
     * executed to render each frame of the application.
     * 
     * @param applet The applet on which to draw the context.
     * @see dds.view.RenderableContext#drawOn(dds.control.DDSApplet)
     */
    public final void drawOn(final DDSApplet applet) {
        if (this.highlighted) {
            applet.tint(HIGHLIGHT_COLOR);
        }
        applet.image(this.itemImage, this.xPosition, this.yPosition);
        applet.noTint();
    }

    /**
     * Getter for executionMethod.
     * 
     * @return the executionMethod
     */
    public final Method getExecutionMethod() {
        return executionMethod;
    }

    /**
     * Setter for executionMethod.
     * 
     * @param method the executionMethod to set
     */
    public final void setExecutionMethod(final Method method) {
        this.executionMethod = method;
    }

    /**
     * Getter for highlighted.
     * 
     * @return the highlighted
     */
    public final boolean isHighlighted() {
        return highlighted;
    }

    /**
     * Setter for highlighted.
     * 
     * @param isHighlighted the highlighted to set
     */
    public final void setHighlighted(final boolean isHighlighted) {
        this.highlighted = isHighlighted;
    }

    /**
     * Getter for itemImage.
     * 
     * @return the itemImage
     */
    public final PImage getItemImage() {
        return itemImage;
    }

    /**
     * Setter for itemImage.
     * 
     * @param image the itemImage to set
     */
    public final void setItemImage(final PImage image) {
        this.itemImage = image;
    }

    /**
     * Getter for parent.
     * 
     * @return the parent
     */
    public final AbstractListMenuModule getParent() {
        return parent;
    }

    /**
     * Setter for parent.
     * 
     * @param module the parent to set
     */
    public final void setParent(final AbstractListMenuModule module) {
        this.parent = module;
    }

    /**
     * Getter for xPosition.
     * 
     * @return the xPosition
     */
    public final int getXPosition() {
        return xPosition;
    }

    /**
     * Setter for xPosition.
     * 
     * @param position the xPosition to set
     */
    public final void setXPosition(final int position) {
        xPosition = position;
    }

    /**
     * Getter for yPosition.
     * 
     * @return the yPosition
     */
    public final int getYPosition() {
        return yPosition;
    }

    /**
     * Setter for yPosition.
     * 
     * @param position the yPosition to set
     */
    public final void setYPosition(final int position) {
        yPosition = position;
    }

    /**
     * The getWidth.
     * 
     * @return The width of this item.
     */
    public final int getWidth() {
        return this.itemImage.width;
    }

    /**
     * The getHeight.
     * 
     * @return The height of this item.
     */
    public final int getHeight() {
        return this.itemImage.height;
    }

}
