/**
 * SimulationModule.java.
 */
package dds.control.user.modules;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collection;

import processing.core.PImage;
import dds.control.DDSApplet;
import dds.control.simulator.SimCommand;
import dds.control.simulator.SimCommandQueue;
import dds.control.simulator.SimController;
import dds.control.user.AbstractModule;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Simulation;
import dds.model.simulation.Vector3D;
import dds.view.simulation.AbstractMovableObjectRenderer;
import dds.view.simulation.AbstractSimObjectRenderer;
import dds.view.simulation.renderers.GenericVehicleRenderer;
import dds.view.simulation.renderers.RendererType;

/**
 * The SimulationModule class.
 * 
 * @author Jacob Bolton
 */
public class SimulationModule extends AbstractModule {
    /**
     * The ACCELERATE command type.
     */
    public static final String ACCELERATE = "accelerate";

    /**
     * The ACCELERATE_CHAR.
     */
    private static final char ACCELERATE_CHAR = 'w';

    /**
     * The BAD_ARROW_COLOR.
     */
    private static final int BAD_ARROW_COLOR = 0x77ff0000;

    /**
     * The BOUNDS_HEIGHT is the default height that is used to render the
     * boundary planes for the map.
     */
    private static final float BOUNDS_HEIGHT = 100;

    /**
     * The BRAKE.
     */
    private static final String BRAKE = "brake";

    /**
     * The BRAKE_CHAR.
     */
    private static final char BRAKE_CHAR = 's';

    /**
     * The CLIPPING_FAR is the value of the far clipping plane.
     */
    private static final float CLIPPING_FAR = 963.862f;

    /**
     * The CLIPPING_NEAR is the value for the near clippping plane.
     */
    private static final float CLIPPING_NEAR = 0.1f;

    /**
     * The COMPASS_SIZE.
     */
    private static final int COMPASS_SIZE = 20;

    /**
     * The CTRL_TOGGLE_CHAR.
     */
    private static final char CTRL_TOGGLE_CHAR = 'j';

    /**
     * The DEFAULT_CAMERAZ is the default height for the camera. This doesn't
     * really belong here since the camera height should be determined by the
     * driver's vehicle.
     */
    // TODO Move this value into the Car class.
    private static final float DEFAULT_CAMERAZ = 3.75f;

    /**
     * The DRIVER of the simulation.
     */
    public static final String DRIVER = "driver";

    /**
     * The FOV is the default viewing angle for the perspective projection in
     * the simulation. It is currently set to PI / 2.
     */
    private static final float FOV = HALF_PI;

    /**
     * The GOOD_ARROW_COLOR.
     */
    private static final int GOOD_ARROW_COLOR = 0x7700ff00;

    /**
     * The GROUND_COLOR is the default color used for drawing the ground plane.
     */
    private static final int GROUND_COLOR = 0xff007700;

    /**
     * The GROUND_Z is the distance below the map that the ground plane should
     * be drawn.
     */
    private static final float GROUND_Z = -0.1f;

    /**
     * The GUIDE_COLOR is the color to use for drawing the the control guides on
     * to the screen.
     */
    private static final int GUIDE_COLOR = 0x44ffffff;

    /**
     * The HIGH_CAMERAZ is the distance to move the camera upwards for viewing
     * from a third-person perspective.
     */
    private static final float HIGH_CAMERAZ = 50;

    /**
     * The LOOK_LEFT_CHAR.
     */
    private static final char LOOK_LEFT_CHAR = 'q';

    /**
     * The LOOK_RIGHT_CHAR.
     */
    private static final char LOOK_RIGHT_CHAR = 'e';

    /**
     * The NULL_SIMULATION is simply null cast as a Simulation. It is used for
     * initialization as well as cheking for non-existant Simulations.
     */
    private static final Simulation NO_SIMULATION = null;

    /**
     * The OVERHEAD_CHAR.
     */
    private static final char OVERHEAD_CHAR = ' ';

    /**
     * Krystian's paused menu magic school bus.
     */
    private static final char PAUSE = 'p';

    /**
     * The POINTLIGHT is the 'color' to use for the point light that gives
     * illumination to provide more realistic shading.
     */
    private static final float POINTLIGHT = 128;

    /**
     * The SKY_COLOR is the default color used for drawing the sky..
     */
    private static final int SKY_COLOR = 0xff9a9aff;

    /**
     * The STOP_ACCELERATE.
     */
    private static final String STOP_ACCELERATE = "stopAcceleration";

    /**
     * The STOP_BRAKE.
     */
    private static final String STOP_BRAKE = "stopBraking";

    /**
     * The TOGGLE_COLLISION_DETECTION.
     */
    private static final char COLLISION_CHAR = 'c';

    /**
     * The TURN command type.
     */
    public static final String TURN = "setTurn";

    /**
     * The TURN_LEFT_CHAR.
     */
    private static final char TURN_LEFT_CHAR = 'a';

    /**
     * The TURN_RIGHT_CHAR.
     */
    private static final char TURN_RIGHT_CHAR = 'd';

    /**
     * The VIEWTURN_ANGLE is the amount to rotate the camera to look left or
     * right.
     */
    private static final float VIEWTURN_ANGLE = 1.047f;

    /**
     * The WARN_ARROW_COLOR.
     */
    private static final int WARN_ARROW_COLOR = 0x77ffff00;

    /**
     * The ZONE_FRACTION.
     */
    private static final float ZONE_FRACTION = 10.0f;

    /**
     * The TURN_AMOUNT.
     */
    private static final float TURN_AMOUNT = 0.5f;

    /**
     * The INFO_FONT_SIZE constant defines the font size (in pixels) to use for
     * displaying the extra information.
     */
    private static final float INFO_FONT_SIZE = 16;

    /**
     * The TEXTURE_SCALE constant is the constant to use in order to scale
     * textures to an agreeble dimension on the screen for display.
     */
    public static final float TEXTURE_SCALE = 4.0f;

    /**
     * The ZONEY_SCALE is the appropriate scalar to use for the zoneY variable
     * that determines the steering guides.
     */
    private static final float ZONEY_SCALE = 1.5f;

    /**
     * The SCREENSHOT_CHAR constant defines the key used to create a screenshot.
     */
    private static final char SCREENSHOT_CHAR = 'i';

    /**
     * The COMPAS_CENTER constant contains the value used to to calculate a
     * horizontally centering transform to place the path compass in the middle
     * of the screen.
     */
    private static final float COMPAS_CENTER = 4f;

    /**
     * The borderTexture is the PImage to use to texture the map boundary lines
     * with.
     */
    private final transient PImage borderTexture;

    /**
     * The cameraZ the amount to reposition the camera along the z-axis (extra
     * height). This is essentially used to determine which perspective to use
     * for the driver (first or third).
     */
    // TODO Consider refactoring this as a boolean and incorporate it into the
    // driver's renderer.
    private transient float cameraZ;

    /**
     * The driver is the Renderer that draws the driver's vehicle. It is the
     * object through which information about the driver is passed into the
     * SimContext for use.
     */
    private GenericVehicleRenderer driver;

    /**
     * The grassTexture is the PImage to use to texture the ground plane with.
     */
    private final transient PImage grassTexture;

    /**
     * The guideMaxX describes where to put the right vertical guide.
     */
    private transient float guideMaxX;

    /**
     * The guideMaxY describes where to put the bottom horizontal guide.
     */
    private transient float guideMaxY;

    /**
     * The guideMinX describes where to put the left vertical guide.
     */
    private transient float guideMinX;

    /**
     * The guideMinY describes where to put the top horizontal guide.
     */
    private transient float guideMinY;

    /**
     * The jarekControls.
     */
    private transient boolean jarekControls;

    /**
     * The mousePressed.
     */
    private transient boolean mouseStillPressed;

    /**
     * The movingObjects is a set of objects that move during the simulation. If
     * used correctly, this list will allow for the interpolation of these
     * objects if the simulation hasn't had time to update their positions
     * itself. This way there can be a more seamless display of the simulation.
     */
    private Collection<AbstractMovableObjectRenderer> movingObjects;

    /**
     * The pressedX.
     */
    private transient int pressedX;

    /**
     * The pressedY.
     */
    private transient int pressedY;

    /**
     * The showGuides boolean simply flags this Context to tell it whether or
     * not to draw the control lines.
     */
    private transient boolean showGuides;

    /**
     * The sim is the current Simulation that this Context is responsible for
     * displaying.
     */
    private Simulation sim;

    /**
     * The staticObjects is a set of objects that should not change or move
     * during the simulation.
     */
    private Collection<AbstractSimObjectRenderer> staticObjects;

    /**
     * The useOverheadView flags the Context to tell it whether or not to use
     * the overhead, third-person perspective.
     */
    private boolean useOverheadView;

    /**
     * The viewRotation is the number of radians to rotate the camera's view
     * around the z-axis. This is used for looking to the left and right while
     * driving.
     */
    private transient float viewRotation;

    /**
     * Constructor for SimulationModule.
     * 
     * @param applet The parent DDSApplet.
     */
    public SimulationModule(final DDSApplet applet) {
        super(applet);
        this.mouseStillPressed = false;
        this.jarekControls = true;
        // Start with a null Simulation until one is set during initialization
        // from the SimContextFactory.
        this.sim = NO_SIMULATION;
        // this.sim = SimulationController.getSimControl().getCurrentSim();
        // Initialize the fields used for the overhead positioning of the
        // camera.
        this.cameraZ = DEFAULT_CAMERAZ;
        this.useOverheadView = false;
        // Create the lists used to store the diferent renderers that work under
        // this context.
        this.movingObjects = new ArrayList<AbstractMovableObjectRenderer>();
        this.staticObjects = new ArrayList<AbstractSimObjectRenderer>();
        // Preload the texture PImages to reduce object instantiation during the
        // simulation.
        this.grassTexture = DDSApplet.getApplet().loadImage("grass.jpg");
        this.borderTexture = DDSApplet.getApplet().loadImage("border.jpg");
        // // Create a blank SimContext
        // final SimContext context = new SimContext();
        // // Set the simulation that it observes
        // context.setSim(sim);
    }

    /**
     * The removeMovingObject method removes an AbstractMovableObjectRenderer
     * from the simulation.
     * 
     * @param targetObj is the AbstractSimObject whos renderer needs to be
     *            removed.
     */
    public final void removeMovingObject(final AbstractSimObject targetObj) {

        for (AbstractMovableObjectRenderer objRenderer : movingObjects) {

            if (objRenderer.getSimObject().equals(targetObj)) {
                movingObjects.remove(objRenderer);
            } // end if

        } // end for-each

    } // end method removeMovingObject()

    /**
     * The addObjRenderer method adds SimObject renderers to this context. It
     * determines what kind of object the Renderer is suposed to draw and adds
     * it to the apropriate list.
     * 
     * @param renderer The SimObject Renderer to be added.
     */
    public final void addObjRenderer(final AbstractSimObjectRenderer renderer) {
        // Check if the Renderer is for a MovableObject.
        if (renderer instanceof AbstractMovableObjectRenderer) {
            // If so, add to the MovableObjectRenderer list.
            this.movingObjects.add((AbstractMovableObjectRenderer) renderer);
        } else {
            // If not, add to the regular ObjectRenderer list.
            this.staticObjects.add(renderer);
        }
    }

    /**
     * The destruct.
     * 
     * @see dds.control.user.AbstractModule#destruct()
     */
    @Override
    public final void destruct() {
        // System.out.println("destruct called.");
        // TODO Write destruct in SimulationModule

    }

    /**
     * The drawGuides method is reponsible for placing guides on the screen,
     * which help the user with his mouse movements to control the vehicle.
     * 
     * @param applet The DDSAppet to draw onto.
     */
    private void drawGuides(final DDSApplet applet) {
        // Check to see if guides should be displayed.
        if (this.showGuides) {
            // Push the current transform matrix onto the stack to preserve it
            // for later.
            applet.pushMatrix();
            // Perform the transform to screen coordinates.
            applet.tranformToScreen();
            // Set drawing aesthitics.
            applet.stroke(GUIDE_COLOR);
            applet.strokeWeight(1);
            // Draw the guide lines.
            applet.line(this.guideMinX, 0, this.guideMinX, applet.height);
            applet.line(this.guideMaxX, 0, this.guideMaxX, applet.height);
            applet.line(0, this.guideMinY, applet.width, this.guideMinY);
            applet.line(0, this.guideMaxY, applet.width, this.guideMaxY);
            // Restore the original transformation matrix.
            applet.tranformFromScreen();
            applet.popMatrix();
        }
    }

    /**
     * The renderMapBoundary method is responsible for drawing the boundary
     * walls for the map.
     * 
     * @param applet The applet to draw on.
     */
    private void drawMapBoundary(final DDSApplet applet) {
        // Get the map's bounds.
        final Point mapX = sim.getXMapBounds();
        final Point mapY = sim.getYMapBounds();
        // Set some drawing aesthetics.
        applet.fill(-1); // -1 maps to the color white
        applet.noStroke();
        applet.textureMode(NORMALIZED);
        applet.beginShape(DDSApplet.QUADS);
        // Use the appropriate texture.
        applet.texture(this.borderTexture);
        // Draw the four bounding walls.
        // applet.strokeWeight(5);
        // Wall 1
        // applet.stroke(0xffff0000);
        applet.vertex(mapX.x, mapY.y, GROUND_Z, 1, 1);
        applet.vertex(mapX.x, mapY.y, GROUND_Z + BOUNDS_HEIGHT, 1, 0);
        applet.vertex(mapX.y, mapY.y, GROUND_Z + BOUNDS_HEIGHT, 0, 0);
        applet.vertex(mapX.y, mapY.y, GROUND_Z, 0, 1);
        // Wall 2
        // applet.stroke(0xff00ff00);
        applet.vertex(mapX.x, mapY.y, GROUND_Z, 1, 1);
        applet.vertex(mapX.x, mapY.y, GROUND_Z + BOUNDS_HEIGHT, 1, 0);
        applet.vertex(mapX.x, mapY.x, GROUND_Z + BOUNDS_HEIGHT, 0, 0);
        applet.vertex(mapX.x, mapY.x, GROUND_Z, 0, 1);
        // Wall 3
        // applet.stroke(0xff0000ff);
        applet.vertex(mapX.y, mapY.x, GROUND_Z + BOUNDS_HEIGHT, 1, 0);
        applet.vertex(mapX.y, mapY.x, GROUND_Z, 1, 1);
        applet.vertex(mapX.x, mapY.x, GROUND_Z, 0, 1);
        applet.vertex(mapX.x, mapY.x, GROUND_Z + BOUNDS_HEIGHT, 0, 0);
        // Wall 4
        // applet.stroke(0xffffffff);
        applet.vertex(mapX.y, mapY.x, GROUND_Z, 1, 1);
        applet.vertex(mapX.y, mapY.x, GROUND_Z + BOUNDS_HEIGHT, 1, 0);
        applet.vertex(mapX.y, mapY.y, GROUND_Z + BOUNDS_HEIGHT, 0, 0);
        applet.vertex(mapX.y, mapY.y, GROUND_Z, 0, 1);
        // Done drawing the walls.
        applet.endShape();
    }

    /**
     * The renderMapGround draws the ground plane for the simulation.
     * 
     * @param applet The DDSApplet to draw onto.
     */
    private void drawMapGround(final DDSApplet applet) {
        // Get the map's bounds.
        final Point mapXBounds = sim.getXMapBounds();
        final Point mapYBounds = sim.getYMapBounds();
        // Set some aesthetic drawing parameters.
        applet.noStroke();
        applet.fill(GROUND_COLOR);
        // Determine the uv coordinates for texture mapping.
        // TODO When repeating textures are available implement the appropriate
        // mappings for the ground texture.
        // final float texMinX = 0;
        // final float texMaxX = 1;
        // final float texMinY = 0;
        // final float texMaxY = 1;
        final float texMinX = (float) mapXBounds.x * TEXTURE_SCALE
                / (float) this.grassTexture.width;
        final float texMaxX = (float) mapXBounds.y * TEXTURE_SCALE
                / (float) this.grassTexture.width;
        final float texMinY = (float) mapYBounds.x * TEXTURE_SCALE
                / (float) this.grassTexture.width;
        final float texMaxY = (float) mapYBounds.y * TEXTURE_SCALE
                / (float) this.grassTexture.width;
        // Set the texture mode.
        applet.textureMode(NORMALIZED);
        // Draw the ground plane using the grass texture.
        applet.beginShape();
        applet.texture(grassTexture);
        applet.vertex(mapXBounds.x, mapYBounds.y, GROUND_Z, texMinX, texMaxY);
        applet.vertex(mapXBounds.y, mapYBounds.y, GROUND_Z, texMaxX, texMaxY);
        applet.vertex(mapXBounds.y, mapYBounds.x, GROUND_Z, texMaxX, texMinY);
        applet.vertex(mapXBounds.x, mapYBounds.x, GROUND_Z, texMinX, texMinY);
        applet.endShape();
    }

    /**
     * The drawOn method draws each frame of the simulation onto the DDSApplet.
     * It needs to draw all of the SimObjects using the correct perspective,
     * vantage point, and also display any other user interface graphics that
     * are used during the simulation.
     * 
     * @param applet The applet on which to draw this Context's graphics.
     */
    public final void drawOn(final DDSApplet applet) {
        // Clear the previous frame with the sky color
        applet.background(SKY_COLOR);

        // Set up the perspective matrix
        applet.perspective(FOV, applet.width / applet.height, CLIPPING_NEAR,
                CLIPPING_FAR);

        // Get camera position from the driver object.
        // System.out.println(this.driver);
        final Vector3D cam = this.driver.getCamPosition(this.cameraZ);
        // Get the camera look-at position from the driver object.
        final Vector3D camAim = this.driver.getCamAimPosition(this.cameraZ,
                this.viewRotation);

        // System.out.println("Cam: " + cam);
        // System.out.println("Cam aim: " + camAim);

        // Position the camera.
        applet.camera(cam.getXpos(), cam.getYpos(), cam.getZpos(), camAim
                .getXpos(), camAim.getYpos(), camAim.getZpos(), 0, 0, -1);

        // Turn on Processing-default lighting scheme.
        applet.lights();
        // Add a point light at the camera position to enhance environment
        // shading
        applet.pointLight(POINTLIGHT, POINTLIGHT, POINTLIGHT, cam.getXpos(),
                cam.getYpos(), cam.getZpos()/* + cameraZ */);

        // Push model matrix for drawing.
        applet.pushMatrix();

        // Draw the ground plane.
        this.drawMapGround(applet);
        // Draw the map boundary walls.
        this.drawMapBoundary(applet);
        // Draw the static objects.
        for (AbstractSimObjectRenderer obj : this.staticObjects) {
            applet.noStroke();
            obj.drawOn(applet);
        }
        // Draw the Dynamic objects.
        for (AbstractMovableObjectRenderer obj : this.movingObjects) {
            applet.noStroke();
            obj.drawOn(applet);
        }
        // Draw the driver
        driver.drawOn(applet);

        this.drawInfo(applet);

        // Draw the on-screen guides that assist with steering.
        this.drawGuides(applet);

        // Draws the arrow that points to the next path point for the driver.
        this.drawPathGuide(applet);

        // Pop the model matrix because drawing is complete.
        applet.popMatrix();

    }

    /**
     * The drawInfo method draws additional debugging information onto the
     * screen during a simulation. This information currently consists of the
     * driver's location and the framerate.
     * 
     * @param applet The applet to draw onto.
     */
    private void drawInfo(final DDSApplet applet) {
        // System.out.println("Drawing info.");
        applet.pushMatrix();
        applet.tranformToScreen();
        applet.fill(-1);
        applet.textAlign(LEFT);
        applet.textSize(INFO_FONT_SIZE);
        final String infoString = "Driver at: ("
                + (int) driver.getSimObject().getPosition().getXpos() + ", "
                + (int) driver.getSimObject().getPosition().getYpos() + ", "
                + (int) driver.getSimObject().getPosition().getZpos() + ")";
        applet.text(infoString, 0, INFO_FONT_SIZE, 0);
        applet.text("FPS: " + applet.frameRate, 0, INFO_FONT_SIZE * 2, 0);
        applet.tranformFromScreen();
        applet.popMatrix();
        // System.out.println(applet.frameRate);
    }

    /**
     * The drawPathGuide method is responsible for displaying the on-screen
     * compass, which assists the driver with following the path.
     * 
     * @param applet The DDSApplet to draw the compass onto.
     */
    private void drawPathGuide(final DDSApplet applet) {
        // TODO Comment ME!!
        if (SimController.getSimControl().getNextPathPoint() == null) {
            return;
        }
        final Vector3D orientation = this.driver.getOrientation().duplicate();
        orientation.normalize();
        final Vector3D position = SimController.getSimControl()
                .getNextPathPoint().getPosition();
        final Vector3D direction = Vector3D.subtract(position, this.driver
                .getPosition());
        direction.normalize();
        final float deviateOrient = (float) Math.atan2(orientation.getYpos(),
                orientation.getXpos());
        final float deviateDirection = (float) Math.atan2(direction.getYpos(),
                direction.getXpos());
        // float deviation = orientation.dotProduct(direction);
        final float deviation = deviateDirection - deviateOrient;
        final float size = applet.height / COMPASS_SIZE;
        int fillColor = GOOD_ARROW_COLOR;
        // System.out.println(deviation / PI);
        if ((deviation > QUARTER_PI) || (deviation < (-QUARTER_PI))) {
            if ((deviation > HALF_PI + QUARTER_PI)
                    || (deviation < -(HALF_PI + QUARTER_PI))) {
                fillColor = BAD_ARROW_COLOR;
            } else {
                fillColor = WARN_ARROW_COLOR;

            }
        }
        // System.out.println("-----------------");
        // System.out.println(orientation);
        // System.out.println(direction);
        // System.out.println("Vectors produce an offset of " + deviation + "
        // radians");
        applet.pushMatrix();
        applet.tranformToScreen();
        applet.translate((float) applet.width / COMPAS_CENTER, applet.height
                / COMPASS_SIZE);
        applet.rotateZ(deviation);
        // applet.noStroke();
        applet.strokeWeight(1);
        applet.stroke(0);
        applet.fill(fillColor);
        // applet.triangle(-size, 0, 0,-size,size,0);
        // applet.rect(-size / 2.0f, 0, size, size);
        applet.beginShape();
        applet.vertex(-size, 0);
        applet.vertex(0, -size);
        applet.vertex(size, 0);
        applet.vertex(size / 2.0f, 0);
        applet.vertex(size / 2.0f, size);
        applet.vertex(-size / 2.0f, size);
        applet.vertex(-size / 2.0f, 0);
        applet.endShape(CLOSE);
        applet.fill(-1);
        // applet.textSize(size);
        // applet.text(SimulationController.getSimControl().getNextPathPoint()
        // .getIdStr(), 0, 0, 0);
        applet.tranformFromScreen();
        applet.popMatrix();
    }

    /**
     * Getter for driver.
     * 
     * @return the driver
     */
    public final GenericVehicleRenderer getDriver() {
        return driver;
    }

    /**
     * Getter for movingObjects.
     * 
     * @return the movingObjects
     */
    public final Collection<AbstractMovableObjectRenderer> getMovingObjects() {
        return movingObjects;
    }

    /**
     * Getter for sim.
     * 
     * @return the sim
     */
    public final Simulation getSim() {
        return sim;
    }

    /**
     * Getter for staticObjects.
     * 
     * @return the staticObjects
     */
    public final Collection<AbstractSimObjectRenderer> getStaticObjects() {
        return staticObjects;
    }

    /**
     * Getter for useOverheadView.
     * 
     * @return the useOverheadView
     */
    public final boolean isUseOverheadView() {
        return useOverheadView;
    }

    /**
     * The keyPressed.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     */
    // TODO Design a way to reduce this complexity.
    public final void keyPressed(final char key, final int keyCode) {
        switch (Character.toLowerCase(key)) {
        case ACCELERATE_CHAR:
            this.sendCommand(ACCELERATE, SimCommand.NO_ARGS);
            break;
        case BRAKE_CHAR:
            this.sendCommand(BRAKE, SimCommand.NO_ARGS);
            break;
        case CTRL_TOGGLE_CHAR:
            this.jarekControls ^= true;
            this.unsetGuides();
            break;
        case OVERHEAD_CHAR:
            this.toggleOverheadView();
            break;
        case LOOK_LEFT_CHAR:
            this.setSideView(true);
            break;
        case LOOK_RIGHT_CHAR:
            this.setSideView(false);
            break;
        case TURN_LEFT_CHAR:
            this.sendCommand(TURN, new Object[] { -TURN_AMOUNT });
            break;
        case TURN_RIGHT_CHAR:
            this.sendCommand(TURN, new Object[] { TURN_AMOUNT });
            break;
        case PAUSE:
            // TODO Add pause menu in here.
            break;

        case COLLISION_CHAR:
            SimController.getSimControl().setDetectCollisions(
                    !(SimController.getSimControl().getDetectCollisions()));
            break;

        case SCREENSHOT_CHAR:
            if (this.getParent().isStandAlone()) {
                this.getParent().saveFrame("dds-#####.png");
            }
            break;
        default:
        }
        switch (keyCode) {
        case DDSApplet.LEFT:
            this.setSideView(true);
            break;
        case DDSApplet.RIGHT:
            this.setSideView(false);
            break;
        default:
        }
    }

    /**
     * The keyReleased.
     * 
     * @param key The key pressed.
     * @param keyCode The keyCode for the pressed key.
     */
    public final void keyReleased(final char key, final int keyCode) {
        // final SimContext context = (SimContext) DDSApplet.getApplet()
        // .getContext();
        switch (Character.toLowerCase(key)) {
        case ACCELERATE_CHAR:
            this.sendCommand(STOP_ACCELERATE, SimCommand.NO_ARGS);
            break;
        case BRAKE_CHAR:
            this.sendCommand(STOP_BRAKE, SimCommand.NO_ARGS);
            break;
        case LOOK_LEFT_CHAR:
        case LOOK_RIGHT_CHAR:
            this.unsetSideView();
            break;
        case TURN_LEFT_CHAR:
        case TURN_RIGHT_CHAR:
            this.sendCommand(TURN, new Object[] { 0.0f });
            break;
        default:
        }
        switch (keyCode) {
        case DDSApplet.LEFT:
            this.unsetSideView();
            break;
        case DDSApplet.RIGHT:
            this.unsetSideView();
            break;
        default:
        }
    }

    /**
     * The mouseDragged.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    public final void mouseDragged(final int mouseX, final int mouseY,
            final int mouseButton) {
        final float width = (float) DDSApplet.getApplet().width;
        final float height = (float) DDSApplet.getApplet().height;
        final float zoneX = width / 2.0f / ZONE_FRACTION;
        float zoneY = height / ZONEY_SCALE / ZONE_FRACTION;
        if (!this.jarekControls) {
            this.pressedX = (int) width / 2;
            if (!this.mouseStillPressed) {
                zoneY = height;
            }
        }
        this.setGuides(this.pressedX - zoneX, this.pressedX + zoneX,
                this.pressedY - zoneY, this.pressedY + zoneY);
        if (mouseX > (this.pressedX + zoneX)) {
            sendCommand(TURN, new Object[] { (mouseX - this.pressedX)
                    / (width - this.pressedX + zoneX) });
        } else if (mouseX < (this.pressedX - zoneX)) {
            sendCommand(TURN, new Object[] { (mouseX - this.pressedX + zoneX)
                    / (this.pressedX - zoneX) });
        } else {
            sendCommand(TURN, new Object[] { 0.0f });
        }

        if (mouseY > (this.pressedY + zoneY)) {
            sendCommand(BRAKE, new Object[] { (mouseY - this.pressedY)
                    / (height - this.pressedY + zoneY) });
            this.sendCommand(STOP_ACCELERATE, SimCommand.NO_ARGS);
        } else if (mouseY < (this.pressedY - zoneY)) {
            sendCommand(ACCELERATE,
                    new Object[] { -(mouseY - this.pressedY + zoneY)
                            / (this.pressedY - zoneY) });
            this.sendCommand(STOP_BRAKE, SimCommand.NO_ARGS);
        } else {
            this.sendCommand(STOP_BRAKE, SimCommand.NO_ARGS);
            this.sendCommand(STOP_ACCELERATE, SimCommand.NO_ARGS);
        }
    }

    /**
     * The mouseMoved.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     */
    public final void mouseMoved(final int mouseX, final int mouseY) {
        if (!this.jarekControls) {
            this.mouseDragged(mouseX, mouseY, mouseX);
        }
    }

    /**
     * The mousePressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    public final void mousePressed(final int mouseX, final int mouseY,
            final int mouseButton) {
        this.mouseStillPressed = true;
        this.pressedX = mouseX;
        this.pressedY = mouseY;
    } // end method mousePressed()

    /**
     * The mouseReleased.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     */
    public final void mouseReleased(final int mouseX, final int mouseY,
            final int mouseButton) {
        this.mouseStillPressed = false;
        // if (this.jarekControls) {
        // final SimContext context = (SimContext) DDSApplet.getApplet()
        // .getContext();
        this.unsetGuides();
        // }
        this.sendCommand(STOP_BRAKE, SimCommand.NO_ARGS);
        this.sendCommand(STOP_ACCELERATE, SimCommand.NO_ARGS);
    }

    /**
     * The sendCommand.
     * 
     * @param method The method to execute.
     * @param args The args to pass to the method.
     */
    private void sendCommand(final String method, final Object[] args) {
        // System.out.println("COMMAND: " + target + " " + method);
        SimCommandQueue.enQueue(SimCommand.getCommand(DRIVER, method, args));
    }

    /**
     * Setter for driver.
     * 
     * @param renderer the driver to set
     */
    public final void setDriver(final GenericVehicleRenderer renderer) {
        this.driver = renderer;
    }

    /**
     * The setGuides method sets the parameters for drawing the control guides.
     * 
     * @param minX The guide's minimum x coordiate in screen coordinates.
     * @param maxX The guide's maximum x coordiate in screen coordinates.
     * @param minY The guide's minimum y coordiate in screen coordinates.
     * @param maxY The guide's maximum y coordiate in screen coordinates.
     */
    public final void setGuides(final float minX, final float maxX,
            final float minY, final float maxY) {
        this.showGuides = true;
        this.guideMinX = minX;
        this.guideMaxX = maxX;
        this.guideMinY = minY;
        this.guideMaxY = maxY;
    }

    /**
     * Setter for movingObjects.
     * 
     * @param objRenderers the movingObjects to set
     */
    public final void setMovingObjects(
            final Collection<AbstractMovableObjectRenderer> objRenderers) {
        this.movingObjects = objRenderers;
    }

    /**
     * The setOverheadView.
     */
    public final void setOverheadView() {
        this.cameraZ = HIGH_CAMERAZ;
    }

    /**
     * The setSideView method sets this Context to use a view that looks either
     * right or left from the driver's perspective.
     * 
     * @param lookRight True if the camera is to look to the driver's right.
     */
    public final void setSideView(final boolean lookRight) {
        if (lookRight) {
            this.viewRotation = -VIEWTURN_ANGLE;
        } else {
            this.viewRotation = VIEWTURN_ANGLE;
        }
    }

    /**
     * The setSim.
     * 
     * @param newSim The new simulation to give to this context.
     */
    public final void setSim(final Simulation newSim) {
        this.sim = newSim;
    }

    /**
     * Setter for staticObjects.
     * 
     * @param objRenderers the staticObjects to set
     */
    public final void setStaticObjects(
            final Collection<AbstractSimObjectRenderer> objRenderers) {
        this.staticObjects = objRenderers;
    }

    /**
     * Setter for useOverheadView.
     * 
     * @param useView the useOverheadView to set
     */
    public final void setUseOverheadView(final boolean useView) {
        this.useOverheadView = useView;
    }

    /**
     * The toggleOverheadView method switches the use of the overhead view on
     * anf off.
     */
    public final void toggleOverheadView() {
        this.useOverheadView ^= true;
        if (this.useOverheadView) {
            this.setOverheadView();
        } else {
            this.unsetOverheadView();
        }
    }

    /**
     * The unsetGuides.
     */
    public final void unsetGuides() {
        this.showGuides = false;
    }

    /**
     * The unsetOverheadView.
     */
    public final void unsetOverheadView() {
        this.cameraZ = DEFAULT_CAMERAZ;
    }

    /**
     * The unsetSideView method removes any camera rotation being used to have
     * the driver look to the right or the left..
     */
    public final void unsetSideView() {
        this.viewRotation = 0;
    }

    /**
     * The setup.
     * 
     * @see dds.control.AbstractDDSAppletChild#setup()
     */
    @Override
    public final void setup() {
        this.jarekControls = true;

        SimController.getSimControl().generateScripts("scenario1.xml");
        SimController.getSimControl().setupSimulation(
                SimController.getSimControl().getCurrentAIManager()
                        .getMapFile());
        this.sim = SimController.getSimControl().getCurrentSim();

        // Populate the context with renderers for all the SimObjects in the
        // Simulation
        // try {
        for (AbstractSimObject obj : sim.getStaticObjects()) {
            this.addObjRenderer(RendererType.createRenderer(obj));
        }
        for (AbstractSimObject obj : sim.getDynamicObjects()) {
            this.addObjRenderer(RendererType.createRenderer(obj));
        }
        // } catch (UnsupportedSimObjectException e) {
        // // TODO Add Logger
        // System.err.println("Unsupported SimObject");
        // }
        // Create and set renderer for the driver object
        this.setDriver(new GenericVehicleRenderer(sim.getDriver()));

        SimController.getSimControl().runSimulation();
    }

}
