package dds.control.user.modules;

import java.awt.Rectangle;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.swing.JOptionPane;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import processing.core.PGraphics;
import dds.control.DDSApplet;
import dds.control.InvalidTypeException;
import dds.control.ResourceManager;
import dds.control.ResourceType;
import dds.control.simulator.MapGenerator;
import dds.control.user.AbstractModule;
import dds.control.user.ModuleType;
import dds.model.simulation.Vector3D;
import dds.model.trapeditor.BNode2D;
import dds.model.trapeditor.BezierCurve;
import dds.model.trapeditor.INode2D;
import dds.model.trapeditor.RNode2D;
import dds.model.trapeditor.TNode2D;
import dds.model.trapeditor.TrapEditorXYMapper;
import dds.view.simplegui.SimpleAbstractGUIElement;
import dds.view.simplegui.SimpleCanvas;
import dds.view.simplegui.SimpleImageContainer;
import dds.view.simplegui.SimpleTextArea;

/**
 * The TrapEditorModule class is the infamous trap editor interface.
 * 
 * @author cyberdan
 */
public class TrapEditorModule extends AbstractModule {
    /**
     * The P0_LABEL constant contains the index of the label used to display the
     * information about control point 0.
     */
    public static final int P0_LABEL = 0;

    /**
     * The P1_LABEL constant contains the index of the label used to display the
     * information about control point 1.
     */
    public static final int P1_LABEL = 1;

    /**
     * The P2_LABEL constant contains the index of the label used to display the
     * information about control point 2.
     */
    public static final int P2_LABEL = 2;

    /**
     * The P3_LABEL constant contains the index of the label used to display the
     * information about control point 3.
     */
    public static final int P3_LABEL = 3;

    /**
     * The default canvas width to use.
     */
    public static final int DEFAULT_WIDTH = ((int) (0.75 * DDSApplet
            .getApplet().width));

    /**
     * The default canvas height to use.
     */
    public static final int DEFAULT_HEIGHT = ((int) (0.75 * DDSApplet
            .getApplet().height));

    /**
     * The default x coordinate of the canvas to use.
     */
    public static final int DEFAULT_CANVAS_X = 0;

    /**
     * The default y coordinate of the canvas to use.
     */
    public static final int DEFAULT_CANVAS_Y = 0;

    /**
     * The mapCanvas is the area used to draw the map.
     */
    private transient SimpleCanvas mapCanvas;

    /**
     * The mapContainer is the container for the map image.
     */
    private transient SimpleImageContainer mapContainer;

    /**
     * The mapTree is a Document representation of the 2D map display.
     */
    private transient Document mapTree;

    /**
     * The mapImage is the created map image from the mapTree.
     */
    private transient PGraphics mapImage;

    /**
     * The mapFile.
     */
    private String mapFile;

    /**
     * The guiElements of the trap editor window.
     */
    private ArrayList<SimpleAbstractGUIElement> guiElements;

    /**
     * The tnodes is a list of TNode2D objects that is created as the 2D map
     * elements are parsed from the XML file. It is stored as a HashMap for
     * convenient and fast non-linear lookup.
     */
    private HashMap<String, TNode2D> tnodes;

    /**
     * The xyMapper is the coordinate mapper to use for all Tnode2D objects
     * associated with this trap editor.
     */
    private TrapEditorXYMapper xyMapper;

    /**
     * The bezierCurve represents the control points (interpolated) that will
     * define this trap's trajectory.
     */
    private BezierCurve bezierCurve;

    /**
     * The startTime is the time the trap starts.
     */
    private int startTime;

    /**
     * The duration is how long the trap has to execute.
     */
    private int duration;

    /**
     * Constructor for TrapEditorModule.
     * 
     * @param dapp the applet onto which this module should be loaded
     */
    public TrapEditorModule(final DDSApplet dapp) {
        super(dapp);
        startTime = 0;
        duration = 10000;
    }

    /**
     * The initGuiElements initializes all the gui components.
     */
    private void initGuiElements() {
        guiElements = new ArrayList<SimpleAbstractGUIElement>();
        tnodes = new HashMap<String, TNode2D>();
        mapCanvas = new SimpleCanvas(
                new Rectangle(TrapEditorModule.DEFAULT_CANVAS_X,
                        TrapEditorModule.DEFAULT_CANVAS_Y,
                        TrapEditorModule.DEFAULT_WIDTH,
                        TrapEditorModule.DEFAULT_HEIGHT));
        loadNewMap("map1.xml");
        guiElements.add(new SimpleTextArea("", "FFScala", 10, new Rectangle(
                TrapEditorModule.DEFAULT_CANVAS_X
                        + TrapEditorModule.DEFAULT_WIDTH + 10,
                TrapEditorModule.DEFAULT_CANVAS_Y + 10, 0, 0)));
        guiElements.add(new SimpleTextArea("", "FFScala", 10, new Rectangle(
                TrapEditorModule.DEFAULT_CANVAS_X
                        + TrapEditorModule.DEFAULT_WIDTH + 10,
                TrapEditorModule.DEFAULT_CANVAS_Y + 22, 0, 0)));
        guiElements.add(new SimpleTextArea("", "FFScala", 10, new Rectangle(
                TrapEditorModule.DEFAULT_CANVAS_X
                        + TrapEditorModule.DEFAULT_WIDTH + 10,
                TrapEditorModule.DEFAULT_CANVAS_Y + 34, 0, 0)));
        guiElements.add(new SimpleTextArea("", "FFScala", 10, new Rectangle(
                TrapEditorModule.DEFAULT_CANVAS_X
                        + TrapEditorModule.DEFAULT_WIDTH + 10,
                TrapEditorModule.DEFAULT_CANVAS_Y + 46, 0, 0)));
    }

    /**
     * The setup method is called outside this class to prepare the module
     * loader.
     * 
     * @see dds.control.AbstractDDSAppletChild#setup()
     */
    public final void setup() {
        initGuiElements();
        this.bezierCurve = new BezierCurve();
    }

    /**
     * The loadNewMap loads in a new map to the trap editor. It also renders the
     * map image so it is buffered and ready to draw. This effectively prepares
     * the mapContainer and mapCanvas for rendering.
     * 
     * @param fileName the name of the new map file to look in
     */
    public final void loadNewMap(final String fileName) {
        if (mapTree != null) {
            final NodeList nlist = mapTree.getChildNodes();
            for (int n = 0; n < nlist.getLength(); n++) {
                mapTree.removeChild(nlist.item(n));
            }
        }

        mapTree = MapGenerator.getMapGen().loadMap(fileName);
        this.mapFile = fileName;
        renderMapImage();
    }

    /**
     * The renderMapImage prepares the buffered image for this map.
     */
    public final void renderMapImage() {
        DDSApplet screen = null;
        Rectangle mapBounds = null;
        Map<String, ArrayList<TNode2D>> bri = null;

        mapBounds = getMapBoundary();
        mapContainer = new SimpleImageContainer(null, new Rectangle(
                mapBounds.x, mapBounds.y, mapBounds.width, mapBounds.height));

        if (mapCanvas.elementInsideMe(mapContainer)
                || mapCanvas.elementFitsInsideMe(mapContainer)) {
            mapContainer.centerMeInside(mapCanvas);
        } else {
            mapContainer.scaleToFit(mapCanvas);
        }
        xyMapper = new TrapEditorXYMapper(mapBounds, mapContainer.getBoundary());

        screen = DDSApplet.getApplet();
        mapImage = screen.createGraphics(mapContainer.getBoundary().width,
                mapContainer.getBoundary().height, DDSApplet.JAVA2D);

        bri = findTNodes();

        mapImage.beginDraw();
        mapImage.colorMode(PGraphics.RGB);

        for (TNode2D bnode : bri.get("blocks")) {
            bnode.drawNode(mapImage);
        }
        for (TNode2D rnode : bri.get("roads")) {
            rnode.drawNode(mapImage);
        }
        for (TNode2D inode : bri.get("intersections")) {
            inode.drawNode(mapImage);
        }
        mapImage.endDraw();

        mapContainer.setImage(mapImage);
        mapCanvas.addImageContainer(mapContainer);
    }

    /**
     * The getMapBoundary method finds the bounding rectangle of the map as
     * specified in the map's XML file.
     * 
     * @return A rectangle representing the map's bounds.
     */
    private Rectangle getMapBoundary() {
        final NamedNodeMap nodeMap = mapTree.getElementsByTagName("map")
                .item(0).getAttributes();
        int xlAttrib, xuAttrib, ylAttrib, yuAttrib;

        xlAttrib = Integer.valueOf(nodeMap.getNamedItem("xl").getNodeValue());
        xuAttrib = Integer.valueOf(nodeMap.getNamedItem("xu").getNodeValue());
        ylAttrib = Integer.valueOf(nodeMap.getNamedItem("yl").getNodeValue());
        yuAttrib = Integer.valueOf(nodeMap.getNamedItem("yu").getNodeValue());

        return new Rectangle(xlAttrib, ylAttrib, xuAttrib - xlAttrib, yuAttrib
                - ylAttrib);
    }

    /**
     * The findTNodes methods parses the map document and initializes the
     * 'tnodes' list of this module.
     * 
     * @return a mapping of roads, blocks, and intersections in 'tnodes'
     */
    private Map<String, ArrayList<TNode2D>> findTNodes() {
        final Map<String, ArrayList<TNode2D>> extractedData = new HashMap<String, ArrayList<TNode2D>>();
        final NodeList children = mapTree.getChildNodes().item(0)
                .getChildNodes();

        Node currentChild;
        int index = 0;

        for (; index < children.getLength(); ++index) {
            currentChild = children.item(index);
            if (!currentChild.getNodeName().equals("#text")) {
                try {
                    parseNode(currentChild);
                } catch (InvalidTypeException e) {
                    System.err.println(e.getMessage());
                }
            }
        }

        try {
            extractedData.put("roads", extractRoads());
            extractedData.put("intersections", extractIntersections());
            extractedData.put("blocks", extractBlocks());
        } catch (ClassNotFoundException e) {
            System.err.println("a fatal error occurred during abstraction!");
            System.err.println(e);
            System.exit(1);
        }

        addRoadVerts(extractedData.get("roads"));
        addBlockVerts(extractedData.get("blocks"));

        return extractedData;
    }

    /**
     * The extractBlocks method iterates though 'tnodes' adding only the BNode2D
     * objects, which represent blocks.
     * 
     * @return all the blocks in 'tnodes'
     * @throws ClassNotFoundException If the class can't be found.
     */
    private ArrayList<TNode2D> extractBlocks() throws ClassNotFoundException {
        return genericExtract(Class.forName("dds.model.trapeditor.BNode2D"));
    }

    /**
     * The extractRoads method iterates though 'tnodes' adding only the RNode2D
     * objects, which represent roads.
     * 
     * @return all the roads in 'tnodes'
     * @throws ClassNotFoundException If the class can't be found.
     */
    private ArrayList<TNode2D> extractRoads() throws ClassNotFoundException {
        return genericExtract(Class.forName("dds.model.trapeditor.RNode2D"));
    }

    /**
     * The extractBlocks method iterates though 'tnodes' adding only the INode2D
     * objects, which represent intersections.
     * 
     * @return all the intersections in 'tnodes'
     * @throws ClassNotFoundException If the class can't be found.
     */
    private ArrayList<TNode2D> extractIntersections()
            throws ClassNotFoundException {
        return genericExtract(Class.forName("dds.model.trapeditor.INode2D"));
    }

    /**
     * The genericExtract method is an abstract of the other extract... methods
     * that takes in the type of object to search for in the HashMap collection
     * 'tnodes'.
     * 
     * @param type the Class to look for
     * @return an ArrayList of objects of type 'type' that were found in
     *         'tnodes'.
     */
    private ArrayList<TNode2D> genericExtract(final Class type) {
        final ArrayList<TNode2D> ret = new ArrayList<TNode2D>();
        for (TNode2D tmp : this.tnodes.values()) {
            if (type.isInstance(tmp)) {
                ret.add(tmp);
            }
        }
        return ret;
    }

    /**
     * The addRoadVerts method iterates through the finished tnodes list and
     * fills in the road vertices to prepare them for rendering.
     * 
     * @param roads the road nodes
     */
    private void addRoadVerts(final ArrayList<TNode2D> roads) {
        for (TNode2D curNode : roads) {
            final RNode2D rnode = (RNode2D) curNode;
            final INode2D intA = (INode2D) tnodes.get("intersection:"
                    + rnode.getIntA());
            final INode2D intB = (INode2D) tnodes.get("intersection:"
                    + rnode.getIntB());

            rnode.addPoint(0, intA.getXcors()[0], intA.getYcors()[0]);
            rnode.addPoint(1, intB.getXcors()[0], intB.getYcors()[0]);
        }
    }

    /**
     * The addBlockVerts method iterates through the finished tnodes list and
     * fills in the block verts to prepare them for rendering.
     * 
     * @param blocks the block nodes
     */
    private void addBlockVerts(final ArrayList<TNode2D> blocks) {
        for (TNode2D curNode : blocks) {
            final BNode2D bnode = (BNode2D) curNode;
            final List<INode2D> blockIntersections = Collections
                    .synchronizedList(new ArrayList<INode2D>());
            final ArrayList<Float> xcorsTmp = new ArrayList<Float>();
            final ArrayList<Float> ycorsTmp = new ArrayList<Float>();
            INode2D inode = null;

            // int i;
            for (String roadName : bnode.getRoads()) {
                blockIntersections.add(((INode2D) (tnodes.get("intersection:"
                        + (((RNode2D) (tnodes.get("road:" + roadName)))
                                .getIntA())))));
                blockIntersections.add(((INode2D) (tnodes.get("intersection:"
                        + (((RNode2D) (tnodes.get("road:" + roadName)))
                                .getIntB())))));
            }

            for (int i = 0; i < blockIntersections.size(); i++) {
                inode = blockIntersections.get(i);
                for (int j = i + 1; j < blockIntersections.size(); j++) {
                    if (blockIntersections.get(j) == inode) {
                        blockIntersections.remove(j);
                    }
                }

                xcorsTmp.add(new Float(inode.getXcors()[0]));
                ycorsTmp.add(new Float(inode.getYcors()[0]));
            }

            bnode.setNpoints(xcorsTmp.size());
            bnode.setXcors(new float[bnode.getNpoints()]);
            bnode.setYcors(new float[bnode.getNpoints()]);

            for (int i = 0; i < bnode.getNpoints(); i++) {
                bnode.addPoint(i, xcorsTmp.get(i).floatValue(), ycorsTmp.get(i)
                        .floatValue());
            }
            bnode.polygonize();
        }
    }

    /**
     * The parseNode method parses an individual node from the xml map
     * specification and delegates the creation of the TNode2D task to the
     * appropriate helper method.
     * 
     * @param node a parent node type in the map file
     * @throws InvalidTypeException when an invalid type is passed in.
     */
    private void parseNode(final Node node) throws InvalidTypeException {
        if (node.getNodeName().equals("intersections")) {
            addINode2D(node);
        } else if (node.getNodeName().equals("roads")) {
            addRNode2D(node);
        } else if (node.getNodeName().equals("blocks")) {
            addBNode2D(node);
        } else {
            throw new InvalidTypeException("Invalid type for XML node!");
        }
    }

    /**
     * The addINode2D adds an intersection 2D node to 'tnodes'.
     * 
     * @param node a parent node
     */
    private void addINode2D(final Node node) {
        final NodeList children = node.getChildNodes();
        Node currentNode;
        TNode2D inode2D = null;
        int index = 0;

        for (; index < children.getLength(); ++index) {
            inode2D = new INode2D();
            currentNode = children.item(index);
            if (!currentNode.getNodeName().equals("#text")) {
                final NamedNodeMap attributes = currentNode.getAttributes();
                try {
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();
                    final float xValue = Float.parseFloat(attributes
                            .getNamedItem("x").getTextContent());
                    final float yValue = Float.parseFloat(attributes
                            .getNamedItem("y").getTextContent());

                    inode2D.setName(name);
                    inode2D.addPoint(0, xValue, yValue);
                    inode2D.setXyMapper(xyMapper);
                    inode2D.remapPoints();
                    tnodes.put("intersection:" + inode2D.getName(), inode2D);
                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                }
            }
        }
    }

    /**
     * The addRNode2D adds a road 2D node to 'tnodes'.
     * 
     * @param node a parent node
     */
    private void addRNode2D(final Node node) {
        final NodeList children = node.getChildNodes();
        Node currentNode;
        TNode2D rnode2D = null;
        int index = 0;

        for (; index < children.getLength(); ++index) {
            currentNode = children.item(index);
            if (!currentNode.getNodeName().equals("#text")) {
                final NamedNodeMap attributes = currentNode.getAttributes();
                try {
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();
                    final String intA = attributes.getNamedItem("ia")
                            .getTextContent();
                    final String intB = attributes.getNamedItem("ib")
                            .getTextContent();

                    rnode2D = new RNode2D();
                    rnode2D.setName(name);
                    ((RNode2D) rnode2D).setIntA(intA);
                    ((RNode2D) rnode2D).setIntB(intB);
                    rnode2D.setXyMapper(xyMapper);

                    tnodes.put("road:" + rnode2D.getName(), rnode2D);
                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                }
            }
        }
    }

    /**
     * The addBNode2D adds a block 2D node to 'tnodes'.
     * 
     * @param node a parent node
     */
    private void addBNode2D(final Node node) {
        final NodeList children = node.getChildNodes();
        Node currentNode;
        TNode2D bnode2D = null;
        int index = 0;

        for (; index < children.getLength(); ++index) {
            currentNode = children.item(index);
            bnode2D = new BNode2D();

            if (!currentNode.getNodeName().equals("#text")) {
                final NamedNodeMap attributes = currentNode.getAttributes();

                try {
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();
                    final String type = attributes.getNamedItem("type")
                            .getTextContent();
                    final StringTokenizer tokenizer = new StringTokenizer(
                            attributes.getNamedItem("roads").getTextContent());

                    bnode2D.setName(name);
                    ((BNode2D) bnode2D).setType(type);
                    while (tokenizer.hasMoreTokens()) {
                        final String token = tokenizer.nextToken(" ");
                        ((BNode2D) bnode2D).addRoad(token);
                    }
                    bnode2D.setXyMapper(xyMapper);
                    tnodes.put("block:" + bnode2D.getName(), bnode2D);
                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                }
            }
        }
    }

    /**
     * The drawOn method renders the user interface for the trap editor.
     * 
     * @param applet the applet onto which to render the gui
     * @see dds.control.user.AbstractModule#drawOn(dds.control.DDSApplet)
     */
    public final void drawOn(final DDSApplet applet) {
        applet.perspective();
        applet.camera();
        applet.colorMode(DDSApplet.RGB);
        applet.background(0, 0, 0);
        applet.fill(0, 0, 0);
        applet.rectMode(DDSApplet.CORNER);

        mapCanvas.draw(applet);
        bezierCurve.draw(applet, mapContainer.getBoundary().x, mapContainer
                .getBoundary().y);

        final ArrayList<Vector3D> displayCoordinates = bezierCurve
                .unmapCoordinates(xyMapper);
        for (int v = 0; v < displayCoordinates.size(); v++) {
            SimpleAbstractGUIElement plabel = null;
            final Vector3D pcoord = displayCoordinates.get(v);

            switch (v) {
            case 0:
                plabel = guiElements.get(TrapEditorModule.P0_LABEL);
                break;
            case 1:
                plabel = guiElements.get(TrapEditorModule.P1_LABEL);
                break;
            case 2:
                plabel = guiElements.get(TrapEditorModule.P2_LABEL);
                break;
            case 3:
                plabel = guiElements.get(TrapEditorModule.P3_LABEL);
                break;
            default:
                // TODO Do something here, like throw or log an error.
            }

            ((SimpleTextArea) plabel).setText("(" + pcoord.getXpos() + ", "
                    + pcoord.getYpos() + ")");
        }

        for (SimpleAbstractGUIElement g : this.guiElements) {
            g.draw(applet);
        }

        applet.noStroke();
        applet.fill(0xff444444);
        applet.rect(0, applet.height - 50, applet.width, 25);
        applet.fill(-1);
        applet.text("Back to Menu", 25, applet.height - 35);
        applet.text("Load Trap", 125, applet.height - 35);
        applet.text("Save Trap", 225, applet.height - 35);
        applet.text("Set Start Time", 325, applet.height - 35);
        applet.text("Set Duration", 425, applet.height - 35);
    }

    /**
     * The destruct cleans up any cached data so that this module can be
     * unloaded safely.
     * 
     * @see dds.control.AbstractDDSAppletChild#destruct()
     */
    public final void destruct() {
        for (SimpleAbstractGUIElement g : this.guiElements) {
            g.dispose();
        }
    }

    /**
     * The keyPressed method handles key events for this module.
     * 
     * @param key the pressed key
     * @param keyCode in caseL: (key == CODED)
     * @see dds.control.user.AbstractModule#keyPressed(char, int)
     */
    @Override
    public final void keyPressed(final char key, final int keyCode) {
        switch (key) {
        case 'w':
            this.saveTrap();
            break;
        case 's':
            // choose a start time
            this.setStartTime();
            break;
        case 'd':
            // choose a duration
            this.setDuration();
            break;
        case 'l':
            this.loadTrap();
            break;
        case 'q':
            this.returnToMenu();
            break;
        default:
        }
    }

    /**
     * The returnToMenu method is responsible for returning to the main menu
     * module.
     */
    private void returnToMenu() {
        this.getParent().loadModule(ModuleType.MAIN_MENU);
    }

    /**
     * The loadTrap method is responsible for getting a trap file name from the
     * user and then loading it into the trap editor.
     */
    private void loadTrap() {
        new Thread(new Runnable() {
            public void run() {
                final String lInput = JOptionPane
                        .showInputDialog("load trap file");
                if (lInput != null) {
                    parseTrapFile(ResourceManager.getLOADER().getFile(lInput,
                            ResourceType.SCRIPT));
                }
            }
        }).start();
    }

    /**
     * The setDuration method is responsible for getting a duration time for the
     * curve from the user and then setting it.
     */
    private void setDuration() {
        new Thread(new Runnable() {
            public void run() {
                final String dInput = JOptionPane
                        .showInputDialog("choose a duration in millisecs");
                if (dInput != null) {
                    try {
                        duration = Integer.parseInt(dInput);
                    } catch (NumberFormatException nfe) {
                        // TODO Add logger and/or handle this.
                        System.err.println(nfe);
                    }
                }
            }
        }).start();
    }

    /**
     * The setStartTime method is responsible for getting a start time for the
     * curve from the user and then setting it.
     */
    private void setStartTime() {
        new Thread(new Runnable() {
            public void run() {
                final String stInput = JOptionPane
                        .showInputDialog("choose a start time");
                if (stInput != null) {
                    try {
                        startTime = Integer.parseInt(stInput);
                    } catch (NumberFormatException nfe) {
                        // TODO Add logger and/or handle this.
                        System.err.println(nfe);
                    }
                }
            }
        }).start();
    }

    /**
     * The saveTrap method is responsible for getting a trap file name from the
     * user and then saving the current trap into the local file repository.
     */
    private void saveTrap() {
        new Thread(new Runnable() {
            public void run() {
                final String fInput = JOptionPane
                        .showInputDialog("save trap to file ");
                if (fInput != null) {
                    final String buffer = bezierCurve.writeToBuffer(xyMapper,
                            startTime, duration);
                    ResourceManager.getLOADER().writeFile(fInput,
                            ResourceType.SCRIPT, buffer);
                }
            }
        }).start();
    }

    /**
     * The parseTrapFile parses a trap file.
     * 
     * @param url The URL of the trap file to parse.
     */
    // TODO Consider refactoring this somewhere in the model for the trapeditor.
    public final void parseTrapFile(final URL url) {
        BufferedReader fileBuffer = null;

        try {
            fileBuffer = new BufferedReader(new InputStreamReader(url
                    .openStream()));
        } catch (IOException ioe) {
            return;
        }

        final ArrayList<String> urlContents = new ArrayList<String>();
        String nextLine = null;
        int lines = 0;

        if (fileBuffer != null) {
            try {
                nextLine = fileBuffer.readLine();
                while (nextLine != null) {
                    urlContents.add(nextLine);
                    ++lines;
                    nextLine = fileBuffer.readLine();
                }
                fileBuffer.close();
            } catch (IOException ioe) {
                return;
            }
        }

        bezierCurve = new BezierCurve();

        final StringTokenizer groupSplitter = new StringTokenizer(urlContents
                .get(0), ":/");
        String nextGroup = null;
        boolean calculatedStartTime = false;
        boolean done = false;

        while (!done) {
            try {
                nextGroup = groupSplitter.nextToken();
                if (nextGroup.charAt(0) == '(') {
                    final StringTokenizer coordST = new StringTokenizer(
                            nextGroup, " (,)");
                    final float x = Float.parseFloat(coordST.nextToken());
                    final float y = Float.parseFloat(coordST.nextToken());
                    final float z = Float.parseFloat(coordST.nextToken());

                    final float[] temp = new float[] { x, y };
                    xyMapper.getMappedCoordinate(temp);

                    bezierCurve.addPoint(temp[0], temp[1]);
                } else {
                    if (calculatedStartTime) {
                        duration = Integer.parseInt(nextGroup);
                        done = true;
                    } else {
                        startTime = Integer.parseInt(nextGroup);
                        calculatedStartTime = true;
                    }
                }
            } catch (Exception e) {
                System.err
                        .println("error parsing trap file: " + url.toString());
                bezierCurve = new BezierCurve();
                done = true;
            }
        }
    }

    /**
     * The mousePressed defines what actions take place when mouse buttons are
     * pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.AbstractModule#mousePressed(int, int, int)
     */
    public final void mousePressed(final int mouseX, final int mouseY,
            final int mouseButton) {
        if (mapContainer.inBounds(mouseX, mouseY)) {
            float mappedX, mappedY;

            mappedX = mouseX - mapContainer.getBoundary().x;
            mappedY = mouseY - mapContainer.getBoundary().y;

            bezierCurve.addPoint(mappedX, mappedY);
        } else {
            this.checkAgainstMenu(mouseX, mouseY);
        }
    }

    /**
     * The checkAgainstMenu method is checks a set of mouse coordinates against
     * the coordinates for.
     * 
     * @param mouseX The mouse x-coordinate.
     * @param mouseY The mouse y-coordinate.
     */
    private void checkAgainstMenu(final int mouseX, final int mouseY) {
        // int width = DDSApplet.getApplet().width;
        final int height = DDSApplet.getApplet().height;
        if ((mouseY >= height - 50) && (mouseY <= height - 25)) {
            switch (mouseX / 100) {
            case 0:
                this.returnToMenu();
                break;
            case 1:
                this.loadTrap();
                break;
            case 2:
                this.saveTrap();
                break;
            case 3:
                this.setStartTime();
                break;
            case 4:
                this.setDuration();
                break;
            default:
            }
        }
    }

    /**
     * The mouseDragged method defines what actions take place when mouse motion
     * is detected while buttons are pressed.
     * 
     * @param mouseX The mouse x coordinate.
     * @param mouseY The mouse y coordinate.
     * @param mouseButton The mouse button pressed.
     * @see dds.control.user.AbstractModule#mouseDragged(int, int, int)
     */
    @Override
    public final void mouseDragged(final int mouseX, final int mouseY,
            final int mouseButton) {
        this.mousePressed(mouseX, mouseY, mouseButton);
    }

    /**
     * Getter for mapFile.
     * 
     * @return the mapFile
     */
    public final String getMapFile() {
        return mapFile;
    }

    /**
     * Setter for mapFile.
     * 
     * @param fileName the mapFile to set
     */
    public final void setMapFile(final String fileName) {
        this.mapFile = fileName;
    }
}
