package dds.model.simulation;

/**
 * The AbstractDynamicObject class is used to model objects which are dynamic
 * within the context of a simulation. This means that they may change in some
 * aspect during the course of a simulation. It is important to make this kind
 * of distinction because this can help increase performance during simulation
 * runtime.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractDynamicObject extends AbstractSimObject {

    /**
     * The NULL_STATE is mapped to null and should be used when a
     * DynamicObject's state needs to be null.
     */
    public static final Object NULL_STATE = null;

    /**
     * The state object keeps track of any pertinent information that this
     * object needs to keep track of to handle itself within the context of the
     * simulation.
     */
    private Object state;

    /**
     * The timeSinceChange was implemented as an accumulator for time passing
     * since this object was last updated.
     */
    private float timeSinceChange;

    /**
     * Constructor for AbstractDynamicObject.
     * 
     * @param idString The UNIQUE ID for this object within the simulation.
     */
    public AbstractDynamicObject(final String idString) {
        super(idString);
        this.state = NULL_STATE;
        this.timeSinceChange = 0;
    }

    /**
     * Getter for state.
     * 
     * @return the state
     */
    public final Object getState() {
        return state;
    }

    /**
     * Getter for timeSinceChange.
     * 
     * @return the timeSinceChange
     */
    public final float getTimeSinceChange() {
        return timeSinceChange;
    }

    /**
     * Setter for state.
     * 
     * @param stateObj the state to set
     */
    public final void setState(final Object stateObj) {
        this.state = stateObj;
    }

    /**
     * Setter for timeSinceChange.
     * 
     * @param time the timeSinceChange to set
     */
    public final void setTimeSinceChange(final float time) {
        this.timeSinceChange = time;
    }

    /**
     * The simulate method is responsible for computing any change that this
     * object may experience during a specified amount time.
     * 
     * @param time The change in time to simulate.
     */
    protected abstract void simulate(final float time);

}
