package dds.model.simulation;

/**
 * The AbstractMovableObject class represents a set of DynamicSimObjects taht
 * are designed to move using Kinetics.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractMovableObject extends AbstractDynamicObject {

    /**
     * Constructor for AbstractMovableObject.
     * 
     * @param idString The UNIQUE ID for this object within the simulation.
     */
    public AbstractMovableObject(final String idString) {
        super(idString);
        this.lastPosition = new Vector3D();
        // Create and set the Kinetics for this object.
        this.kinetics = new Kinetics();
    }

    /**
     * The kinetics stores a Kinetics object which is responsible for modelling
     * physics for this moveable object.
     */
    private Kinetics kinetics;

    /**
     * The lastIntersectionCrossed is a reference to the last Intersection this
     * object has crossed.
     */
    private Intersection lastIntersection = null;

    /**
     * The atIntersection field is a flag that reports if this object is at an
     * Intersection.
     */
    // TODO Consider turning this into a method rather than a field.
    private boolean atIntersection = false;

    /**
     * The driverHitBuilding field flags whether or not this object has collided
     * with a building.
     */
    // TODO Consider turning this into a method instead of a field.
    private boolean driverHitBuilding = false;

    /**
     * The driverHitCar field flags whether or not this object has collided with
     * a car.
     */
    // TODO Consider turning this into a method instead of a field.
    private boolean driverHitCar = false;

    /**
     * The objectUnderMe field stores a reference to the object that is
     * currently under this object.
     */
    private AbstractSimObject objectUnderMe = null;

    /**
     * The lastPosition field keeps track of the last position of this object.
     */
    private final Vector3D lastPosition;

    /**
     * The getter for lastPosition.
     * 
     * @return the lastPosition.
     */
    public final Vector3D getLastPosition() {
        return lastPosition;
    }

    /**
     * The setter for lastPosition.
     * 
     * @param position The lastPosition to set.
     */
    public final void setLastPosition(final Vector3D position) {
        this.lastPosition.setPos(position);
    }

    /**
     * Getter for kinetics.
     * 
     * @return the kinetics
     */
    public final Kinetics getKinetics() {
        return kinetics;
    }

    /**
     * Setter for kinetics.
     * 
     * @param kins the kinetics to set
     */
    public final void setKinetics(final Kinetics kins) {
        this.kinetics = kins;
    }

    /**
     * The getter for objectUnderMe.
     * 
     * @return The objectUnderme.
     */
    public final AbstractSimObject getObjectUnderMe() {
        return this.objectUnderMe;
    }

    /**
     * The setter for objectUnderMe.
     * 
     * @param obj The objectUnderMe to set.
     */
    public final void setObjectUnderMe(final AbstractSimObject obj) {
        this.objectUnderMe = obj;
    }

    /**
     * The accessor method for lastIntersectionCrossed.
     * 
     * @return the lastIntersectionCrossed of this object
     */
    public final Intersection getLastIntersection() {
        return lastIntersection;
    } // end of method getLastIntersectionCrossed

    /**
     * The modifier method for lastIntersectionCrossed. This also update the
     * objectUnderMe variable as necessary.
     * 
     * @param instersection the last intersection crossed
     */
    public final void setLastIntersection(final Intersection instersection) {
        this.lastIntersection = instersection;
        if (this.isAtIntersection()) {
            this.objectUnderMe = instersection;
        }
    } // end of method setLastIntersectionCrossed

    /**
     * The accessor for atIntersection.
     * 
     * @return the atIntersection for this object
     */
    public final boolean isAtIntersection() {
        return atIntersection;
    } // end of method isAtIntersection

    /**
     * The setAtIntersection is the modifier for atIntersection.
     * 
     * @param isAtIntersection updated atIntersection value.
     */
    public final void setAtIntersection(final boolean isAtIntersection) {
        this.atIntersection = isAtIntersection;
    } // end of method setAtIntersection

    /**
     * The isDriverHitBuilding is the accessor for driverHitBuilding.
     * 
     * @return the driverHitBuilding of this object
     */
    public final boolean isDriverHitBuilding() {
        return driverHitBuilding;
    } // end of method isDriverHitBuilding

    /**
     * The setDriverHitBuilding is the modifier for driverHitBuilding.
     * 
     * @param hasHitBuilding Whether or not this has hit a building.
     */
    public final void setDriverHitBuilding(final boolean hasHitBuilding) {
        this.driverHitBuilding = hasHitBuilding;
    } // end of method setDriverHitBuilding

    /**
     * The isDriverHitCar is the accessor for driverHitCar.
     * 
     * @return the driverHitCar of this object.
     */
    public final boolean isDriverHitCar() {
        return driverHitCar;
    } // end of method isDriverHitCar

    /**
     * The setDriverHitCar is the modifier for driverHitCar.
     * 
     * @param hasDriverHitCar Whether or not the driver has hit a car?
     */
    public final void setDriverHitCar(final boolean hasDriverHitCar) {
        this.driverHitCar = hasDriverHitCar;
    } // end of method setDriverHitCar

    /**
     * The move method was supposed to be used for calculating motion, but this
     * functionality should be done in simulate instead.
     * 
     * @param time The delta t for calculating motion.
     */
    public abstract void move(final float time);

    /**
     * The basicMove does a linearly interpolated move based on the kinetics of
     * this object. Then, it notifies the observers of the updated position.
     * 
     * @param time The delta t for calculating motion.
     */
    public final void basicMove(final float time) {
        // System.out.println(this.getIdStr() + ".move("+time+")");
        final Vector3D newPosition = this.getPosition().duplicate();
        this.kinetics.linearMotion(newPosition, time);
        this.setPosition(newPosition);
        this.notifyObservers(SimObjectUpdate.POSITION);
    } // end of method basicMove

    /**
     * The basicSimulate does the basic simulation operations for a moving
     * object. It simply uses this object's Kinetics to affect the position.
     * 
     * @param time The delta t for calculating motion.
     */
    public final void simulate(final float time) {
        // final Vector3D oldPosition = this.getPosition().duplicate();
        this.updateKinetics(time);
        this.move(time);

        /*
         * if (this.getPolyLoop() != null) { this.getPolyLoop().moveMe(
         * this.getPosition().getXpos() - oldPosition.getXpos(),
         * this.getPosition().getYpos() - oldPosition.getYpos()); }
         */
    } // end of method simulate

    /**
     * The updateKinetics method updates the kinetics according to the time
     * passed.
     * 
     * @param time The time passed
     */
    public abstract void updateKinetics(final float time);

    /**
     * The detectCollision detects if the driver has collided with aso. If it
     * has, then it prepares this object for collision handling.
     * 
     * @param obj the object to check collision with
     * @return true if collision; false otherwise
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    public final boolean detectCollision(final AbstractSimObject obj) {
        boolean result = false;

        // if (this.getPolyLoop() != null) {
        // if (obj.getPolyLoop() == null) {
        // obj.computePolyLoop();
        // }

        this.setAtIntersection(false);

        if (obj instanceof Road) {
            result = this.detectCollision((Road) obj);
        } else if (obj instanceof Intersection) {
            result = this.detectCollision((Intersection) obj);
        } else if (obj instanceof Block) {
            result = this.detectCollision((Block) obj);
        } else if (obj instanceof AbstractMovableObject) {
            result = this.detectCollision((AbstractMovableObject) obj);
        }

        // }
        return result;
    } // end of method detectCollision

    /**
     * The detectCollision detects if the driver has collided with aso. If it
     * has, then it prepares this object for collision handling.
     * 
     * @param road the object to check collision with
     * @return true if collision; false otherwise
     */
    public final boolean detectCollision(final Road road) {
        if (this.getPolyLoop().intersectsPoly(road.getPolyLoop())) {
            this.setObjectUnderMe(road);
        }
        return false;
    }

    /**
     * The detectCollision detects if the driver has collided with aso. If it
     * has, then it prepares this object for collision handling.
     * 
     * @param intersection the object to check collision with
     * @return true if collision; false otherwise
     */
    public final boolean detectCollision(final Intersection intersection) {
        boolean result = false;
        final SmartPolygon tmp = intersection.getPolyLoop();
        if (this.getPolyLoop().intersectsPoly(tmp)) {
            this.setAtIntersection(true);
            result = true;
        }
        return result;
    }

    /**
     * The detectCollision detects if the driver has collided with aso. If it
     * has, then it prepares this object for collision handling.
     * 
     * @param block the object to check collision with
     * @return true if collision; false otherwise
     */
    public final boolean detectCollision(final Block block) {
        boolean result = false;
        if (this.getPolyLoop().intersectsPoly(block.getPolyLoop())) {
            for (SmartPolygon nextPoly : block.getBuildingPolyLoops()) {
                if (this.getPolyLoop().intersectsPoly(nextPoly)) {
                    block.setCollidedPolyloop(nextPoly);
                    result = true;
                    break;
                }
            }
        }
        return result;
    }

    /**
     * The detectCollision detects if the driver has collided with aso. If it
     * has, then it prepares this object for collision handling.
     * 
     * @param obj the object to check collision with
     * @return true if collision; false otherwise
     */
    public final boolean detectCollision(final AbstractMovableObject obj) {
        return this.getPolyLoop().intersectsPoly(obj.getPolyLoop());
    }

} // end of class AbstractMovableObject
