package dds.model.simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Observable;

/**
 * The AbstractSimObject class is responsible for modelling the basic
 * information required for object to exist within the simulation. The most
 * basic information objects within the simulation should have is position,
 * orientation, shape, and size. Given these parameters the objects should also
 * be able to provide a way of detecting collisions with it as well as providing
 * the observable part of an observer-observable relationship. As a
 * responsibility of the observer-observable relationship this object will need
 * to send updates to all observers about any changes that occur.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public abstract class AbstractSimObject extends Observable {

    /**
     * The BOUNDING_VERTS is the number of verts used for the bounding box.
     */
    public static final int BOUNDING_VERTS = 8;

    /**
     * The idStr is this object's UNIQUE identifier for the simulation. This is
     * used to recognize and index this individual object amongst the ther
     * objects in the simulation. It is most useful for use in AIScript files
     * that control objects by the ID they have.
     */
    private String idStr;

    /**
     * The orientation is the vector describing this object's orientation within
     * the simulation. This single orientation vector may not be enough to
     * properly position the object within the simulation. It may be necessary
     * to also define an vector that describes the upward direction of this
     * object. It may also be necessary to set default orientation vectors that
     * describ the default orientation and upward direction for the object.
     * These would be used to calculate the different rotations that may have
     * been applied to this object.
     */
    private final Vector3D orientation;

    /**
     * The position is the vector describing this object's location in global
     * coordinates.
     */
    private final Vector3D position;

    /**
     * The scaleFactor is a scalar used to resize this object. However, this
     * singular scaling factor may not be entirely useful since it must scale
     * all three dimensions by the same factor. It may be prudent to use a
     * vector to describe the scaling since that would allow different dimension
     * scalings.
     */
    private float scaleFactor;

    /**
     * The verts is the main list of verts associated with this object. These
     * are useful in determining the shape and standard volume of the object.
     * They can and also are used for rendering.
     */
    private final List<Vector3D> verts;

    /**
     * The vertsCalculated is a flag to describe if the verts have been
     * calculated recently.
     */
    private boolean vertsCalculated;

    /**
     * The polyLoop.
     */
    // TODO Comment me!
    private SmartPolygon polyLoop;

    /**
     * Constructor for AbstractSimObject that initializes the structures that
     * store the vital data describing this object.
     */
    private AbstractSimObject() {
        super();
        // Initialize Vector storage.
        this.orientation = new Vector3D();
        this.position = new Vector3D();
        this.verts = new ArrayList<Vector3D>();

        // Set initialial scaling to 1.0.
        this.scaleFactor = 1.0f;

        // No calculation have occured yet so set flags to false.
        this.vertsCalculated = false;

        this.polyLoop = new SmartPolygon();
    }

    /**
     * Constructor for AbstractSimObject which forces all SimObjects to have
     * their ID set upon creation.
     * 
     * @param idString The UNIQUE ID of this object.
     */
    public AbstractSimObject(final String idString) {
        this();
        this.idStr = idString;
    }

    /**
     * Getter for the polyLoop.
     * 
     * @return The polyLoop.
     */
    public final SmartPolygon getPolyLoop() {
        return this.polyLoop;
    }

    /**
     * The setter for the polyLoop.
     * 
     * @param smartPoly The polyLoop to set.
     */
    public final void setPolyLoop(final SmartPolygon smartPoly) {
        this.polyLoop = smartPoly;
    }

    /**
     * The calculateAll method performs this object's vertex and bounding box
     * calculations. This effectively prepares this object for both rendering
     * and collision detection.
     */
    public final void calculateAll() {
        // Calculate the vertices for this object.
        this.calculateVerts();

        // TODO Comment me!
        this.computePolyLoop();
    }

    /**
     * The calculateVerts method is the method used to calculate verts for a
     * SimObject. This defines the standard procedure for vert calculation,
     * which consists of removeing the old verts, calculating the new ones and
     * then setting the verts calculated flag. The actual calculation of the
     * verts is left up to the abstract method myCalculateVerts so that actual
     * simobjects can have their own system for vert calculation. The definition
     * of this method simply ensures that the process is standardized.
     */
    // TODO Consider sending a message about vert calculation to observers.
    public final void calculateVerts() {
        // Clear any previously calculated verts.
        this.clearVerts();

        // Call on the object-specific vert calculation.
        this.myCalculateVerts();

        // Set the verts calculated flag to reflect that the verts have been
        // calculated.
        this.vertsCalculated = true;
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts. This method does NOT need to clear the older verts OR set
     * the vertsCalculated flag becuase that functionality has been standardized
     * in the calculateVerts method defined in this class. This method just
     * needs to calculate the verts. The verts that are calculate MUST be stored
     * as local coordinates. These coordinates will be offset by the current
     * position when being used for collision detection of rendering.
     */
    protected abstract void myCalculateVerts();

    /**
     * The clearVerts method clears this objects collection of verts and sets
     * the vertsCalculated flag to false. This forces getVerts to calculate the
     * verts on the next access and sets the vertsCalculated flag to notify
     * others. Like the calculateVerts method, this also relies on an abstract
     * method to handle object-specific vert clearing, which may be used to
     * reset any fields necessary.
     */
    // TODO Consider sending a message about vert calculation to observers.
    protected final void clearVerts() {
        // Clear the verts defined in this class.
        this.verts.clear();

        // Set the vertCalculated flag to false since we removed them all.
        this.vertsCalculated = false;

        // Call the object-specific vert clearing so that specific vert
        // implementations can be reset properly.
        this.clearOtherVerts();
    }

    /**
     * The clearOtherVerts method is reposible for clearing any object-specific
     * verts or other calculated fields that a SimObject might have generated
     * during vert calculation.
     */
    protected abstract void clearOtherVerts();

    /**
     * The destruct method is responsible for unlinking all of this object's
     * references to make garbage collection easier.
     */
    // TODO Consider using reflection to scan through the objects fields to do
    // the clearing automatically, but this needs to be weighed against any
    // consequences that may have on the objects stored within and also if it
    // would be a more expensive operation than just statically defining the
    // procedure in each subclass.
    protected abstract void destruct();

    /**
     * The detectCollision.
     * 
     * @param obj Object to detect collision against.
     * @return Return true on collision situation.
     */
    // TODO Comment detectCollision in AbstractSimObject.
    protected abstract boolean detectCollision(final AbstractSimObject obj);

    /**
     * The handleCollision method takes care of what needs to be done when a
     * collision is detected.
     * 
     * @param obj The obj to handle collision against.
     * @param data The data for handling the collision.
     */
    protected abstract void handleCollision(final AbstractSimObject obj,
            final Object data);

    /**
     * The computePolyLoop method sets up the SmartPolygon used for collision
     * detection.
     */
    protected abstract void computePolyLoop();

    /**
     * Getter for id.
     * 
     * @return the id
     */
    public final String getIdStr() {
        return this.idStr;
    }

    /**
     * Getter for orientation.
     * 
     * @return the orientation
     */
    public final Vector3D getOrientation() {
        return orientation;
    }

    /**
     * Getter for position.
     * 
     * @return the position
     */
    public final Vector3D getPosition() {
        return position;
    }

    /**
     * Getter for scaleFactor.
     * 
     * @return the scaleFactor
     */
    public final float getScaleFactor() {
        return scaleFactor;
    }

    /**
     * The getVerts.
     * 
     * @return Returns a set of vertices for this object.
     */
    public final List<Vector3D> getVerts() {
        return this.verts;
    }

    /**
     * Getter for vertsCalculated.
     * 
     * @return the vertsCalculated
     */
    protected final boolean isVertsCalculated() {
        return vertsCalculated;
    }

    /**
     * Setter for id.
     * 
     * @param str the id to set
     */
    public final void setIdStr(final String str) {
        this.idStr = str;
    }

    /**
     * Setter for orientation.
     * 
     * @param vec the orientation to set
     */
    public final void setOrientation(final Vector3D vec) {
        this.orientation.setPos(vec);

        // Notify Observers of the change.
        this.notifyObservers(SimObjectUpdate.ORIENTATION);
    }

    /**
     * Setter for orientation.
     * 
     * @param xCoord The x-axis component for the orientation.
     * @param yCoord The y-axis component for the orientation.
     * @param zCoord The z-axis component for the orientation.
     */
    public final void setOrientation(final float xCoord, final float yCoord,
            final float zCoord) {
        // Set the orientation field.
        this.orientation.setPos(xCoord, yCoord, zCoord);

        // TODO Determine if we should update the polyLoop here.

        // Notify Observers of the change.
        this.notifyObservers(SimObjectUpdate.ORIENTATION);
    }

    /**
     * Setter for position.
     * 
     * @param pos the position to set
     */
    public final void setPosition(final Vector3D pos) {
        this.setPosition(pos.getXpos(), pos.getYpos(), pos.getZpos());
    }

    /**
     * A setter for the Position.
     * 
     * @param objX The x-coordinate.
     * @param objY The y-coordinate.
     * @param objZ The z-coordinate.
     */
    public final void setPosition(final float objX, final float objY,
            final float objZ) {
        // Change the position field.
        this.position.setPos(objX, objY, objZ);

        // Update the polyLoop.
        this.polyLoop.setOrigin(objX, objY, objZ);

        // Notify Observers of the change.
        this.notifyObservers(SimObjectUpdate.POSITION);
    }

    /**
     * Setter for scaleFactor.
     * 
     * @param scale the scaleFactor to set
     */
    public final void setScaleFactor(final float scale) {
        this.scaleFactor = scale;
    }

    /**
     * Setter for verts.
     * 
     * @param newVerts the verts to set
     */
    protected final void setVerts(final Collection<Vector3D> newVerts) {
        // Check to make sure we'r not trying to set the verts to null.
        if (newVerts != null) {
            // Clear out the old verts.
            this.verts.clear();

            // Add the new verts.
            this.verts.addAll(newVerts);
        }
    }

    /**
     * Setter for vertsCalculated.
     * 
     * @param calculated the vertsCalculated to set
     */
    protected final void setVertsCalculated(final boolean calculated) {
        this.vertsCalculated = calculated;
    }

    /**
     * The addVert method adds a single verts to the list of verts for this
     * object. Currently, the vert being added must NOT be in the collection of
     * verts already. However, this may not be necessary or useful. It could
     * potentially add extra computation to vert calculations.
     * 
     * @param vert The vert to add to the list
     */
    // TODO Make sure that we want to check for relatively equals vertsalready
    // existing in the collection before proceding.
    protected final void addVert(final Vector3D vert) {
        // Check to see if the a vert at a position that's reasonably the same
        // as another vert in the vert collection.
        if (!this.getVerts().contains(vert)) {
            // It's not so add it to the collection.
            this.getVerts().add(vert);
        }
    }

    /**
     * The addVerts adds a list of verts to the list of verts for this object.
     * IT does this by adding the verts individually and by going through the
     * addVert method that allow a single vert to be added if it doesn't already
     * exist in the collection.
     * 
     * @param newVerts The vert list to add.
     */
    protected final void addVerts(final List<Vector3D> newVerts) {
        // Iterate through the verts passed in.
        for (Vector3D vert : newVerts) {
            // Add the individual vert using the addVert(Vector3D) method.
            this.addVert(vert);
        }
    }

    /**
     * The notifyObservers method is part of the Observer-Observable contract.
     * This is the method through which updates to observers are sent. Currently
     * it forces the update it receives to be sent immediately. This may not be
     * desireable if the object is able to collect updates and then pass them
     * all at once. This needs to be checked and dealt with appropriately.
     * 
     * @param arg The argument for the update.
     * @see java.util.Observable#notifyObservers(java.lang.Object)
     */
    @Override
    public final void notifyObservers(final Object arg) {
        // Set this observable's state to changed.
        this.setChanged();

        // Call for observers to be notified IMMEDIATELY.
        // TODO Determine if updates can be queued for a single release or if
        // they
        // really need to be sent out immediately. This needs to be determined
        // because it could potentially reduce computation of updates.
        super.notifyObservers(arg);
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object. Currently this is only used for determining
     * whether or not to draw this object in the renderer.
     * 
     * @return The List of bounding verts.
     */
    // TODO Incorporate this into the collision detection and SmartPolygon.
    public abstract List<Vector3D> getBoundingVerts();

    /**
     * The moveMe method alters the position of this SimObject by adding the
     * parameters specified to the current position.
     * 
     * @param deltaX The position change along the x-axis.
     * @param deltaY The position change along the y-axis.
     * @param deltaZ The position change along the z-axis.
     */
    public final void moveMe(final float deltaX, final float deltaY,
            final float deltaZ) {
        final Vector3D currentPosition = this.getPosition();
        final float newX = currentPosition.getXpos() + deltaX;
        final float newY = currentPosition.getYpos() + deltaY;
        final float newZ = currentPosition.getZpos() + deltaZ;
        this.setPosition(newX, newY, newZ);
    }

}
