package dds.model.simulation;

/**
 * The AbstractVehicle class defines a set of MovableObjects that can be driven
 * according to the Drivable interface specification.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractVehicle extends AbstractMovableObject implements
        Drivable {

    // TODO Make certain there aren't any methods defined here which actually
    // belong in the Drivable interface.

    // TODO Consider adding functionality for managing fuel so that it's easier
    // for AI scripts to manage fuel. Perhaps even include some boolean flags
    // that affect fuel consumption and such and write supporting methods.

    // TODO Try to get the basic functionality of this class to simulate the
    // functionality of an actual vehicle.

    // TODO Fully comment this class.

    /**
     * The DEFAULT_FUEL_CAPACITY is the default amount of fuel that a vehicle
     * will contain.
     */
    public static final float DEFAULT_CAPACITY = 25;

    /**
     * The DEFAULT_TOP_SPEED is the default topspeed of a vehicle.
     */
    public static final float DEFAULT_TOP_SPEED = 120;

    /**
     * The BRAKE_ACCEL_SCALE.
     */
    private static final float BRAKE_ACCEL_SCALE = 0.5f;

    /**
     * The BRAKE_VEL_SCALE.
     */
    private static final float BRAKE_VEL_SCALE = 0.1f;

    // /**
    // * The BRAKE_INIT_SCALE.
    // */
    // private static final float BRAKE_INIT_SCALE = 0.8f;

    /**
     * The DEFAULT_MAX_ACCEL.
     */
    private static final float DEFAULT_MAX_ACCEL = 20;

    /**
     * The DEFAULT_REV_ACCEL.
     */
    private static final float DEFAULT_REV_ACCEL = 5;

    /**
     * The DEFAULT_TURN.
     */
    private static final float DEFAULT_TURN = 0.1309f; // PI / 24.0f;

    /**
     * The SPEEDOMETER_X.
     */
    private static final float SPEEDOMETER_X = -0.9f;

    /**
     * The SPEEDOMETER_Y.
     */
    private static final float SPEEDOMETER_Y = -0.4f;

    /**
     * The SPEEDOMETER_Z.
     */
    private static final float SPEEDOMETER_Z = 2.7f;

    /**
     * The ACCELOMETER_X.
     */
    private static final float ACCELOMETER_X = -1.9f;

    /**
     * The ACCELOMETER_Y.
     */
    private static final float ACCELOMETER_Y = -0.4f;

    /**
     * The ACCELOMETER_Z.
     */
    private static final float ACCELOMETER_Z = 2.7f;

    /**
     * The accelerateAmount is the magnitude to scale the orientation vector by
     * in order to get the forward component for acceleration.
     */
    private float acceleration;

    /**
     * The fuelCapacity.
     */
    private float fuelCapacity;

    /**
     * The fuelLevel.
     */
    private float fuelLevel;

    /**
     * The topSpeed.
     */
    private float topSpeed;

    /**
     * The braking.
     */
    private float braking;

    /**
     * The turnAmount is a magnitude to scale the right turn vector by to
     * compute the turning component for the acceleration vector.
     */
    private float turn;

    /**
     * The revAcceleration.
     */
    private float revAcceleration;

    /**
     * The maxAcceleration.
     */
    private float maxAcceleration;

    /**
     * The turnRadians.
     */
    private float turnRadians;

    /**
     * The speedometer.
     */
    private final Vector3D speedometer;

    /**
     * The accelometer.
     */
    private Vector3D accelometer;

    /**
     * Constructor for AbstractVehicle.
     * 
     * @param idString The UNIQUE ID for this object within the simulation.
     */
    public AbstractVehicle(final String idString) {
        super(idString);
        this.fuelCapacity = DEFAULT_CAPACITY;
        this.fuelLevel = DEFAULT_CAPACITY;
        this.topSpeed = DEFAULT_TOP_SPEED;
        this.acceleration = 0;
        this.braking = 0;
        this.maxAcceleration = DEFAULT_MAX_ACCEL;
        this.revAcceleration = DEFAULT_REV_ACCEL;
        this.turn = 0;
        this.turnRadians = DEFAULT_TURN;
        this.speedometer = new Vector3D(SPEEDOMETER_X, SPEEDOMETER_Y,
                SPEEDOMETER_Z);
        this.accelometer = new Vector3D(ACCELOMETER_X, ACCELOMETER_Y,
                ACCELOMETER_Z);
    }

    /**
     * The accelerate method causes the Driavable to maximally accelerate.
     * 
     * @see dds.model.simulation.Drivable#accelerate()
     */
    public final void accelerate() {
        this.accelerate(1);
    }

    /**
     * The basicTurn method just adds an amount to the turn.
     * 
     * @param amount The amount to turn by.
     */
    public final void basicTurn(final float amount) {
        this.turn += amount;
    }

    /**
     * The brake method causes the Driavable to maximally brake.
     * 
     * @see dds.model.simulation.Drivable#brake()
     */
    public final void brake() {
        this.setBraking(1);
    }

    /**
     * The getAcceleration returns the current acceleration.
     * 
     * @return The value of the acceleration.
     * @see dds.model.simulation.Drivable#getAcceleration()
     */
    public final float getAcceleration() {
        return this.acceleration;
    }

    /**
     * Getter for fuelCapacity.
     * 
     * @return the fuelCapacity
     */
    public final float getFuelCapacity() {
        return fuelCapacity;
    }

    /**
     * Getter for fuelLevel.
     * 
     * @return the fuelLevel
     */
    public final float getFuelLevel() {
        return fuelLevel;
    }

    /**
     * Getter for topSpeed.
     * 
     * @return the topSpeed
     */
    public final float getTopSpeed() {
        return topSpeed;
    }

    /**
     * The getTurn gets the current turn value.
     * 
     * @return the current turn value.
     * @see dds.model.simulation.Drivable#getTurn()
     */
    public final float getTurn() {
        return this.turn;
    }

    /**
     * Setter for fuelCapacity.
     * 
     * @param capacity the fuelCapacity to set
     */
    public final void setFuelCapacity(final float capacity) {
        this.fuelCapacity = capacity;
    }

    /**
     * Setter for fuelLevel.
     * 
     * @param level the fuelLevel to set
     */
    public final void setFuelLevel(final float level) {
        this.fuelLevel = level;
    }

    /**
     * Setter for topSpeed.
     * 
     * @param speed the topSpeed to set
     */
    public final void setTopSpeed(final float speed) {
        this.topSpeed = speed;
    }

    /**
     * The setTurn stes the turn amount.
     * 
     * @param amount The amount to set the turn to.
     * @see dds.model.simulation.Drivable#setTurn(float)
     */
    public final void setTurn(final float amount) {
        // System.out.println("amount: " + amount);
        if (amount > 1) {
            this.turn = 1;
        } else if (amount < -1) {
            this.turn = -1;
        } else {
            this.turn = amount;
        }
    }

    /**
     * The basicUpdateKinetics.
     * 
     * @param time The time change.
     */
    public final void basicUpdateKinetics(final float time) {
        // aVec.setPos(this.getRightNormal());
        // aVec.scale(this.turn);
        this.getKinetics().getVelocity().rotateZ(this.getTurnAmount());
        this.getOrientation().rotateZ(this.getTurnAmount());
        // final Vector3D aVec = this.getKinetics().getAcceleration();
        if (this.getVelocity() > Kinetics.mphToMps(this.topSpeed)) {
            this.setAcceleration(0);
        }
        if (this.isAccelerating()) {
            final Vector3D forward = this.getOrientation().duplicate();
            // aVec.add(forward);
            // aVec.normalize();
            forward.scale(Kinetics.mphToMps(this.getAccelerationAmount()));
            this.getKinetics().setAcceleration(forward);
        } else {
            this.getKinetics().getAcceleration().clear();
        }
        if (this.isBraking()) {
            this.getKinetics().getAcceleration().scale(
                    1.0f - (BRAKE_ACCEL_SCALE * this.braking));
            this.getKinetics().getVelocity().scale(
                    1.0f - (BRAKE_VEL_SCALE * this.braking));
        }
    }

    /**
     * The getVelocity returns the current velocity in m/s.
     * 
     * @return The current velocity in m/s.
     */
    private float getVelocity() {
        return this.getKinetics().getVelocity().magnitude();
    }

    /**
     * The getter for the turnAmount.
     * 
     * @return The turnAmount.
     */
    // TODO Really comment me!
    public abstract float getTurnAmount();

    /**
     * The basicGetTurnAmount method provides the most basic way to get the
     * turnAmount.
     * 
     * @return The basic turnAmount.
     */
    // TODO Really comment me!
    public final float basicGetTurnAmount() {
        float result = 0;
        if (Math.round(this.getVelocity()) != 0) {
            result = this.turnRadians * this.turn;
        }
        return result;
    }

    /**
     * The getAccelerationAmount.
     * 
     * @return The accelerationAmount.
     */
    // TODO Really comment me!
    public abstract float getAccelerationAmount();

    /**
     * The basicGetAccelerationAmount.
     * 
     * @return The basic accelerationAmount.
     */
    // TODO Really comment me!
    public final float basicGetAccelerationAmount() {
        float result = 0;
        if (this.acceleration < 0) {
            result = this.acceleration * this.revAcceleration;
        } else if (this.acceleration > 0) {
            result = this.acceleration * this.maxAcceleration;
        }
        return result;
    }

    /**
     * The reorient method reorients this object with its velocity.
     */
    public final void reorient() {
        final Vector3D newOVec = this.getKinetics().getVelocity().duplicate();

        if (newOVec.magnitude() != 0) {
            newOVec.normalize();
            this.setOrientation(newOVec);

//            if (this.getPolyLoop() != null) {
                final Vector3D tmp = new Vector3D(1, 0, 0);
                float theta = 0f;

                theta = (float) Math.acos(tmp.dotProduct(newOVec));
                // special cases 3rd & 4th quadrant:
                if (newOVec.getYpos() >= 0) {
                    theta = (((float) (Math.PI * 2f)) - theta);
                }

                // System.out.println("theta: " + theta);
                this.getPolyLoop().rotateMe(theta);
//            }
            // System.out.println("orientation: " + this.getOrientation());
            this.notifyObservers(SimObjectUpdate.ORIENTATION);
        }
    }

    /**
     * The getRightNormal.
     * 
     * @return The right normal for this object's current orientation
     */
    public final Vector3D getRightNormal() {
        final Vector3D result = Vector3D.cross(this.getOrientation(), Vector3D
                .getZAxis());
        result.normalize();
        result.scale(-1);
        return result;
    }

    /**
     * The isAccelerating method returne id this object is accelerating or not.
     * 
     * @return True if it's accelerating.
     * @see dds.model.simulation.Drivable#isAccelerating()
     */
    public final boolean isAccelerating() {
        return (this.acceleration != 0);
    }

    /**
     * The setAcceleration.
     * 
     * @param accel The acceleration amount as a decimal from -1 to 1.
     * @see dds.model.simulation.Drivable#setAcceleration(float)
     */
    public final void setAcceleration(final float accel) {
        if (accel > 1) {
            this.acceleration = 1;
        } else if (accel < -1) {
            this.acceleration = -1;
        } else {
            this.acceleration = accel;
        }
    }

    /**
     * The stopBraking.
     */
    public final void stopBraking() {
        this.setBraking(0);
    }

    /**
     * The stopAcceleration.
     * 
     * @see dds.model.simulation.Drivable#stopAcceleration()
     */
    public final void stopAcceleration() {
        this.setAcceleration(0);
    }

    /**
     * Getter for braking.
     * 
     * @return the braking
     */
    public final boolean isBraking() {
        return (this.braking > 0);
    }

    /**
     * Setter for braking.
     * 
     * @param amount the braking to set
     */
    public final void setBraking(final float amount) {
        if (amount > 1) {
            this.braking = 1;
        } else if (amount < 0) {
            this.braking = 0;
        } else {
            this.braking = amount;
        }
    }

    /**
     * Getter for maxAcceleration.
     * 
     * @return the maxAcceleration
     */
    public final float getMaxAcceleration() {
        return maxAcceleration;
    }

    /**
     * Setter for maxAcceleration.
     * 
     * @param accel the maxAcceleration to set
     */
    public final void setMaxAcceleration(final float accel) {
        this.maxAcceleration = accel;
    }

    /**
     * Getter for revAcceleration.
     * 
     * @return the revAcceleration
     */
    public final float getRevAcceleration() {
        return revAcceleration;
    }

    /**
     * Setter for revAcceleration.
     * 
     * @param accelAmount the revAcceleration to set
     */
    public final void setRevAcceleration(final float accelAmount) {
        this.revAcceleration = accelAmount;
    }

    /**
     * Getter for turnRadians.
     * 
     * @return the turnRadians
     */
    public final float getTurnRadians() {
        return turnRadians;
    }

    /**
     * Setter for turnRadians.
     * 
     * @param radians the turnRadians to set
     */
    public final void setTurnRadians(final float radians) {
        this.turnRadians = radians;
    }

    /**
     * Getter for braking.
     * 
     * @return the braking
     */
    public final float getBraking() {
        return braking;
    }

    /**
     * Getter for speedometerPosition.
     * 
     * @return the speedometerPosition
     */
    public final Vector3D getSpeedometer() {
        return speedometer;
    }

    /**
     * Setter for speedometerPosition.
     * 
     * @param position the speedometerPosition to set
     */
    public final void setSpeedometer(final Vector3D position) {
        this.speedometer.setPos(position);
    }

    /**
     * The getSpeed.
     * 
     * @return The speed.
     */
    public final float getSpeed() {
        return this.getKinetics().getVelocity().magnitude();
    }

    /**
     * Getter for accelometerPosition.
     * 
     * @return the accelometerPosition
     */
    public final Vector3D getAccelometer() {
        return accelometer;
    }

    /**
     * Setter for accelometerPosition.
     * 
     * @param position the accelometerPosition to set
     */
    public final void setAccelometer(final Vector3D position) {
        this.accelometer = position;
    }

}
