package dds.model.simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * The Block class represents the modelled aspects of a city block within the
 * context of the Simulation.
 * 
 * @author Jacob Bolton
 */
public class Block extends AbstractSimObject {

    /**
     * The THREE constant has a value of 3.
     */
    private static final int THREE = 3;

    /**
     * The buildingList field stores a list of the Buildings attached to this
     * Block.
     */
    private ArrayList<Building> buildingList;

    /**
     * The buildingPolyLoops.
     */
    // TODO Comment the buildingPolyLoops in Block.
    // TODO Consider moving this into the Building class.
    private transient ArrayList<SmartPolygon> buildingPolyLoops = null;

    /**
     * The buildings field stores the number of Buildings that should be placed
     * on this Block on a per Road basis.
     */
    private int buildings;

    /**
     * The collidedPolyloop.
     */
    // TODO Comment the collidedPolyloop in Block.
    private SmartPolygon collidedPolyloop = null;

    /**
     * The intersections is a CCW-ordered collection of Intersections that
     * inscribe this Block.
     */
    private List<Intersection> intersections;

    /**
     * The roadDirections is a Map of Roads to a Boolean that describes if a
     * Road is in the Block's loop as an AB edge (true) or as a BA edge (false).
     */
    private Map<Road, Boolean> roadDirections;

    /**
     * The roads is an CCW-ordered collection of Roads that inscribe this Block.
     */
    private final List<Road> roads;

    /**
     * The textureFile field stores the filename of the image to use to texture
     * this Block.
     */
    private String textureFile;

    /**
     * The type field stores the type of Block that this is supposed to be.
     */
    private BlockType type;

    /**
     * Constructor for Block.
     * 
     * @param idStr The XML id String.
     */
    public Block(final String idStr) {
        super(idStr);
        this.setType(BlockType.RESIDENTIAL);
        this.buildings = 1;
        this.roads = new ArrayList<Road>();
        // this.buildingVerts = new ArrayList<Vector3D>();
        this.roadDirections = new HashMap<Road, Boolean>();
        this.intersections = new ArrayList<Intersection>();
        this.buildingPolyLoops = new ArrayList<SmartPolygon>();
        this.buildingList = new ArrayList<Building>();
    }

    /**
     * Constructor for Block.
     * 
     * @param idStr The XML id String.
     * @param blockType The type of Block this will be.
     * @param numBuildings The number of buildings to be on this block.
     * @param myRoads The Collection of roads associated with this block.
     */
    public Block(final String idStr, final BlockType blockType,
            final int numBuildings, final List<Road> myRoads) {
        this(idStr);
        this.setType(blockType);
        this.buildings = numBuildings;
        this.roads.clear();
        this.roads.addAll(myRoads);
        this.sortRoads();
        this.calculatePosition();
    }

    /**
     * The addRoad method adds a new Road to this Blocks collection of Roads.
     * 
     * @param newRoad The Road to add.
     */
    public final void addRoad(final Road newRoad) {
        // TODO Ensure that all pertinent calculations are done and all
        // variables are set.
        this.roads.add(newRoad);
        this.sortRoads();
        this.calculatePosition();
    }

    /**
     * The calcBuildingVerts calculates the Buildings for this Road. It will
     * generate new Building objects to be placed on this Block along the
     * specified Road.
     * 
     * @param road The Road to add buildings to.
     * @param left True if the buildings should be on the left side, false for
     *            the right.
     */
    private void calcBuildingVerts(final Road road, final boolean left) {
        // Get the road's baseline, the line which buildings are set upon.
        final List<Vector3D> base = road.getBaseline(left);

        // Get a unit vector perpendicular to the road that point into the
        // Block.
        final Vector3D depthUnitVec;
        if (left) {
            depthUnitVec = road.getLeftNormal();
        } else {
            depthUnitVec = road.getRightNormal();
        }

        // Retrieve the Building type.
        final BuildingType buildingType = this.type.getBuildingType();

        // Set the current Building position to be at the beginning of the
        // Road's base for Buildings.
        final Vector3D currentPosition = base.get(0).duplicate();

        // Change the position coordinate from global to local.
        currentPosition.subtract(this.getPosition());

        // Calculate a directional vector along the baseline of Buildings along
        // the Road.
        final Vector3D baseUnitVec = Vector3D
                .subtract(base.get(1), base.get(0));

        // Calculate the length of the Road's base where Buildings are placed.
        final float roadDistance = baseUnitVec.magnitude();

        // Generate a normal vector in the direction of the Road.
        baseUnitVec.normalize();

        // Calculate the width of Buildings along this Road.
        float buildingWidth = roadDistance;
        if (this.buildings > 1) {
            buildingWidth /= (float) (this.buildings + 1);
        }

        // Calculate the spacing between buidings.
        final float spaceDist = buildingWidth / (float) (this.buildings - 1);

        // Calculate the depth of the Buildings along this Road.
        final float buildingDepth = buildingWidth / 2.0f;

        // Calculate the distance from the first edge of one Building to the
        // first edge of the next.
        final Vector3D buildingShift = baseUnitVec.duplicate();
        buildingShift.scale(buildingWidth + spaceDist);

        // Loop for the number of buildings.
        for (int i = 0; i < this.buildings; i++) {
            // Generate a new Building.
            this.generateBuilding(road, i, buildingType, currentPosition,
                    buildingWidth, buildingDepth, baseUnitVec, depthUnitVec);

            // Move to the next building position.
            currentPosition.add(buildingShift);
        }
    }

    /**
     * The calculatePosition method determines the local origin for vertex
     * positioning. This is based on an average of the positions used by the
     * Roads making up this Block.
     */
    private void calculatePosition() {
        // Create storage space for vector calculations.
        final Vector3D newPosition = new Vector3D();

        // Iterate through the Roads.
        for (Road road : this.roads) {
            // Sum the position of this Road with the others.
            newPosition.add(road.getPosition());
        }

        // Divide the sum to form the average.
        newPosition.scale(1.0f / (float) this.roads.size());

        // Set the position to the calculated result.
        this.setPosition(newPosition);
    }

    /**
     * The clearAllVerts method is responsible for clearing all verts that are
     * used by this object.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    protected final void clearOtherVerts() {
        // TODO Do something about the Building object attached to this Block.
        this.buildingList.clear();
    }

    /**
     * The computeBuildingPolyLoops.
     */
    // TODO Fix checkstyle errors.
    private void computeBuildingPolyLoops() {
        // TODO Consider putting this functionality withing the Building class.
        if (this.buildings == 1) {
            final float deltaX = this.getPosition().getXpos();
            final float deltaY = this.getPosition().getYpos();
            final Building newBuilding = this.getBuildingList().get(0);
            ArrayList bverts = (ArrayList) newBuilding.getGroundVerts();

            // TODO This assumption of 3 verts for a one-building block is
            // incorrect. The proper amount is the number of intersections
            // attached to the Block, which is the same as half the intersctions
            // attached to all the Roads of this Block. I suppose it could be
            // just the number of Roads.
            newBuilding.setPolyLoop(new SmartPolygon(THREE));
            newBuilding.getPolyLoop().addPoint(
                    ((Vector3D) bverts.get(0)).getXpos() + deltaX,
                    ((Vector3D) bverts.get(0)).getYpos() + deltaY);
            newBuilding.getPolyLoop().addPoint(
                    ((Vector3D) bverts.get(1)).getXpos() + deltaX,
                    ((Vector3D) bverts.get(1)).getYpos() + deltaY);
            bverts = (ArrayList) this.getBuildingList().get(1).getGroundVerts();
            newBuilding.getPolyLoop().addPoint(
                    ((Vector3D) bverts.get(1)).getXpos() + deltaX,
                    ((Vector3D) bverts.get(1)).getYpos() + deltaY);
            newBuilding.getPolyLoop().validate();
            this.getBuildingPolyLoops().add(newBuilding.getPolyLoop());

            this.getBuildingList().get(1)
                    .setPolyLoop(newBuilding.getPolyLoop());
            this.getBuildingList().get(2)
                    .setPolyLoop(newBuilding.getPolyLoop());
        } else {
            for (Building b : this.getBuildingList()) {
                b.computePolyLoop();
                this.getBuildingPolyLoops().add(b.getPolyLoop());
            }
        }
    }

    /**
     * The computePolyLoop.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    // TODO Comment me.
    protected final void computePolyLoop() {
        SmartPolygon thePoly = null;
        final float pointX = this.getPosition().getXpos();
        final float pointY = this.getPosition().getYpos();

        setPolyLoop(new SmartPolygon(this.getVerts().size()));
        thePoly = getPolyLoop();

        for (Vector3D nextVert : this.getVerts()) {
            thePoly.addPoint(pointX + nextVert.getXpos(), pointY
                    + nextVert.getYpos());
        }
    }

    /**
     * The destruct method is responsible for unlinking all of this object's
     * references to make garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    protected final void destruct() {
        this.roads.clear();
    }

    /**
     * The detectCollision.
     * 
     * @param obj SimObject to detect collision with.
     * @return True on collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    // TODO Comment detectCollision for Block.
    protected final boolean detectCollision(final AbstractSimObject obj) {
        // TODO Write detectCollision for Block.
        return false;
    }

    /**
     * The generateBuilding method creates a new Building object on this Block
     * using the specified parameters.
     * 
     * @param road The road to generate the building on.
     * @param index The building's index.
     * @param buildingType The building's type.
     * @param currentPosition The current position along the road.
     * @param buildingWidth The building's width.
     * @param buildingDepth The building's depth.
     * @param baseUnitVec The unit vector describing the dirction of the
     *            building's base walls.
     * @param depthUnitVec The unit vector describing the direction of the
     *            buildings side walls.
     */
    // TODO Come up with a way to reduce the inputs.
    private void generateBuilding(final Road road, final int index,
            final BuildingType buildingType, final Vector3D currentPosition,
            final float buildingWidth, final float buildingDepth,
            final Vector3D baseUnitVec, final Vector3D depthUnitVec) {
        // Create a new Building object.
        final Building building = new Building(road.getIdStr() + "-" + index,
                this, buildingType);

        // Set the parameters which let it calculate its verts.
        building.setPosition(currentPosition);
        building.setWidth(buildingWidth);
        building.setDepth(buildingDepth);
        building.setBaseVector(baseUnitVec);
        building.setSideVector(depthUnitVec);
        if (index == 0) {
            building.setFirstWall(false);
        }
        if (index == (this.buildings - 1)) {
            building.setSecondWall(false);
        }

        // Calculate the Building's verts.
        // TODO Consider doing this as part of the vert calculation for a Block.
        building.calculateVerts();

        // Add the finished Building to the list of Buildings for this Block.
        this.buildingList.add(building);
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.getVerts();
    }

    /**
     * Getter for buildingList.
     * 
     * @return the buildingList
     */
    public final ArrayList<Building> getBuildingList() {
        return buildingList;
    }

    /**
     * The getBuildingPolyLoops.
     * 
     * @return the poly loops for the Buildings.
     */
    public final ArrayList<SmartPolygon> getBuildingPolyLoops() {
        return buildingPolyLoops;
    }

    /**
     * Getter for buildings. py
     * 
     * @return the buildings
     */
    public final int getBuildings() {
        return buildings;
    }

    /**
     * The getCollidedPolyloop.
     * 
     * @return the colidedPolyLoop.
     */
    public final SmartPolygon getCollidedPolyloop() {
        return collidedPolyloop;
    }

    /**
     * Getter for intersections.
     * 
     * @return the intersections
     */
    public final List<Intersection> getIntersections() {
        return intersections;
    }

    /**
     * Getter for roadDirections.
     * 
     * @return the roadDirections
     */
    public final Map<Road, Boolean> getRoadDirections() {
        return roadDirections;
    }

    /**
     * Getter for roads.
     * 
     * @return the roads
     */
    public final Collection getRoads() {
        return roads;
    }

    /**
     * Getter for textureFile.
     * 
     * @return the textureFile
     */
    public final String getTextureFile() {
        return textureFile;
    }

    /**
     * Getter for type.
     * 
     * @return the type
     */
    public final BlockType getType() {
        return type;
    }

    /**
     * The handleCollision.
     * 
     * @param obj The obj.
     * @param data The data.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    // TODO Comment handleCollision for Block.
    protected void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Decide what, if anything to put here.
    }

    /**
     * The initRoadDirections method looks at ONLY the first two Roads
     * inscribing this Block and determines their orientation.
     */
    private void initRoadDirections() {
        // Clear out previously calculated fields
        this.intersections.clear();
        this.roadDirections.clear();
        final Road roadA = this.roads.get(0);
        final Road roadB = this.roads.get(1);
        final Intersection intAStart = roadA.getStart();
        final Intersection intAEnd = roadA.getEnd();
        final Intersection intBStart = roadB.getStart();
        final Intersection intBEnd = roadB.getEnd();
        // See how intersections of these Roads to determine Road orientation.
        // Also start to populate the list of Intersections that inscribe this
        // Block.
        if (intAEnd.equals(intBStart)) {
            this.roadDirections.put(roadA, true);
            this.roadDirections.put(roadB, true);
            this.intersections.add(intAStart);
            this.intersections.add(intAEnd);
            this.intersections.add(intBEnd);
        } else if (intAStart.equals(intBStart)) {
            this.roadDirections.put(roadA, false);
            this.roadDirections.put(roadB, true);
            this.intersections.add(intAEnd);
            this.intersections.add(intAStart);
            this.intersections.add(intBEnd);
        } else if (intAEnd.equals(intBEnd)) {
            this.roadDirections.put(roadA, true);
            this.roadDirections.put(roadB, false);
            this.intersections.add(intAStart);
            this.intersections.add(intAEnd);
            this.intersections.add(intBStart);
        } else if (intAStart.equals(intBEnd)) {
            this.roadDirections.put(roadA, false);
            this.roadDirections.put(roadB, false);
            this.intersections.add(intAEnd);
            this.intersections.add(intAStart);
            this.intersections.add(intBStart);
        }
        // else {
        // // TODO This branch of the if statement should exist to log
        // // errors since this should be an unreachable branch.
        // }
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts. This method does NOT need to clear the older verts OR set
     * the vertsCalculated flag becuase that functionality has been standardized
     * in the calculateVerts method defined in this class. This method just
     * needs to calculate the verts. The verts that are calculate MUST be stored
     * as local coordinates. These coordinates will be offset by the current
     * position when being used for collision detection of rendering.
     * 
     * @see dds.model.simulation.AbstractSimObject#calculateVerts()
     */
    @Override
    public final void myCalculateVerts() {
        // Iterate through the Roads to ensure all of their verts are
        // calculated.
        for (Road road : this.roads) {
            road.calculateVerts();
        }

        // Iterate through the Roads to get the proper verts for the Block's
        // edges.
        for (Road road : this.roads) {
            // Check in which direction the Road has been defined by
            // Intersections and to the Block. This is needed in order to add
            // the Block's verts in a counter clockwise manner.
            if (this.roadDirections.get(road)) {
                this.addVerts(road.getLeftVerts());
                this.calcBuildingVerts(road, true);
            } else {
                this.addVerts(road.getRightVerts());
                this.calcBuildingVerts(road, false);
            }
        }

        // Iterate through the verts to convert them from global to local.
        for (Vector3D vert : this.getVerts()) {
            vert.subtract(this.getPosition());
        }

        // Compute the building polyloops.
        // TODO Consider moving this into the Building class.
        computeBuildingPolyLoops();
    }

    /**
     * The populateRoadDirections method is called after initRoadDirections to
     * continue through the rest of the roads to determine their order.
     */
    private void populateRoadDirections() {
        // Get the first Intersection.
        final Intersection first = this.intersections.get(0);

        // Get a reference to the last intersection.
        Intersection last = this.intersections
                .get(this.intersections.size() - 1);

        // Iterate through the remaining Roads. The first two should've been
        // handled in the initRoadDirections method.
        for (int i = 2; i < this.roads.size(); i++) {
            // Get a reference to the next Road.
            final Road next = this.roads.get(i);

            // Retrieve the starting and ending verts of the current Road.
            final Intersection start = next.getStart();
            final Intersection end = next.getEnd();

            // Determine how this Road attaches to the the current loop of Roads
            // inscribed by Intersections and get the appropriate next
            // Intersection in order.
            if (last.equals(start)) {
                last = end;
                this.roadDirections.put(next, true);
            } else if (last.equals(end)) {
                last = start;
                this.roadDirections.put(next, false);
            }
            // else {
            // // TODO This branch of the if statement should exist to log
            // // errors since this should be an unreachable branch.
            // }

            // Check to make sure that the chosen Intersection is not the first
            // one.
            if (!first.equals(last)) {
                // It's not so add it to the list.
                this.intersections.add(last);
            }
        }
    }

    /**
     * Setter for buildingList.
     * 
     * @param list the buildingList to set
     */
    public final void setBuildingList(final ArrayList<Building> list) {
        this.buildingList = list;
    }

    /**
     * Setter for buildings.
     * 
     * @param numOf the buildings to set
     */
    public final void setBuildings(final int numOf) {
        this.buildings = numOf;
    }

    /**
     * The setCollidedPolyloop.
     * 
     * @param polyloop the collidedPolyloop.
     */
    public final void setCollidedPolyloop(final SmartPolygon polyloop) {
        this.collidedPolyloop = polyloop;
    }

    /**
     * Setter for intersections.
     * 
     * @param ints the intersections to set
     */
    public final void setIntersections(final List<Intersection> ints) {
        this.intersections = ints;
    }

    /**
     * The setRoadAdjacency method tells the Roads inscribing this Block which
     * side the Block is with respect to their orientation.
     */
    private void setRoadAdjacency() {
        // Iterate through all the Roads.
        for (Road road : this.roads) {
            // Check the Road's direction to determine which side this Block is
            // on.
            if (this.roadDirections.get(road)) {
                road.setLeftBlock(this);
            } else {
                road.setRightBlock(this);
            }
        }
    }

    /**
     * Setter for roadDirections.
     * 
     * @param directions the roadDirections to set
     */
    public final void setRoadDirections(final Map<Road, Boolean> directions) {
        this.roadDirections = directions;
    }

    /**
     * Setter for roads.
     * 
     * @param adjRoads the roads to set
     */
    public final void setRoads(final List<Road> adjRoads) {
        this.roads.clear();
        this.roads.addAll(adjRoads);
        this.calculatePosition();
    }

    /**
     * Setter for textureFile.
     * 
     * @param filename the textureFile to set
     */
    public final void setTextureFile(final String filename) {
        this.textureFile = filename;
    }

    /**
     * Setter for type.
     * 
     * @param blockType the type to set
     */
    public final void setType(final BlockType blockType) {
        this.type = blockType;
        this.textureFile = blockType.pickTexture();
    }

    /**
     * The sortRoads method looks at all the roads associated with this Block
     * assuming they've been added in CCW order. It analizes their data and
     * determines if the Roads' Intersections are ordered AB or BA.
     */
    private void sortRoads() {
        this.initRoadDirections();
        this.populateRoadDirections();
        this.setRoadAdjacency();
    }

    /**
     * The toString.
     * 
     * @return info is a string containing the information for this Block
     *         object.
     * @see java.lang.Object#toString()
     */
    @Override
    public final String toString() {

        String info = "";

        info = "ID String: " + this.getIdStr() + "\n";
        info = info + "# Buildings: " + this.getBuildings() + "\n";
        info = info + "Roads: ";

        for (Road road : roads) {
            info = info + road.getIdStr() + " ";
        } // end for-each

        info = info + "\n";

        return info;

    } // end method toString()
} // end class definition Block
