/**
 * BlockType.java.
 */
package dds.model.simulation;

import java.util.Locale;
import java.util.Random;

/**
 * The BlockType enumeration defines the different types of Blocks that can
 * exist within the simulation.
 * 
 * @author Jacob Bolton
 */
public enum BlockType {
    /**
     * The RESIDENTIAL enumeration defines the properties for a residential
     * block.
     */
    RESIDENTIAL(0xff00c800, BuildingType.RESIDENTIAL, new String[] {
            "grass.jpg", "residential_ground.jpg" }),

    /**
     * The INDUSTRIAL enumeration defines the properties for an industrial
     * block.
     */
    INDUSTRIAL(0xffc8c800, BuildingType.INDUSTRIAL,
            new String[] { "industrial_ground.jpg" }),

    /**
     * The COMMERCIAL enumeration defines the properties for a commercial block.
     */
    COMMERCIAL(0xff0000c8, BuildingType.COMMERCIAL,
            new String[] { "concrete.jpg" }),

    /**
     * The USER enumeration defines the properties for a user-defined block.
     */
    USER(0xff00c8c8, BuildingType.USER, new String[] { "park.jpg" }),

    /**
     * The NONE enumeration defines the properties for a residential block.
     */
    NONE(0xff646464, BuildingType.NONE, new String[] { "broken_concrete.jpg" });

    /**
     * The textures field stores an array of image filenames that can be used to
     * texture a Block of this BlockType.
     */
    private String[] textures;

    /**
     * The color field contains the color to use to fill a Block with if an
     * texture file is not specified or loaded.
     */
    private int color;

    /**
     * The buildingType field stores a reference to the BuildingType that will
     * appear on this block.
     */
    // TODO At a later stage, we may want to have a random selection of
    // different building types.
    private BuildingType buildingType;

    /**
     * The random field stores a random number generator that will be used to
     * randomize the texture picked for an instance of a Block of this type.
     */
    private Random random;

    /**
     * Constructor for BlockType.
     * 
     * @param defaultColor The default color this BlockType.
     * @param type The BuildingType for buildings that may appear on a Block of
     *            this BlockType.
     * @param textureFilenames The array of texture image files that can be used
     *            for this BlockType.
     */
    private BlockType(final int defaultColor, final BuildingType type,
            final String[] textureFilenames) {
        this.textures = textureFilenames;
        this.color = defaultColor;
        this.random = new Random();
        this.buildingType = type;
    }

    /**
     * The pickTexture method randomly selects a texture file from this
     * BuildingType's set of available texture files.
     * 
     * @return A String containing the randomly selected texture file name.
     */
    public String pickTexture() {
        // Assume that we won't return a texture file name.
        String result = null;

        // Check to see if we have defined any textures for this BlockType.
        if (this.textures.length > 0) {
            // Randomly pick a filename from the defined texture files.
            result = this.textures[random.nextInt(this.textures.length)];
        }

        // Return the result.
        return result;
    }

    /**
     * Getter for color.
     * 
     * @return the color
     */
    public final int getColor() {
        return color;
    }

    /**
     * The parseType method tries to resolve an input string to a BlockType.
     * 
     * @param string The String to attemp resolution.
     * @return The corresponding BlockType or null if it cannot be resolved.
     */
    public static BlockType parseType(final String string) {
        // Clean up the provided String so we can compare it with the default
        // toString for enumerations.
        final String lower = string.toUpperCase(Locale.getDefault());

        // Assume that the resulting BlockType is NONE.
        BlockType result = NONE;

        // Iterate through the defined BlockTypes.
        for (BlockType type : BlockType.values()) {
            // Compare the current BlockType's string with the one provided.
            if (lower.equals(type.toString())) {
                // They're the same meaning this is the BlockType we want to
                // return as a result.
                result = type;
                break;
            }
        }

        // Return the result.
        return result;
    }

    /**
     * The getBuildingType is an accessor for the buildingType field.
     * 
     * @return The buildingType field.
     */
    public BuildingType getBuildingType() {
        return this.buildingType;
    }

}
