/**
 * Building.java.
 */
package dds.model.simulation;

import java.util.ArrayList;
import java.util.List;

/**
 * The Building class is responsible for modelling a building in the simulation.
 * 
 * @author Jacob Bolton
 */
public class Building extends AbstractSimObject {

    /**
     * The NO_PARENT constant is an easy-to-read way to set the parent field to
     * null.
     */
    private static final Block NO_PARENT = null;

    /**
     * The NO_TEXTURE constant is an easy-to-read way to set the texture field
     * to null.
     */
    private static final String NO_TEXTURE = null;

    /**
     * The NO_TYPE constant is an easy-to-read way to set the type field to
     * null.
     */
    private static final BuildingType NO_TYPE = null;

    /**
     * The WALL_VERTS constant indicates how many verts are required to inscribe
     * a wall.
     */
    private static final int WALL_VERTS = 4;

    /**
     * The FIRST constant denotes the first vert in a vert array defining a
     * wall.
     */
    private static final int FIRST = 0;

    /**
     * The SECOND constant denotes the second vert in a vert array defining a
     * wall.
     */
    private static final int SECOND = 1;

    /**
     * The THIRD constant denotes the third vert in a vert array defining a
     * wall.
     */
    private static final int THIRD = 2;

    /**
     * The FOURTH constant denotes the fourth vert in a vert array defining a
     * wall.
     */
    private static final int FOURTH = 3;

    /**
     * The parent field is a reference to the Block which this Building belongs
     * to.
     */
    private Block parent;

    /**
     * The dimensions field stores the width, height, and depth of this
     * building. The width is in the x-coordinate, the height in the
     * y-coordinate, and the depth in the z-coordinate.
     */
    private final Vector3D dimensions;

    /**
     * The baseVector field stores the directional unit vector for the base of
     * the Building.
     */
    private Vector3D baseVector;

    /**
     * The sideVector field stores the directional unit vector for the side of
     * the Building.
     */
    private Vector3D sideVector;

    /**
     * The firstWall field flags if the Building should have a 'first' side
     * wall.
     */
    private boolean firstWall;

    /**
     * The secondWall field flags if the Building should have a 'second' side
     * wall.
     */
    private boolean secondWall;

    /**
     * The texture field stores the name of the image file to use to texture the
     * walls of this building.
     */
    private String textureFile;

    /**
     * The type field stores the BuildingType for this Building.
     */
    private BuildingType type;

    /**
     * The groundVerts field contains the set of verts that inscribe the ground
     * of this Building.
     */
    private List<Vector3D> groundVerts;

    /**
     * The firstWallVerts field contains the set of verts that inscribe the
     * first wall of this Building.
     */
    private List<Vector3D> firstWallVerts;

    /**
     * The secondWallVerts field contains the set of verts that inscribe the
     * second wall of this Building.
     */
    private List<Vector3D> secondWallVerts;

    /**
     * The frontWallVerts field contains the set of verts that inscribe the
     * front wall of this Building.
     */
    private List<Vector3D> frontWallVerts;

    /**
     * The backWallVerts field contains the set of verts that inscribe the back
     * wall of this Building.
     */
    private List<Vector3D> backWallVerts;

    /**
     * The ceilingVerts field contains the set of verts that inscribe the
     * ceiling of this Building.
     */
    private List<Vector3D> ceilingVerts;

    /**
     * Constructor for Building.
     * 
     * @param idString The UNIQUE ID of this object.
     */
    private Building(final String idString) {
        super(idString);
        this.parent = NO_PARENT;
        this.textureFile = NO_TEXTURE;
        this.type = NO_TYPE;
        this.dimensions = new Vector3D();
        this.baseVector = new Vector3D();
        this.sideVector = new Vector3D();
        this.firstWall = true;
        this.secondWall = true;
        this.groundVerts = new ArrayList<Vector3D>();
        this.ceilingVerts = new ArrayList<Vector3D>();
        // this.firstWallVerts = new Vector3D[WALL_VERTS];
        this.firstWallVerts = new ArrayList<Vector3D>(WALL_VERTS);
        this.secondWallVerts = new ArrayList<Vector3D>(WALL_VERTS);
        this.frontWallVerts = new ArrayList<Vector3D>(WALL_VERTS);
        this.backWallVerts = new ArrayList<Vector3D>(WALL_VERTS);
    }

    /**
     * The initializeWallVerts method clears the given List of verts and then
     * adds a null entry for every vert that is expected to be in a wall.
     * 
     * @param vertList The List of verts to initialize.
     */
    private void initializeWallVerts(final List<Vector3D> vertList) {
        vertList.clear();
        for (int i = 0; i < WALL_VERTS; i++) {
            vertList.add(null);
        }
    }

    /**
     * Constructor for Building.
     * 
     * @param idString The UNIQUE ID of this object.
     * @param parentBlock The block which is supposed to own this building.
     * @param buildingType The type of building that this is supposed to be.
     */
    public Building(final String idString, final Block parentBlock,
            final BuildingType buildingType) {
        this(idString);
        this.parent = parentBlock;
        this.setType(buildingType);
    }

    /**
     * The clearOtherVerts method is reposible for clearing any other verts that
     * a SimObject might have.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    @Override
    protected final void clearOtherVerts() {
        this.groundVerts.clear();
        this.ceilingVerts.clear();
    }

    /**
     * The computePolyLoop.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    // TODO Write the javadoc properly for computePolyLoop in Building.
    @Override
    protected final void computePolyLoop() {
        final float posX = parent.getPosition().getXpos();
        final float posY = parent.getPosition().getYpos();

        this.setPolyLoop(new SmartPolygon(this.getGroundVerts().size()));
        for (Vector3D vert : this.getGroundVerts()) {
            this.getPolyLoop().addPoint(posX + vert.getXpos(),
                    posY + vert.getYpos());
        }
        this.getPolyLoop().validate();
    }

    /**
     * The destruct method unlinks all of this object's references to make
     * garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    @Override
    protected void destruct() {
        // TODO Write destruct in Building.
    }

    /**
     * The detectCollision.
     * 
     * @param obj Object to detect collision against.
     * @return Return true on collision situation.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    // TODO Write the javadoc properly for detectCollision in Building.
    @Override
    protected final boolean detectCollision(final AbstractSimObject obj) {
        // TODO Write detectCollision in Building.
        return false;
    }

    /**
     * The handleCollision.
     * 
     * @param obj Object to handle collision against.
     * @param data Supplemental data used to handle the collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    // TODO Write the javadoc properly for handleCollision in Building.
    @Override
    protected void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Write handleCollision in Building.
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts.
     * 
     * @see dds.model.simulation.AbstractSimObject#myCalculateVerts()
     */
    @Override
    protected final void myCalculateVerts() {
        // TODO Comment internals of myCalculateVerts in Building.
        this.initializeWallVerts(this.firstWallVerts);
        this.initializeWallVerts(this.secondWallVerts);
        this.initializeWallVerts(this.frontWallVerts);
        this.initializeWallVerts(this.backWallVerts);

        final Vector3D upVec = Vector3D.getZAxis();
        upVec.scale(this.getHeight());
        final Vector3D baseVec = this.baseVector.duplicate();
        baseVec.scale(this.getWidth());
        final Vector3D inVec = this.sideVector.duplicate();
        inVec.scale(this.getDepth());

        final Vector3D basePoint = this.getPosition().duplicate();
        final Vector3D topPoint = basePoint.duplicate();
        topPoint.add(upVec);
        final Vector3D topBack = topPoint.duplicate();
        topBack.add(inVec);
        final Vector3D baseBack = basePoint.duplicate();
        baseBack.add(inVec);
        // Check to see if we need to have verts to draw the first wall.
        if (this.firstWall) {
            this.firstWallVerts.set(FIRST, topPoint.duplicate());
            this.firstWallVerts.set(SECOND, topBack.duplicate());
            this.firstWallVerts.set(THIRD, baseBack.duplicate());
            this.firstWallVerts.set(FOURTH, basePoint.duplicate());
            this.groundVerts.add(baseBack.duplicate());
            this.ceilingVerts.add(topBack.duplicate());
        }

        // Draw the main wall
        this.groundVerts.add(basePoint.duplicate());
        this.ceilingVerts.add(topPoint.duplicate());
        this.frontWallVerts.set(SECOND, topPoint.duplicate());
        this.frontWallVerts.set(THIRD, basePoint.duplicate());
        basePoint.add(baseVec);
        topPoint.add(baseVec);
        this.groundVerts.add(basePoint.duplicate());
        this.ceilingVerts.add(topPoint.duplicate());
        this.frontWallVerts.set(FIRST, topPoint.duplicate());
        this.frontWallVerts.set(FOURTH, basePoint.duplicate());
        this.backWallVerts.set(FIRST, this.frontWallVerts.get(SECOND));
        this.backWallVerts.set(SECOND, this.frontWallVerts.get(FIRST));
        this.backWallVerts.set(THIRD, this.frontWallVerts.get(FOURTH));
        this.backWallVerts.set(FOURTH, this.frontWallVerts.get(THIRD));
        // Check to see if we need to draw the back wall too
        if (this.firstWall) {
            this.backWallVerts.set(FIRST, topBack.duplicate());
            this.backWallVerts.set(FOURTH, baseBack.duplicate());
        }
        baseBack.add(baseVec);
        topBack.add(baseVec);
        if (this.secondWall) {
            this.backWallVerts.set(SECOND, topBack.duplicate());
            this.backWallVerts.set(THIRD, baseBack.duplicate());
        }

        // Check to see if we need to draw the second wall
        if (this.secondWall) {
            this.secondWallVerts.set(FIRST, topBack.duplicate());
            this.secondWallVerts.set(SECOND, topPoint.duplicate());
            this.secondWallVerts.set(THIRD, basePoint.duplicate());
            this.secondWallVerts.set(FOURTH, baseBack.duplicate());
            this.groundVerts.add(baseBack.duplicate());
            this.ceilingVerts.add(topBack.duplicate());
        }
    }

    /**
     * Getter for height.
     * 
     * @return the height
     */
    public final float getHeight() {
        return this.dimensions.getYpos();
    }

    /**
     * Setter for height.
     * 
     * @param buildingHeight the height to set
     */
    public final void setHeight(final float buildingHeight) {
        this.dimensions.setYpos(buildingHeight);
        this.setVertsCalculated(false);
    }

    /**
     * Getter for parent.
     * 
     * @return the parent
     */
    public final Block getParent() {
        return parent;
    }

    /**
     * Setter for parent.
     * 
     * @param parentBlock the parent to set
     */
    public final void setParent(final Block parentBlock) {
        this.parent = parentBlock;
        this.setVertsCalculated(false);
    }

    /**
     * Getter for texture.
     * 
     * @return the texture
     */
    public final String getTextureFile() {
        return textureFile;
    }

    /**
     * Setter for texture.
     * 
     * @param buildingTexture the texture to set
     */
    public final void setTextureFile(final String buildingTexture) {
        this.textureFile = buildingTexture;
    }

    /**
     * Getter for type.
     * 
     * @return the type
     */
    public final BuildingType getType() {
        return type;
    }

    /**
     * Setter for type.
     * 
     * @param buildingType the type to set
     */
    public final void setType(final BuildingType buildingType) {
        this.type = buildingType;
        this.dimensions.setYpos(this.type.pickHeight());
        this.textureFile = this.type.pickTexture();
        this.setVertsCalculated(false);
    }

    /**
     * Getter for baseVector.
     * 
     * @return the baseVector
     */
    public final Vector3D getBaseVector() {
        return baseVector;
    }

    /**
     * Setter for baseVector.
     * 
     * @param vector the baseVector to set
     */
    public final void setBaseVector(final Vector3D vector) {
        this.baseVector.setPos(vector);
        this.setVertsCalculated(false);
    }

    /**
     * Getter for depth.
     * 
     * @return the depth
     */
    public final float getDepth() {
        return this.dimensions.getZpos();
    }

    /**
     * Setter for depth.
     * 
     * @param depthValue the depth to set
     */
    public final void setDepth(final float depthValue) {
        this.dimensions.setZpos(depthValue);
        this.setVertsCalculated(false);
    }

    /**
     * Getter for firstWall.
     * 
     * @return the firstWall
     */
    public final boolean isFirstWall() {
        return firstWall;
    }

    /**
     * Setter for firstWall.
     * 
     * @param enabled the firstWall to set
     */
    public final void setFirstWall(final boolean enabled) {
        this.firstWall = enabled;
        this.setVertsCalculated(false);
    }

    /**
     * Getter for secondWall.
     * 
     * @return the secondWall
     */
    public final boolean isSecondWall() {
        return secondWall;
    }

    /**
     * Setter for secondWall.
     * 
     * @param enabled the secondWall to set
     */
    public final void setSecondWall(final boolean enabled) {
        this.secondWall = enabled;
        this.setVertsCalculated(false);
    }

    /**
     * Getter for sideVector.
     * 
     * @return the sideVector
     */
    public final Vector3D getSideVector() {
        return sideVector;
    }

    /**
     * Setter for sideVector.
     * 
     * @param vector the sideVector to set
     */
    public final void setSideVector(final Vector3D vector) {
        this.sideVector.setPos(vector);
        this.setVertsCalculated(false);
    }

    /**
     * Getter for width.
     * 
     * @return the width
     */
    public final float getWidth() {
        return this.dimensions.getXpos();
    }

    /**
     * Setter for width.
     * 
     * @param widthValue the width to set
     */
    public final void setWidth(final float widthValue) {
        this.dimensions.setXpos(widthValue);
        this.setVertsCalculated(false);
    }

    /**
     * Getter for boundingVerts.
     * 
     * @return the boundingVerts
     */
    public final List<Vector3D> getGroundVerts() {
        return groundVerts;
    }

    /**
     * Setter for boundingVerts.
     * 
     * @param verts the boundingVerts to set
     */
    public final void setGroundVerts(final List<Vector3D> verts) {
        this.groundVerts = verts;
    }

    /**
     * Getter for backWallVerts.
     * 
     * @return the backWallVerts
     */
    public final List<Vector3D> getBackWallVerts() {
        return backWallVerts;
    }

    /**
     * Setter for backWallVerts.
     * 
     * @param verts the backWallVerts to set
     */
    public final void setBackWallVerts(final Vector3D[] verts) {
        System.arraycopy(verts, 0, this.backWallVerts, 0, WALL_VERTS);
    }

    /**
     * Getter for ceilingVerts.
     * 
     * @return the ceilingVerts
     */
    public final List<Vector3D> getCeilingVerts() {
        return ceilingVerts;
    }

    /**
     * Setter for ceilingVerts.
     * 
     * @param verts the ceilingVerts to set
     */
    public final void setCeilingVerts(final List<Vector3D> verts) {
        this.ceilingVerts = verts;
    }

    /**
     * Getter for firstWallVerts.
     * 
     * @return the firstWallVerts
     */
    public final List<Vector3D> getFirstWallVerts() {
        return firstWallVerts;
    }

    /**
     * Setter for firstWallVerts.
     * 
     * @param verts the firstWallVerts to set
     */
    public final void setFirstWallVerts(final Vector3D[] verts) {
        System.arraycopy(verts, 0, this.firstWallVerts, 0, WALL_VERTS);
    }

    /**
     * Getter for frontWallVerts.
     * 
     * @return the frontWallVerts
     */
    public final List<Vector3D> getFrontWallVerts() {
        return frontWallVerts;
    }

    /**
     * Setter for frontWallVerts.
     * 
     * @param verts the frontWallVerts to set
     */
    public final void setFrontWallVerts(final Vector3D[] verts) {
        System.arraycopy(verts, 0, this.frontWallVerts, 0, WALL_VERTS);
    }

    /**
     * Getter for secondWallVerts.
     * 
     * @return the secondWallVerts
     */
    public final List<Vector3D> getSecondWallVerts() {
        return secondWallVerts;
    }

    /**
     * Setter for secondWallVerts.
     * 
     * @param verts the secondWallVerts to set
     */
    public final void setSecondWallVerts(final Vector3D[] verts) {
        System.arraycopy(verts, 0, this.secondWallVerts, 0, WALL_VERTS);
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.getGroundVerts();
    }

    /**
     * Getter for dimensions.
     * 
     * @return the dimensions
     */
    public final Vector3D getDimensions() {
        return dimensions;
    }

    /**
     * Setter for dimensions.
     * 
     * @param newDimensions the dimensions to set
     */
    public final void setDimensions(final Vector3D newDimensions) {
        this.dimensions.setPos(newDimensions);
        this.setVertsCalculated(false);
    }

}
