package dds.model.simulation;

import java.util.Locale;
import java.util.Random;

/**
 * The BuildingType enum class is responsible for providing information about
 * the different types of buildings. Each enumeration of this class is imbued
 * with knowledge about the height and texturing for it's particular building
 * type. Building types are defined by a default height, a height variance, and
 * an array of texture file names. This information is combined with a random
 * factor to produce more individual sets of data. The default height is altered
 * by a random application of the variance to produce a final height for a
 * building. The texture for a building is a random choice from the filenames
 * within the array. This class is also able to parse a String into a
 * BuildingType. This function should be useful for processing text from an XML
 * file.
 * 
 * @author Jacob Bolton
 */
public enum BuildingType {
    /**
     * The RESIDENTIAL enumeration defines the data used for rendering
     * residential buildings.
     */
    RESIDENTIAL(15.0f, 5.0f, new String[] { "residential1.jpg",
            "residential2.jpg", "residential3.jpg", "residential4.jpg",
            "residential5.jpg", "residential6.jpg", "residential7.jpg",
            "residential8.jpg", "residential9.jpg", "residential10.jpg" }),

    /**
     * The INDUSTRIAL enumeration defines the data used for rendering industrial
     * buildings.
     */
    INDUSTRIAL(20.0f, 5.0f, new String[] { "industrial1.jpg",
            "industrial2.jpg", "industrial3.jpg", "industrial4.jpg",
            "industrial5.jpg", "industrial6.jpg", "industrial7.jpg",
            "industrial8.jpg", "industrial9.jpg", "industrial10.jpg" }),

    /**
     * The COMMERCIAL enumeration defines the data used for rendering
     * user-defined buildings.
     */
    COMMERCIAL(18.0f, 5.0f, new String[] { "commercial1.jpg",
            "commercial2.jpg", "commercial3.jpg", "commercial4.jpg",
            "commercial5.jpg", "commercial6.jpg", "commercial7.jpg",
            "commercial8.jpg", "commercial9.jpg", "commercial10.jpg" }),

    /**
     * The USER enumeration defines the data used for rendering user-defined
     * buildings. It is important to note that this enumeration will always
     * return a null textures file. This is because the user-defined building
     * should have its texture file set from the XML and not from our library of
     * standard textures.
     */
    USER(12.0f, 3.0f, null),

    /**
     * The NONE enumeration defines the data used for rendering buildings
     * without a type defined. It should be noted that these buildings are
     * txetured as abandoned or run-down.
     */
    NONE(16.0f, 4.0f, new String[] { "none1.jpg", "none2.jpg", "none3.jpg",
            "none4.jpg" }),

    /**
     * The UNDEFINED enumeration defines a BuildingType that represents a
     * problematic input for the type. This is usually returned as an error when
     * attempting to parse a String into a BuildingType.
     */
    UNDEFINED(0.0f, 0.0f, null);

    /**
     * The NO_TEXTURE constant is returned as the texture file name when there
     * is no texture file name to be returned.
     */
    private static final String NO_TEXTURE = null;

    /**
     * The textures field is where the array of texture file names is stored
     * within this enumeration.
     */
    private String[] textures;

    /**
     * The height field is where the default height is stored within this
     * enumeration.
     */
    private float height;

    /**
     * The variance field is where the array height variance amount is stored
     * within this enumeration.
     */
    private float variance;

    /**
     * The random field is used as a reference to a random number generator.
     */
    private Random random;

    /**
     * Constructor for BuildingType is responsible for setting up the parameters
     * of the BuildingType enumerations.
     * 
     * @param defaultHeight The default height for the BuildingType.
     * @param heightVariance The amount by which the height can vary. Currently
     *            this can affect the height by increasing or decreasing it.
     * @param textureFiles This is an array of Strings that contains all of the
     *            potential image files that can be used for textures.
     */
    private BuildingType(final float defaultHeight, final float heightVariance,
            final String[] textureFiles) {
        // Set the data contained within the enumeration.
        this.height = defaultHeight;
        this.variance = heightVariance;
        this.textures = textureFiles;
        this.random = new Random();
    }

    /**
     * The pickTexture method randomly selects a texture file from this
     * BuildingType's set of available texture files.
     * 
     * @return A String containing the randomly selected texture file name.
     */
    public String pickTexture() {
        // Set result to nothing in case there's nothing to choose from.
        String result = NO_TEXTURE;

        // Check to see if there are textures defined to pick from.
        if ((this.textures != null) && (this.textures.length > 0)) {
            // Randomly select a texure file name to be returned.
            result = this.textures[random.nextInt(this.textures.length)];
        }
        // Return the reuslt.
        return result;
    }

    /**
     * The pickHeight method generates a random height based on this BlockType's
     * standard height and an applied variance.
     * 
     * @return A height to be used for this Block.
     */
    public float pickHeight() {
        // Pick a random value between 0.0f and 1.0f to be used to scale the
        // height variance.
        float result = this.random.nextFloat();

        // Randomly select whether or not the variance should be added or
        // subtracted from the default height by randomly choosing a boolean and
        // setting up for subtraction if it is true.
        if (this.random.nextBoolean()) {
            result *= -1.0f;
        }

        // Return the default height combined with the randomly scaled variance
        // as the final height for a particular building.
        return this.height + (result * this.variance);
    }

    /**
     * The parseType method tries to resolve an input String to a BlockType by
     * comparing it's UPPERCASE value to the toStrings for each defined
     * enumeration.
     * 
     * @param string The String to attemp resolution.
     * @return The corresponding BlockType or null if it cannot be resolved.
     */
    public static BuildingType parseType(final String string) {
        // Convert the String to UPPERCASE so that it can be compared to the
        // Strings returned by the toString method for enumerations.
        final String upper = string.toUpperCase(Locale.getDefault());

        // Start by seting the result to UNDEFINED for the eventuality when the
        // type can NOT be determined from the input String.
        BuildingType result = UNDEFINED;

        // Check the input against all defined BuildingTypes. Iterating through
        // the defined types allows easier maintainence than using a set of if
        // statements.
        for (BuildingType type : BuildingType.values()) {
            // Check to see if the BuildingType's toString and the UPPERCASE
            // input String match.
            if (upper.equals(type.toString())) {
                // Set our result to this type.
                result = type;

                // Break out of the loop since we've gotten our result.
                break;
            }
        }

        // Return the final result, even if noother type can be determined
        // UNDEFINED will be returned.
        return result;
    }
}
