package dds.model.simulation;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import dds.control.InvalidTypeException;

/**
 * The ConcreteSimulationFactory class is responsible for instantiating all
 * objects that are required to simulate a given scenario.
 * 
 * @author Mike Sanders
 */
// TODO Make sure that this can be used to create multiple simulations at the
// same time. This will ensure a good design that can be used in future, more
// complicated implementations.
public final class ConcreteSimulationFactory implements SimulationFactory {

    /**
     * The factory field contains a reference to the singleton instance of the
     * factory.
     */
    private static ConcreteSimulationFactory factory;

    /**
     * The TEXT_NODE constant defines the String to expect for nodes in the XML
     * DOM?
     */
    // TODO Verify comment.
    private static final String TEXT_NODE = "#text";

    static {
        // Create the Factory singleton.
        factory = new ConcreteSimulationFactory();
    }

    /**
     * The getFactory.
     * 
     * @return The singleton factory.
     */
    public static ConcreteSimulationFactory getFactory() {
        return factory;
    } // end method getFactory()

    /**
     * The map field stores a reference to the current map document.
     */
    // TODO Consider moving this into a method as a local variable.
    private transient Document map;

    /**
     * The sim field stores a reference to the surrent simulation being
     * constructed.
     */
    // TODO Consider moving this into a method as a local variable.
    private transient Simulation sim;

    /**
     * Constructor for ConcreteSimulationFactory.
     */
    private ConcreteSimulationFactory() {
        super();
    } // end method ConcreteSimulationFactory()

    /**
     * The addBlockObject.
     * 
     * @param rootNode is the root node of the Block object tree
     * @throws InvalidTypeException if the block has an invalid type
     */
    private void addBlockObject(final Node rootNode)
            throws InvalidTypeException {

        // get the child nodes to create the objects from
        final NodeList children = rootNode.getChildNodes();

        // the current node being parsed
        Node currentNode;

        // the current intersection being added
        Block currBlock;

        // index for iterating the list
        int index = 0;

        // iterate the list and add the objects
        for (; index < children.getLength(); ++index) {

            currentNode = children.item(index);

            if (!currentNode.getNodeName().equals(TEXT_NODE)) {

                // get the attributes
                final NamedNodeMap attributes = currentNode.getAttributes();

                // attempt to get the values for the attributes
                // and set them in the new Block object
                try {

                    // get the name
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();

                    // get the block type
                    final String type = attributes.getNamedItem("type")
                            .getTextContent();

                    // get the number of buildings on the block
                    final int buildings = Integer.parseInt(attributes
                            .getNamedItem("buildings").getTextContent());

                    // this collection and tokenizer will be used to get the
                    // relevant Road objects from the sim and add them to the
                    // Block object as an ArrayList
                    final List<Road> roads = parseRoadIDs(attributes
                            .getNamedItem("roads").getTextContent());

                    // create the new block object
                    currBlock = createBlock(name, BlockType.parseType(type),
                            buildings, roads);

                    // add the new intersection to the sim
                    sim.addObject(currBlock);

                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                } // end try-catch

            } // end if

        } // end for

    } // end method addBlockObject()

    /**
     * The addIntersectionObject.
     * 
     * @param rootNode is the root node of the tree to be parsed
     */
    /**
     * The addIntersectionObject.
     * 
     * @param rootNode is the root node of the Intersection tree
     */
    private void addIntersectionObject(final Node rootNode) {

        // get the child nodes to create the objects from
        final NodeList children = rootNode.getChildNodes();

        // the current node being parsed
        Node currentNode;

        // the current intersection being added
        Intersection currInx;

        // index for iterating the list
        int index = 0;

        // iterate the list and add the objects
        for (; index < children.getLength(); ++index) {

            currentNode = children.item(index);

            if (!currentNode.getNodeName().equals(TEXT_NODE)) {

                // get the attributes
                final NamedNodeMap attributes = currentNode.getAttributes();

                // attempt to get the values for the attributes
                // and create the new Intersection object
                try {

                    // get the name
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();

                    // get the x coordinate
                    final float xValue = Float.parseFloat(attributes
                            .getNamedItem("x").getTextContent());

                    // get the y coordinate
                    final float yValue = Float.parseFloat(attributes
                            .getNamedItem("y").getTextContent());

                    // get the z coordinate
                    final float zValue = Float.parseFloat(attributes
                            .getNamedItem("z").getTextContent());

                    // create the new intersection and set the attributes
                    currInx = createIntersection(name, xValue, yValue, zValue);
                    // currInx.setIdStr(name);
                    // System.out.println("I just made: " + currInx);
                    // System.out.println("It uses these coordinates: " + xValue
                    // + ", " + yValue + ", " + zValue);

                    // add the new intersection to the sim
                    sim.addObject(currInx);

                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                } // end try-catch

            } // end if

        } // end for

    } // end method addIntersectionObject()

    /**
     * The addRoadObject.
     * 
     * @param root is the root node of the Road object tree
     * @throws InvalidTypeException if the direction value is invalid
     */
    private void addRoadObject(final Node root) throws InvalidTypeException {

        // get the child nodes to create the objects from
        final NodeList children = root.getChildNodes();

        // the current node being parsed
        Node currentNode;

        // the current intersection being added
        Road currRoad = null;

        // index for iterating the list
        int index = 0;

        // iterate the list and add the objects
        for (; index < children.getLength(); ++index) {

            currentNode = children.item(index);

            if (!currentNode.getNodeName().equals(TEXT_NODE)) {

                // get the attributes
                final NamedNodeMap attributes = currentNode.getAttributes();

                // attempt to get the numerical values for the attributes
                // and set them in the new Road object
                try {

                    // get the name
                    final String name = attributes.getNamedItem("id")
                            .getTextContent();

                    final String intA = attributes.getNamedItem("ia")
                            .getTextContent();

                    final String intB = attributes.getNamedItem("ib")
                            .getTextContent();

                    currRoad = createRoad(name, (Intersection) sim
                            .getObject(intA), (Intersection) sim
                            .getObject(intB));

                    final String direction = attributes.getNamedItem(
                            "direction").getTextContent();

                    if (direction.equals(Road.TRAFFIC_A_TO_B)
                            || direction.equals(Road.TRAFFIC_B_TO_A)
                            || direction.equals(Road.TRAFFIC_BI)) {
                        currRoad.setDirection(direction);
                    } else {
                        throw new InvalidTypeException(
                                "Invalid value for direction!\n");
                    } // end if-else

                    final Float thickness = Float.parseFloat(attributes
                            .getNamedItem("thickness").getTextContent());

                    final int speedlimit = Integer.parseInt(attributes
                            .getNamedItem("speedlimit").getTextContent());

                    final int lanes = Integer.parseInt(attributes.getNamedItem(
                            "lanes").getTextContent());

                    // add the new attributes
                    currRoad.setThickness(thickness);
                    currRoad.setSpeedlimit(speedlimit);
                    currRoad.setLanes(lanes);

                    sim.addObject(currRoad);
                    currRoad.addRoadSigns();

                    for (RoadSign rs : currRoad.getRoadSigns()) {
                        sim.addObject(rs);
                    }
                } catch (NumberFormatException e) {
                    System.err.println(e.getMessage());
                } // end try-catch

            } // end if

        } // end for
    } // end method addRoadObject()

    /**
     * The addSimObjects.
     */
    private void addSimObjects() {

        // the list of the root node's children
        final NodeList children = map.getChildNodes().item(0).getChildNodes();

        // the current node being examined
        Node currentChild;

        // index for iterating the node list
        int index = 0;

        // cycle through each node and parse them if necessary
        for (; index < children.getLength(); ++index) {
            currentChild = children.item(index);

            // parse the node if it is named
            if (!currentChild.getNodeName().equals(TEXT_NODE)) {

                try {
                    parseNode(currentChild);
                } catch (InvalidTypeException e) {
                    System.err.println(e.getMessage());
                } // end try-catch

            } // end if

        } // end for

    } // end method addSimObjects()

    /**
     * The createBlock method creates a new Block with the specified parameters.
     * 
     * @param name The name of the Block object.
     * @param type The BlockType of the Block.
     * @param buildings The number of buildings to place along each road
     *            inscribing the Block.
     * @param roads A list of Roads that inscribe the Block.
     * @return A new Block object.
     */
    private Block createBlock(final String name, final BlockType type,
            final int buildings, final List<Road> roads) {
        return new Block(name, type, buildings, roads);
    }

    /**
     * The createIntersection method creates a new Intersection object.
     * 
     * @param name is a String representing the name of the Intersection.
     * @param xValue is the x location of the Intersection.
     * @param yValue is the y location of the Intersection.
     * @param zValue is the z location of the Intersection.
     * @return the newly created Intersection object.
     */
    private Intersection createIntersection(final String name,
            final float xValue, final float yValue, final float zValue) {

        return new Intersection(name, new Vector3D(xValue, yValue, zValue));

    } // end method createIntersection()

    /**
     * The createRoad method creates a new Road from the specified parameters.
     * 
     * @param name The name of the Road object.
     * @param intA The starting Intersection.
     * @param intB The ending Intersection.
     * @return A new Road.
     */
    private Road createRoad(final String name, final Intersection intA,
            final Intersection intB) {
        return new Road(name, intA, intB);
    }

    /**
     * The createSimulation.
     * 
     * @param dom is the Document Object Model from the sim controller
     * @return sim is the simulation that was created from the dom
     * @see dds.model.simulation.SimulationFactory
     *      #createSimulation(org.w3c.dom.Document)
     */
    public Simulation createSimulation(final Document dom) {

        // the document object module from the XML file
        this.map = dom;

        // create the new simulation
        sim = new Simulation();

        // add the objects to the simulation
        addSimObjects();

        final Node mapXMLEntry = (map.getElementsByTagName("map")).item(0);
        final NamedNodeMap nodeMap = mapXMLEntry.getAttributes();
        String xlAttrib, xuAttrib, ylAttrib, yuAttrib;

        xlAttrib = nodeMap.getNamedItem("xl").getNodeValue();
        xuAttrib = nodeMap.getNamedItem("xu").getNodeValue();
        ylAttrib = nodeMap.getNamedItem("yl").getNodeValue();
        yuAttrib = nodeMap.getNamedItem("yu").getNodeValue();

        sim.setTerrainBoundaries(Integer.valueOf(xlAttrib), Integer
                .valueOf(xuAttrib), Integer.valueOf(ylAttrib), Integer
                .valueOf(yuAttrib));

        // System.out.println(sim.toString());

        // return the new sim with the objects from the map file added
        return sim;

    } // end method createSumulation()

    /**
     * The parseNode.
     * 
     * @param node is the node to parse
     * @throws InvalidTypeException if the node type is not recognized
     */
    private void parseNode(final Node node) throws InvalidTypeException {

        // add the node objects to the sim based on the
        // type of node given. At this point it is important
        // to add the objects to the sim in the order given
        // below.
        if (node.getNodeName().equals("intersections")) {
            addIntersectionObject(node);
        } else if (node.getNodeName().equals("roads")) {
            addRoadObject(node);
        } else if (node.getNodeName().equals("blocks")) {
            addBlockObject(node);
        } else {
            throw new InvalidTypeException("Invalid type for XML node!");
        } // end if-then-else

    } // end method parseNode()

    /**
     * The parseRoadIDs method takes in a space-delimited string of Road ID
     * Strings and parses the list so it can return a List of Roads with
     * matching IDs.
     * 
     * @param roadIDs The space-delimited list of Road IDs.
     * @return A List containing the matching Road objects from the Simulation.
     */
    private List<Road> parseRoadIDs(final String roadIDs) {
        // this collection and tokenizer will be used to get the
        // relevant Road objects from the sim and add them to the
        // Block object as an ArrayList
        final List<Road> roads = new ArrayList<Road>();
        final StringTokenizer tokenizer = new StringTokenizer(roadIDs);

        while (tokenizer.hasMoreTokens()) {
            final String token = tokenizer.nextToken(" ");
            roads.add((Road) sim.getObject(token));
        } // end while return null;
        return roads;
    }

} // end class definition ConcreteSimulationFactory()
