/**
 * GenericMovableObject.java.
 */
package dds.model.simulation;

import java.util.List;

import dds.control.DDSOBJModel;

/**
 * The GenericMovableObject class is responsible for modelling a movable object
 * that relies on an DDSOBJModel for shape data.
 * 
 * @author Jacob Bolton
 */
public class GenericMovableObject extends AbstractMovableObject implements
        GenericObjectInterface {

    /**
     * The genObj is the GenericObject that has the model for this Object.
     */
    private GenericObject genObj;

    /**
     * Constructor for GenericMovableObject.
     * 
     * @param idString The unique ID for this object within the simulation.
     * @param modelFile The name of the model file to use.
     */
    public GenericMovableObject(final String idString, final String modelFile) {
        super(idString);
        this.genObj = new GenericObject(idString, modelFile);
    }

    /**
     * The updateKinetics method updates the kinetics according to the time
     * passed.
     * 
     * @param time The time passed
     * @see dds.model.simulation.AbstractMovableObject#updateKinetics(float)
     */
    @Override
    public void updateKinetics(final float time) {
        // TODO Maybe do something with this?
    }

    /**
     * The clearOtherVerts method is reposible for clearing any object-specific
     * verts or other calculated fields that a SimObject might have generated
     * during vert calculation.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    protected void clearOtherVerts() {
        // TODO Determine if we need to do anything here. Perhaps pass this to
        // the GenericObject.
    }

    /**
     * The destruct method is responsible for unlinking all of this object's
     * references to make garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    @Override
    protected final void destruct() {
        this.clearVerts();
        this.genObj.destruct();
    }

    /**
     * The handleCollision method takes care of what needs to be done when a
     * collision is detected.
     * 
     * @param obj The obj to handle collision against.
     * @param data The data for handling the collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    protected void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Determine if we should fill this in or leave it open for
        // extension.
    }

    /**
     * The computePolyLoop method sets up the SmartPolygon used for collision
     * detection.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    protected final void computePolyLoop() {
        this.genObj.computePolyLoop();
        this.setPolyLoop(this.genObj.getPolyLoop());
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts. This method does NOT need to clear the older verts OR set
     * the vertsCalculated flag becuase that functionality has been standardized
     * in the calculateVerts method defined in this class. This method just
     * needs to calculate the verts. The verts that are calculate MUST be stored
     * as local coordinates. These coordinates will be offset by the current
     * position when being used for collision detection of rendering.
     * 
     * @see dds.model.simulation.AbstractSimObject#myCalculateVerts()
     */
    @Override
    protected final void myCalculateVerts() {
        this.genObj.calculateVerts();
        this.setVerts(this.genObj.getVerts());
    }

    /**
     * The getModel3D() method returns the externally defined model.
     * 
     * @return The DDSOBJModel associated with this object.
     * @see dds.model.simulation.GenericObjectInterface#getModel3D()
     */
    public final DDSOBJModel getModel3D() {
        return this.genObj.getModel3D();
    }

    /**
     * The getModelVert gets a Vector3D representation of the specified Vertex
     * in the DDSOBJModel.
     * 
     * @param vertNum The index of the Vertex to get.
     * @return A Vector3D representation of the Vertex.
     * @see dds.model.simulation.GenericObjectInterface#getModelVert(int)
     */
    public final Vector3D getModelVert(final int vertNum) {
        return this.genObj.getModelVert(vertNum);
    }

    /**
     * The loadModelFile method loads the specified file from the ResourceLoader
     * as the current 3D model to use for this object.
     * 
     * @param modelFile The .obj file to use. It MUST be in the proper location
     *            in the data repository.
     * @see dds.model.simulation.GenericObjectInterface
     *      #loadModelFile(java.lang.String)
     */
    public final void loadModelFile(final String modelFile) {
        this.genObj.loadModelFile(modelFile);
    }

    /**
     * The scale method allows a GenericObject to be resized.
     * 
     * @param xScalar The factor to scale along the x-axis.
     * @param yScalar The factor to scale along the y-axis.
     * @param zScalar The factor to scale along the z-axis.
     * @see dds.model.simulation.GenericObjectInterface#scale(float, float,
     *      float)
     */
    public final void scale(final float xScalar, final float yScalar,
            final float zScalar) {
        this.genObj.scale(xScalar, yScalar, zScalar);
    }

    /**
     * The scaleTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param xDim The dimension to scale the size along the x-axis to.
     * @param yDim The dimension to scale the size along the y-axis to.
     * @param zDim The dimension to scale the size along the z-axis to.
     * @see dds.model.simulation.GenericObjectInterface#scaleTo(float, float,
     *      float)
     */
    public final void scaleTo(final float xDim, final float yDim,
            final float zDim) {
        this.genObj.scaleTo(xDim, yDim, zDim);
    }

    /**
     * The scaleX method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the x-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleX(float, boolean)
     */
    public final void scaleX(final float scalar, final boolean keepAspect) {
        this.genObj.scaleX(scalar, keepAspect);
    }

    /**
     * The scaleXTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the x-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleXTo(float, boolean)
     */
    public final void scaleXTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleXTo(dim, keepAspect);
    }

    /**
     * The scaleY method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the y-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleY(float, boolean)
     */
    public final void scaleY(final float scalar, final boolean keepAspect) {
        this.genObj.scaleY(scalar, keepAspect);
    }

    /**
     * The scaleYTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the y-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleYTo(float, boolean)
     */
    public final void scaleYTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleYTo(dim, keepAspect);
    }

    /**
     * The scaleZ method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the z-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZ(float, boolean)
     */
    public final void scaleZ(final float scalar, final boolean keepAspect) {
        this.genObj.scaleZ(scalar, keepAspect);
    }

    /**
     * The scaleZTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the z-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZTo(float, boolean)
     */
    public final void scaleZTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleZTo(dim, keepAspect);
    }

    /**
     * The setDefaultOrientation method sets the default orientation for a
     * GenericObject so that comparison with the current orientation can serve
     * as a calculation for rotation of this GenericObject.
     * 
     * @param orientation The normalized direction that the GenericObject should
     *            be oriented originaly.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultOrientation(dds.model.simulation.Vector3D)
     */
    public final void setDefaultOrientation(final Vector3D orientation) {
        this.genObj.setDefaultOrientation(orientation);
    }

    /**
     * The setDefaultOrientation method sets the default upwards vector for a
     * GenericObject so that comparison with the current upwards vector can
     * serve as a calculation for rotation of this GenericObject.
     * 
     * @param upVector The normalized direction that the GenericObject should
     *            use for up.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultUp(dds.model.simulation.Vector3D)
     */
    public final void setDefaultUp(final Vector3D upVector) {
        this.genObj.setDefaultUp(upVector);
    }

    /**
     * Getter for genObj.
     * 
     * @return the genObj
     */
    public final GenericObject getGenObj() {
        return genObj;
    }

    /**
     * Setter for genObj.
     * 
     * @param obj the genObj to set
     */
    public final void setGenObj(final GenericObject obj) {
        this.genObj = obj;
    }

    /**
     * The move method was supposed to be used for calculating motion, but this
     * functionality should be done in simulate instead.
     * 
     * @param time The delta t for calculating motion.
     * @see dds.model.simulation.AbstractMovableObject#move(float)
     */
    @Override
    public final void move(final float time) {
        this.basicMove(time);
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.genObj.getBoundingVerts();
    }

}
