package dds.model.simulation;

import java.util.ArrayList;
import java.util.List;
import java.util.Observable;

import saito.objloader.Vertex;
import dds.control.DDSOBJModel;

/**
 * The GenericObject class is responsible for simulating objects that are
 * supplied by the user. These objects are typically inactive in the simulation
 * and are drawn according to an Alias .obj file.
 * 
 * @author Jacob Bolton
 */
public class GenericObject extends AbstractSimObject implements
        GenericObjectInterface {
    /**
     * The NO_MODEL is an easy-to-read way to set the model to null.
     */
    private static final DDSOBJModel NO_MODEL = null;

    /**
     * The currentModel is storage for the current DDSOBJModel being used for
     * this object.
     */
    private DDSOBJModel currentModel;

    /**
     * The currentOrient field stores the current orientation for this object.
     */
    // TODO Determine if this is already defined in the superclass.
    private Vector3D currentOrient;

    /**
     * The currentUp field stores the current upwards vector for this object.
     */
    // TODO Determine if this is already defined in the superclass.
    private Vector3D currentUp;

    /**
     * The xDimension is the current x-axis dimension for this object.
     */
    private float xDimension;

    /**
     * The xScale is the current x-axis scaling for this object.
     */
    private float xScale;

    /**
     * The yDimension is the current y-axis dimension for this object.
     */
    private float yDimension;

    /**
     * The yScale is the current y-axis scaling for this object.
     */
    private float yScale;

    /**
     * The zDimension is the current z-axis dimension for this object.
     */
    private float zDimension;

    /**
     * The zScale is the current z-axis scaling for this object.
     */
    private float zScale;

    /**
     * The boundingVerts field stores the verts which inscribe the bounding of
     * this object.
     */
    private final transient List<Vector3D> boundingVerts;

    /**
     * Constructor for AbstractGenericObject.
     * 
     * @param idString The unique ID for this object within the simulation.
     */
    private GenericObject(final String idString) {
        super(idString);
        this.currentModel = NO_MODEL;
        this.boundingVerts = new ArrayList<Vector3D>();
        this.xScale = 1.0f;
        this.yScale = 1.0f;
        this.zScale = 1.0f;
        this.xDimension = 1.0f;
        this.yDimension = 1.0f;
        this.zDimension = 1.0f;
    }

    /**
     * Constructor for AbstractGenericObject.
     * 
     * @param idString The unique ID for this object within the simulation.
     * @param modelFile The .obj filename of the model in the data repository to
     *            use for this object.
     */
    public GenericObject(final String idString, final String modelFile) {
        this(idString);
        this.currentModel = DDSOBJModel.getOBJModel(modelFile);
        this.calculateModelDimensions();
    }

    /**
     * The calculateModelDimensions method determines the default dimensions for
     * the current DDSOBJModel and sets the scaling to 1.0.
     */
    public final void calculateModelDimensions() {
        // Start the min's and max's at the values for the first vert.
        float minX = this.currentModel.getVertex(0).vx;
        float minY = this.currentModel.getVertex(0).vy;
        float minZ = this.currentModel.getVertex(0).vz;
        float maxX = this.currentModel.getVertex(0).vx;
        float maxY = this.currentModel.getVertex(0).vy;
        float maxZ = this.currentModel.getVertex(0).vz;

        // Iterate through the rest of the verts.
        for (int i = 1; i < this.currentModel.getVertexsize(); i++) {
            // Get the next vert.
            final Vertex vert = this.currentModel.getVertex(i);

            // Compare the vert's coordinates with the current values and update
            // when appropriate.
            if (vert.vx < minX) {
                minX = vert.vx;
            }
            if (vert.vy < minY) {
                minY = vert.vy;
            }
            if (vert.vz < minZ) {
                minZ = vert.vz;
            }
            if (vert.vx > maxX) {
                maxX = vert.vx;
            }
            if (vert.vy > maxY) {
                maxY = vert.vy;
            }
            if (vert.vz > maxZ) {
                maxZ = vert.vz;
            }
        }

        // Set scaling to 1, which is normal size.
        this.xScale = 1.0f;
        this.yScale = 1.0f;
        this.zScale = 1.0f;

        // Calculate dimensions.
        this.xDimension = maxX - minX;
        this.yDimension = maxY - minY;
        this.zDimension = maxZ - minZ;

        // Clear bounding verts and then recalculate and add them.
        this.boundingVerts.clear();
        this.boundingVerts.add(new Vector3D(minX, minY, minZ));
        this.boundingVerts.add(new Vector3D(minX, minY, maxZ));
        this.boundingVerts.add(new Vector3D(minX, maxY, minZ));
        this.boundingVerts.add(new Vector3D(minX, maxY, maxZ));
        this.boundingVerts.add(new Vector3D(maxX, minY, minZ));
        this.boundingVerts.add(new Vector3D(maxX, minY, maxZ));
        this.boundingVerts.add(new Vector3D(maxX, maxY, minZ));
        this.boundingVerts.add(new Vector3D(maxX, maxY, maxZ));
    }

    /**
     * The clearOtherVerts method is reposible for clearing any object-specific
     * verts or other calculated fields that a SimObject might have generated
     * during vert calculation.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    @Override
    protected void clearOtherVerts() {
        // TODO Consider clearing bounding verts here.
        // There are no other verts to clear for this object.
    }

    /**
     * The destruct method is responsible for unlinking all of this object's
     * references to make garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    protected final void destruct() {
        this.clearVerts();
    }

    /**
     * The detectCollision.
     * 
     * @param obj SimObject to detect collision with.
     * @return True on collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    // TODO Comment detectCollision in GenericObject.
    protected final boolean detectCollision(final AbstractSimObject obj) {
        return false;
    }

    /**
     * The handleCollision method takes care of what needs to be done when a
     * collision is detected.
     * 
     * @param obj The object to handle collision against?
     * @param data The collision data?
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    // TODO Verify these comments.
    protected void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Decide what should go here and if this method should be final.
    }

    /**
     * The computePolyLoop.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    // TODO Comment computePolyLoop in GenericObject.
    protected void computePolyLoop() {
        // TODO Decide what should go here and if this method should be final.
    }

    /**
     * Getter for currentModel.
     * 
     * @return the currentModel
     */
    public final DDSOBJModel getCurrentModel() {
        return currentModel;
    }

    /**
     * Getter for currentOrient.
     * 
     * @return the currentOrient
     */
    public final Vector3D getCurrentOrient() {
        return currentOrient;
    }

    /**
     * Getter for currentUp.
     * 
     * @return the currentUp
     */
    public final Vector3D getCurrentUp() {
        return currentUp;
    }

    /**
     * The getModel3D() method returns the externally defined model.
     * 
     * @return The DDSOBJModel associated with this object.
     * @see dds.model.simulation.GenericObjectInterface#getModel3D()
     */
    public final DDSOBJModel getModel3D() {
        return this.getCurrentModel();
    }

    /**
     * The getModelVert gets a Vector3D representation of the specified Vertex
     * in the DDSOBJModel.
     * 
     * @param vertNum The index of the Vertex to get.
     * @return A Vector3D representation of the Vertex.
     * @see dds.model.simulation.GenericObjectInterface#getModelVert(int)
     */
    public final Vector3D getModelVert(final int vertNum) {
        final Vertex vert = this.currentModel.getVertex(vertNum);
        return new Vector3D(vert.vx, vert.vy, vert.vz);
    }

    /**
     * Getter for xDimension.
     * 
     * @return the xDimension
     */
    public final float getXDimension() {
        return xDimension;
    }

    /**
     * Getter for xScale.
     * 
     * @return the xScale
     */
    public final float getXScale() {
        return xScale;
    }

    /**
     * Getter for yDimension.
     * 
     * @return the yDimension
     */
    public final float getYDimension() {
        return yDimension;
    }

    /**
     * Getter for yScale.
     * 
     * @return the yScale
     */
    public final float getYScale() {
        return yScale;
    }

    /**
     * Getter for zDimension.
     * 
     * @return the zDimension
     */
    public final float getZDimension() {
        return zDimension;
    }

    /**
     * Getter for zScale.
     * 
     * @return the zScale
     */
    public final float getZScale() {
        return zScale;
    }

    /**
     * The loadModelFile method loads the specified file from the ResourceLoader
     * as the current 3D model to use for this object.
     * 
     * @param modelFile The .obj file to use. It MUST be in the proper location
     *            in the data repository.
     * @see dds.model.simulation.GenericObjectInterface
     *      #loadModelFile(java.lang.String)
     */
    public final void loadModelFile(final String modelFile) {
        this.currentModel.load(modelFile);
    }

    /**
     * The myCalculateVerts.
     * 
     * @see dds.model.simulation.AbstractSimObject#myCalculateVerts()
     */
    @Override
    public final void myCalculateVerts() {
        for (int i = 0; i < this.currentModel.getVertexsize(); i++) {
            this.getVerts().add(this.getModelVert(i));
        }
    }

    /**
     * The scale method allows a GenericObject to be resized.
     * 
     * @param xScalar The factor to scale along the x-axis.
     * @param yScalar The factor to scale along the y-axis.
     * @param zScalar The factor to scale along the z-axis.
     * @see dds.model.simulation.GenericObjectInterface#scale(float, float,
     *      float)
     */
    public final void scale(final float xScalar, final float yScalar,
            final float zScalar) {
        this.xScale *= xScalar;
        this.xDimension *= xScalar;
        this.yScale *= yScalar;
        this.yDimension *= yScalar;
        this.zScale *= zScalar;
        this.zDimension *= zScalar;
    }

    /**
     * The scaleTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param xDim The dimension to scale the size along the x-axis to.
     * @param yDim The dimension to scale the size along the y-axis to.
     * @param zDim The dimension to scale the size along the z-axis to.
     * @see dds.model.simulation.GenericObjectInterface#scaleTo(float, float,
     *      float)
     */
    public final void scaleTo(final float xDim, final float yDim,
            final float zDim) {
        this.scaleXTo(xDim, false);
        this.scaleYTo(yDim, false);
        this.scaleZTo(zDim, false);
    }

    /**
     * The scaleX method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the x-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleX(float, boolean)
     */
    public final void scaleX(final float scalar, final boolean keepAspect) {
        if (keepAspect) {
            this.scale(scalar, scalar, scalar);
        } else {
            this.scale(scalar, 1.0f, 1.0f);
        }
    }

    /**
     * The scaleXTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the x-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleXTo(float, boolean)
     */
    public final void scaleXTo(final float dim, final boolean keepAspect) {
        final float ratio = dim / this.xDimension;
        this.scaleX(ratio, keepAspect);
    }

    /**
     * The scaleY method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the y-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleY(float, boolean)
     */
    public final void scaleY(final float scalar, final boolean keepAspect) {
        if (keepAspect) {
            this.scale(scalar, scalar, scalar);
        } else {
            this.scale(1.0f, scalar, 1.0f);
        }
    }

    /**
     * The scaleYTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the y-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleYTo(float, boolean)
     */
    public final void scaleYTo(final float dim, final boolean keepAspect) {
        final float ratio = dim / this.yDimension;
        this.scaleY(ratio, keepAspect);
    }

    /**
     * The scaleZ method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the z-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZ(float, boolean)
     */
    public final void scaleZ(final float scalar, final boolean keepAspect) {
        if (keepAspect) {
            this.scale(scalar, scalar, scalar);
        } else {
            this.scale(1.0f, 1.0f, scalar);
        }
    }

    /**
     * The scaleZTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the z-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZTo(float, boolean)
     */
    public final void scaleZTo(final float dim, final boolean keepAspect) {
        final float ratio = dim / this.zDimension;
        this.scaleZ(ratio, keepAspect);
    }

    /**
     * Setter for currentModel.
     * 
     * @param model the currentModel to set
     */
    public final void setCurrentModel(final DDSOBJModel model) {
        this.currentModel = model;
    }

    /**
     * Setter for currentOrient.
     * 
     * @param orientation the currentOrient to set
     */
    public final void setCurrentOrient(final Vector3D orientation) {
        this.currentOrient.setPos(orientation);
    }

    /**
     * Setter for currentUp.
     * 
     * @param upVector the currentUp to set
     */
    public final void setCurrentUp(final Vector3D upVector) {
        this.currentUp.setPos(upVector);
    }

    /**
     * The setDefaultOrientation method sets the default orientation for a
     * GenericObject so that comparison with the current orientation can serve
     * as a calculation for rotation of this GenericObject.
     * 
     * @param orientation The normalized direction that the GenericObject should
     *            be oriented originaly.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultOrientation(dds.model.simulation.Vector3D)
     */
    public final void setDefaultOrientation(final Vector3D orientation) {
        this.setCurrentOrient(orientation);
    }

    /**
     * The setDefaultOrientation method sets the default upwards vector for a
     * GenericObject so that comparison with the current upwards vector can
     * serve as a calculation for rotation of this GenericObject.
     * 
     * @param upVector The normalized direction that the GenericObject should
     *            use for up.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultUp(dds.model.simulation.Vector3D)
     */
    public final void setDefaultUp(final Vector3D upVector) {
        this.setCurrentUp(upVector);
    }

    /**
     * Setter for xDimension.
     * 
     * @param dimension the xDimension to set
     */
    public final void setXDimension(final float dimension) {
        this.scaleXTo(dimension, false);
    }

    /**
     * Setter for xScale.
     * 
     * @param scale the xScale to set
     */
    public final void setXScale(final float scale) {
        this.scaleX(scale / this.xScale, false);
    }

    /**
     * Setter for yDimension.
     * 
     * @param dimension the yDimension to set
     */
    public final void setYDimension(final float dimension) {
        this.scaleYTo(dimension, false);
    }

    /**
     * Setter for yScale.
     * 
     * @param scale the yScale to set
     */
    public final void setYScale(final float scale) {
        this.scaleY(scale / this.yScale, false);
    }

    /**
     * Setter for zDimension.
     * 
     * @param dimension the zDimension to set
     */
    public final void setZDimension(final float dimension) {
        this.scaleZTo(dimension, false);
    }

    /**
     * Setter for zScale.
     * 
     * @param scale the zScale to set
     */
    public final void setZScale(final float scale) {
        this.scaleZ(scale / this.zScale, false);
    }

    /**
     * The update method allows GenericObjects to receive updates from
     * SimObjects that own them.
     * 
     * @param object The Object calling the update.
     * @param arg The argument of the update.
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public final void update(final Observable object, final Object arg) {
        if ((object instanceof AbstractSimObject)
                && (arg instanceof SimObjectUpdate)) {
            final AbstractSimObject simobj = (AbstractSimObject) object;
            final SimObjectUpdate update = (SimObjectUpdate) arg;
            switch (update) {
            case POSITION:
                this.setPosition(simobj.getPosition());
                break;
            case ORIENTATION:
                this.setOrientation(simobj.getOrientation());
                break;
            case UPVECTOR:
                // TODO: Add upVector functionality
                break;
            default:
                this.notifyObservers(update);
            }
        }
        // TODO: Add logger?
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.boundingVerts;
    }
}
