package dds.model.simulation;

import dds.control.DDSOBJModel;

/**
 * The GenericObjectInterface interface. This interface describes how SimObjects
 * that use external models (.obj files) should pass and receive this
 * information.
 * 
 * @author Jacob Bolton
 */
public interface GenericObjectInterface {

    /**
     * The getModel3D() method returns the externally defined model.
     * 
     * @return The DDSOBJModel associated with this object.
     */
    DDSOBJModel getModel3D();

    /**
     * The loadModelFile method loads the specified file from the ResourceLoader
     * as the current 3D model to use for this object.
     * 
     * @param modelFile The .obj file to use. It MUST be in the proper location
     *            in the data repository.
     */
    void loadModelFile(String modelFile);

    /**
     * The getModelVert gets a Vector3D representation of the specified Vertex
     * in the DDSOBJModel.
     * 
     * @param vertNum The index of the Vertex to get.
     * @return A Vector3D representation of the Vertex.
     */
    Vector3D getModelVert(int vertNum);

    /**
     * The scale method allows a GenericObject to be resized.
     * 
     * @param xScalar The factor to scale along the x-axis.
     * @param yScalar The factor to scale along the y-axis.
     * @param zScalar The factor to scale along the z-axis.
     */
    void scale(float xScalar, float yScalar, float zScalar);

    /**
     * The scaleX method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the x-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleX(float scalar, boolean keepAspect);

    /**
     * The scaleY method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the y-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleY(float scalar, boolean keepAspect);

    /**
     * The scaleZ method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the z-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleZ(float scalar, boolean keepAspect);

    /**
     * The scaleTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param xDim The dimension to scale the size along the x-axis to.
     * @param yDim The dimension to scale the size along the y-axis to.
     * @param zDim The dimension to scale the size along the z-axis to.
     */
    void scaleTo(float xDim, float yDim, float zDim);

    /**
     * The scaleXTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the x-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleXTo(float dim, boolean keepAspect);

    /**
     * The scaleYTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the y-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleYTo(float dim, boolean keepAspect);

    /**
     * The scaleZTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the z-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     */
    void scaleZTo(float dim, boolean keepAspect);

    /**
     * The setDefaultOrientation method sets the default orientation for a
     * GenericObject so that comparison with the current orientation can serve
     * as a calculation for rotation of this GenericObject.
     * 
     * @param orientation The normalized direction that the GenericObject should
     *            be oriented originaly.
     */
    void setDefaultOrientation(Vector3D orientation);

    /**
     * The setDefaultOrientation method sets the default upwards vector for a
     * GenericObject so that comparison with the current upwards vector can
     * serve as a calculation for rotation of this GenericObject.
     * 
     * @param upVector The normalized direction that the GenericObject should
     *            use for up.
     */
    void setDefaultUp(Vector3D upVector);
}
