package dds.model.simulation;

import java.awt.geom.Line2D;
import java.util.List;

import dds.control.DDSOBJModel;

/**
 * The GenericVehicle class is responsible for modelling vehicles that are based
 * on Alias Wavefront .obj 3D models.
 * 
 * @author Jacob Bolton
 */
public class GenericVehicle extends AbstractVehicle implements
        GenericObjectInterface {

    /**
     * The POLYLOOP_VERTS constant is used to indicate how many verts are in the
     * poly loop.
     */
    // TODO Enhance the poly loop calculation to generate the poly loop from the
    // generic object's verts.
    private static final int POLYLOOP_VERTS = 4;

    /**
     * The VEHICLE_LENGTH.
     */
    private static final float VEHICLE_LENGTH = 5.5f;

    /**
     * The VEHICLE_WIDTH.
     */
    private static final float VEHICLE_WIDTH = 3.5f;

    /**
     * The handleCollision method responds to a detected collision. It
     * collaborates with SmartPolygon by invoking accessor 5.55.5on information
     * modified when intersectsPoly was last called on obj. This method also
     * does cleanup with cached vector calculations stored in class
     * SmartPolygon.
     * 
     * @param obj the object that collided with me
     * @param data in this case, my position prior to collision (Vector3D)
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    public final void handleCollision(final Road obj, final Object data) {
        this.updateObjectUnderMe(obj);
    }

    /**
     * The handleCollision method responds to a detected collision. It
     * collaborates with SmartPolygon by invoking accessor 5.55.5on information
     * modified when intersectsPoly was last called on obj. This method also
     * does cleanup with cached vector calculations stored in class
     * SmartPolygon.
     * 
     * @param intersection the object that collided with me
     * @param data in this case, my position prior to collision (Vector3D)
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    public final void handleCollision(final Intersection intersection,
            final Object data) {
        setLastIntersection((Intersection) intersection);
        this.updateObjectUnderMe(intersection);
    }

    /**
     * The updateObjectUnderMe method is responsible for updating the object
     * under this object.
     * 
     * @param obj The object to set to be under this one.
     */
    private void updateObjectUnderMe(final AbstractSimObject obj) {
        if (this.getObjectUnderMe() == null) {
            this.setObjectUnderMe(obj);
        }
    }

    /**
     * The handleCollision method responds to a detected collision. It
     * collaborates with SmartPolygon by invoking accessor 5.55.5on information
     * modified when intersectsPoly was last called on obj. This method also
     * does cleanup with cached vector calculations stored in class
     * SmartPolygon.
     * 
     * @param obj the object that collided with me
     * @param data in this case, my position prior to collision (Vector3D)
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    public final void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Comment the internals of handleCollision in GenericVehicle.
        if (obj instanceof Road) {
            this.handleCollision((Road) obj, data);
        } else if (obj instanceof Intersection) {
            this.handleCollision((Intersection) obj, data);
        } else {
            if (data != null) {
                final SmartPolygon myPoly = this.getPolyLoop();
                SmartPolygon collidedPoly = null;
                final Vector3D lastPosition = (Vector3D) data;

                if (obj instanceof Block) {
                    collidedPoly = ((Block) obj).getCollidedPolyloop();
                    this.setDriverHitBuilding(true);
                } else {
                    collidedPoly = obj.getPolyLoop();
                    if (obj instanceof GenericVehicle) {
                        // TODO Do we need to know if the driver hit a car?
                        // TODO This should probably be in the Simulation since
                        // it has to do with the driver?
                        this.setDriverHitCar(true);
                    }
                }

                if (myPoly.getInsidePoly()) {
                    final Line2D.Float testLine = new Line2D.Float(lastPosition
                            .getXpos(), lastPosition.getYpos(), this
                            .getPosition().getXpos(), this.getPosition()
                            .getXpos());
                    // final Line2D.Float testLine = new Line2D.Float(lastO
                    // .getXpos(), lastO.getYpos(), polyB.getOriginX(),
                    // polyB.getOriginY());
                    collidedPoly.findInternalNormal(testLine);
                }

                final Line2D.Float normal = collidedPoly.getBrokenNormal();

                // This is the adjustment for the polyloop collision.
                // lastO.setXpos(polyB.getOriginX());
                // lastO.setYpos(polyB.getOriginY());
                // polyB.moveMe(normal.x2, normal.y2);
                this.moveMe(normal.x2, normal.y2, 0);

                collidedPoly.unbreak();
            }
        }
    } // end of method handleCollision

    /**
     * The computePolyLoop method computes the 2D bounding polyloop for this
     * object. This effectively sets up this object for collision detection.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    protected final void computePolyLoop() {
        // TODO Comment the internals of computePolyLoop in GenericVehicle.
        // TODO Write a calculation to compute a poly loop from the generic's
        // verts rather than depend on a static calculation.

        this.setPolyLoop(new SmartPolygon(POLYLOOP_VERTS));

        final float xCoord = this.getPosition().getXpos();
        final float yCoord = this.getPosition().getYpos();

        this.getPolyLoop().addPoint(xCoord - VEHICLE_LENGTH,
                yCoord - VEHICLE_WIDTH);
        this.getPolyLoop().addPoint(xCoord + VEHICLE_LENGTH,
                yCoord - VEHICLE_WIDTH);
        this.getPolyLoop().addPoint(xCoord + VEHICLE_LENGTH,
                yCoord + VEHICLE_WIDTH);
        this.getPolyLoop().addPoint(xCoord - VEHICLE_LENGTH,
                yCoord + VEHICLE_WIDTH);
    } // end of method computePolyLoop

    /**
     * The clearOtherVerts.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    // TODO Comment clearOtherVerts in GenericVehicle.
    @Override
    protected void clearOtherVerts() {
        // TODO Determine how to write clearOtherVerts in GenericVehicle.
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts. This method does NOT need to clear the older verts OR set
     * the vertsCalculated flag becuase that functionality has been standardized
     * in the calculateVerts method defined in this class. This method just
     * needs to calculate the verts. The verts that are calculate MUST be stored
     * as local coordinates. These coordinates will be offset by the current
     * position when being used for collision detection of rendering.
     * 
     * @see dds.model.simulation.AbstractSimObject#myCalculateVerts()
     */
    @Override
    public final void myCalculateVerts() {
        for (int i = 0; i < this.genObj.getModel3D().getVertexsize(); i++) {
            this.getVerts().add(this.getModelVert(i));
        }
    }

    /**
     * The accelerate method causes the Drivable to change it's forward
     * acceleraion by a specified amount.
     * 
     * @param amount The amount to accelerate by. This amount is additive, NOT
     *            multiplicative.
     * @see dds.model.simulation.Drivable#accelerate(float)
     */
    public final void accelerate(final float amount) {
        this.setAcceleration(amount);
    }

    /**
     * The turn method causes the Drivable to change it's direction by a
     * specified amount.
     * 
     * @param amount Amount to turn. Values < 0 are for left turns; values > 0
     *            are for right turn.
     * @see dds.model.simulation.Drivable#turn(float)
     */
    public final void turn(final float amount) {
        this.basicTurn(amount);
    }

    /**
     * The updateKinetics method updates the kinetics according to the time
     * passed.
     * 
     * @param time The time passed
     * @see dds.model.simulation.AbstractMovableObject#updateKinetics(float)
     */
    @Override
    public final void updateKinetics(final float time) {
        this.basicUpdateKinetics(time);
    }

    /**
     * The genObj is the GenericObject that has the model for this Object.
     */
    private GenericObject genObj;

    /**
     * Constructor for GenericMovableObject.
     * 
     * @param idString The unique ID for this object within the simulation.
     * @param modelFile The name of the model file to use.
     */
    public GenericVehicle(final String idString, final String modelFile) {
        super(idString);
        this.genObj = new GenericObject(idString, modelFile);
    }

    /**
     * The destruct method unlinks all of this object's references to make
     * garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    @Override
    protected final void destruct() {
        this.clearVerts();
        this.genObj.destruct();
    }

    /**
     * The getModel3D() method returns the externally defined model.
     * 
     * @return The DDSOBJModel associated with this object.
     * @see dds.model.simulation.GenericObjectInterface#getModel3D()
     */
    public final DDSOBJModel getModel3D() {
        return this.genObj.getModel3D();
    }

    /**
     * The getModelVert gets a Vector3D representation of the specified Vertex
     * in the DDSOBJModel.
     * 
     * @param vertNum The index of the Vertex to get.
     * @return A Vector3D representation of the Vertex.
     * @see dds.model.simulation.GenericObjectInterface#getModelVert(int)
     */
    public final Vector3D getModelVert(final int vertNum) {
        return this.genObj.getModelVert(vertNum);
    }

    /**
     * The loadModelFile method loads the specified file from the ResourceLoader
     * as the current 3D model to use for this object.
     * 
     * @param modelFile The .obj file to use. It MUST be in the proper location
     *            in the data repository.
     * @see dds.model.simulation.GenericObjectInterface
     *      #loadModelFile(java.lang.String)
     */
    public final void loadModelFile(final String modelFile) {
        this.genObj.loadModelFile(modelFile);
    }

    /**
     * The scale method allows a GenericObject to be resized.
     * 
     * @param xScalar The factor to scale along the x-axis.
     * @param yScalar The factor to scale along the y-axis.
     * @param zScalar The factor to scale along the z-axis.
     * @see dds.model.simulation.GenericObjectInterface#scale(float, float,
     *      float)
     */
    public final void scale(final float xScalar, final float yScalar,
            final float zScalar) {
        this.genObj.scale(xScalar, yScalar, zScalar);
    }

    /**
     * The scaleTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param xDim The dimension to scale the size along the x-axis to.
     * @param yDim The dimension to scale the size along the y-axis to.
     * @param zDim The dimension to scale the size along the z-axis to.
     * @see dds.model.simulation.GenericObjectInterface#scaleTo(float, float,
     *      float)
     */
    public final void scaleTo(final float xDim, final float yDim,
            final float zDim) {
        this.genObj.scaleTo(xDim, yDim, zDim);
    }

    /**
     * The scaleX method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the x-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleX(float, boolean)
     */
    public final void scaleX(final float scalar, final boolean keepAspect) {
        this.genObj.scaleX(scalar, keepAspect);
    }

    /**
     * The scaleXTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the x-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleXTo(float, boolean)
     */
    public final void scaleXTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleXTo(dim, keepAspect);
    }

    /**
     * The scaleY method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the y-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleY(float, boolean)
     */
    public final void scaleY(final float scalar, final boolean keepAspect) {
        this.genObj.scaleY(scalar, keepAspect);
    }

    /**
     * The scaleYTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the y-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleYTo(float, boolean)
     */
    public final void scaleYTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleYTo(dim, keepAspect);
    }

    /**
     * The scaleZ method allows a GenericObject to be resized.
     * 
     * @param scalar The factor to scale along the z-axis.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZ(float, boolean)
     */
    public final void scaleZ(final float scalar, final boolean keepAspect) {
        this.genObj.scaleZ(scalar, keepAspect);
    }

    /**
     * The scaleZTo method allows a GenericObject to be resized to a specific
     * size.
     * 
     * @param dim The dimension to scale the size along the z-axis to.
     * @param keepAspect True if the x:y:z ratio should be preserved.
     * @see dds.model.simulation.GenericObjectInterface#scaleZTo(float, boolean)
     */
    public final void scaleZTo(final float dim, final boolean keepAspect) {
        this.genObj.scaleZTo(dim, keepAspect);
    }

    /**
     * The setDefaultOrientation method sets the default orientation for a
     * GenericObject so that comparison with the current orientation can serve
     * as a calculation for rotation of this GenericObject.
     * 
     * @param orientation The normalized direction that the GenericObject should
     *            be oriented originaly.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultOrientation(dds.model.simulation.Vector3D)
     */
    public final void setDefaultOrientation(final Vector3D orientation) {
        this.genObj.setDefaultOrientation(orientation);
    }

    /**
     * The setDefaultOrientation method sets the default upwards vector for a
     * GenericObject so that comparison with the current upwards vector can
     * serve as a calculation for rotation of this GenericObject.
     * 
     * @param upVector The normalized direction that the GenericObject should
     *            use for up.
     * @see dds.model.simulation.GenericObjectInterface
     *      #setDefaultUp(dds.model.simulation.Vector3D)
     */
    public final void setDefaultUp(final Vector3D upVector) {
        this.genObj.setDefaultUp(upVector);
    }

    /**
     * Getter for genObj.
     * 
     * @return the genObj
     */
    public final GenericObject getGenObj() {
        return genObj;
    }

    /**
     * Setter for genObj.
     * 
     * @param obj the genObj to set
     */
    public final void setGenObj(final GenericObject obj) {
        this.genObj = obj;
    }

    /**
     * The brake.
     * 
     * @param amount The amount to brake.
     * @see dds.model.simulation.Drivable#brake(float)
     */
    public final void brake(final float amount) {
        this.setBraking(amount);
    }

    /**
     * The getAccelerationAmount.
     * 
     * @return the accelerationAmount.
     * @see dds.model.simulation.AbstractVehicle#getAccelerationAmount()
     */
    @Override
    public final float getAccelerationAmount() {
        return this.basicGetAccelerationAmount();
    }

    /**
     * The getTurnAmount.
     * 
     * @return The turnAmount.
     * @see dds.model.simulation.AbstractVehicle#getTurnAmount()
     */
    @Override
    public final float getTurnAmount() {
        return this.basicGetTurnAmount();
    }

    /**
     * The move method was supposed to be used for calculating motion, but this
     * functionality should be done in simulate instead.
     * 
     * @param time The delta t for calculating motion.
     * @see dds.model.simulation.AbstractMovableObject#move(float)
     */
    @Override
    public final void move(final float time) {
        this.basicMove(time);
        this.reorient();
        this.calculateAll();
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.genObj.getBoundingVerts();
    }

}
