package dds.model.simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * The Intersection class. This class is responsible for modelling an
 * intersection of roads. It is defined as a point that the roads(lines) connect
 * to, but for rendering and other detection the vertices need to be calculated
 * in order to give the intersection three dimensions.
 * 
 * @author Jacob Bolton
 */
public class Intersection extends AbstractSimObject 
        implements Comparator<Road> {

    /**
     * The roads that connect to this intersection.
     */
    private final SortedSet<Road> roads;

    /**
     * The roadVerts are the verts according to the Road they belong with.
     */
    private Map<Road, List<Vector3D>> roadVerts;

    /**
     * The trafficLights associated with this intersection.
     */
    private final Collection<TrafficLight> trafficLights;

    /**
     * Constructor for Intersection.
     * 
     * @param idString The UNIQUE ID for this object within the simulation.
     * @param pos The position of this Intersection
     */
    public Intersection(final String idString, final Vector3D pos) {
        super(idString);
        this.roads = new TreeSet<Road>(this);
        this.trafficLights = new ArrayList<TrafficLight>();
        this.roadVerts = new HashMap<Road, List<Vector3D>>();
        this.setVertsCalculated(false);
        this.setPosition(pos);
    }

    /**
     * The addRoad method allows a new road to be associated with an
     * intersection.
     * 
     * @param newRoad A new Road to associate with.
     */
    public final void addRoad(final Road newRoad) {
        this.roads.add(newRoad);
    }

    /**
     * The calculateVerts method calculates the verts for this Intersection.
     * 
     * @see dds.model.simulation.AbstractSimObject#calculateVerts()
     */
    public final void myCalculateVerts() {
        // The intersection's verts depend on the conecting roads. Start
        // calculation by iterating though the sorted list of roads. The sorting
        // causes this traversal to take place in a counter clockwise direction
        // around the intersection's center.
        for (Road road : this.roads) {
            // Get the two roads next to the current one.
            final Vector3D roadCCW = this.getIntersection(road, true);
            final Vector3D roadCW = this.getIntersection(road, false);

            // Calculate the rotations of the adjacent Roads fom this one.
            final float magCCW = roadCCW
                    .dotProduct(road.getDirectionFrom(this));
            final float magCW = roadCW.dotProduct(road.getDirectionFrom(this));

            // Add the verts from the intersections.
            this.addVert(road.getEndsFrom(this, magCCW, magCW).get(0));
            this.addVerts(road.getEndsFrom(this, magCCW, magCW));
        }
    }

    /**
     * The clearAllVerts method clears any and all vert lists/collections for
     * this object.
     */
    protected final void clearOtherVerts() {
        this.roadVerts.clear();
    }

    /**
     * The compare method provides this Object with the ability to sort Roads
     * based on their angle coming off of this Intersection.
     * 
     * @param roadA The first Road.
     * @param roadB The second Road.
     * @return 0 if they have the same angle, -1 if A < B, and 1 if A > B.
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public final int compare(final Road roadA, final Road roadB) {
        // Compare the Roads using a difference of their angles.
        float result = roadA.getAngle(this) - roadB.getAngle(this);

        // Check to see if the result is zero to prevent a division by zero
        // problem.
        if (result != 0) {
            // Result is not zero so scale the resulting angle difference to
            // [0-1].
            result = result / Math.abs(result);
        }

        // Return the result.
        return (int) result;
    }

    /**
     * The destruct method decomposes this object for easier garbage collection.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    protected final void destruct() {
        this.roads.clear();
        this.trafficLights.clear();
        this.roadVerts.clear();
    }

    /**
     * The detectCollision method checks this object's space against another's
     * to determine if there has been a collision.
     * 
     * @param obj SimObject to detect collision with.
     * @return True on collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    public final boolean detectCollision(final AbstractSimObject obj) {
        // TODO Determine if this needs to be written out.
        return false;
    }

    /**
     * The handleCollision.
     * 
     * @param obj The obj.
     * @param data The data.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    protected void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Determine if this needs to be written out.
    }

    /**
     * The computePolyLoop.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    // TODO Comment me!
    protected final void computePolyLoop() {
        SmartPolygon thePoly = null;
        final float pointX = this.getPosition().getXpos();
        final float pointY = this.getPosition().getYpos();

        setPolyLoop(new SmartPolygon(this.getVerts().size()));
        thePoly = getPolyLoop();

        for (Vector3D nextVert : this.getVerts()) {
            thePoly.addPoint(pointX + nextVert.getXpos(), pointY
                    + nextVert.getYpos());
        }
    }

    /**
     * The getIntersection method will get the point of intersection of the
     * specified Road and the one around this Intersection in either the CCW or
     * CW direction.
     * 
     * @param road The Road.
     * @param ccw True for CCW false for CW.
     * @return The intersection point bewteen the two roads.
     */
    private Vector3D getIntersection(final Road road, final boolean ccw) {
        // Get the adjacent Road in the specified direction.
        final Road next = this.getNext(road, ccw);

        // Get the each Road's directional vector.
        final Vector3D roadM = road.getDirectionFrom(this);
        final Vector3D nextM = next.getDirectionFrom(this);

        // Get each Road's offset vector.
        final Vector3D roadB = road.getNormalAround(this, ccw);
        final Vector3D nextB = next.getNormalAround(this, !ccw);

        // Scale the offset vectors for the Road's thickness.
        roadB.scale(road.getThickness());
        nextB.scale(next.getThickness());

        // Calculate the ray intersection.
        Vector3D result = Vector3D.rayIntersection(roadM, roadB, nextM, nextB);

        // Check to see if the intersection calculation produced a result.
        if (result == null) {
            // It didn't so compare the thickness of the Roads and use the
            // thickest one's coordinates for the intersection coordnates.
            if (road.getThickness() > next.getThickness()) {
                result = roadM;
                result.scale(road.getThickness());
                result.add(roadB);
            } else {
                result = nextM;
                result.scale(-next.getThickness());
                result.add(nextB);
            }
        }

        // Return the result.
        return result;
    }

    /**
     * The getNext method gets the Road next to the specified one in either the
     * CCW or CW direction.
     * 
     * @param road The Road to start from.
     * @param ccw True for the CCW direction, false for CW.
     * @return The adjacent Road.
     */
    private Road getNext(final Road road, final boolean ccw) {
        // Get a List of the Roads attached to this intersection so they can be
        // numerically indexed.
        final List<Road> roadList = new ArrayList<Road>(this.roads);

        // Get the ndex of the scpecified Road.
        int roadIndex = roadList.indexOf(road);

        // Increment or decrement the roadIndex based on whether or not we're
        // looking for the counterclockwise or clockwise Road.
        if (ccw) {
            roadIndex++;
        } else {
            roadIndex--;
        }

        // Make sure the new index is within the bounds of the road list. If
        // not, then correct it.
        if (roadIndex < 0) {
            roadIndex += roadList.size();
        } else if (roadIndex >= roadList.size()) {
            roadIndex -= roadList.size();
        }

        // Return the Road at the resulting index.
        return roadList.get(roadIndex);
    }

    /**
     * The getRoad method returns the Road at the specified index.
     * 
     * @param index The index.
     * @return The road.
     */
    public final Road getRoad(final int index) {
        // Get a List of the Roads attached to this intersection so they can be
        // numerically indexed.
        final List<Road> rList = new ArrayList<Road>(this.roads);

        // Return the Road at the specified index.
        return rList.get(index);
    }

    /**
     * Getter for roads.
     * 
     * @return the roads
     */
    public final Collection<Road> getRoads() {
        return roads;
    }

    /**
     * Getter for roadVerts.
     * 
     * @return the roadVerts
     */
    public final Map<Road, List<Vector3D>> getRoadVerts() {
        return roadVerts;
    }

    /**
     * The getRoadVerts returns the verts of this Intersection that are
     * associated with a particular Road.
     * 
     * @param road The Road to get verts for.
     * @return A List of verts for the Road. There will only be two in the list,
     *         and the first will be for the left side, and the second the right
     *         side.
     */
    public final List<Vector3D> getRoadVerts(final Road road) {
        // Check to see if verts have been calculated.
        if (!this.isVertsCalculated()) {
            // They haven't so calculate them.
            this.calculateVerts();
        }

        // Return the road verts for the specified Road.
        return this.roadVerts.get(road);
    }

    /**
     * The getThickness method returns the greatest thickness from all the Roads
     * associated with this Intersection.
     * 
     * @return The greatest Road thickness.
     */
    public final float getThickness() {
        // Start with 0 thickness.
        float thickness = 0.0f;

        // Iterate through the Roads.
        for (Road road : this.roads) {
            // If the current Road's thickness is greater than the current
            // thickness then set the current thickness to this Road's.
            if (road.getThickness() > thickness) {
                thickness = road.getThickness();
            }
        }

        // Return the resulting thickness.
        return thickness;
    }

    /**
     * Getter for trafficLights.
     * 
     * @return the trafficLights
     */
    public final Collection<TrafficLight> getTrafficLights() {
        return trafficLights;
    }

    /**
     * The removeRoad method removes an association with a particular Road.
     * 
     * @param road The Road to dissociate from.
     */
    public final void removeRoad(final Road road) {
        // Remove the Road and recalculate the verts.
        this.roads.remove(road);
        this.calculateVerts();
    }

    /**
     * Setter for the roads.
     * 
     * @param theRoads the roads to set
     */
    @Deprecated
    public final void setRoads(final Collection<Road> theRoads) {
        this.roads.clear();
        this.roads.addAll(theRoads);
    }

    /**
     * Setter for roadVerts.
     * 
     * @param rVerts the roadVerts to set
     */
    @Deprecated
    public final void setRoadVerts(final Map<Road, List<Vector3D>> rVerts) {
        this.roadVerts = rVerts;
    }

    /**
     * Setter for trafficLights.
     * 
     * @param tlights the trafficLights to set
     */
    @Deprecated
    public final void setTrafficLights(final Collection<TrafficLight> tlights) {
        this.trafficLights.clear();
        this.trafficLights.addAll(tlights);
    }

    /**
     * The toString.
     * 
     * @return A String representation of this object.
     * @see java.lang.Object#toString()
     */
    @Override
    public final String toString() {
        return "Intersection: " + this.getIdStr() + " " + this.getPosition();
    }

    /**
     * The getBoundingVerts method returns the boundings verts for this object,
     * which happen to be the same as the verts used for this object. So this
     * implementation just reurns the object's verts.
     * 
     * @return The object's verts as the bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        // Return the object's verts as the bounding verts.
        return this.getVerts();
    }

} // end class definition Intersection
