package dds.model.simulation;

/**
 * The Kinetics class is responsible for tracking and calculating the kinematics
 * of the simulation. It's goal is to emulate real-enough physics that exude
 * enough reality to be effectively believable. The kinetics class is currently
 * able to track information pertaining to acceleration mass and velociy. Since
 * linear motion is simple enough, we're using a true-to-life model for motion
 * using a linear interpolation of velocity and acceleration to produce changes
 * in position. This may need to change if we add more complex physical
 * calculations. This class is also repsonsible for for providing unit
 * conversion classes to go between the meters per second units used to do
 * kinetics calculations and the miles per hour units which are used for
 * relating speeds to the user.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public class Kinetics {
    /**
     * The MPS_TO_MPH is ratio of meters-per-second to miles-per-hour.
     */
    public static final float MPS_TO_MPH = 2.24f;

    /**
     * The MPH_TO_MPS is the ratio of miles-per-hour to meters-per-second.
     */
    public static final float MPH_TO_MPS = 0.447f;

    /**
     * The SIM_DISTANCE defines the conversion factor from going between map
     * units to meters. Right now this is constant; however in future versions
     * it may be useful to make this a changeable parameter.
     */
    // TODO Consider making this a changeable parameter, perhaps a parameter
    // that can be set from the map XML.
    public static final float SIM_DISTANCE = 2.0f;

    /**
     * The acceleration vector for this Kinematics (in m / s^2 units).
     */
    private final Vector3D acceleration;

    /**
     * The registerA is a Vector3D that is used for temporary storage during
     * vector calculations. Since vector calculations are probably going to be
     * very abundant in this class, we provide a set of registers, which can be
     * used a temp variables for calculations. By reuseing these register
     * objects we reduce object creation and destruction, which will hopefullly
     * improve performance.
     */
    private transient Vector3D registerA;

    /**
     * The registerB is a Vector3D that is used for temporary storage during
     * vector calculations. Since vector calculations are probably going to be
     * very abundant in this class, we provide a set of registers, which can be
     * used a temp variables for calculations. By reuseing these register
     * objects we reduce object creation and destruction, which will hopefullly
     * improve performance.
     */
    private transient Vector3D registerB;

    /**
     * The mass of the Kinematics (kg).
     */
    private float mass;

    /**
     * The velocity for this Kinematics (m / s).
     */
    private final Vector3D velocity;

    /**
     * Constructor for Kinetics that initializes all the storage necessary for
     * this instance. After this is executed, this class should NOT cause any
     * further object to be created or destroyed for any reason. This is for
     * performance benefits.
     */
    public Kinetics() {
        // Mandatory call to super().
        super();

        // Allocate and initialize storage space.
        this.acceleration = new Vector3D();
        this.velocity = new Vector3D();
        this.registerA = new Vector3D();
        this.registerB = new Vector3D();

        // Set default mass.
        this.mass = 1;
    }

    /**
     * Constructor for Kinetics that allows for the setting of the parameters.
     * 
     * @param vvec Velocity vector in m/s.
     * @param mkg Mass in kg
     * @param avec Acceleration vector in m/s^2
     */
    public Kinetics(final Vector3D vvec, final float mkg, final Vector3D avec) {
        // Call to this() in order to initialize storage and call super()
        // through code re-use.
        this();

        // Set the parameters of the Kinetics from the data passed in.
        this.velocity.setPos(vvec);
        this.acceleration.setPos(avec);
        this.mass = mkg;
    }

    /**
     * The duplicate method creates a copy of this kinetics object. This is a
     * convenient way to copy this object, but it results in object creation. It
     * IS, however, necessary for use in the view in order to send a copy of the
     * kinetics to the renderer for interpolation between simulation updates.
     * 
     * @return A new copy of this Kinetics.
     */
    public final Kinetics duplicate() {
        return new Kinetics(this.velocity, this.mass, this.acceleration);
    }

    /**
     * Getter for acceleration.
     * 
     * @return the acceleration
     */
    public final Vector3D getAcceleration() {
        return acceleration;
    }

    /**
     * Getter for mass.
     * 
     * @return the mass
     */
    public final float getMass() {
        return mass;
    }

    /**
     * Getter for velocity.
     * 
     * @return the velocity
     */
    public final Vector3D getVelocity() {
        return velocity;
    }

    /**
     * The linearMotion method takes in a position and manipulates it according
     * to linear kinematics. It should be noted that this will affect the
     * velocity of this Kinematics. The provided position coordinate will be
     * updated according to the calculated position change and scaled
     * appropriately for the map.
     * 
     * @param position The position to update.
     * @param time The change in time.
     */
    public final void linearMotion(final Vector3D position, final float time) {
        // Preserve velocity for the change in distance calculation later.
        this.registerB.setPos(this.velocity);

        // This is the change in velocity calculation. It follows the
        // mathematical form of: v(t) = v_0 + at
        // registerA is designated to contain the at component of the equation.
        this.registerA.setPos(this.acceleration);
        this.registerA.scale(time);
        // So simply add the at component to the original velocity to get the
        // result.
        this.velocity.add(this.registerA);

        // This is the change in position calculation. It follows the
        // mathematical form of: x(t) = x_0 + vt + (at^2) / 2
        // Since the at component is already in registerA all we need to do is
        // scale it by t/2 in order to form the (at^2)/2 component. Since this
        // is a change in position is must be scaled to map coordinates, thus
        // the additional scaling of SIM_DISTANCE.
        this.registerA.scale(SIM_DISTANCE * time / 2.0f);
        // The original velocity is still stored in registerB so it just needs
        // to be scaled by t in order to get the vt component. Since this is a
        // change in position is must be scaled to map coordinates, thus the
        // additional scaling of SIM_DISTANCE.
        this.registerB.scale(SIM_DISTANCE * time);
        // Add the positional changes to the original position.
        position.add(this.registerA);
        position.add(this.registerB);
    }

    /**
     * Setter for acceleration.
     * 
     * @param vec the acceleration to set
     */
    public final void setAcceleration(final Vector3D vec) {
        // Reuse the velocity vector to prevent object creation/destruction.
        this.acceleration.setPos(vec);
    }

    /**
     * Setter for mass.
     * 
     * @param kgs the mass to set
     */
    public final void setMass(final float kgs) {
        this.mass = kgs;
    }

    /**
     * Setter for velocity.
     * 
     * @param vec the velocity to set
     */
    public final void setVelocity(final Vector3D vec) {
        // Reuse the velocity vector to prevent object creation/destruction.
        this.velocity.setPos(vec);
    }

    /**
     * The mphToMps method converts miles-per-hour to meters-per-second.
     * 
     * @param mph The number of miles / hour.
     * @return The number of meters / second.
     */
    public static final float mphToMps(final float mph) {
        // Convert to meters per second using the defined conversion constant.
        return mph * MPH_TO_MPS;
    }

    /**
     * The mpsToMph method converts meters-per-second to miles-per-hour.
     * 
     * @param mps The number of meters / second.
     * @return The number of miles / hour.
     */
    public static final float mpsToMph(final float mps) {
        // Convert to miles per hour using the defined conversion constant.
        return mps * MPS_TO_MPH;
    }
}
