package dds.model.simulation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import dds.control.DDSApplet;

/**
 * The Road class represents the model's conceptualization of a road within the
 * context of the simulation.
 * 
 * @author Jacob Bolton
 */
public class Road extends AbstractSimObject implements Comparable {

    /**
     * The DEFAULT_NAME constant defines which street name should be given to a
     * road by default..
     */
    public static final String DEFAULT_NAME = "Peachtree St.";

    /**
     * The DEFAULT_SPEED_LIMIT is the default speed limit to use for new roads.
     */
    public static final int DEFAULT_SPEED = 25;

    /**
     * The DEFAULT_THICKNESS constant defines what thickness name should be
     * given to a road by default..
     */
    public static final float DEFAULT_THICKNESS = 5.0f;

    /**
     * The POLYLOOP_SIZE constant defines the size of a Road's SmartPolygon,
     * which is used for collision detection.
     */
    private static final int POLYLOOP_SIZE = 4;

    /**
     * The THREE constant has a value of 3.
     */
    private static final int THREE = 3;

    /**
     * The TRAFFIC_A_TO_B specifies constant that the expected traffic flow on
     * this road is one-way from intersection A towards intersection B.
     */
    public static final String TRAFFIC_A_TO_B = "ab";

    /**
     * The TRAFFIC_B_TO_A constant specifies that the expected traffic flow on
     * this road is one-way from intersection B towards intersection A.
     */
    public static final String TRAFFIC_B_TO_A = "ba";

    /**
     * The TRAFFIC_BIDIRECTIONAL specifies that the expected traffic flow on
     * this road is bidirectional.
     */
    public static final String TRAFFIC_BI = "bi";

    /**
     * The traffic direction of the road.
     */
    private String direction;

    /**
     * The ending Intersection that defines this Road.
     */
    private Intersection end;

    /**
     * The endVerts is the List of verts that this Road contributes to the
     * ending Intersection.
     */
    private List<Vector3D> endVerts;

    /**
     * The number of lanes that appear on this road.
     */
    private int lanes;

    /**
     * The leftBlock is the Block on the left as per a winged-edge definition of
     * this Road.
     */
    private Block leftBlock;

    /**
     * The name of the Street. This is different from the ID since there can be
     * many streets with the same name, but all must have different IDs.
     */
    private String name;

    /**
     * The rightBlock is the Block on the right as per a winged-edge definition
     * of this Road.
     */
    private Block rightBlock;

    /**
     * The rightNormal the normalized vector that is directed towards the right
     * of this road.
     */
    private Vector3D rightNormal;

    /**
     * The roadSigns Collection contains all of the road signs that are attached
     * to this road.
     */
    private Collection<RoadSign> roadSigns;

    /**
     * The speedlimit of the road in MPH.
     */
    private int speedlimit;

    /**
     * The starting Intersection that defines this Road.
     */
    private Intersection start;

    /**
     * The startVerts is the List of verts that this Road contributes to the
     * starting Intersection.
     */
    private List<Vector3D> startVerts;

    /**
     * The thickness of the Road from its center to edge (in meters).
     */
    private float thickness;

    /**
     * Constructor for Road that takes in an ID String.
     * 
     * @param idStr ID string from XML.
     */
    private Road(final String idStr) {
        super(idStr);
        this.roadSigns = new ArrayList<RoadSign>();
        this.startVerts = new ArrayList<Vector3D>();
        this.endVerts = new ArrayList<Vector3D>();
    }

    /**
     * Constructor for Road. Does NOT set individual values for lanes,
     * direction, street name, thickness, or speed limit. These values are not
     * considerred critical and must be set using their appropriate setters.
     * 
     * @param idString The XML id of the SimObject
     * @param intA The starting Intersection.
     * @param intB The ending Intersection.
     */
    public Road(final String idString, final Intersection intA,
            final Intersection intB) {
        this(idString);
        this.direction = TRAFFIC_BI;
        this.end = intB;
        this.lanes = 1;
        this.name = DEFAULT_NAME;
        this.speedlimit = DEFAULT_SPEED;
        this.start = intA;
        this.thickness = DEFAULT_THICKNESS;
        // Set the Road's position to the midpoint between Intersections
        this.setPosition(Vector3D.midpoint(intA.getPosition(), intB
                .getPosition()));
        // Set the Road's orientation i.e. a normalized vector from intA to intB
        this.setOrientation(Vector3D.subtract(intB.getPosition(), intA
                .getPosition()));
        this.getOrientation().normalize();
        // Set the Road's rightNormal, an orthoganol Vector pointing right of
        // the orientation.
        this.rightNormal = Vector3D.cross(this.getOrientation(), Vector3D
                .getZAxis());
        this.rightNormal.normalize();
        // Add an association to this Road to it's starting and ending
        // Intersections
        intA.addRoad(this);
        intB.addRoad(this);
    }

    /**
     * The addRoadSigns method adds a stop sign at each intersection where
     * necessary for this road. It also calls the appropriate helper methods to
     * compute the locations of these stop signs.
     */
    public final void addRoadSigns() {
        final RoadSign startStopSign = new RoadSign("stop sign: "
                + start.toString());
        final RoadSign endStopSign = new RoadSign("stop sign: "
                + end.toString());

        startStopSign.setType(RoadSign.STOP_SIGN);
        endStopSign.setType(RoadSign.STOP_SIGN);

        if (this.direction.equals(Road.TRAFFIC_A_TO_B)) {
            // put a stop sign at intersection b
            this.roadSigns.add(endStopSign);
            this.calculateRoadSignPosition(endStopSign, end, 0);
        } else if (this.direction.equals(Road.TRAFFIC_B_TO_A)) {
            // put a stop sign at intersection a
            this.roadSigns.add(startStopSign);
            this.calculateRoadSignPosition(startStopSign, start, 1);
        } else if (this.direction.equals(Road.TRAFFIC_BI)) {
            // put a stop sign at intersection a & b
            this.roadSigns.add(startStopSign);
            this.roadSigns.add(endStopSign);
            this.calculateRoadSignPosition(startStopSign, start, 1);
            this.calculateRoadSignPosition(endStopSign, end, 0);
        }
    }

    /**
     * The calculateRoadSignPosition.
     * 
     * @param sign sign
     * @param ixt ixt
     * @param rightOrLeft rightOrLeft
     */
    private void calculateRoadSignPosition(final RoadSign sign,
            final Intersection ixt, final int rightOrLeft) {
        final List<Vector3D> intVerts = this.getIntVerts(ixt);
        Vector3D unitDir = null;
        Vector3D dir = null;
        Vector3D normal = null;
        final Vector3D roadDir = this.getOrientation().duplicate();

        unitDir = Vector3D.subtract(intVerts.get(1), intVerts.get(0));
        dir = unitDir.duplicate();
        unitDir.normalize();
        normal = unitDir.getNormal();

        if (rightOrLeft == 0) {
            roadDir.scale(-1f);
            dir.add(normal);
        } else {
            dir.scale(-1f);
            dir.add(normal);
        }

        sign.setPosition(dir);
        sign.reorient(roadDir);
    }

    /**
     * The clearOtherVerts.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    @Override
    protected final void clearOtherVerts() {
        // TODO Auto-generated method stub

    }

    /**
     * The compareTo allows this object to be compared with other Objects.
     * 
     * @param obj The Object to compare this one with.
     * @return An int value describing the result of the comparison.
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public final int compareTo(final Object obj) {
        // TODO Implement this for real or find a way to remove it.
        return 0;
    }

    /**
     * The computePolyLoop method sets up the SmartPolygon used for collision
     * detection.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    protected final void computePolyLoop() {
        float xCoord, yCoord;

        this.setPolyLoop(new SmartPolygon(POLYLOOP_SIZE));
        xCoord = this.getPosition().getXpos();
        yCoord = this.getPosition().getYpos();

        this.getPolyLoop().addPoint(xCoord + this.getVerts().get(0).getXpos(),
                yCoord + this.getVerts().get(0).getYpos());
        this.getPolyLoop().addPoint(xCoord + this.getVerts().get(1).getXpos(),
                yCoord + this.getVerts().get(1).getYpos());
        this.getPolyLoop().addPoint(xCoord + this.getVerts().get(2).getXpos(),
                yCoord + this.getVerts().get(2).getYpos());
        this.getPolyLoop().addPoint(
                xCoord + this.getVerts().get(THREE).getXpos(),
                yCoord + this.getVerts().get(THREE).getYpos());
    }

    /**
     * The destruct method unlinks references within this Objects collections to
     * allow for easier garbage collection.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    protected final void destruct() {
        this.roadSigns.clear();
        this.endVerts.clear();
        this.startVerts.clear();
        this.getVerts().clear();
    }

    /**
     * The detectCollision.
     * 
     * @param obj SimObject to detect collision with.
     * @return True on collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    protected final boolean detectCollision(final AbstractSimObject obj) {
        // TODO detectCollision
        return false;
    }

    /**
     * The getAngle method returns the angle of this Road off of a given
     * Intersection.
     * 
     * @param center The Intersection to use as a rotation point.
     * @return An angle off of the specified Intersection. If the specified
     *         Intersection is NOT part of this Road it will default to the
     *         angle off of this Roads starting Intersection.
     */
    public final float getAngle(final Intersection center) {
        final Vector3D vec = this.getOrientation().duplicate();
        if (center.equals(this.getEnd())) {
            vec.scale(-1);
        }
        final float sin = (float) Math.asin(vec.getYpos());
        final float cos = (float) Math.acos(vec.getXpos());
        float result;
        if (sin < 0) {
            result = (float) ((2.0 * DDSApplet.PI) - cos);
        } else {
            result = cos;
        }
        return result;
    }

    /**
     * The getBaseline creates a List of two verts that represents the base of
     * the road where building can start to appear.
     * 
     * @param left True if the baseline is for the left side of this road, false
     *            for the right.
     * @return A list of two global verts that inscribe the line.
     */
    public final List<Vector3D> getBaseline(final boolean left) {
        final List<Vector3D> result;
        if (left) {
            result = this.getLeftVerts();
        } else {
            result = this.getRightVerts();
        }

        final Intersection intA;
        if (left) {
            intA = this.start;
        } else {
            intA = this.end;
        }
        final Intersection intB;
        if (left) {
            intB = this.end;
        } else {
            intB = this.start;
        }
        final Vector3D vecA = result.get(0);
        final Vector3D vecB = result.get(1);
        vecA.subtract(intA.getPosition());
        vecA.scale(2);
        vecA.add(intA.getPosition());
        vecB.subtract(intB.getPosition());
        vecB.scale(2);
        vecB.add(intB.getPosition());
        return result;
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        return this.getVerts();
    }

    /**
     * Getter for direction.
     * 
     * @return the direction
     */
    public final String getDirection() {
        return direction;
    }

    /**
     * The getDirectionFrom method returns a Vector3D of a normalized direction
     * along this Road away from the specified Intersection.
     * 
     * @param isect The Intersection to get a direction away from.
     * @return A Vector3D along the Road.
     */
    public final Vector3D getDirectionFrom(final Intersection isect) {
        final Vector3D result = this.getOrientation().duplicate();
        if (isect.equals(this.end)) {
            result.scale(-1);
        }
        return result;
    }

    /**
     * Getter for end.
     * 
     * @return the end
     */
    public final Intersection getEnd() {
        return end;
    }

    /**
     * The getEndsFrom method creates a set of points that are a specified
     * distance from the given Intersection.
     * 
     * @param isect The intersection to move down the Road from.
     * @param distance The distance to travel down the Road.
     * @return A List of verts.
     */
    public final List<Vector3D> getEndsFrom(final Intersection isect,
            final float distance) {
        return this.getEndsFrom(isect, distance, distance);
    }

    /**
     * The getEndsFrom method creates a set of points that are a specified
     * distance from the given Intersection.
     * 
     * @param isect The intersection to move down the Road from.
     * @param distLeft The distance to travel down the Road on the left.
     * @param distRight The distance to travel down the Road on the right.
     * @return A List of verts.
     */
    public final List<Vector3D> getEndsFrom(final Intersection isect,
            final float distLeft, final float distRight) {
        final boolean forward = isect.equals(this.start);
        final List<Vector3D> result = new ArrayList<Vector3D>();
        Vector3D directLeft = this.getOrientation().duplicate();
        Vector3D directRight = this.getOrientation().duplicate();
        directLeft.scale(distLeft);
        directRight.scale(distRight);
        Vector3D leftVert = this.getLeftNormal();
        Vector3D rightVert = this.getRightNormal();
        leftVert.scale(this.thickness);
        rightVert.scale(this.thickness);
        List<Vector3D> isectVerts;
        if (forward) {
            isectVerts = this.startVerts;
        } else {
            isectVerts = this.endVerts;
            directLeft.scale(-1);
            directRight.scale(-1);
            final Vector3D temp = directLeft;
            directLeft = directRight;
            directRight = temp;
        }
        leftVert.add(directLeft);
        rightVert.add(directRight);
        if (forward) {
            result.add(rightVert);
            result.add(leftVert);
        } else {
            result.add(leftVert);
            result.add(rightVert);
        }
        isectVerts.clear();
        leftVert = leftVert.duplicate();
        rightVert = rightVert.duplicate();
        leftVert.add(isect.getPosition());
        rightVert.add(isect.getPosition());
        isectVerts.add(leftVert);
        isectVerts.add(rightVert);
        return result;
    }

    /**
     * Getter for endVerts.
     * 
     * @return the endVerts
     */
    public final List<Vector3D> getEndVerts() {
        return endVerts;
    }

    /**
     * The getIntVerts method return this Road's List of verts for a particular
     * Intersection.
     * 
     * @param intersection The Intersection to get verts for.
     * @return The verts for the Intersection, but it will default to the verts
     *         for the starting Intersection if the one specified is not valid.
     */
    public final List<Vector3D> getIntVerts(final Intersection intersection) {
        List<Vector3D> result = this.startVerts;
        if (intersection.equals(this.end)) {
            result = this.endVerts;
        }
        return result;
    }

    /**
     * Getter for lanes.
     * 
     * @return the lanes
     */
    public final int getLanes() {
        return lanes;
    }

    /**
     * Getter for leftBlock.
     * 
     * @return the leftBlock
     */
    public final Block getLeftBlock() {
        return leftBlock;
    }

    /**
     * Getter for leftNormal.
     * 
     * @return the leftNormal
     */
    public final Vector3D getLeftNormal() {
        final Vector3D leftNormal = this.rightNormal.duplicate();
        leftNormal.scale(-1);
        return leftNormal;
    }

    /**
     * The getLeftVerts method gets the verts for the left side of this Road.
     * 
     * @return A Collection of verts.
     */
    public final List<Vector3D> getLeftVerts() {
        final List<Vector3D> verts = new ArrayList<Vector3D>();
        verts.add(this.startVerts.get(0).duplicate());
        verts.add(this.endVerts.get(0).duplicate());
        return verts;
    }

    /**
     * Getter for name.
     * 
     * @return the name
     */
    public final String getName() {
        return name;
    }

    /**
     * The getNormalAround method return the left or right normal of this Road
     * depending on which is CCW or CW around the given Intersection..
     * 
     * @param center The Intersection to look around.
     * @param ccw True for looking CCW false for CW.
     * @return The appropriate normal orthoganol to this Road.
     */
    protected final Vector3D getNormalAround(final Intersection center,
            final boolean ccw) {
        boolean left = ccw;
        if (center.equals(this.end)) {
            left ^= true;
        }
        Vector3D result;
        if (left) {
            result = this.getLeftNormal();
        } else {
            result = this.getRightNormal();
        }
        return result;
    }

    /**
     * Getter for rightBlock.
     * 
     * @return the rightBlock
     */
    public final Block getRightBlock() {
        return rightBlock;
    }

    /**
     * Getter for rightNormal.
     * 
     * @return the rightNormal
     */
    public final Vector3D getRightNormal() {
        return rightNormal.duplicate();
    }

    /**
     * The getRightVerts method gets the verts for the right side of this Road.
     * 
     * @return A Collection of verts.
     */
    public final List<Vector3D> getRightVerts() {
        final List<Vector3D> verts = new ArrayList<Vector3D>();
        verts.add(this.endVerts.get(1).duplicate());
        verts.add(this.startVerts.get(1).duplicate());
        return verts;
    }

    /**
     * Getter for roadSigns.
     * 
     * @return the roadSigns
     */
    public final Collection<RoadSign> getRoadSigns() {
        return roadSigns;
    }

    /**
     * Getter for speedlimit.
     * 
     * @return the speedlimit
     */
    public final int getSpeedlimit() {
        return speedlimit;
    }

    /**
     * Getter for start.
     * 
     * @return the start
     */
    public final Intersection getStart() {
        return start;
    }

    /**
     * Getter for startVerts.
     * 
     * @return the startVerts
     */
    public final List<Vector3D> getStartVerts() {
        return startVerts;
    }

    /**
     * Getter for thickness.
     * 
     * @return the thickness
     */
    public final float getThickness() {
        return thickness;
    }

    /**
     * The handleCollision method takes care of what needs to be done when a
     * collision is detected.
     * 
     * @param obj The obj to handle collision against.
     * @param data The data for handling the collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    protected final void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Write handleCollision in Road.
    }

    /**
     * The calculateVerts method calculates the verts for this Road.
     * 
     * @see dds.model.simulation.AbstractSimObject#calculateVerts()
     */
    public final void myCalculateVerts() {
        this.start.calculateVerts();
        this.end.calculateVerts();
        final Vector3D vertA = this.startVerts.get(0).duplicate();
        final Vector3D vertB = this.startVerts.get(1).duplicate();
        final Vector3D vertC = this.endVerts.get(1).duplicate();
        final Vector3D vertD = this.endVerts.get(0).duplicate();
        vertA.subtract(this.getPosition());
        vertB.subtract(this.getPosition());
        vertC.subtract(this.getPosition());
        vertD.subtract(this.getPosition());
        this.getVerts().add(vertA);
        this.getVerts().add(vertB);
        this.getVerts().add(vertC);
        this.getVerts().add(vertD);
    }

    /**
     * Setter for direction.
     * 
     * @param dir the direction to set
     */
    public final void setDirection(final String dir) {
        this.direction = dir;
    }

    /**
     * Setter for end.
     * 
     * @param intB the end to set
     */
    @Deprecated
    public final void setEnd(final Intersection intB) {
        this.end = intB;
    }

    /**
     * Setter for endVerts.
     * 
     * @param verts the endVerts to set
     */
    @Deprecated
    public final void setEndVerts(final List<Vector3D> verts) {
        this.endVerts.clear();
        this.endVerts.addAll(verts);
    }

    /**
     * Setter for lanes.
     * 
     * @param laneCount the lanes to set
     */
    public final void setLanes(final int laneCount) {
        this.lanes = laneCount;
    }

    /**
     * Setter for leftBlock.
     * 
     * @param left the leftBlock to set
     */
    public final void setLeftBlock(final Block left) {
        this.leftBlock = left;
    }

    /**
     * Setter for name.
     * 
     * @param nameStr the name to set
     */
    public final void setName(final String nameStr) {
        this.name = nameStr;
    }

    /**
     * Setter for rightBlock.
     * 
     * @param right the rightBlock to set
     */
    public final void setRightBlock(final Block right) {
        this.rightBlock = right;
    }

    /**
     * Setter for rightNormal.
     * 
     * @param vec the rightNormal to set
     */
    @Deprecated
    public final void setRightNormal(final Vector3D vec) {
        this.rightNormal.setPos(vec);
    }

    /**
     * Setter for roadSigns.
     * 
     * @param signs the roadSigns to set
     */
    @Deprecated
    public final void setRoadSigns(final Collection<RoadSign> signs) {
        this.roadSigns.clear();
        this.roadSigns.addAll(signs);
    }

    /**
     * Setter for speedlimit.
     * 
     * @param speed the speedlimit to set
     */
    public final void setSpeedlimit(final int speed) {
        this.speedlimit = speed;
    }

    /**
     * Setter for start.
     * 
     * @param intA the start to set
     */
    @Deprecated
    public final void setStart(final Intersection intA) {
        this.start = intA;
    }

    /**
     * Setter for startVerts.
     * 
     * @param verts the startVerts to set
     */
    @Deprecated
    public final void setStartVerts(final List<Vector3D> verts) {
        this.startVerts.clear();
        this.startVerts.addAll(verts);
    }

    /**
     * Setter for thickness.
     * 
     * @param size the thickness to set
     */
    public final void setThickness(final float size) {
        this.thickness = size;
        // this.start.setVertsCalculated(false);
        // this.end.setVertsCalculated(false);
        this.start.calculateVerts();
        this.end.calculateVerts();
    }

    /**
     * The toString.
     * 
     * @return A String representation of this Object.
     * @see java.lang.Object#toString()
     */
    @Override
    public final String toString() {
        return "Road: " + this.getIdStr() + " " + "[" + this.start.getIdStr()
                + " " + this.end.getIdStr() + "] " + this.getPosition();
    }
}
