package dds.model.simulation;

import java.util.List;

/**
 * The RoadSign class is currently responsible for modelling road signs in the
 * simulation. It is currently capable of providing a variety of signs.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
// TODO Refactor functionality so that the sign is defined by the number of side
// and text to display OR a texture to use for rendering. This will allow this
// to be very versatile and we will no longer have to worry about maintaining
// variation-specific methods.
public class RoadSign extends AbstractSimObject {

    /**
     * The STOP_SIGN is used to index the stop sign texture.
     */
    public static final int STOP_SIGN = 0;

    /**
     * The YEILD_SIGN is used to index the yeild sign texture.
     */
    public static final int YEILD_SIGN = 1;

    /**
     * The WARNING_SIGN is used to index the warning sign.
     */
    public static final int WARNING_SIGN = 2;

    /**
     * The OTHER is used to index the 'other' sign.
     */
    public static final int OTHER = 3;

    /**
     * The Z_OFFSET constant defines the z-axis offset to use for road signs?.
     */
    // Check the comment of Z_OFFSET.
    private static final float Z_OFFSET = 2f;

    /**
     * The STOP_LONG constant defines the larger co-ordinate used to create the
     * stop sign's octagon.
     */
    private static final float STOP_LONG = 5.0f;

    /**
     * The STOP_SHORT constant defines the smaller co-ordinate used to create
     * the stop sign's octagon.
     */
    private static final float STOP_SHORT = 1.66f;

    /**
     * Constructor for RoadSign.
     * 
     * @param idString The UNIQUE ID for this object within the simulation.
     */
    public RoadSign(final String idString) {
        super(idString);
    }

    /**
     * The type field designates what type of RoadSign this object should be.
     */
    // TODO Consider using an enumeration for this provided there's enough
    // functionality to put into this.
    private int type;

    /**
     * The label field contains the text to use for this RoadSign.
     */
    private String label;

    /**
     * Getter for label.
     * 
     * @return the label
     */
    public final String getLabel() {
        return label;
    }

    /**
     * Setter for label.
     * 
     * @param text the label to set
     */
    public final void setLabel(final String text) {
        this.label = text;
    }

    /**
     * Getter for type.
     * 
     * @return the type
     */
    public final int getType() {
        return type;
    }

    /**
     * Setter for type.
     * 
     * @param signType the type to set
     */
    public final void setType(final int signType) {
        this.type = signType;
    }

    /**
     * The destruct method is responsible for unlinking all of this object's
     * references to make garbage collection easier.
     * 
     * @see dds.model.simulation.AbstractSimObject#destruct()
     */
    protected final void destruct() {
        this.clearVerts();
    }

    /**
     * The detectCollision.
     * 
     * @param obj SimObject to detect collision with.
     * @return True on collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #detectCollision(dds.model.simulation.AbstractSimObject)
     */
    // TODO Comment detectCollision in RoadSign.
    protected final boolean detectCollision(final AbstractSimObject obj) {
        return false;
    }

    /**
     * The handleCollision method takes care of what needs to be done when a
     * collision is detected.
     * 
     * @param obj The obj to handle collision against.
     * @param data The data for handling the collision.
     * @see dds.model.simulation.AbstractSimObject
     *      #handleCollision(dds.model.simulation.AbstractSimObject,
     *      java.lang.Object)
     */
    @Override
    protected final void handleCollision(final AbstractSimObject obj,
            final Object data) {
        // TODO Write handleCollision in RoadSign.
    }

    /**
     * The computePolyLoop method sets up the SmartPolygon used for collision
     * detection.
     * 
     * @see dds.model.simulation.AbstractSimObject#computePolyLoop()
     */
    protected void computePolyLoop() {
        // TODO Write computePolyLoop in RoadSign.
    }

    /**
     * The clearOtherVerts method is reposible for clearing any object-specific
     * verts or other calculated fields that a SimObject might have generated
     * during vert calculation. There are currently no other verts to clear for
     * a RoadSign.
     * 
     * @see dds.model.simulation.AbstractSimObject#clearOtherVerts()
     */
    @Override
    protected final void clearOtherVerts() {
        // There are currently no other verts to clear for a RoadSign.
    }

    /**
     * The myCalculateVerts houses the object-specfic calculations for this
     * object's verts. This method does NOT need to clear the older verts OR set
     * the vertsCalculated flag becuase that functionality has been standardized
     * in the calculateVerts method defined in this class. This method just
     * needs to calculate the verts. The verts that are calculate MUST be stored
     * as local coordinates. These coordinates will be offset by the current
     * position when being used for collision detection of rendering.
     * 
     * @see dds.model.simulation.AbstractSimObject#myCalculateVerts()
     */
    @Override
    protected final void myCalculateVerts() {
        switch (type) {
        case RoadSign.STOP_SIGN:
            verticizeStopSign();
            break;
        case RoadSign.YEILD_SIGN:
            verticizeYeildSign();
            break;
        case RoadSign.WARNING_SIGN:
            verticizeWarningSign();
            break;
        case RoadSign.OTHER:
            verticizeOther();
            break;
        default:
            // TODO Put a logger or some error or default handling here.
        }
    }

    /**
     * The verticizeStopSign calculates the vertices of a stop sign.
     */
    private void verticizeStopSign() {
        final List<Vector3D> verts = this.getVerts();

        // Add the eight points of the stop sign's octagon.

        // TODO Tweak these values to get a desirable result OR rework how the
        // verts should be defined.
        verts.add(new Vector3D(STOP_SHORT, -STOP_LONG, Z_OFFSET));
        verts.add(new Vector3D(STOP_LONG, -STOP_SHORT, Z_OFFSET));
        verts.add(new Vector3D(STOP_LONG, STOP_SHORT, Z_OFFSET));
        verts.add(new Vector3D(STOP_SHORT, STOP_LONG, Z_OFFSET));
        verts.add(new Vector3D(-STOP_SHORT, STOP_LONG, Z_OFFSET));
        verts.add(new Vector3D(-STOP_LONG, STOP_SHORT, Z_OFFSET));
        verts.add(new Vector3D(-STOP_LONG, -STOP_SHORT, Z_OFFSET));
        verts.add(new Vector3D(-STOP_SHORT, -STOP_LONG, Z_OFFSET));
    }

    /**
     * The reorient method transforms the vertices of this road sign by rotating
     * relative to the unit normal 'normalFace' passed in. Warning: this method
     * changes 'normalFace'.
     * 
     * @param normalFace Normal ector describing the plane of the sign's face?
     */
    // TODO Check comment for normalFace.
    public final void reorient(final Vector3D normalFace) {
        final Vector3D posZAxis = new Vector3D(0, -1, 0);
        final float rotTheta = (float) Math.acos(normalFace
                .dotProduct(posZAxis));
        for (Vector3D vert : this.getVerts()) {
            vert.rotateAroundPoint(rotTheta, this.getPosition());
        }
    }

    /**
     * The verticizeYeildSign calculates the verticesof a yeild sign.
     */
    private void verticizeYeildSign() {
        // Write verticizeYeildSign in RoadSign.
    }

    /**
     * The verticizeWarningSign calculates the vertices of a warning sign.
     */
    private void verticizeWarningSign() {
        // Write verticizeWarningSign in RoadSign.
    }

    /**
     * The verticizeOther calculates the vertices of an other sign.
     */
    private void verticizeOther() {
        // Write verticizeOther in RoadSign.
    }

    /**
     * The getBoundingVerts method returns a List of verts that inscribe the
     * bounding of this object.
     * 
     * @return The List of bounding verts.
     * @see dds.model.simulation.AbstractSimObject#getBoundingVerts()
     */
    @Override
    public final List<Vector3D> getBoundingVerts() {
        // TODO This needs to be written. I was going to use the object's
        // SmartPolygon, but apparently that only uses global coordinates and
        // doesn't yield information easily.
        return null;
    }
}
