package dds.model.simulation;

import java.awt.Polygon;
import java.awt.geom.Line2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import processing.core.PGraphics;

/**
 * The SmartPolygon class is not just any ordinary polygon; it has the
 * capabilities of doing intelligent collision detection with other objects of
 * this class. This class is used by the collision detection system by providing
 * a 2D polyloop for each 3D object that needs one.
 * 
 * @author cyberdanint
 * @author Jacob Bolton
 */
// TODO Consider using polymorphism to shorten the number of methods.
public class SmartPolygon {

    /**
     * The DEFAULT_POINT_COUNT constant defines how many points will probably be
     * in this polygon. Using this guess does NOT affect how many points can
     * actually be in a new polygon.
     */
    private static final int DEFAULT_POINTS = 4;

    /**
     * The MATRIX_SIZE constant defines the dimension size for the
     * transformation matrix.
     */
    private static final int MATRIX_SIZE = 3;

    /**
     * The brokenNormal is the calculated normal vector to be used by the
     * collision handler.
     */
    private final Line2D.Float brokenNormal;

    /**
     * The brokenSegments stores the line segments that the collision detection
     * detected intersections on in the last call to intersectsPoly().
     */
    private final List<Line2D.Float> brokenSegments;

    /**
     * The insidePoly is a flag indicating that the last call to
     * intersectsPoly() found this object inside the parameter object. This is
     * need for a special case for calculating the brokenNormal vector.
     */
    private boolean insidePoly = false;

    /**
     * The maxX is the largest x coordinate in this polygon.
     */
    private transient float maxX;

    /**
     * The maxY is the largest y coordinate in this polygon.
     */
    private transient float maxY;

    /**
     * The minX is the smallest x coordinate in this polygon.
     */
    private transient float minX;

    /**
     * The minY is the smallest y coordinate in this polygon.
     */
    private transient float minY;

    /**
     * The calculated field indicates whether the data for this polygon have
     * been calculated.
     */
    private boolean calculated;

    /**
     * The orientation indicates how much this polygon has been rotated since
     * it's creation. This angle should be in radians.
     */
    private transient float orientation;

    /**
     * The originX is x coordinate of the origin of this polygon.
     */
    private transient float originX;

    /**
     * The originY is the y coordinate of the origin of this polygon.
     */
    private transient float originY;

    /**
     * The points fields stores all of the points associated with this polygon.
     */
    private final List<Vector3D> points;

    /**
     * The segmentNormals stores the unit normal vectors for each segment.
     */
    private final Map<Line2D.Float, Line2D.Float> segmentNormals;

    /**
     * The segments stores the line segments connecting the verts of this
     * object.
     */
    private final List<Line2D.Float> segments;

    /**
     * The polygon field contains a polygon object that assists with certain
     * collision detection operations.
     */
    private final Polygon polygon;

    /**
     * The transformMatrix is an array that is used to store the current
     * transformation information for this polygon.
     */
    // TODO Get the transformation functions to either make use of this all the
    // time or create a Matrix class that will handle transformations.
    private final transient float[][] transformMatrix;

    /**
     * Constructor for SmartPolygon just initializes some book-keeping info.
     * 
     * @param numOfPoints the number of verts that this polygon should
     *            ultimately have
     */
    public SmartPolygon(final int numOfPoints) {
        // this.npoints = numOfPoints;
        // this.xcors = new float[npoints];
        // this.ycors = new float[npoints];
        this.points = new ArrayList<Vector3D>(numOfPoints);
        this.segments = new ArrayList<Line2D.Float>(numOfPoints);
        this.segmentNormals = new HashMap<Line2D.Float, Line2D.Float>(
                numOfPoints);
        this.brokenSegments = new ArrayList<Line2D.Float>(numOfPoints);
        this.brokenNormal = new Line2D.Float(0, 0, 0, 0);
        this.calculated = false;
        this.orientation = 0;
        this.polygon = new Polygon();
        this.transformMatrix = new float[MATRIX_SIZE][MATRIX_SIZE];
    }

    /**
     * Constructor for SmartPolygon.
     */
    public SmartPolygon() {
        // Use one with storage for the default number of points.
        this(DEFAULT_POINTS);
    }

    /**
     * The addPoint adds a new vertex to the polygon.
     * 
     * @param xCoord The x coordinate.
     * @param yCoord The y coordinate.
     * @return -1 if point already exists; 0 if you've exceeded the number of
     *         points you can add; 1 if point was added successfully;
     */
    public final int addPoint(final float xCoord, final float yCoord) {
        // TODO Change functionality to allow continuous adding of points.
        // TODO Change ret into a more descriptive boolean.
        int ret = 1;
        for (Vector3D point : this.points) { // check if point exists
            // for (int p = 0; p < nextAddPoint; p++) { // check if point exists
            if ((point.getXpos() == xCoord) && (point.getYpos() == yCoord)) {
                ret = -1;
                break;
            }
        }

        if (ret != -1) { // if point doesn't exist, then proceed
            // TODO Allow for 3D coordinates.
            final Vector3D newPoint = new Vector3D(xCoord, yCoord, 0);
            this.points.add(newPoint);
            this.polygon.addPoint((int) xCoord, (int) yCoord);
        }

        this.calculated = false;

        return ret;
    } // end of method addPoint

    /**
     * The generateSegments method iterates through all of the points to create
     * segments between each one. Currently this also closes the polygon.
     */
    private void generateSegments() {
        this.segments.clear();
        for (int i = 0; i < this.points.size(); i++) {
            final int next = (i + 1) % this.points.size();
            this.addSegment(i, next);
        }
    }

    /**
     * The calculate performs the all the calculations necessary to start using
     * this polygon.
     */
    private void calculate() {
        this.generateSegments();
        this.updateMinAndMax();
        this.computeNormals();
        this.calculated = true;
    }

    /**
     * The updateMinAndMax recomputes the min, max and origin coordinates for
     * all points.
     */
    private void updateMinAndMax() {
        for (int i = 0; i < this.points.size(); i++) {
            if (i == 0) {
                final Vector3D firstPoint = this.points.get(0);
                minX = firstPoint.getXpos();
                maxX = firstPoint.getXpos();
                minY = firstPoint.getYpos();
                maxY = firstPoint.getYpos();
            } else {
                updateMinAndMax(i);
            }
        }
    }

    /**
     * The addSegment adds a line to the polygon.
     * 
     * @param pi1 index of start point of line
     * @param pi2 index end point of line
     */
    public final void addSegment(final int pi1, final int pi2) {
        final Vector3D point1 = this.points.get(pi1);
        final Vector3D point2 = this.points.get(pi2);
        segments.add(new Line2D.Float(point1.getXpos(), point1.getYpos(),
                point2.getXpos(), point2.getYpos()));
    } // end of method addSegment

    /**
     * The breakNormal calculates the brokenNormal vector of this object based
     * on the last processed collision.
     */
    public final void breakNormal() {
        if (this.brokenSegments.size() > 0) {
            float seglen = 0f;
            for (Line2D.Float segment : this.brokenSegments) {
                final Line2D.Float segmentNormal = this.segmentNormals
                        .get(segment);
                brokenNormal.x2 += (segmentNormal.x2 - segmentNormal.x1);
                brokenNormal.y2 += (segmentNormal.y2 - segmentNormal.y1);
            }
            seglen = segmentLength(brokenNormal);
            if (seglen > 0) {
                brokenNormal.x2 /= seglen;
                brokenNormal.y2 /= seglen;
            } else {
                brokenNormal.x2 = -1;
                brokenNormal.y2 = 0;
            }
        }
    } // end of breakNormal

    /**
     * The breakSegment adds the segment at index si to the broken segment list.
     * 
     * @param index index of segment
     */
    public final void breakSegment(final int index) {
        // if (index < npoints && numBrokenSegments < npoints) {
        brokenSegments.add(this.segments.get(index));
        // ++numBrokenSegments;
        // }
    } // end of method breakSegment

    /**
     * The breakSegment adds the segment to the broken segment list.
     * 
     * @param segment The segment to add.
     */
    private void breakSegment(final Line2D.Float segment) {
        this.brokenSegments.add(segment);
    }

    /**
     * The computeNormals computes the normal vectors of the line segments.
     */
    public final void computeNormals() {
        this.segmentNormals.clear();
        for (Line2D.Float segment : this.segments) {
            // for (int s = 0; s < npoints; s++) {
            segmentNormals.put(segment, segNormal(segment));
        }
    } // end of method computeNormals

    /**
     * The contains calls Polygon's native contains method with the info. given.
     * 
     * @param xCoord x coordinate
     * @param yCoord y coordinate
     * @return whether x, y is in this polygon
     */
    public final boolean contains(final float xCoord, final float yCoord) {
        return this.polygon.contains((double) xCoord, (double) yCoord);
    } // end of method contains

    /**
     * The drawMe method should not be used unless you are debugging collision
     * detection. This draws out the polygon along with its edge normals.
     * 
     * @param applet the context to draw this on
     */
    // TODO Consider removing this method.
    public final void drawMe(final PGraphics applet) {
        applet.beginShape(PGraphics.POLYGON);
        for (int s = 0; s < this.points.size(); s++) {
            applet.vertex(this.points.get(s).getXpos(), this.points.get(s)
                    .getYpos());
        }
        applet.endShape();
    } // end of method drawMe

    /**
     * The findInternalNormal is used instead of segNormal if insidePoly is
     * true.
     * 
     * @param oLine a line to check intersections against
     */
    public final void findInternalNormal(final Line2D oLine) {
        if (!this.calculated) {
            this.calculate();
        }
        int target = 0;
        for (int s = 0; s < this.points.size(); s++) {
            if (oLine.intersectsLine(segments.get(s))) {
                target = s;
                break;
            }
        }
        breakSegment(target); // let's break this poly!
        breakNormal();
    } // end of method findInternalNormal

    /**
     * The finishLoop is called if you decide to terminate this polygon before
     * all of its vertices have been added. This does update the npoints var.
     */
    public final void finishLoop() {
        // npoints = points.size();
        // addSegment(points.size() - 1, 0);
        // updateMinAndMax(points.size() - 1); // final book-keeping
        // computeNormals();
        if (!this.calculated) {
            this.calculate();
        }
    } // end of method finishLoop

    /**
     * Accessor for brokenNormal.
     * 
     * @return the brokenNormal for this object
     */
    public final Line2D.Float getBrokenNormal() {
        return brokenNormal;
    }

    /**
     * Accessor for insidePoly Do not use this unless you know what you are
     * doing!
     * 
     * @return true if the last call to intersectsPoly() encountered a collision
     *         where the polygon was inside the other.
     */
    public final boolean getInsidePoly() {
        return this.insidePoly;
    }

    /**
     * Accessor for npoints.
     * 
     * @return npoints for this object
     */
    public final int getNpoints() {
        return this.points.size();
    }

    /**
     * Accessor for numBrokenSegments.
     * 
     * @return the numBrokeSegments for this object
     */
    public final int getNumBrokenSegments() {
        return this.brokenSegments.size();
    }

    /**
     * Accessor for originX.
     * 
     * @return the originX for this object
     */
    public final float getOriginX() {
        return this.originX;
    }

    /**
     * Accessor for originY.
     * 
     * @return the originY for this object
     */
    public final float getOriginY() {
        return this.originY;
    }

    /**
     * The getRotationMatrix computes a rotation matrix of dtheta.
     * 
     * @param dtheta the angle of rotation
     * @return the rotation matrix
     */
    private float[][] getRotationMatrix(final float dtheta) {
        this.resetMatrix();
        final float sin = (float) Math.sin(dtheta); // return new float[][] {
        // { (float) Math.cos(dtheta), (float) -Math.sin(dtheta), 0 },
        // { (float) Math.sin(dtheta), (float) Math.cos(dtheta), 0 },
        // { 0, 0, 1 } };

        final float cos = (float) Math.cos(dtheta);
        this.transformMatrix[0][0] = cos;
        this.transformMatrix[0][1] = -sin;
        this.transformMatrix[1][0] = sin;
        this.transformMatrix[1][1] = cos;
        // TODO Have this make changes to the current transformation matrix as a
        // void method.
        return this.transformMatrix;
    } // end of method getRotationMatrix

    /**
     * The resetMatrix method resets the current transformation matrix to the
     * identity matrix.
     */
    private void resetMatrix() {
        for (int i = 0; i < this.transformMatrix.length; i++) {
            for (int j = 0; j < this.transformMatrix[i].length; j++) {
                if (i == j) {
                    this.transformMatrix[i][j] = 1;
                } else {
                    this.transformMatrix[i][j] = 0;
                }
            }
        }
    }

    /**
     * Accessor for segmentNormals.
     * 
     * @return the segmentNormals for this object
     */
    public final List<Line2D.Float> getSegmentNormals() {
        return (List<Line2D.Float>) this.segmentNormals.values();
    }

    /**
     * Accessor for segments.
     * 
     * @return the segments for this object
     */
    public final List<Line2D.Float> getSegments() {
        return this.segments;
    }

    /**
     * The getTranslationMatrix computes a translation matrix.
     * 
     * @param displaceX the x displacement
     * @param displaceY the y displacement
     * @return the computed transformation matrix
     */
    private float[][] getTranslationMatrix(final float displaceX,
            final float displaceY) {
        // TODO Have this make changes to the current transformation matrix as a
        // void method.
        return new float[][] { { 1, 0, displaceX }, { 0, 1, displaceY },
                { 0, 0, 1 } };
    } // end of method getTranslationMatrix

    /**
     * The intersectsLine determines if this polygon intersects the line.
     * 
     * @param seg the line
     * @return true if the polygon intersects the line; false otherwise
     */
    public final boolean intersectsLine(final Line2D.Float seg) {
        if (!this.calculated) {
            this.calculate();
        }
        boolean result = false;
        for (Line2D.Float segment : this.segments) {
            if (segment.intersectsLine(seg)) {
                result = true;
                break;
            }
        }
        return result;
    } // end of method intersectsLine

    /**
     * The intersectsPoly determines if this SmartPolygon intersects with poly.
     * An intersection means that at least one pair of edges intersects, or the
     * origin of this object is contained in poly.
     * 
     * @param poly the object to check collision with
     * @return true if intersection; false otherwise
     */
    public final boolean intersectsPoly(final SmartPolygon poly) {
        // Ensure that this polygon is calculated.
        if (!this.calculated) {
            this.calculate();
        }
        // Ensure that the colliding polygon is calculated.
        if (!poly.isCalculated()) {
            poly.calculate();
        }
        // Assume there won't be a collision.
        boolean result = false;
        // Assume this isn't inside the other polygon.
        this.insidePoly = false;
        // Reset the broken segment mapping.
        poly.brokenSegments.clear();

        // Iterate through the other polygon's segments.
        for (Line2D.Float segment : poly.getSegments()) {
            if (this.intersectsLine(segment)) {
                poly.breakSegment(segment);
                result = true;
                poly.breakNormal();
                break;
            }
        }

        if (!result) {
            // No segments intersected, check to see if the origin is contained.
            result = this.isInsidePoly(poly);
            this.insidePoly = result;
        }

        return result;
    } // end of method intersectsPoly

    /**
     * The isBestNormal determines if this normal is indeed the best one.
     * 
     * @param normal the normal to test
     * @return true if it's the best; false otherwise
     */
    private boolean isBestNormal(final Line2D.Float normal) {
        return (!contains(normal.x2, normal.y2));
    } // end of method isBestNormal

    /**
     * Determines if this is inside poly.
     * 
     * @param poly the poly
     * @return true if inside; false otherwise
     */
    public final boolean isInsidePoly(final SmartPolygon poly) {
        return poly.contains(originX, originY);
    } // end of method isInsidePoly

    /**
     * The lineToVec converts a Line2D.Float to a float[] vector.
     * 
     * @param line the line to convert
     * @return the new vector
     */
    private float[] lineToVec(final Line2D.Float line) {
        return new float[] { line.x2 - line.x1, line.y2 - line.y1 };
    } // end of method lineToVec

    /**
     * The moveMe moves this object by the specified amount. It also calls the
     * necessary update functions. Use this instead of the matrix functions.
     * 
     * @param deltaX the x displacement
     * @param deltaY the y displacement
     */
    public final void moveMe(final float deltaX, final float deltaY) {
        final Vector3D delta = new Vector3D(deltaX, deltaY, 0);
        for (int p = 0; p < this.points.size(); p++) {
            this.points.get(p).add(delta);
            updateSegments(p);
        }
        minX += deltaX;
        maxX += deltaX;
        minY += deltaY;
        maxY += deltaY;
        originX += deltaX;
        originY += deltaY;
    } // end of method moveMe

    /**
     * The normalizeSegment method normalizes the vector described by seg.
     * 
     * @param seg a float[] of size 2 (x & y)
     */
    private void normalizeSegment(final float[] seg) {
        final float seglen = segmentLength(seg);
        seg[0] /= seglen;
        seg[1] /= seglen;
    } // end of method normalizeSegment

    /**
     * The properLineIntersects determines if two line segments meet at any
     * other point besides their endpoints, hence the name "proper" line
     * intersections.
     * 
     * @param line1 first line
     * @param line2 second line
     * @return true if they intersect properly; false otherwise
     */
    private boolean properLineIntersects(final Line2D.Float line1,
            final Line2D.Float line2) {
        float coordA = ((line2.x2 - line2.x1) * (line1.y1 - line2.y1))
                - ((line2.y2 - line2.y1) * (line1.x1 - line2.x1));
        float coordB = ((line1.x2 - line1.x1) * (line1.y1 - line2.y1))
                - ((line1.y2 - line1.y1) * (line1.x1 - line2.x1));

        // this will store the denominator needed for the ua, ub
        final float denominator = (line2.y2 - line2.y1) * (line1.x2 - line1.x1)
                - (line2.x2 - line2.x1) * (line1.y2 - line1.y1);

        boolean result = false;
        if (denominator != 0) { // if the denominator is 0, then we can't divide
            // it!
            coordA /= denominator; // otherwise update ua & ub
            coordB /= denominator;

            result = (coordA > 0) && (coordA < 1) && (coordB > 0)
                    && (coordB < 1);
        }
        return result;
    } // end of method properLineIntersects

    /**
     * The rotateMe rotates this object and updates its orientation.
     * 
     * @param dtheta the angle of rotation
     */
    public final void rotateMe(final float dtheta) {
        final float rotAngle = dtheta - orientation;
        orientation = dtheta;
        if (orientation < 0) {
            orientation += ((float) Math.PI * 2f);
        } else if (orientation > Math.PI * 2f) {
            orientation %= ((float) Math.PI * 2f);
        }
        // System.out.println("rotating by: " + rotAngle);
        rotateMeAroundOrigin(-rotAngle, new float[] { originX, originY });
        // System.out.println("orientation: " + orientation);
    } // end of method rotateMe

    /**
     * The rotateMeAroundOrigin rotates this object about the origin defined by
     * 'origin'.
     * 
     * @param dtheta the angle to rotate by
     * @param origin the rotation axis
     */
    public final void rotateMeAroundOrigin(final float dtheta,
            final float[] origin) {
        final float[][] rMatrix = getRotationMatrix(dtheta);
        final float[][] tMatrix = getTranslationMatrix(-origin[0], -origin[1]);
        final float[][] tMatrix2 = getTranslationMatrix(origin[0], origin[1]);

        for (int p = 0; p < this.points.size(); p++) {
            transformPoint(tMatrix, p);
            transformPoint(rMatrix, p);
            transformPoint(tMatrix2, p);
            updateMinAndMax(p);
            updateSegments(p);
        }
        for (int p = 0; p < this.points.size(); p++) {
            transformNormal(rMatrix, p);
        }
    } // end of method rotateMeAroundOrigin

    /**
     * The rotateVector rotates the vector defined by v.
     * 
     * @param vec the vector to rotate
     * @param theta the angle to rotate v by
     */
    private void rotateVector(final float[] vec, final float theta) {
        final float[][] matrix = getRotationMatrix(theta);
        final float oldX = vec[0];
        final float oldY = vec[1];
        // final float[] oldCor = new float[] { vec[0], vec[1] };

        vec[0] = (matrix[0][0] * oldX) + (matrix[0][1] * oldY) + matrix[0][2];
        vec[1] = (matrix[1][0] * oldX) + (matrix[1][1] * oldY) + matrix[1][2];
    } // end of method rotateVector

    /**
     * The segmentLength computes the length of the segment represented by a
     * Line2D.Float.
     * 
     * @param line vector as a Line2D.Float or null
     * @return a float indicating the length
     */
    private float segmentLength(final Line2D.Float line) {
        return this.segmentLength(line.x2 - line.x1, line.y2 - line.y1, 0);
    }

    /**
     * The segmentLength computes the length of the segment represented by a
     * float[].
     * 
     * @param vectorArray vector as a float[] or null
     * @return a float indicating the length
     */
    private float segmentLength(final float[] vectorArray) {
        return this.segmentLength(vectorArray[0], vectorArray[1], 0);
    }

    /**
     * The segmentLength computes the length of the segment represented as a set
     * of floats.
     * 
     * @param deltaX The change along the x-axis.
     * @param deltaY The change along the y-axis.
     * @param deltaZ The change along the z-axis.
     * @return a float indicating the length
     */
    private float segmentLength(final float deltaX, final float deltaY,
            final float deltaZ) {
        return (float) Math.sqrt(Math.pow((double) deltaX, 2.0)
                + Math.pow((double) deltaY, 2.0)
                + Math.pow((double) deltaZ, 2.0));
    }

    /**
     * The segNormal is the default method for calculating the collision's
     * normal vector.
     * 
     * @param seg the segment to calculate the normal for
     * @return the new calculated normal vector as a Line2D.Float
     */
    private Line2D.Float segNormal(final Line2D.Float seg) {
        final float[] norm1 = lineToVec(seg); // n1 and n2 are unit normals
        final float[] norm2 = norm1.clone(); // Clone is probably faster than
        // making a new array
        float[] segMidpoint = null;
        Line2D.Float nrml1 = null;
        Line2D.Float nrml2 = null;

        segMidpoint = new float[] { (seg.x1 + seg.x2) / 2f,
                (seg.y1 + seg.y2) / 2f };
        // normalize n1 & n2
        normalizeSegment(norm1);
        normalizeSegment(norm2);

        // now let's make them true normal vectors
        rotateVector(norm1, ((float) Math.PI / 2.0f));
        rotateVector(norm2, ((float) -Math.PI / 2.0f));

        // move them to the line they belong to and determine the best one
        nrml1 = new Line2D.Float(segMidpoint[0], segMidpoint[1], segMidpoint[0]
                + norm1[0], segMidpoint[1] + norm1[1]);
        nrml2 = new Line2D.Float(segMidpoint[0], segMidpoint[1], segMidpoint[0]
                + norm2[0], segMidpoint[1] + norm2[1]);

        final Line2D.Float result;

        if (isBestNormal(nrml1)) {
            nrml1.x1 = 0;
            nrml1.y1 = 0;
            nrml1.x2 = norm1[0];
            nrml1.y2 = norm1[1];
            result = nrml1;
        } else {
            nrml2.x1 = 0;
            nrml2.y1 = 0;
            nrml2.x2 = norm2[0];
            nrml2.y2 = norm2[1];
            result = nrml2;
        }
        return result;
    } // end of method segNormal

    /**
     * Modifier for npoints. This is currently not supported!
     * 
     * @param numOfPoints the new npoints
     */
    public final void setNpoints(final int numOfPoints) {
        throw new UnsupportedOperationException();
    }

    /**
     * The swapVerts swaps two vertices in the polygon and updates their
     * respective edges.
     * 
     * @param indexA first vert index
     * @param indexB second vert index
     */
    public final void swapVerts(final int indexA, final int indexB) {
        final Vector3D temp = this.points.get(indexA);
        final float tempX = temp.getXpos();
        final float tempY = temp.getYpos();
        final float tempZ = temp.getZpos();
        this.points.get(indexA).setPos(this.points.get(indexB));
        this.points.get(indexB).setPos(tempX, tempY, tempZ);
        updateSegments(indexA); // now we need to update segments & normals
        updateSegments(indexB);
        computeNormals();
    } // end of method swapVerts

    /**
     * The toString method for this class.
     * 
     * @return the string representation of this
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        String result = super.toString() + "{";
        for (Vector3D point : this.points) {
            // TODO Determine if the \n is necessary.
            result = result + "\n" + point.toString();
        }
        return result;
    } // end of method toString

    /**
     * The transformNormal translates the normal vector at index pi based on the
     * contents of matrix.
     * 
     * @param matrix the transformation matrix
     * @param index the index of the normal vector to transform
     */
    private void transformNormal(final float[][] matrix, final int index) {
        this.transformNormal(matrix, this.segments.get(index));
    } // end of method transformNormal

    /**
     * The transformNormal translates the normal vector at index pi based on the
     * contents of matrix.
     * 
     * @param matrix the transformation matrix
     * @param segment The segment which has a normal to transform.
     */
    private void transformNormal(final float[][] matrix,
            final Line2D.Float segment) {
        // final Line2D.Float oldNormal = new Line2D.Float();
        final Line2D.Float normal = segmentNormals.get(segment);
        final float oldStartX = normal.x1;
        final float oldStartY = normal.y1;
        final float oldEndX = normal.x2;
        final float oldEndY = normal.y2;

        // oldNormal.setLine(normal);
        normal.x1 = (matrix[0][0] * oldStartX) + (matrix[0][1] * oldStartY)
                + matrix[0][2];
        normal.y1 = (matrix[1][0] * oldStartX) + (matrix[1][1] * oldStartY)
                + matrix[1][2];
        normal.x2 = (matrix[0][0] * oldEndX) + (matrix[0][1] * oldEndY)
                + matrix[0][2];
        normal.y2 = (matrix[1][0] * oldEndX) + (matrix[1][1] * oldEndY)
                + matrix[1][2];
    } // end of method transformNormal

    /**
     * The transformPoint moves the point at index pi based on the numbers in
     * the transformation matrix.
     * 
     * @param matrix the transformation matrix
     * @param pointIndex the index of the point to transform
     */
    private void transformPoint(final float[][] matrix, final int pointIndex) {
        final Vector3D point = this.points.get(pointIndex);
        final float oldX = point.getXpos();
        final float oldY = point.getYpos();
        final float newX = (matrix[0][0] * oldX) + (matrix[0][1] * oldY)
                + matrix[0][2];
        final float newY = (matrix[1][0] * oldX) + (matrix[1][1] * oldY)
                + matrix[1][2];
        point.setPos(newX, newY, 0);
    } // end of method transformPoint

    /**
     * The unbreak method restores the brokenNormal data to its default. This
     * should always be called when brokenNormal won't be used anymore.
     */
    public final void unbreak() {
        this.brokenSegments.clear();
        brokenNormal.x1 = 0;
        brokenNormal.y1 = 0;
        brokenNormal.x2 = 0;
        brokenNormal.y2 = 0;
    } // end of method unbreak

    /**
     * The updateMinAndMax recomputes the min, max and origin coordinates for a
     * specified point.
     * 
     * @param index the index of the vert to check against
     */
    private void updateMinAndMax(final int index) {
        final float xCoord = this.points.get(index).getXpos();
        final float yCoord = this.points.get(index).getYpos();
        // TODO Consider using Math.min() and Math.max() instead.
        if (xCoord < minX) { // first update min & max
            minX = xCoord;
        }
        if (xCoord > maxX) {
            maxX = xCoord;
        }
        if (yCoord < minY) {
            minY = yCoord;
        }
        if (yCoord > maxY) {
            maxY = yCoord;
        }

        if (index == this.points.size() - 1) { // now update the origin
            originX = (minX + maxX) / 2.0f;
            originY = (minY + maxY) / 2.0f;
        }
    } // end of method updateMinAndMax

    /**
     * The updateSegments updates the necessary segments since the last add.
     * 
     * @param pointIndex the index of the most recently added vertex
     */
    public final void updateSegments(final int pointIndex) {
        if (!this.calculated) {
            this.calculate();
        }
        final Line2D.Float line1 = segments.get(pointIndex);
        Line2D.Float line2 = null;

        line1.x1 = this.points.get(pointIndex).getXpos();
        line1.y1 = this.points.get(pointIndex).getYpos();

        final int segIndex;
        if (pointIndex == 0) {
            segIndex = this.points.size() - 1;
        } else {
            segIndex = pointIndex - 1;
        }
        line2 = segments.get(segIndex);

        line2.x2 = this.points.get(pointIndex).getXpos();
        line2.y2 = this.points.get(pointIndex).getYpos();
    } // end of method updateSegments

    /**
     * The validate reorders the verts of the polygon in such a way that there
     * will be no crossing at edges. In other words, it "unfolds" this polygon.
     */
    public final void validate() {
        if (!this.calculated) {
            this.calculate();
        }
        boolean hasIntersections = false;
        do {
            hasIntersections = false;
            for (int s1 = 0; s1 < this.points.size(); s1++) {
                for (int s2 = 0; s2 < this.points.size(); s2++) {
                    if (s2 != s1) {
                        final Line2D.Float line1 = segments.get(s1);
                        final Line2D.Float line2 = segments.get(s2);
                        if (properLineIntersects(line1, line2)) {
                            swapVerts((s1 + 1) % this.points.size(), s2);
                            hasIntersections = true;
                        }
                    }
                }
            }
        } while (hasIntersections);
    } // end of method validate

    /**
     * Getter for calculated.
     * 
     * @return the calculated
     */
    public final boolean isCalculated() {
        return calculated;
    }

    /**
     * Setter for calculated.
     * 
     * @param isCalculated the calculated to set
     */
    public final void setCalculated(final boolean isCalculated) {
        // Check to see if this is trying to be set to true and it currently is
        // false.
        if (isCalculated && !this.calculated) {
            // It is, which means we need to invoke a calculation.
            this.calculate();
        } else {
            // Every other condition can just be resolved here since it doesn't
            // require a calculation to occur.
            this.calculated = isCalculated;
        }
    }

    /**
     * Getter for brokenSegments.
     * 
     * @return the brokenSegments
     */
    public final List<Line2D.Float> getBrokenSegments() {
        return brokenSegments;
    }

    /**
     * Setter for brokenSegments.
     * 
     * @param brokenSegs the brokenSegments to set
     */
    @Deprecated
    public final void setBrokenSegments(final List<Line2D.Float> brokenSegs) {
        throw new UnsupportedOperationException();
    }

    /**
     * Setter for brokenNormal.
     * 
     * @param normal the brokenNormal to set
     */
    @Deprecated
    public final void setBrokenNormal(final Line2D.Float normal) {
        throw new UnsupportedOperationException();
    }

    /**
     * Setter for insidePoly.
     * 
     * @param isInside the insidePoly to set
     */
    // TODO Determine if we should allow this.
    public final void setInsidePoly(final boolean isInside) {
        this.insidePoly = isInside;
    }

    /**
     * Getter for points.
     * 
     * @return the points
     */
    public final List<Vector3D> getPoints() {
        return points;
    }

    /**
     * Setter for points.
     * 
     * @param pointList the points to set
     */
    @Deprecated
    public final void setPoints(final List<Vector3D> pointList) {
        throw new UnsupportedOperationException("Use addPoint instead.");
    }

    /**
     * Setter for segmentNormals.
     * 
     * @param segNormals the segmentNormals to set
     */
    @Deprecated
    public final void setSegmentNormals(final List<Line2D.Float> segNormals) {
        throw new UnsupportedOperationException("Use addPoint instead.");
    }

    /**
     * Setter for segments.
     * 
     * @param segList the segments to set
     */
    @Deprecated
    public final void setSegments(final List<Line2D.Float> segList) {
        throw new UnsupportedOperationException("Use addPoint instead.");
    }

    /**
     * The setOrigin method changes the origin of this polygon by shifting all
     * of the points and segments attached to it.
     * 
     * @param objX The new x-coordinate.
     * @param objY The new y-coordinate.
     * @param objZ The new z-coordinate.
     */
    public final void setOrigin(final float objX, final float objY,
            final float objZ) {
        final float deltaX = objX - this.getOriginX();
        final float deltaY = objY - this.getOriginY();
        this.moveMe(deltaX, deltaY);
    }

    /**
     * Getter for polygon.
     * 
     * @return the polygon
     */
    public final Polygon getPolygon() {
        return polygon;
    }

    /**
     * Setter for polygon.
     * 
     * @param poly the polygon to set
     */
    @Deprecated
    public final void setPolygon(final Polygon poly) {
        throw new UnsupportedOperationException();
    }
} // end of class
