package dds.model.simulation;

import java.awt.geom.Line2D;

/**
 * The Vector3D class. This class is responsible for modeling 3D direction and
 * location data. It should be optimized to prevent Object creation and
 * destruction wherever possible. This will improve performance, and the only
 * exception to this rule is for accessors since we don't want to share the
 * actual array, but instead a nice copy of it. The clone() method should
 * provide the most efficient mechanism for this.
 * 
 * @author Jacob Bolton
 */
public class Vector3D {

    /**
     * The toString.
     * 
     * @return A string representing this object.
     * @see java.lang.Object#toString()
     */
    @Override
    public final String toString() {
        return super.toString() + " : [" + this.getXpos() + ", "
                + this.getYpos() + ", " + this.getZpos() + "]";
    }

    /**
     * The NULL_VECTOR is just null, but it is used to enhance readability and
     * comprehensibility of code.
     */
    public static final Vector3D NULL_VECTOR = null;

    /**
     * The MAX_ERROR_PERCENT is the maximum deviation from the expected true
     * value that is allowed.
     */
    public static final float MAX_ERROR_PERCENT = 1.1f;

    /**
     * The MIN_ERROR_PERCENT is the minimum deviation from the expected true
     * value that is allowed.
     */
    public static final float MIN_ERROR_PERCENT = 0.9f;

    /**
     * The HALF constant is the value needed to scale a Vector3D in half.
     */
    public static final float HALF = 0.5f;

    /**
     * The index for the X component of the Vector3D.
     */
    public static final int X_INDEX = 0;

    /**
     * The index for the Y component of the Vector3D.
     */
    public static final int Y_INDEX = 1;

    /**
     * The ARRAY_SIZE the size used for the float[] internal to a Vector3D.
     */
    public static final int ARRAY_SIZE = 3;

    /**
     * The index for the Z component of the Vector3D.
     */
    public static final int Z_INDEX = 2;

    /**
     * The getXAxis returns a normalized Vector3D of the x-axis.
     * 
     * @return A normalized Vector3D of the x-axis.
     */
    public static Vector3D getXAxis() {
        return new Vector3D(1, 0, 0);
    }

    /**
     * The getYAxis returns a normalized Vector3D of the y-axis.
     * 
     * @return A normalized Vector3D of the y-axis.
     */
    public static Vector3D getYAxis() {
        return new Vector3D(0, 1, 0);
    }

    /**
     * The getZAxis returns a normalized Vector3D of the z-axis.
     * 
     * @return A normalized Vector3D of the z-axis.
     */
    public static Vector3D getZAxis() {
        return new Vector3D(0, 0, 1);
    }

    /**
     * The rayIntersection method calculates where, if at all, two line segments
     * intersect. The line segments are specified by pairs of Vector3D's.
     * 
     * @param segAm Starting point for segment A
     * @param segAb Ending point for segment A
     * @param segBm Starting point for segment B
     * @param segBb Ending point for segment B
     * @return The Vector3D representing the point of intersection, null if the
     *         segments don't intersect.
     */
    public static Vector3D rayIntersection(final Vector3D segAm,
            final Vector3D segAb, final Vector3D segBm, final Vector3D segBb) {
        Vector3D result;
        final Vector3D aMCrossBM = Vector3D.cross(segAm, segBm);
        final float crossMag = aMCrossBM.magnitude();
        if (crossMag == 0) {
            result = NULL_VECTOR;
        } else {
            final Vector3D bBMinusAB = Vector3D.subtract(segBb, segAb);
            final float segAt = Vector3D.determinant(bBMinusAB, segBm,
                    aMCrossBM)
                    / (crossMag * crossMag);
            result = segAm.duplicate();
            result.scale(segAt);
            result.add(segAb);
        }
        return result;
    }

    /**
     * The determinant method calculates the determinant of three Vector3D put
     * together as a 3x3 matrix.
     * 
     * @param vecA The first vector.
     * @param vecB The second vector.
     * @param vecC The third vector.
     * @return The determinant of matrix [vecA vecB vecC].
     */
    public static float determinant(final Vector3D vecA, final Vector3D vecB,
            final Vector3D vecC) {
        final float valA = vecA.getXpos();
        final float valB = vecB.getXpos();
        final float valC = vecC.getXpos();
        final float valD = vecA.getYpos();
        final float valE = vecB.getYpos();
        final float valF = vecC.getYpos();
        final float valG = vecA.getZpos();
        final float valH = vecB.getZpos();
        final float valI = vecC.getZpos();
        return (valA * ((valE * valI) - (valH * valF)))
                - (valB * ((valD * valI) - (valG * valF)))
                + (valC * ((valD * valH) - (valG * valE)));
    }

    /**
     * The subtract static method produces a new Vector3D from the subtraction
     * vecA - vecB.
     * 
     * @param vecA A Vector3D for subtraction.
     * @param vecB A Vector3D for subtraction.
     * @return The resultant Vector3D object.
     */
    public static Vector3D subtract(final Vector3D vecA, final Vector3D vecB) {
        final Vector3D result = vecA.duplicate();
        result.subtract(vecB);
        return result;
    }

    /**
     * The position array. We use this instead of having three floats. It should
     * allow for faster object creation when getting the array because it can
     * use Object's clone() method. pos[0] = xpos;pos[0] = ypos; pos[0] = zpos;
     */
    private float[] pos; // use this instead of three floats

    /**
     * Constructor for Vector3D. Sets all coordinate data to 0.
     */
    public Vector3D() {
        super();
        this.pos = new float[Z_INDEX + 1];
        this.pos[X_INDEX] = 0;
        this.pos[Y_INDEX] = 0;
        this.pos[Z_INDEX] = 0;
    }

    /**
     * Constructor for Vector3D. Allows setting the coordinate data.
     * 
     * @param xcoord X coordinate.
     * @param ycoord Y coordinate.
     * @param zcoord Z coordinate.
     */
    public Vector3D(final float xcoord, final float ycoord, 
            final float zcoord) {
        this();
        this.pos[X_INDEX] = xcoord;
        this.pos[Y_INDEX] = ycoord;
        this.pos[Z_INDEX] = zcoord;
    }

    /**
     * Constructor for Vector3D. Sets coordinates to contents of the float[]
     * passed in.
     * 
     * @param posArray Coordinates to use.
     */
    public Vector3D(final float[] posArray) {
        super();
        if (posArray.length == Z_INDEX + 1) {
            this.pos = posArray.clone();
            // System.arraycopy(posArray, 0, this.pos, 0, this.pos.length);
        } else {
            throw new InstantiationError(
                    "Vector3D expects array length of 3, not "
                            + posArray.length);
        }
    }

    /**
     * Constructor for Vector3D.
     * 
     * @param vec Vector to use to make a duplicate Vector3D
     */
    public Vector3D(final Vector3D vec) {
        this(vec.pos);
    }

    /**
     * The dotProduct method performs the dot product operation of this Vector3D
     * with another.
     * 
     * @param vec The other Vector3D.
     * @return The resulting dot product.
     */
    public final float dotProduct(final Vector3D vec) {
        float prod = this.pos[X_INDEX] * vec.pos[X_INDEX];
        prod += this.pos[Y_INDEX] * vec.pos[Y_INDEX];
        prod += this.pos[Z_INDEX] * vec.pos[Z_INDEX];
        return prod;
    }

    /**
     * The clone method clones a Vector3D.
     * 
     * @return An identical copy of this object.
     */
    public final Vector3D duplicate() {
        return new Vector3D(this.pos);
    }

    /**
     * The getNormal return a new Vector3D that contains this Vector3D
     * normalized.
     * 
     * @return A copy of this Vector3D normalized.
     */
    public final Vector3D getNormal() {
        final Vector3D nvec = this.duplicate();
        nvec.normalize();
        return nvec;
    }

    /**
     * The getPos.
     * 
     * @return the position array.
     */
    public final float[] getPos() {
        return this.pos.clone();
    }

    /**
     * The getValueArray. This method has been deprecated since the switch from
     * storing coordinate data in separate floats to storing it in a float[].
     * 
     * @return Array containing the coordinates. {x, y, z}
     */
    @Deprecated
    public final float[] getValueArray() {
        return pos.clone();
    }

    /**
     * Getter for x.
     * 
     * @return the x
     */
    public final float getXpos() {
        return pos[X_INDEX];
    }

    /**
     * Getter for y.
     * 
     * @return the y
     */
    public final float getYpos() {
        return pos[Y_INDEX];
    }

    /**
     * Getter for z.
     * 
     * @return the z
     */
    public final float getZpos() {
        return pos[Z_INDEX];
    }

    /**
     * The isOnSegment method checks this point to see if it's on the line
     * segment inscribed by the two points provided.
     * 
     * @param ptA The first point of the line segment.
     * @param ptB The second point of the line segment.
     * @return True if the point is on the segment, false if not.
     */
    public final boolean isOnSegment(final Vector3D ptA, final Vector3D ptB) {
        float xtime, ytime, ztime, segLength;
        final Vector3D shifted = Vector3D.subtract(this, ptA);
        final Vector3D segment = Vector3D.subtract(ptB, ptA);
        segLength = segment.magnitude();
        segment.normalize();
        xtime = shifted.pos[X_INDEX] / segment.pos[X_INDEX];
        ytime = shifted.pos[Y_INDEX] / segment.pos[Y_INDEX];
        ztime = shifted.pos[Z_INDEX] / segment.pos[Z_INDEX];
        return ((xtime == ytime) && (ytime == ztime) && (xtime == segLength));
    }

    /**
     * The magnitude method returns the magnitude of this Vector3D.
     * 
     * @return Magnitude as a float.
     */
    public final float magnitude() {
        return (float) java.lang.Math.sqrt(this.dotProduct(this));
    }

    /**
     * The normalize method normalizes a Vector3D's components so that its
     * magnitude is 1.0.
     */
    public final void normalize() {
        final float mag = this.magnitude();
        if (mag != 0) {
            this.scale(1.0f / this.magnitude());
        }
    }

    /**
     * The scale method scales all coordinates of the vector by the same
     * multiplicative factor..
     * 
     * @param scaler Value to multiply the Vector3D by.
     */
    public final void scale(final float scaler) {
        this.pos[X_INDEX] *= scaler;
        this.pos[Y_INDEX] *= scaler;
        this.pos[Z_INDEX] *= scaler;
    }

    /**
     * The setter for pos.
     * 
     * @param posArr The new position data to use.
     */
    public final void setPos(final float[] posArr) {
        System.arraycopy(posArr, 0, this.pos, 0, posArr.length);
    }

    /**
     * The setter for pos.
     * 
     * @param posVec The new position data to use.
     */
    public final void setPos(final Vector3D posVec) {
        this.setPos(posVec.getPos());
    }

    /**
     * The setPosNoDup updates the position associated with this vector by
     * changing out the x, y, and z coordinates.
     * 
     * @param pvec new position vector
     */
    public final void setPosNoDup(final Vector3D pvec) {
        this.setXpos(pvec.getXpos());
        this.setYpos(pvec.getYpos());
        this.setZpos(pvec.getZpos());
    }

    /**
     * Setter for x.
     * 
     * @param coord the x to set
     */
    public final void setXpos(final float coord) {
        this.pos[X_INDEX] = coord;
    }

    /**
     * Setter for y.
     * 
     * @param coord the y to set
     */
    public final void setYpos(final float coord) {
        this.pos[Y_INDEX] = coord;
    }

    /**
     * Setter for z.
     * 
     * @param coord the z to set
     */
    public final void setZpos(final float coord) {
        this.pos[Z_INDEX] = coord;
    }

    /**
     * The subtract instance method does subtraction on this Vector3D. It
     * subtracts the coordinates from the specified Vector3D from those of this
     * one.
     * 
     * @param vec The Vector3D to suntract from this one.
     */
    public final void subtract(final Vector3D vec) {
        this.pos[X_INDEX] -= vec.pos[X_INDEX];
        this.pos[Y_INDEX] -= vec.pos[Y_INDEX];
        this.pos[Z_INDEX] -= vec.pos[Z_INDEX];
    }

    /**
     * The midpoint static method calculates the midpoint between two Vector3D
     * objects and returns it as its own Vector3D.
     * 
     * @param position The first 3D position.
     * @param position2 The second 3D position.
     * @return The midpoint between to given positions.
     */
    public static Vector3D midpoint(final Vector3D position,
            final Vector3D position2) {
        final Vector3D midpoint = Vector3D.add(position, position2);
        midpoint.scale(HALF);
        return midpoint;
    }

    /**
     * The add static method adds the two provided Vector3D's and stores the
     * result in a new Vector3D.
     * 
     * @param vecA The first Vector3D to add.
     * @param vecB The second Vector3D to add.
     * @return The sum of the two Vector3D's.
     */
    public static Vector3D add(final Vector3D vecA, final Vector3D vecB) {
        final Vector3D result = vecA.duplicate();
        result.add(vecB);
        return result;
    }

    /**
     * The cross static method returns a new Vector3D object that represents the
     * result of the cross product between two vectors.
     * 
     * @param vecA The first Vector3D of the cross product.
     * @param vecB The second Vector3D of the cross product.
     * @return A new Vector3D that is the result of the cross product.
     */
    public static Vector3D cross(final Vector3D vecA, final Vector3D vecB) {
        final Vector3D cross = new Vector3D();
        cross.pos[X_INDEX] = vecA.pos[Y_INDEX] * vecB.pos[Z_INDEX]
                - vecA.pos[Z_INDEX] * vecB.pos[Y_INDEX];
        cross.pos[Y_INDEX] = vecA.pos[Z_INDEX] * vecB.pos[X_INDEX]
                - vecA.pos[X_INDEX] * vecB.pos[Z_INDEX];
        cross.pos[Z_INDEX] = vecA.pos[X_INDEX] * vecB.pos[Y_INDEX]
                - vecA.pos[Y_INDEX] * vecB.pos[X_INDEX];
        return cross;
    }

    /**
     * The add method adds the coordinates from another Vector3D to this one.
     * 
     * @param vec The Vector3D to bbe added.
     */
    public final void add(final Vector3D vec) {
        this.pos[X_INDEX] += vec.pos[X_INDEX];
        this.pos[Y_INDEX] += vec.pos[Y_INDEX];
        this.pos[Z_INDEX] += vec.pos[Z_INDEX];
    }

    /**
     * The acceptablyEqual method determines if two values are acceptably equal
     * according to their ratio and the specified MIN and MAX error.
     * 
     * @param valueA A value.
     * @param valueB A value.
     * @return True if they're close enough to being equal, otherwise false.
     */
    private static boolean acceptablyEqual(final float valueA,
            final float valueB) {
        final float ratio = Math.abs(valueA - valueB);
        return ratio < MIN_ERROR_PERCENT;
    }

    /**
     * The equals.
     * 
     * @param obj The Object to compare against
     * @return True if they're equal.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public final boolean equals(final Object obj) {
        boolean result = false;
        if (obj instanceof Vector3D) {
            final Vector3D vec = (Vector3D) obj;
            result = acceptablyEqual(this.getXpos(), vec.getXpos())
                    && acceptablyEqual(this.getYpos(), vec.getYpos())
                    && acceptablyEqual(this.getZpos(), vec.getZpos());
        }
        return result;
    }

    /**
     * The hashCode.
     * 
     * @return The Hashcode of this object.
     * @see java.lang.Object#hashCode()
     */
    @Override
    public final int hashCode() {
        return super.hashCode();
    }

    /**
     * The setPos sets this Vector3D's values from floats.
     * 
     * @param xCoord The x coordinate.
     * @param yCoord The y coordinate.
     * @param zCoord The z coordinate.
     */
    public final void setPos(final float xCoord, final float yCoord,
            final float zCoord) {
        this.pos[X_INDEX] = xCoord;
        this.pos[Y_INDEX] = yCoord;
        this.pos[Z_INDEX] = zCoord;
    }

    /**
     * The rotateZ rotates this vector so many radians around the z-axis.
     * 
     * @param radians The number of radians to use for rotation.
     */
    public final void rotateZ(final float radians) {
        final float xPos = this.pos[X_INDEX];
        final float yPos = this.pos[Y_INDEX];
        this.pos[X_INDEX] = (xPos * (float) Math.cos(radians))
                - (yPos * (float) Math.sin(radians));
        this.pos[Y_INDEX] = (xPos * (float) Math.sin(radians))
                + (yPos * (float) Math.cos(radians));
    }

    /**
     * The rotateAroundPoint method rotates this vector around the given point.
     * 
     * @param theta the angle of rotation
     * @param point the rotation axis
     */
    public final void rotateAroundPoint(final float theta, 
            final Vector3D point) {
        point.scale(-1f);
        this.add(point);
        this.rotateZ(theta);
        point.scale(-1f);
        this.add(point);
    }

    /**
     * The traceLine method computes a Line2D that extends from the origin
     * specified by (ox, oy) along this vector. This method does not modify this
     * vector's attributes, so it's safe to use externally; although this vector
     * should be normalized first!
     * 
     * @param line the line to store the result in
     * @param magnitude the length of the line segment to trace
     * @param originX the origin's x coordinate
     * @param originY the origin's y coordinate
     */
    public final void traceLine(final Line2D.Float line, final float magnitude,
            final float originX, final float originY) {
        line.x1 = originX;
        line.y1 = originY;
        line.x2 = originX + (magnitude * this.getXpos());
        line.y2 = originY + (magnitude * this.getYpos());
    }

    /**
     * The clear method zeroes out the position for this Vector3D.
     */
    public final void clear() {
        this.pos[X_INDEX] = 0;
        this.pos[Y_INDEX] = 0;
        this.pos[Z_INDEX] = 0;
    }
}
