package dds.model.trapeditor;

import java.util.ArrayList;

import dds.model.simulation.BlockType;
import dds.model.simulation.SmartPolygon;

import processing.core.PGraphics;

/**
 * The BNode2D class represents a block node in 2D.
 * 
 * @author cyberdan
 */
public class BNode2D extends TNode2D {

    /**
     * The blockType.
     */
    private BlockType blockType = null;

    /**
     * The type of block represented by this node.
     */
    private String type = null;

    /**
     * The roads is the list of roads that define this block. The roads are
     * identified by their 'name' field in their respective nodes.
     */
    private ArrayList<String> roads = null;

    /**
     * The properBoundary is what we use to draw this block onto the map.
     */
    private SmartPolygon properBoundary = null;

    /**
     * The default constructor for BNode2D.
     */
    public BNode2D() {
        // TODO Make sure there's a super() defined in TNode2D
        super();
        this.roads = new ArrayList<String>();
    }

    /**
     * Constructor for BNode2D. Just initializes the coordinate arrays.
     * 
     * @param npoints The number ofpoints?
     */
    public BNode2D(final int npoints) {
        this();
        this.setNpoints(npoints);
        this.setXcors(new float[npoints]);
        this.setYcors(new float[npoints]);
    }

    /**
     * The toString.
     * 
     * @return A String representation of this object.
     * @see java.lang.Object#toString()
     */
    public final String toString() {
        return ("BNode2D: " + this.getName() + ": " + super.toString());
    }

    /**
     * The getType is the accessor for type.
     * 
     * @return type
     */
    public final String getType() {
        return type;
    }

    /**
     * The addRoad method adds a road to the list of roads that define this
     * block.
     * 
     * @param rid a new road name
     */
    public final void addRoad(final String rid) {
        this.roads.add(rid);
    }

    /**
     * The getRoads is the accessor for roads.
     * 
     * @return roads
     */
    public final ArrayList<String> getRoads() {
        return roads;
    }

    /**
     * The setRoads is the modifier for roads.
     * 
     * @param roadList the new roads
     */
    public final void setRoads(final ArrayList<String> roadList) {
        this.roads = roadList;
    }

    /**
     * The setType is the modifier for type.
     * 
     * @param typeStr the new type
     */
    public final void setType(final String typeStr) {
        this.type = typeStr;
        blockType = BlockType.parseType(typeStr);
    }

    /**
     * The polygonize prepares this block node for rendering.
     */
    public final void polygonize() {
        properBoundary = new SmartPolygon(getNpoints());

        for (int v = 0; v < getNpoints(); v++) {
            properBoundary.addPoint(getXcors()[v], getYcors()[v]);
        }

        properBoundary.finishLoop();
        properBoundary.validate();
    }

    /**
     * The inBounds method determines if the point (x, y) is contained in the
     * region defined by this TNode2D.
     * 
     * @param xCoord the x-coordinate
     * @param yCoord the y-coordinate
     * @return true if in bounds; false otherwise
     */
    public final boolean inBounds(final int xCoord, final int yCoord) {
        return properBoundary.contains((float) xCoord, (float) yCoord);
    }

    /**
     * The drawNode method draws the child object's node. A node can represent
     * an intersection, road, or block.
     * 
     * @param context the graphics context
     */
    @Override
    public final void drawNode(final PGraphics context) {
        context.fill(blockType.getColor());
        context.beginShape(PGraphics.POLYGON);
        for (int v = 0; v < getNpoints(); v++) {
            properBoundary.drawMe(context);
        }
        context.endShape();
    }
}
