package dds.model.trapeditor;

import java.util.ArrayList;

import dds.control.DDSApplet;
import dds.model.simulation.Vector3D;

/**
 * The BezierCurve class is responsible for modelling a bezier curve for the
 * trap editor. It will be used to create traps that utilize the BezierPosition
 * AI command.
 * 
 * @author cyberdan
 */
// TODO Consider refactoring parts of this class into a super class that defines
// behaviors for all AI commands that can be modeled through the trap editor.
// TODO Add detailed comments to the code in this BezierCurve.
// TODO Verify javadocs in BezierCurve.
public class BezierCurve {
    /**
     * The MAX_POINTS constant defines how many control points can exist in a
     * BezierCurve.
     */
    public static final int MAX_POINTS = 4;

    /**
     * The controlPoints field stores the control points for this bezier curve.
     */
    private final ArrayList<Vector3D> controlPoints;

    /**
     * The existingPoints.
     */
    // TODO Javadoc existingPoints.
    private int existingPoints;

    /**
     * Constructor for BezierCurve.
     */
    public BezierCurve() {
        this.controlPoints = new ArrayList<Vector3D>();
        existingPoints = 0;
    }

    /**
     * The addPoint method adds a new control point to the control points. If
     * there's no more room for control points, it repositions the closest
     * control point to the new coordinate.
     * 
     * @param xCoord The x coordinate.
     * @param yCoord The x coordinate.
     */
    public final void addPoint(final float xCoord, final float yCoord) {
        if (existingPoints < MAX_POINTS) {
            this.controlPoints.add(new Vector3D(xCoord, yCoord, 0f));
            ++existingPoints;
        } else {
            // pick the point closest to (x, y)
            final int closestPt = getIndexOfClosestPoint(xCoord, yCoord);
            movePointToAbsolutePosition(closestPt, xCoord, yCoord);
        }
    }

    /**
     * The unmapCoordinates.
     * 
     * @param mapper The TrapEditorXYMapper.
     * @return A list of unmapped points?
     */
    // Verify javadoc of unmapCoordinates.
    public final ArrayList<Vector3D> unmapCoordinates(
            final TrapEditorXYMapper mapper) {
        final ArrayList<Vector3D> unmapped = new ArrayList<Vector3D>();
        for (Vector3D vec : controlPoints) {
            final float[] umapf = new float[] { vec.getXpos(), vec.getYpos() };
            mapper.getUnmappedCoordinate(umapf);
            unmapped.add(new Vector3D(umapf[0], umapf[1], vec.getZpos()));
        }

        return unmapped;
    }

    /**
     * The writeToBuffer method generates a String containing the trap script
     * that this curve represents.
     * 
     * @param mapper The XYMapper.
     * @param startTime The starting time for this curve.
     * @param duration The duration of this curve.
     * @return A String containing trap script commands.
     */
    public final String writeToBuffer(final TrapEditorXYMapper mapper,
            final int startTime, final int duration) {
        finalizeCurve();
        final ArrayList<Vector3D> unmapped = unmapCoordinates(mapper);
        String comment = "//";

        for (int i = 0; i < MAX_POINTS; i++) {
            final Vector3D point = unmapped.get(i);
            comment = comment
                    + ("(" + point.getXpos() + ", " + point.getYpos() + ", "
                            + point.getZpos() + "):");
        }
        comment = comment + (startTime + ":");
        comment = comment + (duration + "");

        final String p0Str = "new Vector3D(" + unmapped.get(0).getXpos()
                + "f, " + unmapped.get(0).getYpos() + "f, "
                + unmapped.get(0).getZpos() + "f)";
        final String p1Str = "new Vector3D(" + unmapped.get(1).getXpos()
                + "f, " + unmapped.get(1).getYpos() + "f, "
                + unmapped.get(1).getZpos() + "f)";
        final String p2Str = "new Vector3D(" + unmapped.get(2).getXpos()
                + "f, " + unmapped.get(2).getYpos() + "f, "
                + unmapped.get(2).getZpos() + "f)";
        final String p3Str = "new Vector3D(" + unmapped.get(3).getXpos()
                + "f, " + unmapped.get(3).getYpos() + "f, "
                + unmapped.get(3).getZpos() + "f)";

        final String createSimObjStr = "CreateSimObject("
                + p0Str.substring(13, p0Str.length() - 1)
                + ", 0.0f, 1.0f, 0.0f);";
        final String bezierPositionStr = "BezierPosition(" + p0Str + ",\n"
                + p1Str + ",\n" + p2Str + ",\n" + p3Str + ", " + startTime
                + ", " + duration + ", 33);\n}\n";
        final String isTriggeredStr = "boolean isTriggered() {\n"
                + "return TriggerOnce(WithinDistance(50));\n}";

        return (comment + "\nvoid setup() {\n" + "trapLocation = " + p0Str
                + ";\n" + createSimObjStr + "\n" + bezierPositionStr + "\n" + isTriggeredStr);
    }

    /**
     * The finalizeCurve.
     */
    // TODO Javadoc finalizeCurve.
    private void finalizeCurve() {
        if (existingPoints < MAX_POINTS) {
            int pointIndex;
            if (existingPoints == 0) {
                for (pointIndex = 0; pointIndex < MAX_POINTS; pointIndex++) {
                    controlPoints.add(new Vector3D(0f, 0f, 0f));
                    ++existingPoints;
                }
            }
            for (pointIndex = existingPoints; pointIndex < MAX_POINTS; pointIndex++) {
                addPoint(controlPoints.get(existingPoints - 1).getXpos(),
                        controlPoints.get(existingPoints - 1).getYpos());
            }
        }
    }

    /**
     * The getIndexOfClosestPoint method determines the control point closest to
     * the coordinates specified.
     * 
     * @param xCoord The x-coordinate.
     * @param yCoord The x-coordinate.
     * @return The index of the closest control point.
     */
    public final int getIndexOfClosestPoint(final float xCoord,
            final float yCoord) {
        double distPtoP = 0f;
        int minIndex = 0;
        Vector3D curPoint = null;

        curPoint = getPoint(minIndex);
        distPtoP = Math.sqrt(Math.pow(xCoord - curPoint.getXpos(), 2)
                + Math.pow(yCoord - curPoint.getYpos(), 2));

        for (int i = 1; i < existingPoints; i++) {
            final double tdist = distPtoP;
            curPoint = getPoint(i);
            distPtoP = Math.sqrt(Math.pow(xCoord - curPoint.getXpos(), 2)
                    + Math.pow(yCoord - curPoint.getYpos(), 2));

            if (distPtoP >= tdist) {
                distPtoP = tdist;
            } else {
                minIndex = i;
            }
        }

        return minIndex;
    }

    /**
     * The getPoint method retrieves the control point at the specified index
     * from the List of control points.
     * 
     * @param index The index of the control point to retrieve.
     * @return The control point.
     */
    public final Vector3D getPoint(final int index) {
        return controlPoints.get(index);
    }

    /**
     * The removePointAtIndex method removes the control point at the specified
     * index from the List of control points.
     * 
     * @param index The index of the control point to retrieve.
     */
    public final void removePointAtIndex(final int index) {
        if (controlPoints.remove(index) != null) {
            --existingPoints;
        }
    }

    /**
     * The removePoint method removes the specified control point from the List
     * of control points.
     * 
     * @param point The poin t to remove from the List.
     */
    public final void removePoint(final Vector3D point) {
        if (controlPoints.remove(point)) {
            --existingPoints;
        }
    }

    /**
     * The movePointAtIndex method changes the coordinates of the control point
     * at the specified index within the List of control points.
     * 
     * @param index The index of the control point to move.
     * @param deltaX The change along the x-axis.
     * @param deltaY The change along the y-axis.
     */
    public final void movePointAtIndex(final int index, final float deltaX,
            final float deltaY) {
        final float[] oldPt = controlPoints.get(index).duplicate().getPos();
        controlPoints.get(index).setPos(oldPt[Vector3D.X_INDEX] + deltaX,
                oldPt[Vector3D.Y_INDEX] + deltaY, oldPt[Vector3D.Z_INDEX]);
    }

    /**
     * The movePoint method changes the coordinates of the specified control
     * point in the List of control points.
     * 
     * @param point The control point to move.
     * @param deltaX The change along the x-axis.
     * @param deltaY The change along the y-axis.
     */
    public final void movePoint(final Vector3D point, final float deltaX,
            final float deltaY) {
        this.movePointAtIndex(controlPoints.indexOf(point), deltaX, deltaY);
    }

    /**
     * The movePointToAbsolutePosition method changes the coordinates of the
     * control point at the specified index within the List of control points.
     * 
     * @param index The index of the control point to move.
     * @param xPosition The absolute x-coordinate.
     * @param yPosition The absolute y-coordinate.
     */
    public final void movePointToAbsolutePosition(final int index,
            final float xPosition, final float yPosition) {
        controlPoints.get(index).setPos(xPosition, yPosition, 0);
    }

    /**
     * The draw method draws this BezierCurve onto the specified DDSApplet at
     * the specified position.
     * 
     * @param applet The DDSApplet to draw onto.
     * @param deltaX The x-coordinate chage to draw the curve onto the
     *            DDSApplet.
     * @param deltaY The y-coordinate chage to draw the curve onto the
     *            DDSApplet.
     */
    // TODO Refactor this into another class that is in the view package.
    public final void draw(final DDSApplet applet, final float deltaX,
            final float deltaY) {
        applet.fill(0, 230, 55);
        if (existingPoints == 0) {
            return;
        }
        applet.ellipse(getPoint(0).getXpos() + deltaX, getPoint(0).getYpos()
                + deltaY, 15, 15);
        applet.fill(0, 33, 230);
        for (int p = 1; p < existingPoints - 1; p++) {
            final Vector3D nextPt = getPoint(p);
            applet.ellipse(nextPt.getXpos() + deltaX,
                    nextPt.getYpos() + deltaY, 10, 10);
        }
        applet.fill(230, 55, 0);
        applet.ellipse(getPoint(existingPoints - 1).getXpos() + deltaX,
                getPoint(existingPoints - 1).getYpos() + deltaY, 15, 15);

        if (existingPoints == MAX_POINTS) {
            final Vector3D point0 = getPoint(0);
            final Vector3D point1 = getPoint(1);
            final Vector3D point2 = getPoint(2);
            final Vector3D point3 = getPoint(3);
            Vector3D alpha = null;
            Vector3D beta = null;
            Vector3D delta = null;
            final Vector3D bezierPt = new Vector3D(0f, 0f, 0f);
            final Vector3D lastPt = point0.duplicate();

            delta = Vector3D.subtract(point1, point0);
            delta.scale(3);

            beta = Vector3D.subtract(point2, point1);
            beta.scale(3);
            beta.subtract(delta);

            alpha = Vector3D.subtract(point3, point0);
            alpha.subtract(delta);
            alpha.subtract(beta);

            lastPt.setXpos(lastPt.getXpos() + deltaX);
            lastPt.setYpos(lastPt.getYpos() + deltaY);
            for (float t = 0f; t <= 1f; t += 0.01f) {
                bezierPt.setXpos((alpha.getXpos() * t * t * t)
                        + (beta.getXpos() * t * t) + (delta.getXpos() * t)
                        + point0.getXpos() + deltaX);
                bezierPt.setYpos((alpha.getYpos() * t * t * t)
                        + (beta.getYpos() * t * t) + (delta.getYpos() * t)
                        + point0.getYpos() + deltaY);
                applet.line(lastPt.getXpos(), lastPt.getYpos(), bezierPt
                        .getXpos(), bezierPt.getYpos());
                lastPt.setXpos(bezierPt.getXpos());
                lastPt.setYpos(bezierPt.getYpos());
            }

            applet.line(point0.getXpos() + deltaX, point0.getYpos() + deltaY,
                    point1.getXpos() + deltaX, point1.getYpos() + deltaY);
            applet.line(point2.getXpos() + deltaX, point2.getYpos() + deltaY,
                    point3.getXpos() + deltaX, point3.getYpos() + deltaY);
        }
    }
}
