package dds.view.simplegui;

import java.awt.Rectangle;
import java.awt.geom.Line2D;

import dds.control.DDSApplet;
import dds.model.simulation.Vector3D;

/**
 * The SimpleAbstractGUIElement class.
 * 
 * @author cyberdan
 */
// TODO Rename the class to AbstractSimpleGUIElement as per naming conventions.
public abstract class SimpleAbstractGUIElement {
    /**
     * The elementID is a unique numerical identifier for this component.
     */
    private int elementID = -1;

    /**
     * The description is a textual description of this component.
     */
    private String description = null;

    /**
     * The boundary describes the rectangular border of this component.
     */
    private Rectangle boundary = null;

    /**
     * The getBoundary method returns the boundary of this component.
     * 
     * @return boundary
     */
    public final Rectangle getBoundary() {
        return boundary;
    }

    /**
     * The setBoundary method changes the current boundary.
     * 
     * @param boundaryRect the new boundary
     */
    public final void setBoundary(final Rectangle boundaryRect) {
        this.boundary = boundaryRect;
    }

    /**
     * The getDescription gets the description.
     * 
     * @return description
     */
    public final String getDescription() {
        return description;
    }

    /**
     * The setDescription changes the description.
     * 
     * @param descriptionStr the new description
     */
    public final void setDescription(final String descriptionStr) {
        this.description = descriptionStr;
    }

    /**
     * The getElementID returns the elementID of this component.
     * 
     * @return elementID
     */
    public final int getElementID() {
        return elementID;
    }

    /**
     * The setElementID changes the elementID of this component.
     * 
     * @param idNum the new elementID
     */
    public final void setElementID(final int idNum) {
        this.elementID = idNum;
    }

    /**
     * The setup method setups up any book-keeping info. on a subclass.
     */
    public abstract void setup();

    /**
     * The dispose method tries to free up any book-keeping info. for this
     * subclass.
     */
    public abstract void dispose();

    /**
     * The draw method draws this component to the specified applet.
     * 
     * @param app the drawing screen
     */
    public abstract void draw(DDSApplet app);

    /**
     * The inBounds determines whether the point (px, py) is contained within
     * the boundary of this component.
     * 
     * @param pointX x coordinate
     * @param pointY y coordinate
     * @return true if in bounds; false otherwise
     */
    public final boolean inBounds(final int pointX, final int pointY) {
        boolean result = false;
        if (this.boundary != null) {
            result = new Rectangle(pointX, pointY, 1, 1)
                    .intersects(this.boundary);
        }
        return result;
    }

    /**
     * This does the same thing as the other method of this name, only it takes
     * in a boundary.
     * 
     * @param bound bound
     * @return boolean
     */
    // TODO Javadoc elementInsideMe.
    public final boolean elementInsideMe(final Rectangle bound) {
        boolean smallEnough = true;

        if (this.inBounds(bound.x, bound.y)) {
            if (this.inBounds(bound.x + bound.width, bound.y)) {
                if (inBounds(bound.x + bound.width, bound.y + bound.height)) {
                    if (!this.inBounds(bound.x, bound.y + bound.height)) {
                        smallEnough = false;
                    }
                } else {
                    smallEnough = false;
                }
            } else {
                smallEnough = false;
            }
        } else {
            smallEnough = false;
        }

        return smallEnough;
    }

    /**
     * The elementInsideMe method determines if 'sage' is inside the boundary of
     * this component.
     * 
     * @param sage sage
     * @return boolean
     */
    // TODO Write elementInsideMe
    public final boolean elementInsideMe(final SimpleAbstractGUIElement sage) {
        return this.elementInsideMe(sage.getBoundary());
    }

    /**
     * The elementFitsInsideMe determines if 'sage' is small enough to be fit
     * inside this component.
     * 
     * @param bound a boundary
     * @return true if element fits inside the area 'bound'
     */
    public final boolean elementFitsInsideMe(final Rectangle bound) {
        return (bound.width < this.getBoundary().width)
                && (bound.height < this.getBoundary().height);
    }

    /**
     * The elementFitsInsideMe determines if 'sage' is small enough to be fit
     * inside this component.
     * 
     * @param guiElement sage
     * @return true if element fits inside the boundary of 'sage'
     */
    public final boolean elementFitsInsideMe(
            final SimpleAbstractGUIElement guiElement) {
        return this.elementFitsInsideMe(guiElement.getBoundary());
    }

    /**
     * The centerMeInside.
     * 
     * @param bound the boundary
     * @see SimpleAbstractGuiElement.centerMeInside(SimpleAbstractGuiElement)
     */
    public final void centerMeInside(final Rectangle bound) {
        final int deltaX = (bound.width - this.boundary.width) / 2;
        final int deltaY = (bound.height - this.boundary.height) / 2;

        this.boundary.setLocation(bound.x + deltaX, bound.y + deltaY);
    }

    /**
     * The centerMeInside method centers this component inside the boundary of
     * 'sage.' This method assumes that this component is already small enough
     * to fit! If it's not small enough, it will just center it over the 'sage's
     * boundary.
     * 
     * @param sage sage
     */
    public final void centerMeInside(final SimpleAbstractGUIElement sage) {
        this.centerMeInside(sage.getBoundary());
    }

    /**
     * The scaleToFit.
     * 
     * @param bound the boundary to fit this component inside
     * @see AbstractGuiElement.scaleToFit(AbstractGuiElement)
     */
    public final void scaleToFit(final Rectangle bound) {
        Vector3D shrinkVector = null;
        final Line2D.Float shrinkLine = new Line2D.Float();
        final float theta = (float) Math.atan(boundary.getWidth()
                / boundary.getHeight());
        final float tWidth = (float) boundary.getWidth();
        final float tHeight = (float) boundary.getHeight();
        float deltaX = 0f;
        float deltaY = 0f;

        final Rectangle sbound = bound;
        final Line2D topEdge = new Line2D.Float(sbound.x, sbound.y, sbound.x
                + sbound.width, sbound.y);

        this.centerMeInside(bound);

        shrinkVector = new Vector3D(0, 1, 0);
        shrinkVector.rotateZ(-theta);
        shrinkVector.traceLine(shrinkLine, (float) Math.max(boundary.width,
                boundary.height), boundary.x, boundary.y);
        if (shrinkLine.intersectsLine(topEdge)) {
            deltaY = (float) topEdge.getY1() - boundary.y;
            deltaX = deltaY * (float) Math.tan(theta);
        } else {
            deltaX = (float) topEdge.getX1() - boundary.x;
            deltaY = deltaX / (float) Math.tan(theta);
        }

        boundary.x += deltaX;
        boundary.y += deltaY;
        boundary.width = Math.round(tWidth - 2f * deltaX);
        boundary.height = Math.round(tHeight - 2f * deltaY);
    }

    /**
     * The scaleToFit method scales this component's boundary to fit inside the
     * boundary of 'sage.' This method assumes that this component is larger in
     * dimensions than the 'sage' parameter's.
     * 
     * @param sage the element to fit this component inside
     */
    public final void scaleToFit(final SimpleAbstractGUIElement sage) {
        Vector3D shrinkVector = null;
        final Line2D.Float shrinkLine = new Line2D.Float();
        final float theta = (float) Math.atan(boundary.getWidth()
                / boundary.getHeight());
        final float tWidth = (float) boundary.getWidth();
        final float tHeight = (float) boundary.getHeight();
        float deltaX = 0f;
        float deltaY = 0f;

        final Rectangle sbound = sage.getBoundary();
        final Line2D topEdge = new Line2D.Float(sbound.x, sbound.y, sbound.x
                + sbound.width, sbound.y);

        this.centerMeInside(sage);

        shrinkVector = new Vector3D(0, 1, 0);
        shrinkVector.rotateZ(-theta);
        shrinkVector.traceLine(shrinkLine, (float) Math.max(boundary.width,
                boundary.height), boundary.x, boundary.y);
        if (shrinkLine.intersectsLine(topEdge)) {
            deltaY = (float) topEdge.getY1() - boundary.y;
            deltaX = deltaY * (float) Math.tan(theta);
        } else {
            deltaX = (float) topEdge.getX1() - boundary.x;
            deltaY = deltaX / (float) Math.tan(theta);
        }

        boundary.x += deltaX;
        boundary.y += deltaY;
        boundary.width = Math.round(tWidth - 2f * deltaX);
        boundary.height = Math.round(tHeight - 2f * deltaY);
    }

    /**
     * The handleMouseClick handles a mouse clicked event for this component.
     * 
     * @param mouseX mouseX
     * @param mouseY mouseY
     * @param mouseButton mouseButton
     */
    public abstract void handleMouseClick(final int mouseX, final int mouseY,
            final int mouseButton);
}
