package dds.view.simulation;

import java.util.List;
import java.util.Observable;
import java.util.Observer;

import processing.core.PImage;
import dds.control.DDSApplet;
import dds.control.user.modules.SimulationModule;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.SimObjectUpdate;
import dds.model.simulation.Vector3D;
import dds.view.RenderableContext;

/**
 * The AbstractSimObjectRenderer class defines the general way in which
 * AbstractSimObjects should be rendered in the view. The Renderers have an
 * ovserver-observable relationship with the SimObjects they draw. This helps us
 * properly abstract the model from the view, and by making the view responsive
 * only to updates issued by the model we are also preparing for the possiblity
 * of moving to a client-server version of the program.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public abstract class AbstractSimObjectRenderer implements RenderableContext,
        Observer {

    /**
     * The location field contains this renderer's copy of the most recent
     * SimObject's position.
     */
    private Vector3D position;

    /**
     * The orientation field contains this renderer's copy of the most recent
     * SimObject's orientation.
     */
    private Vector3D orientation;

    /**
     * The simObject field contains the reference to the SimObject to be
     * rendered.
     */
    private AbstractSimObject simObject;

    /**
     * Constructor for AbstractSimObjectRenderer.
     */
    private AbstractSimObjectRenderer() {
        super();
        this.position = new Vector3D();
        this.orientation = new Vector3D();
    }

    /**
     * Constructor for AbstractSimObjectRenderer.
     * 
     * @param obj The SimObject to use for coordinates'n'such.
     */
    public AbstractSimObjectRenderer(final AbstractSimObject obj) {
        this();
        this.setSimObject(obj);
    }

    /**
     * Getter for location.
     * 
     * @return the location
     */
    public final Vector3D getPosition() {
        // Return this renderer's copy of position instead of grabbing the
        // model's directly.
        return position;
    }

    /**
     * Getter for orientation.
     * 
     * @return the orientation
     */
    public final Vector3D getOrientation() {
        // Return this renderer's copy of orientation instead of grabbing the
        // model's directly.
        return orientation;
    }

    /**
     * Getter for simObject.
     * 
     * @return the simObject
     */
    public final AbstractSimObject getSimObject() {
        return simObject;
    }

    /**
     * Setter for position.
     * 
     * @param loc the position to set
     */
    public final void setPosition(final Vector3D loc) {
        this.position.setPos(loc);
    }

    /**
     * Setter for position.
     * 
     * @param xpos The x-coordinate.
     * @param ypos The y-coordinate.
     * @param zpos The z-coordinate.
     */
    public final void setPosition(final float xpos, final float ypos,
            final float zpos) {
        this.position.setPos(xpos, ypos, zpos);
    }

    /**
     * Setter for orientation.
     * 
     * @param vector the orientation to set
     */
    public final void setOrientation(final Vector3D vector) {
        this.orientation.setPos(vector);
    }

    /**
     * Setter for simObject.
     * 
     * @param simobj the simObject to set
     */
    public final void setSimObject(final AbstractSimObject simobj) {
        // Check to see if there already was a SimObject assigned.
        if (this.simObject != null) {
            // There was so remove this as an observer from it.
            this.simObject.deleteObserver(this);
        }

        // Set the new SimObject reference.
        this.simObject = simobj;

        // Update the position and orientation.
        this.setPosition(this.simObject.getPosition());
        this.setOrientation(this.simObject.getOrientation());

        // Set this as an observer of the SimObject.
        this.simObject.addObserver(this);
    }

    /**
     * The update method handles updates from an Observable.
     * 
     * @param obj The Object that called the update.
     * @param arg The argument of the update.
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public final void update(final Observable obj, final Object arg) {
        final AbstractSimObject updatedObj = (AbstractSimObject) obj;
        final SimObjectUpdate update = (SimObjectUpdate) arg;
        switch (update) {
        case POSITION:
            // Object's position has changed.
            this.setPosition(updatedObj.getPosition());
            break;
        case ORIENTATION:
            // Object's orientation has changed.
            this.setOrientation(updatedObj.getOrientation());
            break;
        default:
            // Object type-specific update received, handle in more specific
            // renderer.
            this.updateFromSim(arg);
        }
    }

    /**
     * The updateFromSim updates this renderer to match the data within the
     * simulation.
     * 
     * @param arg The argument of the update.
     */
    public abstract void updateFromSim(Object arg);

    /**
     * The drawOn method performs all of the operations necessary to render the
     * SimObject.
     * 
     * @param applet The DDSApplet to render on.
     * @see dds.view.RenderableContext#drawOn(dds.control.DDSApplet)
     */
    public final void drawOn(final DDSApplet applet) {
        // Get the position of the SimObject.
        final float posX = this.getPosition().getXpos();
        final float posY = this.getPosition().getYpos();
        final float posZ = this.getPosition().getZpos();

        // Turn off stroke and push the transformation matrix.
        applet.noStroke();
        applet.pushMatrix();

        // Translate to the position of the SimObject. This allows us to use
        // local coordinates for drawing.
        applet.translate(posX, posY, posZ);

        // Apply the orientation.
        this.applyOrientation(applet);

        // Check to see if this object's location is visible on screen, and
        // failing that if any bounding verts are visible.
        // TODO Improve this to check for bounding lines so that far-spaced
        // verts won't prevent drawing.
        if (this.isPointOnScreen(0, 0, 0, applet)
                || this.isBoundingOnScreen(applet)) {
            // Since it's apparently visible, draw the SimObject.
            this.drawModel(applet);
        }

        // Restore the transformation matrix.
        applet.popMatrix();
    }

    /**
     * The isBoundingOnScreen method determines if this object has bounding
     * coordinates that will appear on the screen given the DDSApplet's current
     * transformation matrix.
     * 
     * @param applet The applet to infer screen coordinate from.
     * @return True if the bounding of the SimObject will appear on the screen.
     */
    private boolean isBoundingOnScreen(final DDSApplet applet) {
        // Guess that the result will be false.
        boolean result = false;

        // Get the object's bounding verts.
        final List<Vector3D> boundingVerts = this.simObject.getBoundingVerts();

        // Check to make sure we received bounding verts.
        if (boundingVerts == null) {
            // TODO Fix getBoundingVerts() method for this object.
            // TODO RoadSign's broken.

            // Uncomment the following two lines to discover what object has a
            // broken getBoundingVerts method.
            // System.err.println(this.getSimObject().getIdStr()
            // + " has no bounding verts!");

            // This is set just so checkstyle stops complaining about empty if
            // statements.
            result = false;
        } else {
            Vector3D vert;
            // Iterate through the list of bounding verts.
            for (int i = 0; i < boundingVerts.size(); i++) {
                vert = boundingVerts.get(i);
                // Check to see if the vert appears on the screen.
                if (this.isPointOnScreen(vert.getXpos(), vert.getYpos(), vert
                        .getZpos(), applet)) {
                    // It does so set the reslt to true and break the loop.
                    result = true;
                    break;
                }
                int nextVertIndex = i + 1;
                if (nextVertIndex >= boundingVerts.size()) {
                    nextVertIndex = 0;
                }
                // Now try to see if this line is visible.
                if (this.isLineOnScreen(vert, boundingVerts.get(nextVertIndex),
                        applet)) {
                    // It does so set the reslt to true and break the loop.
                    result = true;
                    break;
                }
            }
        }

        // Return the result.
        return result;
    }

    /**
     * The isLineOnScreen method determines if the line segment inscribed by the
     * specified points will be visible on the screen.
     * 
     * @param vertA The first vert of the line segment.
     * @param vertB The second vert of the line segment.
     * @param applet The applet to get screen coordinates from.
     * @return True if the line is going to appear onscreen.
     */
    private boolean isLineOnScreen(final Vector3D vertA, final Vector3D vertB,
            final DDSApplet applet) {
        boolean result = false;
        final float vertAX = applet.screenX(vertA.getXpos(), vertA.getYpos(),
                vertA.getZpos());
        final float vertAY = applet.screenY(vertA.getXpos(), vertA.getYpos(),
                vertA.getZpos());
        final float vertBX = applet.screenX(vertB.getXpos(), vertB.getYpos(),
                vertB.getZpos());
        final float vertBY = applet.screenY(vertB.getXpos(), vertB.getYpos(),
                vertB.getZpos());

        final float deltaX = vertBX - vertAX;
        final float deltaY = vertBY - vertAY;

        final float topX = this
                .calcIntersect(0, deltaY, vertAY, deltaX, vertAX);
        final float bottomX = this.calcIntersect(applet.height, deltaY, vertAY,
                deltaX, vertAX);
        final float leftY = this.calcIntersect(0, deltaX, vertAX, deltaY,
                vertAY);
        final float rightY = this.calcIntersect(applet.width, deltaX, vertAX,
                deltaY, vertAY);

        result = ((topX >= 0) && (topX <= applet.width))
                || ((bottomX >= 0) && (bottomX <= applet.width))
                || ((leftY >= 0) && (leftY <= applet.height))
                || ((rightY >= 0) && (rightY <= applet.height));

        return result;
    }

    /**
     * The calcIntersect method is a parameterization of the solution for line
     * segment intersection with a horizontal or vertical line.
     * 
     * @param known The known value of the horizontal or vertical line.
     * @param knownDelta The known change of the line segment along that axis.
     * @param knownStart The known starting point of the line segment.
     * @param unknownDelta The change of the unknown variable.
     * @param unknownStart The starting point for the unknown variable.
     * @return The point of intersection, -1 if there isn't one and 0 if the
     *         knownDelta is 0.
     */
    private float calcIntersect(final float known, final float knownDelta,
            final float knownStart, final float unknownDelta,
            final float unknownStart) {
        float result = -1;
        if (knownDelta == 0) {
            result = 0;
        } else {
            final float time = (known - knownStart) / knownDelta;
            if ((time >= 0) && (time <= 1)) {
                result = unknownDelta * time + unknownStart;
            }
        }
        return result;
    }

    /**
     * The isPointOnScreen method determines if an arbitrary 3D point will
     * appear on the screen given the DDSApplet's current transformation matrix.
     * 
     * @param posX The x-coordinate.
     * @param posY The y-coordinate.
     * @param posZ The z-coordinate.
     * @param applet The applet to infer screen coordinate from.
     * @return True if the point will appear on the screen.
     */
    private boolean isPointOnScreen(final float posX, final float posY,
            final float posZ, final DDSApplet applet) {
        // Get the screen coordinates for the 3D model point using the current
        // transformation matrix in the specified DDSApplet.
        final float screenX = applet.screenX(posX, posY, posZ);
        final float screenY = applet.screenY(posX, posY, posZ);
        final float screenZ = applet.screenZ(posX, posY, posZ);

        // Check to see if the coordinate is within the screen dimensions and in
        // front of the camera.
        return (screenX >= 0) && (screenX <= applet.width) && (screenY >= 0)
                && (screenY <= applet.height) && (screenZ >= 0)
                && (screenZ <= 1);
    }

    /**
     * The getMapping method calculates a uv-coordinate for a vertex from the
     * model.
     * 
     * @param vert The vertex to get a uv mapping for.
     * @param isUCoordinate True if a u-coordinate is desired.
     * @param applet The DDSApplet to derive screen coordinates from.
     * @param texture The PImage being used for texturing.
     * @return The uv-coordinate.
     */
    public final float getMapping(final Vector3D vert,
            final boolean isUCoordinate, final DDSApplet applet,
            final PImage texture) {
        // Find the greater screen dimension.
        final float maxScreen = (float) Math.max(applet.width, applet.height);

        // Declare space for the result and the appropriate screen:image ratio.
        float result;
        float screenImageRatio;

        // Determine if we're trying to find a u or v coordinate.
        if (isUCoordinate) {
            // U-coordinates should be based on image width.
            screenImageRatio = maxScreen / (float) texture.width;
            result = vert.getXpos() / (float) texture.width;
        } else {
            // V-coordinates should be based on image height.
            screenImageRatio = maxScreen / (float) texture.height;
            result = vert.getYpos() / (float) texture.height;
        }

        // Return the results.
        return result * screenImageRatio * SimulationModule.TEXTURE_SCALE;
    }

    /**
     * The applyOrientation method handles the use of an object's orientation
     * during rendering. Since not all objects need their orientation handled,
     * not all renderers will use a significant implementation of this.
     * 
     * @param applet The DDSApplet to draw onto.
     */
    public abstract void applyOrientation(DDSApplet applet);

    /**
     * The drawModel method draws JUST the model to the screen in local
     * coordinates. The drawOn() method should have performed the appropriate
     * transforms for this to work.
     * 
     * @param applet The DDSApplet to draw onto.
     */
    public abstract void drawModel(DDSApplet applet);

}
