/**
 * AbstractVehicleRenderer.java.
 */
package dds.view.simulation;

import processing.core.PConstants;
import processing.core.PFont;
import dds.control.DDSApplet;
import dds.model.simulation.AbstractMovableObject;
import dds.model.simulation.AbstractVehicle;
import dds.model.simulation.Kinetics;
import dds.model.simulation.Vector3D;

/**
 * The AbstractVehicleRenderer class is responsible for defining the basic
 * functionality of all Vehicle renderers. These renderers need to be able to
 * properly draw any Vehicle from the simulation.
 * 
 * @author Jacob Bolton
 */
public abstract class AbstractVehicleRenderer extends
        AbstractMovableObjectRenderer implements PConstants {

    /**
     * The GUAGE_BACK_COLOR is the default color to use for rendering the 'back'
     * of a guage.
     */
    private static final int GUAGE_BACK_COLOR = 0;

    /**
     * The NEEDLE_COLOR is the default color to use for rendering the 'needle'
     * of a guage.
     */
    private static final int NEEDLE_COLOR = 0xff999999;

    /**
     * The GUAGE_START the position in radians of where the have the starting
     * point for the guage needle.
     */
    private static final float GUAGE_START = PI - QUARTER_PI;

    /**
     * The GUAGE_COVERAGE the number of radians that the needle can traverse.
     */
    private static final float GUAGE_COVERAGE = TWO_PI - HALF_PI;

    /**
     * The GUAGE_SIZE the size used for drawing the guage's arcs.
     */
    private static final float GUAGE_SIZE = 0.3f;

    /**
     * The NEEDLE_SIZE the size of the needle in radians. This describes how
     * thick the needle should be with respect to the guage's radians.
     */
    private static final float NEEDLE_SIZE = 0.0981748f;

    /**
     * The GUAGE_TEXT_SIZE the size to use for drawing the text on a guage.
     */
    private static final float GUAGE_TEXT_SIZE = 0.1f;

    /**
     * The TEXT_Z_OFFEST the z-axis offset to use for drawing text on the guage.
     * this is used to prevent drawing problems that can occur when two planes
     * are rendered in the same place and the renderer needs to pick which one
     * to display.
     */
    private static final float TEXT_Z_OFFEST = 0.01f;

    /**
     * The guageFont field stores the PFont to use for rendering text on guages.
     */
    private final transient PFont guageFont;

    /**
     * Constructor for AbstractVehicleRenderer.
     * 
     * @param obj The object that this renderer will observe.
     */
    public AbstractVehicleRenderer(final AbstractMovableObject obj) {
        super(obj);

        // Load up the font for guages.
        this.guageFont = DDSApplet.getApplet().loadFont(
                "LucidaBright-Demi-48.vlw");
    }

    /**
     * The interpolate method handles making changes to this renderer's
     * positional data based on the last known motion data for this moveable
     * object. This will hopefully compensate for some lag that may occur
     * between updates from the model and the view.
     * 
     * @param time The time change since last draw.
     */
    @Override
    public final void interpolate(final float time) {
        this.basicInterpolate(time);
    }

    /**
     * The updateFromSim method handles updates coming from within the
     * simulation. Based on the update argument it will appropriately update the
     * view for the change in the model.
     * 
     * @param arg The argument for the update.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #updateFromSim(java.lang.Object)
     */

    public void updateFromSim(final Object arg) {
        // TODO Write and comment me!

    }

    /**
     * The applyOrientation method is responsible for using the Vehicle's
     * orientation field to determine the rotation that should be applied for
     * the Vehicle to be rendered correctly. After calculating the appropriate
     * rotation, this method also applies the rotation tranformation.
     * 
     * @param applet The DDSApplet containing the transformation matrix to apply
     *            a rotation to.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #applyOrientation(dds.control.DDSApplet)
     */
    @Override
    public final void applyOrientation(final DDSApplet applet) {
        // Calculate the angle of rotation to get from y-axis to the Vehicle's
        // orientation.
        final float angle = (float) Math.atan2(this.getOrientation().getYpos(),
                this.getOrientation().getXpos())
                + (DDSApplet.PI / 2.0f);
        // Apply that angle of rotation to the current transformation matrix so
        // the Vehicle gets rendered properly.
        applet.rotateZ(angle);

    }

    /**
     * The getRightNormal method goes to the model to retrieve this Vehicle's
     * right normal based on the current orientation.
     * 
     * @return A normalized Vector3D representing the right normal.
     */
    public final Vector3D getRightNormal() {
        // Go to the model and get the Vehicle's right normal.
        // TODO Make this so it doesn't explicitly depend on calls into the
        // model. This kind of abstraction will prepare the eventuality where
        // the applet becomes a client to a simulation server.
        return ((AbstractVehicle) this.getSimObject()).getRightNormal();
    }

    /**
     * The getCamPosition method calculates what the camera's position should be
     * if it is viewing the simulation from this Vehicle's driver's position.
     * 
     * @param cameraZ The additional height to apply to the camera.
     * @return A Vector3D coordinate for the camera's position.
     */
    public final Vector3D getCamPosition(final float cameraZ) {
        // Get the coordinates for the Vehicle.
        final float posX = this.getPosition().getXpos();
        final float posY = this.getPosition().getYpos();
        final float posZ = this.getPosition().getZpos();

        // Get the orientation of the Vehicle.
        final float orientX = this.getOrientation().getXpos();
        final float orientY = this.getOrientation().getYpos();

        // Get the right normal of the Vehicle.
        final float normX = -this.getRightNormal().getXpos();
        final float normY = -this.getRightNormal().getYpos();

        // Set the amounts to travel along the orientation and right normals.
        // TODO Refactor this to use a localized camera positioning system that
        // should be in the model of the Vehicle.
        final float oShift = 1.0f;
        final float nShift = 1.5f;

        // Calculate the coordinates for the camera.
        final float camX = posX - (oShift * orientX) + (nShift * normX);
        final float camY = posY - (oShift * orientY) + (nShift * normY);
        final float camZ = posZ + cameraZ;

        // Return a new Vector3D containing the coordinates for the camera.
        return new Vector3D(camX, camY, camZ);
    }

    /**
     * The getCamAimPosition method calculates what the camera's look at
     * coordinate should be if it is viewing the simulation from this Vehicle's
     * driver's position.
     * 
     * @param cameraZ The additional height to apply to the camera.
     * @param viewRotation The number of radians to deviate from looking
     *            directly in front of the Vehicle.
     * @return A Vector3D coordinate for the camera's position.
     */
    // TODO Determine if cameraZ is really necessary for this method.
    public final Vector3D getCamAimPosition(final float cameraZ,
            final float viewRotation) {
        // Get the position of the Vehicle.
        final float posX = this.getPosition().getXpos();
        final float posY = this.getPosition().getYpos();
        final float posZ = this.getPosition().getZpos();

        // Get the right normal of the Vehicle.
        final float normX = -this.getRightNormal().getXpos();
        final float normY = -this.getRightNormal().getYpos();

        // Make a copy of the Vehicle's orientation vector.
        final Vector3D orient = this.getOrientation().duplicate();

        // Set the amounts to shift along the normals.
        final float aimDist = 30;
        final float nShift = 1.5f;

        // Determine if we're supposed to rotate the view away from in front of
        // the vehicle.
        if (viewRotation != 0) {
            // If so apply the rotation to the orientation vector.
            orient.rotateZ(viewRotation);
        }

        // Calculate the aim at position.
        final float aimX = posX + (orient.getXpos() * aimDist)
                + (nShift * normX);
        final float aimY = posY + (orient.getYpos() * aimDist)
                + (nShift * normY);
        final float aimZ = posZ + (orient.getZpos() * aimDist);

        // Return a new Vector3D containing the aim-at coordinates.
        return new Vector3D(aimX, aimY, aimZ);
    }

    /**
     * The drawGuage method draws a guage graphics at the specified position
     * with the specified attributes.
     * 
     * @param applet The DDSApplet to draw onto.
     * @param position The 3D position of the guage.
     * @param ratio The ratio of completion for the guage.
     * @param maxValue The maximum value represented by a totally completed
     *            guage.
     * @param needleColor The color used for drawing the needle.
     * @param backColor The color used for drawing the background.
     */
    public final void drawGuage(final DDSApplet applet,
            final Vector3D position, final float ratio, final float maxValue,
            final int needleColor, final int backColor) {
        // Push the transformation matrix to preserve previous transformations.
        applet.pushMatrix();
        // TODO Work on a way to change the font without slowing the framerate.
        // applet.loadFont("fonts/LucidaBright-Demi-48.vlw");
        // Translate to the guage's position
        applet.translate(position.getXpos(), position.getYpos(), position
                .getZpos());
        // Rotate the guage so that it's 'perpendicular' to the Vehicle's
        // orientation and thus the camera.
        applet.rotateX(-HALF_PI);
        // Set the fill color for the background of the guage.
        applet.fill(backColor);
        // Draw the background part of the guage.
        applet.arc(0, 0, GUAGE_SIZE, GUAGE_SIZE, GUAGE_START
                + (GUAGE_COVERAGE * ratio), GUAGE_START
                + (GUAGE_COVERAGE * ratio) - NEEDLE_SIZE);
        // Set the fill color to the needle color.
        applet.fill(needleColor);
        // Draw the needle part of the guage.
        applet.arc(0, 0, GUAGE_SIZE, GUAGE_SIZE, GUAGE_START
                + (GUAGE_COVERAGE * ratio) - NEEDLE_SIZE, GUAGE_START
                + (GUAGE_COVERAGE * ratio));
        // Set the fill color to white.
        applet.fill(-1);
        // Set some text rendering parameters.
        applet.textFont(this.guageFont);
        applet.textMode(MODEL);
        applet.textAlign(CENTER);
        applet.textSize(GUAGE_TEXT_SIZE);
        // Draw the text that needs to be shown on the guage.
        applet.text(Integer.toString((int) (ratio * maxValue)), 0, 0,
                TEXT_Z_OFFEST);
        // Restore the previous transformation matrix.
        applet.popMatrix();
    }

    /**
     * The drawSpeedometer draws a speedometer for the Vehicle.
     * 
     * @param applet The DDSApplet to draw the speedometer in.
     */
    public final void drawSpeedometer(final DDSApplet applet) {
        // TODO Figure out a way to not depend directly on the model for
        // positioning and such. This abstraction wil prepare the system for the
        // eventuality where the we use a client to connect to a simulation
        // server.
        // Get the position of the speedometer from the model.
        final Vector3D pos = ((AbstractVehicle) this.getSimObject())
                .getSpeedometer();
        // Get the current speed of the Vehicle from the model.
        float speed = ((AbstractVehicle) this.getSimObject()).getSpeed();
        // Convert the units from m/s to miles per hour.
        speed = Kinetics.mpsToMph(speed);
        // Get the top speed of the Vehicle from the model. This is currently in
        // miles per hour so no conversion is needed.
        final float topSpeed = ((AbstractVehicle) this.getSimObject())
                .getTopSpeed();
        // Calculate the ratio of speed to topSpeed. This will be used to
        // position the needle on the guage.
        final float ratio = speed / topSpeed;
        // Draw the guage.
        this.drawGuage(applet, pos, ratio, topSpeed, NEEDLE_COLOR,
                GUAGE_BACK_COLOR);
    }

    /**
     * The drawAccelometer draws a accelometer for the Vehicle.
     * 
     * @param applet The DDSApplet to draw the accelometer in.
     */
    public final void drawAccelometer(final DDSApplet applet) {
        // TODO Figure out a way to not depend directly on the model for
        // positioning and such. This abstraction wil prepare the system for the
        // eventuality where the we use a client to connect to a simulation
        // server.
        // Get the position of the accelometer from the model.
        final Vector3D pos = ((AbstractVehicle) this.getSimObject())
                .getAccelometer();
        // Get the current amount of acceleration from the model.
        final float accel = ((AbstractVehicle) this.getSimObject())
                .getAccelerationAmount();
        // Get the maximum acceleration from the model.
        final float maxAccel = ((AbstractVehicle) this.getSimObject())
                .getMaxAcceleration();
        // Calculate the ratio of current acceleration to maximuym acceleration.
        // this will be used to position the needle on the guage.
        final float ratio = accel / maxAccel;
        // Draw the guage.
        this.drawGuage(applet, pos, ratio, maxAccel, NEEDLE_COLOR,
                GUAGE_BACK_COLOR);
    }
}
