package dds.view.simulation.renderers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import processing.core.PConstants;
import processing.core.PImage;
import dds.control.DDSApplet;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Block;
import dds.model.simulation.BlockType;
import dds.model.simulation.Building;
import dds.model.simulation.Vector3D;
import dds.view.simulation.AbstractSimObjectRenderer;

/**
 * The BlockRenderer class is responsible for drawing a Block from the
 * simulation.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public class BlockRenderer extends AbstractSimObjectRenderer implements
        PConstants {

    /**
     * The NO_TEXTURE constant is used to initialize the texture to null to
     * indicate that there's no texture set.
     */
    private static final PImage NO_TEXTURE = null;

    /**
     * The WHITE constant defines the int value that represents the color white.
     */
    private static final int WHITE = 0xffffffff;

    /**
     * The verts field is the renderer's copy of the verts for the Block. The
     * renderer retains its own copy of verts to be more independent of the
     * model. This may also be useful in the event we decide to separate this
     * program into a client and server.
     */
    private final transient Collection<Vector3D> verts;

    /**
     * The buildings field contains a set of BuildingRenderers, which are
     * responsible for rendering the Buildings that appear on the Block.
     */
    private transient List<BuildingRenderer> buildings;

    /**
     * The texture field is storage for the loaded texture that will be used on
     * this Block. It is stored here to improve performance since it will be
     * loaded once and in one instance rather than creating new instances of the
     * texture at each frame.
     */
    private transient PImage texture;

    /**
     * Getter for verts.
     * 
     * @return the verts
     */
    private Collection<Vector3D> getVerts() {
        return verts;
    }

    /**
     * Setter for verts.
     * 
     * @param vertices the verts to set
     */
    private void setVerts(final Collection<Vector3D> vertices) {
        // Clear out the old verts and add the new ones without creating a new
        // Collection or losing an old one.
        this.verts.clear();
        this.verts.addAll(vertices);
    }

    /**
     * The updateFromSim method handles updates coming from within the
     * simulation. Based on the update argument it will appropriately update the
     * view for the change in the model.
     * 
     * @param arg The argument for the update.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #updateFromSim(java.lang.Object)
     */
    public void updateFromSim(final Object arg) {
        // Blocks shouldn't really be doing anything in the simulation so we
        // don't realy need to handle any updates. Thus this should be blank.
    }

    /**
     * The drawModel method draws JUST the model to the screen in local
     * coordinates. The drawOn() method should have performed the appropriate
     * transforms for this to work. The BlockRenderer will draw a ground polygon
     * and then have it's BuildingRenderers draw buildings in their appropriate
     * places on the ground polygon. If no texture is set then the BockRenderer
     * will default to using its BlockType's default color. If there is a
     * texture, it will be repeated on the ground polygon unless the BlockType
     * is USER. The USER Blocktype will map the texture so that the polygon's
     * shape will cover the area of the texture's image only once. USER
     * BlockTypes need to be textured differently because the user may want to
     * map a special image (such as a park) that has a certain layout that
     * should only occur once on the entire block.
     * 
     * @param applet The DDSApplet to draw onto.
     */
    public final void drawModel(final DDSApplet applet) {
        // Get the Block's BlockType.
        // TODO Consider storing this in a field within this class for better
        // model-view separation.
        final BlockType blockType = ((Block) this.getSimObject()).getType();

        // Begin a POLYGON for drawing the ground.
        applet.beginShape();

        // Determine if a texture has been set.
        if (this.texture == null) {
            // There's no texture so use the Blocktype's default color.
            applet.fill(blockType.getColor());
        } else {
            // Set the texture for this POLYGON.
            applet.texture(this.texture);

            // Use a white fill color so that the texture being used is not
            // discolored from being mixed with the fill color.
            applet.fill(WHITE);
        }

        // Create a variable to keep track of the index of the vertex being
        // textured. This will be used for uv coordinate calculation for the
        // USER BuildingType.
        int textureVertNum = 0;

        // Get the total number of verts for the ground polygon. This will also
        // be used to uv coordinate calculation of USER BlockTypes.
        final int textureVertCount = this.getVerts().size();

        // Set a flag describing if the Block being drawn is a USER BlockType.
        final boolean isUserType = blockType == BlockType.USER;

        // Iterate through all of the verts making up the ground polygon.
        for (Vector3D vert : this.getVerts()) {
            // Check to see if there's a texture assigned to this Block.
            if (this.texture == null) {
                // When there's no texture we can ignore uv coordinate mapping.
                applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos());
            } else {
                // When there IS a texture we need to specify uv coordinates for
                // each vert. The getMapping method is a nice abstraction to
                // make the coordinate calculation process easier to read and
                // maintain.
                applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos(),
                        this.getMapping(vert, textureVertNum, textureVertCount,
                                true, isUserType, applet), this.getMapping(
                                vert, textureVertNum, textureVertCount, false,
                                isUserType, applet));
                // Increment the current vert number so that the next uv
                // coordinate calculation can proceed nicely if the BlockType is
                // USER.
                textureVertNum++;
            }
        }

        // Finish drawing the ground polygon.
        applet.endShape();

        // Iterate through all of the BuildingRenderers attached to this.
        for (BuildingRenderer building : this.buildings) {
            // Make the Buildingrenderer render its building.
            building.drawOn(applet);
        }
    }

    /**
     * The getMapping method is a nice abstraction to make the coordinate
     * calculation process easier to read and maintain. This version for
     * BlockRenderer allows two different kinds of calculations to take place:
     * one for USER BlockTypes and on for every other type. The USER-based
     * calculation uses the current vert and the total number of verts to
     * determine pick a point around the edge of the texture to use. This
     * coordinate proceeds along the exterior of the image according to the
     * ratio of index to total. As index approaches total a full revolution
     * around the texture image. The non-USER type approach simply defaults to
     * the getMapping method defined in AbstractSimObjectRenderer.
     * 
     * @param vert The Vector3D vertex for which a u or v coordinate is desired.
     * @param index The index of the vertex from the collection of verts which
     *            this one belongs.
     * @param total The total number of verts in the collection that makes up
     *            the polygon to be textured.
     * @param isUCoordinate True if the coordinate desired is a u coordinate;
     *            false for a v coordinate.
     * @param isUserType True if the BlockType is USER; otherwise false.
     * @param applet The DDSApplet which is being rendered on.
     * @return A float texture coordinate for either u or v.
     */
    private float getMapping(final Vector3D vert, final int index,
            final int total, final boolean isUCoordinate,
            final boolean isUserType, final DDSApplet applet) {
        // Declare storage for the final result of this method.
        float result;

        // Check to see if the BlockType is set to USER
        if (isUserType) {
            // If it is, start by determining the number of radians to use per
            // vertex to completely encircle the texture image.
            final float radsPerIndex = TWO_PI / ((float) total);

            // Calculate the number of radians to use for this particular
            // vertex.
            final float rads = radsPerIndex * ((float) index)
                    + (PI / (float) total);

            // Use trigonometry to get some uv coordinates from a circle.
            final float resultU = (float) Math.cos(rads);
            final float resultV = (float) Math.sin(rads);

            // Since texture images are NOT circular we need to map the circular
            // coordinates to the bounding rectangle of the image. To do this we
            // use a simple calculation using the following formula:
            //
            // vCoordinate = (resultV / resultU) * uCoordinate
            //
            // To map to the edges of the for one coordinate image we choose
            // the largest coordinate value as a divisor and then divide the
            // circular coordinates by the magnitude of that divisor.
            final float divisor;

            // Determine which circular coordinate is larger.
            if (resultU > resultV) {
                // Use the magintude of U to be the divisor.
                divisor = Math.abs(resultU);
            } else {
                // Use the magintude of V to be the divisor.
                divisor = Math.abs(resultV);
            }

            // Determine which coordinate we want to calculate.
            if (isUCoordinate) {
                // Scale the U coordinate since we want it.
                result = resultU / divisor;
            } else {
                // Scale the V coordinate since we want it.
                result = resultV / divisor;
            }

            // This process has given use values that range from -1 to 1, but in
            // order to have the texture's center centered and not repeating we
            // need to shift that range to 0 to 1 and store that as the result.
            result = (result + 1.0f) / 2.0f;
        } else {
            // Since this is not a USER BlockType, default to the getMapping
            // method defined in the superclass.
            result = this.getMapping(vert, isUCoordinate, applet, texture);
        }

        // Return the final result.
        return result;
    }

    /**
     * Constructor for BlockRenderer.
     * 
     * @param obj The Block that this will be responsible for rendering.
     */
    public BlockRenderer(final Block obj) {
        // Use the superclass constructor to initialize superclass fields.
        super(obj);

        // Create storage for the ground verts of the Block.
        this.verts = new ArrayList<Vector3D>();

        // Force the Block to calculate its verts and then store them into the
        // verts Collection.
        obj.calculateVerts();
        this.setVerts(this.getSimObject().getVerts());

        // Create space to store the List of BuildingRenderers attached to this
        // Block.
        this.buildings = new ArrayList<BuildingRenderer>();

        // Call the method that creates and populates the Buildingrenderer List.
        this.initializeBuildings();

        // Grab the texture filename that should be used for the ground polygon.
        String textureFile = obj.getTextureFile();

        // Check to make sure we actually got a name.
        if (textureFile == null) {
            // If not, initialize the texture to null.
            this.texture = NO_TEXTURE;
        } else {
            // Otherwise try to load the texture image from the data repository.
            this.texture = DDSApplet.getApplet().loadImage(textureFile);
        }
    }

    /**
     * The initializeBuildings method sets up the list of Buildingrenderers for
     * all of the Buildings that are attached to this Block and should be drawn
     * via this BlockRenderer.
     */
    private void initializeBuildings() {
        // Make sure the list fo buildings is clear before we populate it. We
        // don't want any older buildings sneaking in here for any reason.
        this.buildings.clear();

        // Iterate through the Block's set of Buildings.
        for (Building building : ((Block) this.getSimObject())
                .getBuildingList()) {
            // Create and add a new BuildingRenderer for this particular
            // Building.
            this.createBuildingRenderer(building);
        }
    }

    /**
     * The createBuildingRenderer creates a new Buildingrenderer for a given
     * building and then adds it to this BlockRenderer's list of
     * Buildingrenderers.
     * 
     * @param building The Building to create a renderer for.
     */
    private void createBuildingRenderer(final Building building) {
        // Create and add a new buildingrenderer to the list.
        this.buildings.add(new BuildingRenderer(building));
    }

    /**
     * Constructor for BlockRenderer.
     * 
     * @param obj The Block that this will be responsible for rendering.
     */
    public BlockRenderer(final AbstractSimObject obj) {
        // Make sure we're getting a Block to build this renderer from and NOT
        // some other SimObject.
        this((Block) obj);
    }

    /**
     * The applyOrientation method handles the use of an object's orientation
     * during rendering. Since not all objects need their orientation handled,
     * not all renderers will use a significant implementation of this.
     * 
     * @param applet The DDSApplet to draw onto.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #applyOrientation(dds.control.DDSApplet)
     */
    @Override
    public void applyOrientation(final DDSApplet applet) {
        // Blocks don't have any orientation data that needs to be applied for
        // rendering because they should/do NOT move.
    }

}
