/**
 * BuildingRenderer.java.
 */
package dds.view.simulation.renderers;

import java.util.List;

import processing.core.PImage;
import dds.control.DDSApplet;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Building;
import dds.model.simulation.Vector3D;
import dds.view.simulation.AbstractSimObjectRenderer;

/**
 * The BuildingRenderer class is responsible for rendering Buildings from the
 * simulation.
 * 
 * @author Jacob Bolton
 */
public class BuildingRenderer extends AbstractSimObjectRenderer {

    /**
     * The BLACK constant defines the color black.
     */
    private static final int BLACK = 0;

    /**
     * The FIRST constant defines the index of the first vert in an array of
     * verts that define a wall.
     */
    private static final int FIRST = 0;

    /**
     * The LAST constant defines the index of the last vert in an array of verts
     * that define a wall.
     */
    private static final int LAST = 3;

    /**
     * The NO_TEXTURE constant is an easy-to-read way to set the texture to
     * null.
     */
    private static final PImage NO_TEXTURE = null;

    /**
     * The SECOND constant defines the index of the second vert in an array of
     * verts that define a wall.
     */
    private static final int SECOND = 1;

    /**
     * The THIRD constant defines the index of the third vert in an array of
     * verts that define a wall.
     */
    private static final int THIRD = 2;

    /**
     * The WHITE constant defines the color white.
     */
    private static final int WHITE = -1;

    /**
     * The texture field contains the PImage used to texture this Building.
     */
    private transient PImage texture;

    /**
     * Constructor for BuildingRenderer.
     * 
     * @param obj The Block to render.
     */
    public BuildingRenderer(final AbstractSimObject obj) {
        this((Building) obj);
    }

    /**
     * Constructor for BuildingRenderer.
     * 
     * @param obj The Block to render.
     */
    public BuildingRenderer(final Building obj) {
        super(obj);

        // Get the texture file from the model.
        // TODO Consider moving this functionality into the view.
        String textureFile = obj.getTextureFile();

        // Check to see if there is a texture assigned.
        if (textureFile == null) {
            // There isn't so set our texture apropriately.
            this.texture = NO_TEXTURE;
        } else {
            // There is so load the texture file into a PImage.
            this.texture = DDSApplet.getApplet().loadImage(textureFile);
        }

        // Set the position of this Building to the origina because this is
        // drawn via the BlockRenderer and the vert coordinates are relative to
        // the Block's origin.
        this.setPosition(0, 0, 0);
    }

    /**
     * The applyOrientation method handles the use of an object's orientation
     * during rendering. Since not all objects need their orientation handled,
     * not all renderers will use a significant implementation of this.
     * 
     * @param applet The DDSApplet to draw onto.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #applyOrientation(dds.control.DDSApplet)
     */
    @Override
    public final void applyOrientation(final DDSApplet applet) {
        // TODO Determine if applyOrientation in BuildingRenderer needs to be
        // written. Currently I'm thinking no.
    }

    /**
     * The drawCeiling method draws the ceiling of the Building.
     * 
     * @param applet The DDSApplet to draw onto.
     * @param ceilingVerts The List of verts to draw.
     */
    // TODO Consider refactoring this so that it can be used to draw arbitrary
    // building faces that are defined by vert Lists.
    private void drawCeiling(final DDSApplet applet,
            final List<Vector3D> ceilingVerts) {
        // Turn off stroke.
        applet.noStroke();

        // Set the fill to black.
        applet.fill(BLACK);

        // Begin a shape for the ceiling.
        applet.beginShape();

        // Iterate over all of the ceiling verts.
        for (Vector3D vert : ceilingVerts) {
            // Draw the current vertex.
            applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos());
        }
        // Complete the Ceiling shape.
        applet.endShape();
    }

    /**
     * The drawModel method draws JUST the model to the screen in local
     * coordinates. The drawOn() method should have performed the appropriate
     * transforms for this to work.
     * 
     * @param applet The DDSApplet to draw onto.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #drawModel(dds.control.DDSApplet)
     */
    @Override
    public final void drawModel(final DDSApplet applet) {
        // Get the Building that we draw.
        final Building building = (Building) this.getSimObject();

        // Draw the building's front wall;
        this.drawWall(applet, building.getFrontWallVerts());

        // Check to see if it has a first wall.
        if (building.isFirstWall()) {
            // It does, so draw it.
            this.drawWall(applet, building.getFirstWallVerts());
        }

        // Check to see if it has a second wall.
        if (building.isSecondWall()) {
            // It does, so draw it.
            this.drawWall(applet, building.getSecondWallVerts());
        }

        // Check to see if it should have a back wall, which occurs when there
        // is a wall other than the front wall defined.
        if (building.isFirstWall() || building.isSecondWall()) {
            // It does, so draw the back wall.
            this.drawWall(applet, building.getBackWallVerts());
        }

        // Draw the Ceiling.
        this.drawCeiling(applet, building.getCeilingVerts());
    }

    /**
     * The drawWall method draws a wall of the Building.
     * 
     * @param applet The DDSApplet to draw onto.
     * @param verts The List that holds the verts.
     */
    private void drawWall(final DDSApplet applet, final List<Vector3D> verts) {
        // Turn off stroke.
        applet.noStroke();

        // Begin a QUAD for the wall.
        applet.beginShape(DDSApplet.QUADS);
        // int textureStart = 2;

        // Check to see if there's a texture defined.
        if (this.texture == null) {
            // There isn't a texure, set the fill to the parent Block's default
            // color.
            applet.fill(((Building) this.getSimObject()).getParent().getType()
                    .getColor());
        } else {
            // There is a texture, set the fill to white so color blending works
            // correctly.
            applet.fill(WHITE);
        }

        // Check to see if there's a texture defined.
        if (this.texture == null) {
            // There isn't a texure, just draw the verts.
            // Iterate through the verts.
            for (Vector3D vert : verts) {
                // Draw each one with UV mapping.
                applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos());
            }
        } else {
            // There is a texture, apply it.
            applet.texture(this.texture);

            // Draw the verts.
            applet.vertex(verts.get(FIRST).getXpos(), verts.get(FIRST)
                    .getYpos(), verts.get(FIRST).getZpos(), 0, 0);
            applet.vertex(verts.get(SECOND).getXpos(), verts.get(SECOND)
                    .getYpos(), verts.get(SECOND).getZpos(), 1, 0);
            applet.vertex(verts.get(THIRD).getXpos(), verts.get(THIRD)
                    .getYpos(), verts.get(THIRD).getZpos(), 1, 1);
            applet.vertex(verts.get(LAST).getXpos(), verts.get(LAST).getYpos(),
                    verts.get(LAST).getZpos(), 0, 1);
        }

        // Complete the QUADS shape.
        applet.endShape();
    }

    /**
     * The updateFromSim updates this renderer to match the data within the
     * simulation.
     * 
     * @param arg The argument of the update.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #updateFromSim(java.lang.Object)
     */
    @Override
    public final void updateFromSim(final Object arg) {
        // TODO Write updateFromSim in Buildingrenderer.

    }

}
