package dds.view.simulation.renderers;

import java.util.ArrayList;
import java.util.Collection;

import processing.core.PConstants;
import processing.core.PImage;
import dds.control.DDSApplet;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Intersection;
import dds.model.simulation.Vector3D;
import dds.view.simulation.AbstractSimObjectRenderer;

/**
 * The IntersectionRenderer class is responsible for drawing a Intersection on
 * the screen.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
// TODO Fully comment IntersectionRenderer.
public class IntersectionRenderer extends AbstractSimObjectRenderer implements
        PConstants {

    /**
     * The DEFAULT_COLOR constant defines the color to use for rendering an
     * Intersection.
     */
    private static final int DEFAULT_COLOR = 0xff777777;

    /**
     * The texture to use for redering the intersection.
     */
    private final transient PImage texture;

    /**
     * The verts is the collection of verts used to draw this object.
     */
    private final Collection<Vector3D> verts;

    /**
     * The vertsSetFromSim flag to tell if verts have been loaded. This object's
     * verts don't change so this is useful to prevent extra Collection
     * operations from trying to refresh all the verts from the model.
     */
    private boolean vertsSetFromSim;

    /**
     * Constructor for IntersectionRenderer.
     * 
     * @param simObj The SimObject from the model (for updates).
     */
    public IntersectionRenderer(final AbstractSimObject simObj) {
        this((Intersection) simObj);
    }

    /**
     * Constructor for IntersectionRenderer.
     * 
     * @param obj The SimObject from the model (for updates).
     */
    public IntersectionRenderer(final Intersection obj) {
        super(obj);
        this.vertsSetFromSim = false;
        obj.calculateVerts();
        this.verts = new ArrayList<Vector3D>();
        this.verts.addAll(obj.getVerts());
        this.texture = DDSApplet.getApplet().loadImage("road.jpg");
    }

    /**
     * The applyOrientation method handles the use of an object's orientation
     * during rendering. Since not all objects need their orientation handled,
     * not all renderers will use a significant implementation of this.
     * 
     * @param applet The DDSApplet to draw onto.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #applyOrientation(dds.control.DDSApplet)
     */
    @Override
    public void applyOrientation(final DDSApplet applet) {
        // TODO Write applyOrientation in IntersectionRenderer

    }

    /**
     * The drawOn method is responsible for rendering an intersection.
     * 
     * @param applet The applet on which to draw the context.
     */
    public final void drawModel(final DDSApplet applet) {
        applet.fill(DEFAULT_COLOR);
        applet.beginShape(DDSApplet.TRIANGLE_FAN);
        applet.texture(this.texture);
        for (Vector3D vert : this.verts) {
            applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos(),
                    getMapping(vert, true, applet, texture), getMapping(vert,
                            false, applet, texture));
        }
        applet.endShape();
    }

    /**
     * Getter for verts.
     * 
     * @return the verts
     */
    public final Collection<Vector3D> getVerts() {
        return verts;
    }

    /**
     * Getter for vertsSetFromSim.
     * 
     * @return the vertsSetFromSim
     */
    public final boolean isVertsSetFromSim() {
        return vertsSetFromSim;
    }

    /**
     * Setter for verts.
     * 
     * @param vertices the verts to set
     */
    public final void setVerts(final Collection<Vector3D> vertices) {
        this.verts.clear();
        this.verts.addAll(vertices);
    }

    /**
     * Setter for vertsSetFromSim.
     * 
     * @param isSetFromSim the vertsSetFromSim to set
     */
    public final void setVertsSetFromSim(final boolean isSetFromSim) {
        this.vertsSetFromSim = isSetFromSim;
    }

    /**
     * The updateFromSim method handles updates coming from within the
     * simulation. Based on the update argument it will appropriately update the
     * view for the change in the model.
     * 
     * @param arg The argument for the update.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #updateFromSim(java.lang.Object)
     */

    public final void updateFromSim(final Object arg) {
        if (!this.vertsSetFromSim) {
            this.setVerts(this.getSimObject().getVerts());
            this.vertsSetFromSim = true;
        }
    }
}
