/**
 * RendererType.java.
 */
package dds.view.simulation.renderers;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Block;
import dds.model.simulation.Building;
import dds.model.simulation.Car;
import dds.model.simulation.GenericMovableObject;
import dds.model.simulation.GenericObject;
import dds.model.simulation.GenericVehicle;
import dds.model.simulation.Intersection;
import dds.model.simulation.Road;
import dds.model.simulation.RoadSign;
import dds.model.simulation.TrafficLight;
import dds.view.simulation.AbstractSimObjectRenderer;

/**
 * The RendererType enumeration class is responsible for defining the SimObject
 * to SimObjectRenderer mapping. It basicaly defines which renderer to use to
 * draw the different types of SimObjects.
 * 
 * @author Jacob Bolton
 */
public enum RendererType {
    /**
     * The BLOCK enumeration defines which renderer to use for the Block
     * SimObject.
     */
    BLOCK(Block.class, BlockRenderer.class),

    /**
     * The BUILDING enumeration defines which renderer to use for the Building
     * SimObject.
     */
    BUILDING(Building.class, BuildingRenderer.class),

    /**
     * The GENERIC_MOVEABLE enumeration defines which renderer to use for the
     * GenericMovableObject SimObject.
     */
    GENERIC_MOVEABLE(GenericMovableObject.class, GenericMovableRenderer.class),

    /**
     * The GENERIC_OBJECT enumeration defines which renderer to use for the
     * GenericObject SimObject.
     */
    GENERIC_OBJECT(GenericObject.class, GenericObjectRenderer.class),

    /**
     * The GENERIC_VEHICLE enumeration defines which renderer to use for the
     * GenericVehicle SimObject.
     */
    GENERIC_VEHICLE(GenericVehicle.class, GenericVehicleRenderer.class),

    /**
     * The CAR enumeration defines which renderer to use for the Car SimObject.
     */
    CAR(Car.class, GenericVehicleRenderer.class),

    /**
     * The INTERSECTION enumeration defines which renderer to use for the
     * Intersection SimObject.
     */
    INTERSECTION(Intersection.class, IntersectionRenderer.class),

    /**
     * The ROAD enumeration defines which renderer to use for the Road
     * SimObject.
     */
    ROAD(Road.class, RoadRenderer.class),

    /**
     * The ROAD_SIGN enumeration defines which renderer to use for the RoadSign
     * SimObject.
     */
    ROAD_SIGN(RoadSign.class, RoadSignRenderer.class),

    /**
     * The TRAFFIC_LIGHT enumeration defines which renderer to use for the
     * TrafficLight SimObject.
     */
    TRAFFIC_LIGHT(TrafficLight.class, TrafficLightRenderer.class);

    /**
     * The simObjectClass is the Class of the SimObject that is to be mapped to
     * a renderer.
     */
    private final Class simObjectClass;

    /**
     * The rendererClass is the Class of the SimObjectRenderer that is used for
     * the specified SimObject.
     */
    private final Class rendererClass;

    /**
     * The constructor is the Constructor to use to create new instances of the
     * specified renderer.
     */
    private Constructor constructor;

    /**
     * Constructor for RendererType.
     * 
     * @param objectClass The SimObject Class to map to a renderer.
     * @param renderClass The SimObjectRenderer Class to use to render.
     */
    private RendererType(final Class objectClass, final Class renderClass) {
        // Assign the Classes for the SimObject/SimObjectRenderer mapping.
        this.simObjectClass = objectClass;
        this.rendererClass = renderClass;

        // Double check the mappings provided to make sure they make sense by
        // ensuring there is a SimObject Class and a SimObjectRenderer Class
        // provided. This is mainly a sanity check for developers so that it's
        // easier to trace improper Class mappings.
        if ((!AbstractSimObject.class.isAssignableFrom(this.simObjectClass))
                || (!AbstractSimObjectRenderer.class
                        .isAssignableFrom(this.rendererClass))) {
            // One is not a valid subclass, throw an exception.
            throw new UnsupportedOperationException("Definition of RenderType."
                    + this.toString()
                    + " requires Class definitions that are subclasses"
                    + " of AbstractSimObject and AbstractSimObjectRenderer");
        }

        try {
            // Try to get the appropriate constructor for the SimObjectRenderer.
            this.constructor = this.rendererClass
                    .getConstructor(new Class[] { AbstractSimObject.class });
        } catch (SecurityException e) {
            // TODO Add logger.
            System.out.println(e);
        } catch (NoSuchMethodException e) {
            // TODO Add logger.
            System.out.println(e);
        }
    }

    /**
     * The parseType method tries to determine which enumerated mapping to use
     * for a particular Object.
     * 
     * @param object The SimObject to find the RendererType for.
     * @return The RendererType for the specified SimObject or null if one
     *         cannot be found.
     */
    public static RendererType parseType(final AbstractSimObject object) {
        // Set default result to null.
        RendererType result = null;

        // Make sure the object passed in is not null.
        if (object != null) {
            // Loop through all defined enumeration types.
            for (RendererType type : RendererType.values()) {
                // Check that the object's Class matches the Class in the
                // current RendererType.
                if (object.getClass().equals(type.simObjectClass)) {
                    // A match has been found so set the result and break out of
                    // the loop.
                    result = type;
                    break;
                }
            }
        }

        // Check to see if the result is null.
        if (result == null) {
            // It is, which is bad since this means we couldn't find a matching
            // RendererType.
            // TODO Add logger.
            new UnsupportedOperationException("Parse failed on class: "
                    + object.getClass());
        }

        // Return the result.
        return result;
    }

    /**
     * The createRenderer method tries to create a new SimObjectRenderer for the
     * specified SimObject.
     * 
     * @param object The SimObject to create a renderer for.
     * @return A new SimObjectRenderer or null if one cannot be dtermined.
     */
    public static AbstractSimObjectRenderer createRenderer(
            final AbstractSimObject object) {
        // Set default result to null.
        AbstractSimObjectRenderer result = null;

        // Check to see if the specified SimObject is not null.
        if (object != null) {
            // It's not so try to get the RendererType for it.
            final RendererType type = parseType(object);

            // Check to see if a RendererType was able to be resolved.
            if (type != null) {
                // It was so try to instantiate a new renderer.
                try {
                    result = (AbstractSimObjectRenderer) type.constructor
                            .newInstance(new Object[] { object });
                } catch (IllegalArgumentException e) {
                    // TODO Add logger.
                    System.err.println(e);
                } catch (InstantiationException e) {
                    // TODO Add logger.
                    System.err.println(e);
                } catch (IllegalAccessException e) {
                    // TODO Add logger.
                    System.err.println(e);
                } catch (InvocationTargetException e) {
                    // TODO Add logger.
                    System.err.println(e);
                }
            }
        }
        // TODO Consider logging and/or throwing errors for unresolved
        // RedndererTypes and bad instantiation.

        // Return the result.
        return result;
    }

}
