package dds.view.simulation.renderers;

import java.util.ArrayList;
import java.util.Collection;

import processing.core.PImage;
import dds.control.DDSApplet;
import dds.model.simulation.AbstractSimObject;
import dds.model.simulation.Road;
import dds.model.simulation.Vector3D;
import dds.view.simulation.AbstractSimObjectRenderer;

/**
 * The RoadRenderer class is responsible for drawing a Road and it's attached
 * objects to the screen.
 * 
 * @author Jacob Bolton
 * @author cyberdan
 */
public class RoadRenderer extends AbstractSimObjectRenderer {

    /**
     * The ARBITRARY_N constant is an arbitrary constant that helps determine
     * the length of the dashes that make up the lane lines.
     */
    private static final float ARBITRARY_N = 0.2f;

    /**
     * The DASH_LINE_LENGTH constant defines the default length for a lane line
     * segment.
     */
    private static final float DASH_LINE_LENGTH = 4.0f;

    /**
     * The FIRST constant defines the index of the first vert that defines a
     * Road.
     */
    private static final int FIRST = 0;

    /**
     * The LANE_COLOR constant defines the color used to fill the lane lines.
     */
    private static final int LANE_COLOR = 0x64ffff00;

    /**
     * The LAST constant defines the index of the last vert that defines a Road.
     */
    private static final int LAST = 3;

    /**
     * The ROAD_COLOR constant defines the color to use for rendering the road's
     * asphault.
     */
    private static final int ROAD_COLOR = 0xff777777;

    /**
     * The SECOND constant defines the index of the second vert that defines a
     * Road.
     */
    private static final int SECOND = 1;

    /**
     * The THIRD constant defines the index of the third vert that defines a
     * Road.
     */
    private static final int THIRD = 2;

    /**
     * The VERT_COUNT constant defines the number of verts that are used to
     * define a Road.
     */
    private static final int VERT_COUNT = 4;

    /**
     * The Z_OFFSET constant defines the distance used to place the lane lines
     * over the road so that they display nicely.
     */
    private static final float Z_OFFSET = 0.1f;

    /**
     * The lanes field describes how many lanes are on this Road.
     */
    // TODO Determine if this field is necessary in the renderer.
    private int lanes;

    /**
     * The laneVerts field stores the vertices that defines the polygnos used
     * for drawing the lane lines..
     */
    private final Collection<Vector3D> laneVerts;

    /**
     * The texture field stores the texture to use for rendering the road.
     */
    private final transient PImage texture;

    /**
     * The verts field stores the vertices used to render the Road polygon.
     */
    private final Collection<Vector3D> verts;

    /**
     * Constructor for RoadRenderer.
     * 
     * @param simObj The Road in the simulation that this instance renders.
     */
    public RoadRenderer(final AbstractSimObject simObj) {
        this((Road) simObj);
    }

    /**
     * Constructor for RoadRenderer.
     * 
     * @param obj The Road in the simulation that this instance renders.
     */
    public RoadRenderer(final Road obj) {
        super(obj);

        // Get data from the Road object
        this.lanes = obj.getLanes();
        this.verts = new ArrayList<Vector3D>();

        // Force object to calculate the verts.
        obj.calculateVerts();
        // Get the verts from the object.
        this.setVerts(obj.getVerts());

        // Create storage space for lane verts
        this.laneVerts = new ArrayList<Vector3D>();

        // Calculate them here in the view since they are NOT needed for in the
        // simulation model.
        this.calculateLaneVerts();

        // Load the texture to use for rendering the Road.
        this.texture = DDSApplet.getApplet().loadImage("road.jpg");
    }

    /**
     * The applyOrientation method handles the use of an object's orientation
     * during rendering. Since not all objects need their orientation handled,
     * not all renderers will use a significant implementation of this.
     * 
     * @param applet The DDSApplet to draw onto.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #applyOrientation(dds.control.DDSApplet)
     */
    @Override
    public void applyOrientation(final DDSApplet applet) {
        // TODO Determine if applyOrientation in RoadRenderer needs to be
        // written.
    }

    /**
     * The calculateLaneVerts computes the QUADS for the dashed lanes. It should
     * only be called once after the verts for the Road have been assigned.
     */
    // TODO Comment details of calculateLaneVerts
    private void calculateLaneVerts() {
        final Road theRoad = (Road) this.getSimObject();
        Vector3D[] arrVerts = new Vector3D[VERT_COUNT];

        arrVerts[FIRST] = theRoad.getVerts().get(FIRST);
        arrVerts[SECOND] = theRoad.getVerts().get(LAST);
        arrVerts[THIRD] = theRoad.getVerts().get(THIRD);
        arrVerts[LAST] = theRoad.getVerts().get(SECOND);

        final Vector3D dir = Vector3D.subtract(arrVerts[SECOND],
                arrVerts[FIRST]);
        final Vector3D shiftDir = Vector3D.subtract(arrVerts[LAST],
                arrVerts[FIRST]);
        final Vector3D endDir = Vector3D.subtract(arrVerts[THIRD],
                arrVerts[SECOND]);
        float roadLength = 0f;
        int totalDashCount = 0;

        for (int l = 1; l <= lanes - 1; l++) {
            final Vector3D shiftStart = walkFractionInDir(arrVerts[0],
                    shiftDir, ((float) l) / ((float) lanes));
            final Vector3D shiftEnd = walkFractionInDir(arrVerts[1], endDir,
                    ((float) l) / ((float) lanes));
            float tdc = 0f;
            Vector3D shiftVector = shiftStart.duplicate();
            roadLength = Vector3D.subtract(shiftEnd, shiftStart).magnitude();
            tdc = roadLength / (2f * DASH_LINE_LENGTH);
            totalDashCount = Math.round(tdc);

            for (int dc = 0; dc < totalDashCount; dc++) {
                dash(shiftVector, dir, shiftDir, dir.magnitude());
                shiftVector = walkNUnits(shiftVector, dir, dir.magnitude(),
                        2 * DASH_LINE_LENGTH);
            }
        }
    } // end of method calculateLaneVerts

    /**
     * The dash method creates a new dash on the road based on the supplied
     * vectors indicating road direction & magnitude.
     * 
     * @param origin the origin
     * @param dir the direction
     * @param dirLen the magnitude of the dir vector
     * @param shift The offset?
     */
    // TODO Properly comment offset.
    // TODO Comment details of dash
    private void dash(final Vector3D origin, final Vector3D dir,
            final Vector3D shift, final float dirLen) {
        final float shiftMag = shift.magnitude();
        Vector3D dashStartTop, dashStartBottom, dashEndTop, dashEndBottom;
        dashStartTop = walkNUnits(origin, shift, shiftMag, -ARBITRARY_N);
        dashStartBottom = walkNUnits(origin, shift, shiftMag, ARBITRARY_N);
        dashEndTop = walkNUnits(walkNUnits(origin, dir, dirLen,
                DASH_LINE_LENGTH), shift, shiftMag, -ARBITRARY_N);
        dashEndBottom = walkNUnits(walkNUnits(origin, dir, dirLen,
                DASH_LINE_LENGTH), shift, shiftMag, ARBITRARY_N);

        laneVerts.add(new Vector3D(dashStartTop.getXpos(), dashStartTop
                .getYpos(), dashStartTop.getZpos() + Z_OFFSET));
        laneVerts.add(new Vector3D(dashEndTop.getXpos(), dashEndTop.getYpos(),
                dashEndTop.getZpos() + Z_OFFSET));
        laneVerts.add(new Vector3D(dashEndBottom.getXpos(), dashEndBottom
                .getYpos(), dashEndBottom.getZpos() + Z_OFFSET));
        laneVerts.add(new Vector3D(dashStartBottom.getXpos(), dashStartBottom
                .getYpos(), dashStartBottom.getZpos() + Z_OFFSET));
    } // end of method dash

    /**
     * The drawOn draws the textured road onto the applet.
     * 
     * @param applet The applet on which to draw the context.
     */
    public final void drawModel(final DDSApplet applet) {
        // Don't use a stroke
        applet.noStroke();

        // Set the fill color for the road.
        applet.fill(ROAD_COLOR);

        // Begin drawing the polygon.
        applet.beginShape(DDSApplet.POLYGON);

        // Apply the road's texture.
        applet.texture(this.texture);

        // Iterate through the verts.
        for (Vector3D vert : this.getVerts()) {
            // Draw the current vert using the appropriate texture mapping.
            applet.vertex(vert.getXpos(), vert.getYpos(), vert.getZpos(),
                    getMapping(vert, true, applet, texture), getMapping(vert,
                            false, applet, texture));
        }

        // End the polygon.
        applet.endShape();

        // Render the lane lines.
        this.renderLanes(applet);
    } // end of method drawModel

    /**
     * Getter for lanes.
     * 
     * @return the lanes
     */
    public final int getLanes() {
        return lanes;
    }

    /**
     * Getter for laneVerts.
     * 
     * @return the laneVerts
     */
    public final Collection<Vector3D> getLaneVerts() {
        return laneVerts;
    }

    /**
     * Getter for verts.
     * 
     * @return the verts
     */
    public final Collection<Vector3D> getVerts() {
        return verts;
    }

    /**
     * The renderLanes renders the pretty yellow dashes for lanes on the roads.
     * 
     * @param applet the context to draw on
     */
    private void renderLanes(final DDSApplet applet) {
        // Set the fill to the lane line's color.
        applet.fill(LANE_COLOR);

        // Begin drawing the line verts as quads.
        applet.beginShape(DDSApplet.QUADS);

        // Iterate through the verts.
        for (Vector3D v : laneVerts) {
            // Draw the current vert.
            applet.vertex(v.getXpos(), v.getYpos(), v.getZpos());
        }

        // End the set of QUADS
        applet.endShape();
    } // end of method renderLanes

    /**
     * Setter for lanes.
     * 
     * @param numlanes the lanes to set
     */
    public final void setLanes(final int numlanes) {
        this.lanes = numlanes;
    }

    /**
     * Setter for laneVerts.
     * 
     * @param newVerts the laneVerts to set
     */
    public final void setLaneVerts(final Collection<Vector3D> newVerts) {
        this.laneVerts.clear();
        this.laneVerts.addAll(newVerts);
    }

    /**
     * Setter for verts.
     * 
     * @param vertices the verts to set
     */
    public final void setVerts(final Collection<Vector3D> vertices) {
        // Clear out the old verts so we can reuse the Collection.
        this.verts.clear();

        // Add all of the new verts.
        this.verts.addAll(vertices);
    }

    /**
     * The updateFromSim method handles updates coming from within the
     * simulation. Based on the update argument it will appropriately update the
     * view for the change in the model.
     * 
     * @param arg The argument for the update.
     * @see dds.view.simulation.AbstractSimObjectRenderer
     *      #updateFromSim(java.lang.Object)
     */
    public void updateFromSim(final Object arg) {
        // TODO Write updateFromSim in RoadRenderer
    }

    /**
     * The walkFractionInDir moves along the vector dir by a fraction of the
     * magnitude.
     * 
     * @param start the origin to start from
     * @param dir the direction vector
     * @param fraction the fraction of the vector to move along
     * @return the calculated displacement vector
     */
    // TODO Comment details of walkFractionInDir
    private Vector3D walkFractionInDir(final Vector3D start,
            final Vector3D dir, final float fraction) {
        final float dirMag = dir.magnitude();
        return walkNUnits(start, dir, dirMag, dirMag * fraction);
    } // end of method walkFractionInDir

    /**
     * The walkNUnits moves N units along the vector dir, starting from point o.
     * 
     * @param start the origin
     * @param dir the direction of traversal
     * @param dirLen the magnitude of dir
     * @param units the number of units to move
     * @return Vector3D representing the point after the walk.
     */
    // TODO Comment details of walkNUnits
    private Vector3D walkNUnits(final Vector3D start, final Vector3D dir,
            final float dirLen, final float units) {
        final Vector3D dirScaled = dir.duplicate();
        dirScaled.scale(units / dirLen);
        return Vector3D.add(start, dirScaled);
    } // end of method walkNUnits
}
